#ifndef WIMESH_CAALGOS
#define WIMESH_CAALGOS


#include "WMtypes.h"



/*
 * Hierarchy of classes that can be used to run
 * different channel (re-)assignment algorithms
 */


/** Base class (does nothing) **/
class CAalgo
{
public:
  /* static method to create a new object of a derived class.
   * NOTE This method (defined in 1channel.cpp) must be updated every time a new derived
   * class is added
   */
  static CAalgo* Create (string algo, const list<pair<string,string> >& params);
  
  // should return a non-zero value in case of errors, 0 otherwise
  virtual double Run(Topology* T)  { return 0; }
  virtual const char* name()  { return "N"; }
  virtual vector<float>* getPowerConstraints () { return NULL; }
  virtual void printFurtherInfo (ostream& ofs) {}
  virtual ~CAalgo() {}
};


/*** Channel Assignment algorithms ***/

/** OneChannel **/
class OneChannel : public CAalgo
{
public:
  virtual double Run(Topology* T);
  virtual const char* name()  { return "ONE"; }
};


/** IATC **/
class IATC: public CAalgo
{
public:
  virtual double Run(Topology* T);
  virtual const char* name()  { return "IATC"; }
};


/** LACA **/
class LACA: public CAalgo
{
public:
  virtual double Run(Topology* T);
  virtual const char* name()  { return "LACA"; }
};


/** FCPRA **/
class FCPRA : public CAalgo
{
public:
  FCPRA (bool Optimize, bool turbo, bool RatePowerAdapt) :
    m_optimize(Optimize),
    m_turbo(turbo),
    m_ratepoweradapt(RatePowerAdapt) {}

  FCPRA (const list<pair<string,string> >& params);

  virtual double Run(Topology* T);
  virtual const char* name();

private:
  bool m_optimize;
  bool m_turbo;
  bool m_ratepoweradapt;
};


/** MVCRA **/
class MVCRA : public CAalgo
{
public:
  MVCRA (bool RateAdapt, unsigned int MaxNumChanges, string ccf = "", string lp = "", string lpp = "") :
    m_rateadapt(RateAdapt),
    m_maxnumchanges(MaxNumChanges),
    m_ccf(ccf),
    m_lp(lp),
    m_lpp(lpp)
    {}
  
  MVCRA (const list<pair<string,string> >& params);
  
  virtual double Run(Topology* T);
  virtual const char* name();

private:
  bool m_rateadapt;
  unsigned int m_maxnumchanges;
  string m_ccf, m_lp, m_lpp;
};


/** MP-CARA **/
class MPCARA : public CAalgo
{
public:
  MPCARA (RndVar* npc, double seed, bool reset, bool fcpra, bool old, double u0 = 0.5) :
    m_npc(npc),
    m_seed(seed),
    m_reset(reset),
    m_fcpra(fcpra),
    m_old(old),
    m_u0(u0),
    m_nodepc(0)
    {}

    
  MPCARA (const list<pair<string,string> >& params);
  
  virtual double Run(Topology* T);
  virtual const char* name();
  virtual vector<float>* getPowerConstraints () { return m_nodepc; }
  virtual ~MPCARA() { if (m_nodepc) delete m_nodepc; };

private:
  double doRun(Topology* T);

  RndVar* m_npc;
  double m_seed;
  bool m_reset;
  bool m_fcpra;
  bool m_old;
  double m_u0;
  vector<float>* m_nodepc;
};


/** MPCARA-OPT **/
class MPCARA_OPT : public CAalgo
{
public:
  MPCARA_OPT (RndVar* npc, double seed, double u0, long to1, long to2) :
    m_npc(npc),
    m_seed(seed),
    m_u0(u0),
    m_to1(to1),
    m_to2(to2),
    m_nodepc(0)
    {}

  MPCARA_OPT (const list<pair<string,string> >& params);
    
  virtual double Run(Topology* T);
  virtual const char* name();
  virtual vector<float>* getPowerConstraints () { return m_nodepc; }
  virtual void printFurtherInfo (ostream& ofs) { ofs << endl << "# Highest lower bound found: " << m_lbound << endl; }
  virtual ~MPCARA_OPT() { if (m_nodepc) delete m_nodepc; };

private:
  RndVar* m_npc;
  double m_seed;
  double m_u0;
  long m_to1;
  long m_to2;
  vector<float>* m_nodepc;
  double m_lbound;
};


/** PACA2 **/
class PACA2 : public CAalgo
{
public:
  PACA2 (bool reset, bool fcpra = false, double u0 = 0.5) :
    m_reset(reset),
    m_fcpra(fcpra),
    m_u0(u0)
    {}

  PACA2 (const list<pair<string,string> >& params);

  virtual double Run(Topology* T);
  virtual const char* name();

private:
  bool m_reset;
  bool m_fcpra;
  double m_u0;
};


/** PACA **/
class PACA : public CAalgo
{
public:
  PACA (double u0 = 0.5) : m_u0(u0)  {}

  PACA (const list<pair<string,string> >& params);

  virtual double Run(Topology* T);
  virtual const char* name()  { return "E2CARA"; }
    
private:
  double m_u0;
};


/** E2CAR-OPT **/
class E2CAR_OPT : public CAalgo
{
public:
  E2CAR_OPT (bool relax, double u0 = 0.5) :
    m_relax(relax),
    m_u0(u0)
    {}

  E2CAR_OPT (const list<pair<string,string> >& params);

  virtual double Run(Topology* T);
  virtual const char* name()  { return "E2CAR-OPT"; }
    
private:
  bool m_relax;
  double m_u0;
};




#endif
