#ifndef WIMESH_TYPES
#define WIMESH_TYPES

#include <vector>
#include <list>
#include <set>
#include <map>
#include <cstdio>
#include <limits>
#include <fstream>
#include <cmath>
#include "rndvars.h"



using namespace std;


const float availTXrates[] = {6, 9, 12, 18, 24, 36, 48, 54};

#define NUMRATES sizeof(availTXrates)/sizeof(float)


// P = 30 + 10*log(W)  where P is the power in dbm and W is the power in Watt
// For ns-3 compatibility, values (in dbm) should be equally spaced among the minimum
// and the maximum
// const float availTxPwr[] = {
// 	0.00100f,	//  0 dbm
// 	0.00125f,	//  1 dbm
// 	0.00158f,	//  2 dbm
// 	0.00200f,	//  3 dbm
// 	0.00251f,	//  4 dbm
// 	0.00316f,	//  5 dbm
// 	0.00398f,	//  6 dbm
// 	0.00501f,	//  7 dbm
// 	0.00631f,	//  8 dbm
// 	0.00794f,	//  9 dbm
// 	0.01000f,	// 10 dbm
// 	0.01259f,	// 11 dbm
// 	0.01585f,	// 12 dbm
// 	0.01995f,	// 13 dbm
// 	0.02512f,	// 14 dbm
// 	0.03162f,	// 15 dbm
// 	0.03981f,	// 16 dbm
// 	0.05012f,	// 17 dbm
// 	0.06310f,	// 18 dbm
// 	0.07943f,	// 19 dbm
// 	0.10000f	// 20 dbm
// };
// #define NUMTXPWR sizeof(availTxPwr)/sizeof(float) 

// Tx power levels (dBm)
// For ns-3 compatibility, we specify the minimum and maximum values and the number of levels

const float minTxPowerLevel = 0.f;
const float maxTxPowerLevel = 20.f;
const unsigned int numTxPowerLevels = 21;   // 0, 1, 2, 3, ... 20 dBm

// return the i-th tx power level in dBm (0 <= i <= numTxPowerLevels-1)
inline float TxPowerLevel (unsigned int i) { return minTxPowerLevel + i * (maxTxPowerLevel - minTxPowerLevel) / (numTxPowerLevels - 1); }

// P = 30 + 10*log(W) = 10*log(1000*W) where P is the power in dBm and W is the power in Watt
inline float DbmToW (float dBm) { return pow (10.0f, dBm / 10.0f) / 1000.0f; }

/*
  SINRthresholds are determined such that nodes can communicate
    at 54 Mbps when within 30m
    at 48 Mbps when within 32m
    at 36 Mbps when within 37m
    at 24 Mbps when within 45m
    at 18 Mbps when within 60m
    at 12 Mbps when within 69m
    at  9 Mbps when within 77m
    at  6 Mbps when within 90m
  if transmitting at 15 dbm
*/
const float SINRthresholds[NUMRATES] = {
	0.390f,	// gamma_6
	0.533f,	// gamma_9
	0.664f,	// gamma_12
	0.878f,	// gamma_18
	1.561f,	// gamma_24
	2.309f,	// gamma_36
	3.087f,	// gamma_48
	3.513f	// gamma_54
};


const unsigned int defaultNumChannels = 12;

const float defaultTXrate = availTXrates[NUMRATES-1];

// TODO const float defaultTXpower = availTxPwr[13];

const unsigned int defaultTxPowerLevel = 16;    // 16 dBm

const float default_db_gain = 1.0;

const unsigned int default_min_angle = 0;

const unsigned int default_max_angle = 360;

// Power consumption values (Watts)

typedef struct
{
  float BasePowerConsumption;
  float StandbyPowerConsumption;
  float rho_tx;
  float rho_rx;
  float rho_idle;
  float rho_sleep;
  char  name[30];
} PowerConsumptionModel;


const PowerConsumptionModel availPCmodels[] =
{
  { 2.29f, 0.50f, 3.17f, 2.36f, 1.22f, 0.29f, "Banchs" },    // default
  { 2.29f, 0.50f, 2.10f, 1.60f, 1.45f, 0.10f, "dot11n" }
};


class NodePowerConsumptionProfile
{
public:
  float BasePowerConsumption;
  float StandbyPowerConsumption;
  
  NodePowerConsumptionProfile (float bpc, float spc)
   : BasePowerConsumption (bpc), StandbyPowerConsumption (spc) {}
};

const NodePowerConsumptionProfile availNodePCprofiles[] =
{
  NodePowerConsumptionProfile (2.29f, 0.50f)
};



typedef enum { router, aggregation, gateway } TMeshRouter;

const char MeshTypes[sizeof(TMeshRouter)] = "RAG";


#define almost_null exp(-10.)




/** Classes used by WiMesh **/


/* Abstract base class for radio power consumption in transmission mode function object*/
class foRadioTxPower
{
public:
  virtual ~foRadioTxPower() {}

  virtual foRadioTxPower* Clone () = 0;
  virtual float operator()(float power) = 0;

  virtual void SerializeTo (ofstream& ofs) = 0;
  static foRadioTxPower* DeserializeFrom (ifstream& ifs);
};

/* The radio power consumption is a linear function of the transmission
 * power level (availTxPwr values): a * power + b
 * This is the default method, with a = 0 and b = availPCmodels[0].rho_tx
 * NOTE If you define new classes derived from foRadioTxPower, remember to define 
 * the string name in topology.cpp and update foRadioTxPower::DeserializeFrom
 * (topology.cpp) and WiMeshParser::on_start_element (parsers.cpp)
 */
class foRadioTxPowerLinear: public foRadioTxPower
{
  float a, b;
public:
  foRadioTxPowerLinear (float _a, float _b) : a(_a), b(_b) {}

  float operator()(float power) { return a * power + b; }
  foRadioTxPower* Clone () { return new foRadioTxPowerLinear (*this); }

  void SerializeTo (ofstream& ofs);
  static foRadioTxPowerLinear* DeserializeFrom (ifstream& ifs);

  static const string name;
};


class RadioPowerConsumptionProfile
{
public:
  foRadioTxPower* rho_tx;
  float rho_rx;
  float rho_idle;
  float rho_sleep;
  
  RadioPowerConsumptionProfile (foRadioTxPower* tx, float rx, float idle, float sleep)
   : rho_tx(tx), rho_rx(rx), rho_idle(idle), rho_sleep(sleep) {}
};

const RadioPowerConsumptionProfile availRadioPCprofiles[] =
{
  RadioPowerConsumptionProfile (new foRadioTxPowerLinear(0, 3.17f), 2.36f, 1.22f, 0.29f),         //default
  RadioPowerConsumptionProfile (new foRadioTxPowerLinear(0, 2.10f), 1.60f, 1.45f, 0.10f)
};


class RadioIface
/* contains properties of a radio interface */
{
public:

	unsigned int	channel;	/* 0 stands for unset (for unused after channel assignment) */
	float		DBgain;
	unsigned int	minangle;
	unsigned int	maxangle;
	foRadioTxPower*	rho_tx;
	float		rho_rx;
	float		rho_idle;
	float		rho_sleep;

	/* constructor */
	RadioIface (float _DBgain = default_db_gain,
		    unsigned int _minangle = default_min_angle,
		    unsigned int _maxangle = default_max_angle,
		    foRadioTxPower* _rho_tx = availRadioPCprofiles[0].rho_tx->Clone(),
		    float _rho_rx = availRadioPCprofiles[0].rho_rx,
		    float _rho_idle = availRadioPCprofiles[0].rho_idle,
		    float _rho_sleep = availRadioPCprofiles[0].rho_sleep
	) : 
	  channel (0),
	  DBgain (_DBgain),
	  minangle (_minangle),
	  maxangle (_maxangle),
	  rho_tx (_rho_tx),
	  rho_rx (_rho_rx),
	  rho_idle (_rho_idle),
	  rho_sleep (_rho_sleep)
	{}
	
	/* copy constructor */
	RadioIface (const RadioIface& ri)
	{
	  channel = ri.channel;
	  DBgain = ri.DBgain;
	  minangle = ri.minangle;
	  maxangle = ri.maxangle;
	  rho_tx = ri.rho_tx->Clone();
	  rho_rx = ri.rho_rx;
	  rho_idle = ri.rho_idle;
	  rho_sleep = ri.rho_sleep;
	}

	~RadioIface () { delete rho_tx; }

	int getChannel() const { return channel; }

	void SerializeTo (ofstream& ofs);
	void DeserializeFrom (ifstream& ifs);
};

typedef vector<RadioIface>::iterator RadioIfaceIterator;

class MeshNode
/* contains properties of a mesh node */
{
public:
	const int		id;
	const float		xpos, ypos;
	const TMeshRouter	role;
	float			BasePowerConsumption;
	float			StandbyPowerConsumption;
	vector<RadioIface>	radio;

	/* constructor */
	MeshNode(int _id, float _xpos, float _ypos, TMeshRouter _role, int _numAvailRadios, 
		 float bpc = availNodePCprofiles[0].BasePowerConsumption, float spc = availNodePCprofiles[0].StandbyPowerConsumption) :
	  id (_id),
	  xpos (_xpos),
	  ypos (_ypos),
	  role (_role),
	  BasePowerConsumption (bpc),
	  StandbyPowerConsumption (spc),
	  radio(vector<RadioIface>(_numAvailRadios))
	{}

	/* copy constructor */
// 	MeshNode(const MeshNode& m) : id(m.id), xpos(m.xpos), ypos(m.ypos), role(m.role), numAvailRadios(m.numAvailRadios), radio(new RadioIface[m.numAvailRadios])
// 	{
// 		for (unsigned int i=0; i<m.numAvailRadios; i++)
// 			radio[i] = m.radio[i];
// 	}

	/* destructor */
// 	~MeshNode() { if (radio) delete [] radio; }
	
	unsigned int numAvailRadios() { return radio.size(); }
	
	void SerializeTo (ofstream& ofs);
	static MeshNode* DeserializeFrom (ifstream& ifs);
	
	int countRadiosWithChannel(int c);

	RadioIfaceIterator findRadioWithChannel(int c);

	set<unsigned int> findChannels ();
};

class Topology;

class TGraphNeighbor
/* characterizes a neighbor in Topology */
{
public:
	MeshNode	*info;			/* pointer to neighbor's info structure */
	RadioIface	*radio;			/* pointer to info structure of the radio used to reach the neighbor
						   Must be set to NULL in the initial graph */
	double		availBW;
	double		flow;			/* flow rate on the directed link */
	double		Utot;			/* the total utilization of the collision domain of this directed link */
	float		rate;			/* the transmission rate on this directed link, initialized to the maximum rate, but
						   set according to TXrateFunc (specified in the config file) by buildInitialGraph */
// 	float		power;
	unsigned int	txpower;                /* an index corresponding to the Tx power level in dBm */

	map<pair<int,int>,double> 		f_e2e;	/* amount of flow due to the end-to-end flow between i and j */
										/* The sum of all these values MUST equal the flow field */
	/* constructor */
	TGraphNeighbor(MeshNode *_info, float _rate = defaultTXrate, unsigned int _txpower = defaultTxPowerLevel): info(_info), radio(NULL), availBW(0.), flow(0.), Utot(0.), rate(_rate), txpower(_txpower) {}
	
	void SerializeTo (ofstream& ofs);
	static TGraphNeighbor* DeserializeFrom (ifstream& ifs, Topology* T, int me);
	
	double getFlow() const { return flow; }
};


class TGraphNode
/* characterizes a node in Topology */
{
  public:
	MeshNode		*info;		/* pointer to node's info structure */
	list<TGraphNeighbor>	neighbor;	/* list of neighbors */
	
	/* constructor */
	TGraphNode (MeshNode* _info = NULL) : info(_info) {}
	/* copy constructor */
	TGraphNode(const TGraphNode& n) : info(new MeshNode(*(n.info))), neighbor(n.neighbor) {}
	/* destructor */
	~TGraphNode() { if (info) delete info; }
};




/* Function object that can be used to sort containers of pairs based on the second element of the pairs */
template <class T1, class T2>
class compareSecond {
  public:
    bool operator()(const pair<T1,T2>& x, const pair<T1,T2>& y) { return x.second < y.second; }
};


typedef pair<vector<TGraphNode>::iterator, list<TGraphNeighbor>::iterator> _TLinkItrs;

class Topology;

class TLinkItrs : public _TLinkItrs {
public:
  TLinkItrs() : _TLinkItrs() {}
  TLinkItrs(vector<TGraphNode>::iterator u, list<TGraphNeighbor>::iterator v) : _TLinkItrs(u,v) {}
    
  bool IsInTheCollDomainOf(Topology* T, TLinkItrs l) const;
  
  TGraphNeighbor* getNeighb_ptr() const { return &(*(this->second)); }
};


/* required to compare TLinkItrs objects
   defined in utils.cpp */
bool operator<(const list<TGraphNeighbor>::iterator& x, const list<TGraphNeighbor>::iterator& y);

bool operator<(const TGraphNeighbor& x, const TGraphNeighbor& y);

class TNodeProp
{
public:
	float	distance;
	TLinkItrs prev_link;

	/* constructor */
	TNodeProp() : distance(numeric_limits<float>::max()) {}
};

/* Abstract base class for link cost function objects used by Dijkstra */
class linkCostFunc {
  public:
    virtual float operator()(TLinkItrs l) = 0;
};

/* Default link cost function object: c(l) = 1 */
class linkCost_Unary : public linkCostFunc {
  public:
    float operator()(TLinkItrs l) { return 1.0; }
};

static linkCost_Unary linkCost_Unary_obj;

/* Abstract base class for link pruning function objects used by Dijkstra */
class linkPruneFunc {
  public:
    virtual float operator()(TLinkItrs l) = 0;
};

/* Link pruning function object: availBW < BW 
   BW=0 implies all links are considered */
class linkPrune_Bwd : public linkPruneFunc {
  double bw;
  public:
    linkPrune_Bwd(double _bw) : bw(_bw) {}
    float operator()(TLinkItrs l) { return l.second->availBW < bw; }
};


/* Function which returns the id of a node */
template <class Node>
inline int getNodeID(Node u) { return u.info->id; }

/* Function which returns the channel of a radio */
inline unsigned int getRadioCh(TGraphNeighbor v) { if (v.radio) return v.radio->channel;  return 0; }

/* class that holds power consumption and time usage (tau) of each radio for a given network node */
class TPower
{
public:
  TPower (int r, float p = 0.f) : power(p) { tau.resize(r,0.f); }

  float power;
  vector<float> tau;
};


/* NETWORK TOPOLOGY */
class Topology
{
private:
	void printTopology(FILE *f);

public:
	vector<TGraphNode>	node;
	vector<int>	aggregation_list;
	vector<int>	gateway_list;

	unsigned int numAvailChannels;
	
	float noise;
	
	// parameters used while building the initial grpah
	int maxP;                    // the (index of the) maximum TX power level configurable on radios
	int min_rate;
	int max_rate;

// 	PowerConsumptionModel pm;

	float throughput;			/* used by pfr_MMCF */

	char toponame[40];
	char topopath[100];

	/* constructors */
// 	Topology(unsigned int _numAvailChannels = defaultNumChannels) : numAvailChannels(_numAvailChannels), noise(0.00001), maxP(NUMTXPWR-1), pm(availPCmodels[0]) {}
	Topology(unsigned int _numAvailChannels = defaultNumChannels) : numAvailChannels(_numAvailChannels), noise(0.00001), maxP(numTxPowerLevels-1) {}

	Topology (const char* filename);
	
	/* copy constructor */
	Topology(const Topology&);		/* node information (e.g. MeshNode and RadioIface structures) are not duplicated  */

	/* saves the topology information to a file */
	void SerializeTo (const char *filename); 
	
	/* reads the topology from a file */
	void DeserializeFrom (const char *filename);
	
// 	int getNodeInfo(const char *filename);	/* gets node info from file. Populates the node field, but does not compute neighbors,
// 						   so the neighbor list of the struct TGraphNeighbor is uninitialized */
// 	int getNodeInfo(FILE *f);
// 	int getNodeInfo() { return getNodeInfo(stdin); }

	void addNeighbor(vector<TGraphNode>::iterator u, vector<TGraphNode>::iterator v, float rate, unsigned int power);

	int MaxAllowLinkRate (int u, int v, unsigned int power, int minrate, int maxrate);
	
	void buildGi_FixedPowerMaxAllowRate(unsigned int power, int minrate, int maxrate);
	
	int buildGi_FromFile(const char * filename);

	void findToBePossiblyAddedLinks(int maxp = numTxPowerLevels);

	void printTopology(const char *filename);
	void printTopology() { printTopology(stdout); }

	void drawTopology();
	
// 	void saveConfig(const char* fname, PFR_method* pfr, bool commented);

	/* returns the distance between two nodes */
	float dist(int x, int y);
	
	void setNoise (float n) { noise = n; }

	bool PotInterfLinks(TLinkItrs x, TLinkItrs y);   /* returns true if x potentially interferes with y */

	/* adjust the TX power of each interface so as to just reach the farthest neighbor reachable on the channel that interface is set to */
	// void AdjustPower();
	/* PRE: radios must have been assigned a channel */

	void ResetChannels();
	
	/* Checks whether two nodes u and v can be neighbors using some values for TX power and rate
	  Starts from the highest TX rate and the lowest TX power, then increases the TX power
	  If the SNR is still below the threshold, lowers the TX rate
	  Thus, determines the highest possible capacity
	*/
	bool canBeNeighbors(vector<TGraphNode>::iterator u, vector<TGraphNode>::iterator v, int & c, int & p, int maxp = numTxPowerLevels);

	/* Searches node v among the neighbors of u (using whatever channel)
	 * Returns node[u].neighbor.end() if not found, otherwise an iterator 
	 * pointing to the (first) found TGraphNeighbor element
	 */
	list<TGraphNeighbor>::iterator find_neighbor(int u, int v);

	/* Start the search from a specified neighbor (useful to find successive
	 * links between u and v)
	 */
	list<TGraphNeighbor>::iterator find_neighbor(int u, int v, list< TGraphNeighbor >::iterator n);

	/* Searches node v among the neighbors of u using channel c
	 * Returns node[u].neighbor.end() if not found, otherwise an iterator 
	 * pointing to the (first) found TGraphNeighbor element
	 */
	list<TGraphNeighbor>::iterator find_neighbor(int u, int v, unsigned int c);

	TLinkItrs FindInverseLink(TLinkItrs l, unsigned int channel = 0);    /* Given a TLinkItrs u->v returns the TLinkItrs v->u */

	/* Assign channel c to link l
	PRE: for both end nodes, either a radio assigned to c already exists or an available radio exists */
	void AssignChannel2Link(TLinkItrs l, unsigned int c);

	set<unsigned int> haveCommonChannels (int u, int v);
	
	vector<TLinkItrs> computeNeighborhood(TLinkItrs l);          /* Find all the links that potentially interfere with l */

	void getChannelSet(set<unsigned int> &channelSet);	/* computes the set of channels used in the topology */

	void eraseLinksWithNullFlow();				/* erases links with null flow */

	/* Computes the Utot of all the links and returns true if they all match the value in the Utot field.
	 * If vec is non-null, it points to a vector (that must be de-allocated after use) containing
	 * all the pairs <link, computed Utot>. Values for verbose:
	 * - 0: do not print anything to stdout
	 * - 1: only print a message if the computed Utot differs from the value in the Utot field
	 * - 2: print a message for all the links
	 */
	bool checkUtot (int verbose, vector< std::pair< TLinkItrs, double > >** vec = NULL); 
	
	double computeUtot(TLinkItrs l);  			/* computes the Utot of the collision domain of a link */

	pair<TLinkItrs,double> getUtotMax();              /* returns the maximum among the utilization of the collision domains of all the links */

	/* Update the Utot for every link involved (except l itself) in case l is assigned channel c and rate r.
	* Must be invoked before making the change 
	*/
	void updateUtot (TLinkItrs l, unsigned int c, float r);

	void scaleFlowsByUtotMax();			/* scales all the precomputed flow rates and the total utilizations by the max Utot */

	bool isConnected(bool connect);			/* attempts to make the graph connected if connect is true */

	double getDegree();				/* computes the degree of the network (i.e. average number of neighbors per node) */

	int countRadiosOff();

	int countNodesOff();

	int countNumRadios();

	/* 
	 * Determines how many radios change channel between the current configuration and Tnew 
	 * oldChSet and newChSet contains, for each node, the channels differing between the two configurations
	 */
	int countRadioChanged (Topology* Tnew, vector<multiset<unsigned int> >& oldChSet, vector<multiset<unsigned int> >& newChSet);

	/* Computes the power consumption of all the nodes and the time usage (tau) for each of their radios.
	* Results are stored in the vector consumption (which can also be empty when passed) */  
	void calcAllNodesPowerConsumption (vector<TPower>& consumption);

	/* checks whether the power constraints on all the nodes and the time constraints (tau) on all the radios are satisfied.
	* If the vector of power consumptions is available (i.e., calcAllNodesPowerConsumption has been called), its address
	* can be passed so that calcAllNodesPowerConsumption is not called again
	*/
	bool checkPowerConstraints (vector<float>* nodePowerConstr, vector<TPower>* consumption = NULL);

	/* computes the total power consumption of the whole network.
	* If the vector of power consumptions is available (i.e., calcAllNodesPowerConsumption has been called), its address
	* can be passed so that calcAllNodesPowerConsumption is not called again
	*/
	float TotalPowerConsumption (vector<TPower>* consumption = NULL);

//	vector<TNodeProp> Dijkstra(int source, linkPruneFunc& prune, linkCostFunc& cost = linkCost_Unary_obj);	        /* applies Dijkstra to the topology */
	vector<TNodeProp> Dijkstra(int source, linkPruneFunc& prune, linkCostFunc& cost = linkCost_Unary_obj);	        /* applies Dijkstra to the topology */

//	vector<int> ShortestPath(int source, int dest, linkPruneFunc& prune, linkCostFunc& cost = linkCost_Unary_obj);   /* computes the shortest feasible path between source and dest */
	vector<TLinkItrs>* ShortestPath(int source, int dest, linkPruneFunc& prune, linkCostFunc& cost = linkCost_Unary_obj);

	/*
	* Find the first (at most k) shortest paths with length at most alpha times the length
	* of the shortest path. Returns the set of links on those paths.
	* By default, returns the first shortest path
	*/

	set<pair<int,int> > find_path (pair<int,int> d, float alpha = 1.0f, int k = 1);

	set<pair<int,int> > FindLooplessResilientDAG(int s, int d, float alpha);

	list<TLinkItrs>* changeFlowOnRandomPath(int maxLength, bool increase, double percent);

	/* 
	* If type = 2, swaps all the demands, in the sense that demands with less flow are assigned the flow of the demands with more flow.
	* Each demand is multiplied (type = 0 or 2) or set (type = 1) by a random value derived from the given distribution
	*/
	void changeDemands(RndVar* RV, double seed, unsigned int type = 0);

	void InitRouting();                           /* copies the flow value in the availBW field */

	/* Route flows along the shortest path; AvailBW is used as link capacity; returns the cross-section goodput */
	double RouteFlows(int NumFlows, RndVar *Source, RndVar *Dest, RndVar *Bwdr, double seed);

	/* Checks whether the e2e flows are consistent */
	bool checkE2Eflows(map<pair<int,int>,double>* ret);
	
//	bool PotInterfere(TNodePair x, TNodePair y)	/* returns true if a transmission on x causes interference on y */
//	{ return dist(x.first,y.first)<node[x.first].info->maxIFrange  ||  dist(x.first,y.second)<node[x.first].info->maxIFrange  ||  dist(x.second,y.first)<node[x.second].info->maxIFrange  ||  dist(x.second,y.second)<node[x.second].info->maxIFrange;  }
};



// Utilities for Topology
// 	int createTopology(Topology **T);
	void printCAStats(Topology *T, const char* name, bool verbose);





//REAL solveMaxFtoU(Topology *T, REAL maxF);


/*
    Channel Assignment Algorithms
 */

	void FCCA(Topology *Gi, bool Optimize);
	
	void FCRA(Topology *Gi, bool Optimize, bool turbo, bool RateAdapt);
	
	int solveILP(Topology *Gi);
	
	void FCCA_scale(Topology *T);

	void paca2_MILP(Topology* T, double u0);


/*
    Channel Assignment and Routing
 */

void RanGopChi(Topology *T, int NumFlows, RndVar *Source, RndVar *Dest, RndVar *Bwdr, double seed);


#endif
