#include "dag.h"
#include <limits>


void DAG::addLinks(vector<int>& path)
{
  for (int i=0; i<path.size()-1; i++) {
    nodes.at(path[i]).added = true;
    nodes.at(path[i+1]).added = true;
    nodes[path[i]].adj.insert(path[i+1]);    
  }
}

void DAG::removeLink(int u, int v)
{
  nodes.at(u).adj.erase(v);
}

void DAG::removeNode(int u)
{
  for (int i=0; i<nodes.size(); i++)
    if (i == u) {
      nodes[i].adj.clear();
      nodes[i].added = false;
    }
    else
      nodes[i].adj.erase(u);
}


void DAG::DAGelem::Print(ostream& os)
{
  os << "Adj: ";
  for (set<int>::iterator v=adj.begin(); v!=adj.end(); v++)
    os << *v << " ";
  os << "\t  max_dist_to_dest: " << max_dist_to_dest << endl;
}

void DAG::Print(ostream& os)
{
  if (acyclic)
    os << "The graph is acyclic" << endl;
  if (sortedDAG.empty()) {
    for (vector<DAGelem>::iterator u=nodes.begin(); u!=nodes.end(); u++)
      if (u->added) {
	os << distance(nodes.begin(), u) << ": ";
	u->Print(os);
      }
  } else {
    os << "In sorted order:" << endl;
    for (list<int>::iterator u=sortedDAG.begin(); u!=sortedDAG.end(); u++) {
	os << *u << ": pred(" << nodes[*u].pred << ") ";
	nodes[*u].Print(os);
      }
  }

  os << endl;
}

void DAG::DFS(int s)
{
  acyclic = true;
  sortedDAG.clear();
  for (vector<DAGelem>::iterator u=nodes.begin(); u!=nodes.end(); u++) {
    u->color = DAGelem::white;
    u->pred = numNodes;
  }
  time = 0;
  DFS_Visit(s);
}

void DAG::DFS_Visit(int u)
{
  nodes.at(u).color = DAGelem::gray;
  nodes.at(u).time_d = ++time;
  
  for (set<int>::iterator v=nodes[u].adj.begin(); v!=nodes[u].adj.end(); v++)
    if (nodes.at(*v).color == DAGelem::white) {
      nodes[*v].pred = u;
      DFS_Visit(*v);
    }
    else if (nodes.at(*v).color == DAGelem::gray)   // BACK EDGE !
      acyclic = false;
    
  nodes[u].color = DAGelem::black;
  nodes[u].time_f = ++time;

  if (sortedDAG.empty())
    nodes[u].max_dist_to_dest = 0;   // this is the last node in the topological ordering
  else {
    nodes[u].max_dist_to_dest = 0;
    for (set<int>::iterator v=nodes[u].adj.begin(); v!=nodes[u].adj.end(); v++)
      if (nodes[*v].max_dist_to_dest > nodes[u].max_dist_to_dest - 1)
	nodes[u].max_dist_to_dest = nodes[*v].max_dist_to_dest + 1;
  }
    
  sortedDAG.push_front(u);
}


int DAG::getMaxDistToDest(int u)
{
  return nodes.at(u).max_dist_to_dest;
}


int DAG::getMaxDistFromSource(int n)
{
  if (sortedDAG.empty()) {
    cerr << "DAG not sorted! getMaxDistromSource quits" << endl;
    return -1;
  }
  
  // reset pred and time_d (used as distance)
  for (vector<DAGelem>::iterator u=nodes.begin(); u!=nodes.end(); u++) {
    u->time_d = numeric_limits<int>::max();
    u->pred = numNodes;
  }
  
  // the source node is the first one in the topological order
  nodes[sortedDAG.front()].time_d = 0;

  for (list<int>::iterator u=sortedDAG.begin(); u!=sortedDAG.end(); u++)
    for (set<int>::iterator v=nodes[*u].adj.begin(); v!=nodes[*u].adj.end(); v++)
      // the weight of a link is -1
      if (nodes[*v].time_d > nodes[*u].time_d - 1) {
	nodes[*v].time_d = nodes[*u].time_d - 1;
	nodes[*v].pred = *u;
      }
      
  return -nodes.at(n).time_d;
}
