#include "WMtypes.h"

#include <set>
#include <algorithm>
#include <numeric>
#include <functional>
#include <cassert>

#define CHANNEL(l)              \
l->second->radio->channel

/* computes the maxflow on the initial graph (given by T)
 * and records the flow on each link
 */
//void maxflow(Topology *T);        /* prototype */

/* Compute the new Utot for each link in the neighborhood in case c is assigned to l */
vector<pair<double,TLinkItrs> > computeNewUtotVect(TLinkItrs l, unsigned int c, vector<TLinkItrs> neighborhood);


/* Binary predicate needed to find the link with the maximum Ftot */
inline bool max_flow(const TLinkItrs& x, const TLinkItrs& y) { return x.second->Utot < y.second->Utot; }

/* This function is invoked when we extract i<->j and want to optimize the radios of i (i.e., remove channel 1)*/
void OptimizeRadios(Topology *T, TLinkItrs l, int i, set<TLinkItrs> Q);


/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
void FCCA(Topology *T, bool Optimize)
{
//	Topology* T = new Topology(*Gi);     /* copy of the initial graph */

//	maxflow(T);
//	T->printTopology();

	/* INITIALIZATION */
	//for each node, one of the radios is assigned channel 1
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		u->info->radio[0].channel = 1; 

	// each node can reach its neighbors through that radio
	// (this implies that the channel of each link is set to 1)
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			v->radio = &u->info->radio[0];

	// compute Utot for each link (initially the collision domain equals the neighborhood)
	// for every undirected link u<->v
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			// consider all the links i<->j
			for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
				for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
					// if i<->j is in the neighborhood of u<->v
					if ( T->PotInterfLinks(TLinkItrs(i,j),TLinkItrs(u,v)) )
					// add the flow of i<->j
						v->Utot += j->flow / j->rate;

	// Set containing a pointer to all the links
	set<TLinkItrs> Q;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->id < v->info->id)    /* each link once */
				Q.insert(TLinkItrs(u,v));

	/* END INITIALIZATION */

	set<TLinkItrs>::iterator l;
	vector<TLinkItrs> neighborhood;

	while (!Q.empty())
	{
		// Extract from Q the link with the maximum Utot
		l = max_element(Q.begin(), Q.end(), max_flow);

		if (Optimize)
		{
			OptimizeRadios(T, *l, l->first->info->id, Q);
			OptimizeRadios(T, *l, l->second->info->id, Q);
		}

		set<unsigned int> uChannels, vChannels, S;
		S.clear();
		// Find the set of channels of the end nodes of the link to be extracted
		uChannels = l->first->info->findChannels();
		vChannels = l->second->info->findChannels();

		// Consider all the cases
		// if both end nodes have available radios
		if ( uChannels.size()<l->first->info->numAvailRadios() && vChannels.size()<l->second->info->numAvailRadios() )
			// put all the channels in S
			for (unsigned int i=1; i<=T->numAvailChannels; S.insert(i++));
		// if u has all radios assigned a channel
		else if ( uChannels.size()==l->first->info->numAvailRadios() && vChannels.size()<l->second->info->numAvailRadios() )
			// put the channels of u in S
			copy(uChannels.begin(), uChannels.end(), inserter(S, S.begin()));
		// if v has all radios assigned a channel
		else if ( uChannels.size()<l->first->info->numAvailRadios() && vChannels.size()==l->second->info->numAvailRadios() )
			// put the channels of v in S
			copy(vChannels.begin(), vChannels.end(), inserter(S, S.begin()));
		// if all the radios of both end nodes have been assigned a channel
		else
			// put the common channels of u and v in S
			set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

		// S is now the set of candidate channels to be assigned to the link to be extracted
		assert( !S.empty() );

		/* The links affected by assigning a channel to the extracted link are those in its neighborhood */
		neighborhood = T->computeNeighborhood(*l);

		vector<pair<double,TLinkItrs> > UtotVect, UtotVect_min;
		double Utot_max;
		unsigned int c_min;

		// for each channel c in S
		for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
		{
			// compute the new Utot for each link in the neighborhood in case c is assigned to l
			UtotVect = computeNewUtotVect(*l, *c, neighborhood);

			// sort UtotVect in decreasing order
			sort(UtotVect.begin(), UtotVect.end(), greater<pair<double,TLinkItrs> >() );

			// if this channel is the first one to be evaluated
			if (UtotVect_min.empty())
			{
				// store UtotVect as the vector with the minimum max
				UtotVect_min.resize(UtotVect.size());
				copy(UtotVect.begin(), UtotVect.end(), UtotVect_min.begin());
				c_min = *c;
				Utot_max = UtotVect.begin()->first;
			}
			// if this channel leads to a lower max
			else if (UtotVect.begin()->first < Utot_max)
			{
				// store UtotVect as the vector with the minimum max
				copy(UtotVect.begin(), UtotVect.end(), UtotVect_min.begin());
				c_min = *c;
				Utot_max = UtotVect.begin()->first;
			}
			// if this channel has the same max
			else if (UtotVect.begin()->first == Utot_max)
			{
				/* compare all the components of UtotVect and UtotVect_min */
				unsigned int v=0;
				bool found=false;
				while (v<UtotVect.size() && !found)
					if (UtotVect[v].first!=UtotVect_min[v].first)
						found=true;
					else
						v++;
				// store UtotVect if the nth max is lower than the nth max of UtotVect_min
				if (found && UtotVect[v].first<UtotVect_min[v].first)
				{
					copy(UtotVect.begin(), UtotVect.end(), UtotVect_min.begin());
					c_min = *c;
				}
			}
			// if this channel leads to a higher max do nothing
		}

		// c_min is the channel that minimizes the maximum Utot
		// assign c_min to the extracted link
		T->AssignChannel2Link(*l, c_min);
		T->AssignChannel2Link(T->FindInverseLink(*l), c_min);

		// update Utot for all the links in the neighborhood of the extracted link
		for (vector<pair<double,TLinkItrs> >::iterator i=UtotVect_min.begin(); i!=UtotVect_min.end(); i++)
			i->second.second->Utot = i->first;

		// Erase the extracted link from Q
		Q.erase(l);

	}

}



/* Compute the new Utot for each link in the neighborhood in case c is assigned to l */
vector<pair<double,TLinkItrs> > computeNewUtotVect(TLinkItrs l, unsigned int c, vector<TLinkItrs> neighborhood)
{
	vector<pair<double,TLinkItrs> > UtotVect;
	double Utot;

	/* the new Utot for l itself is given by its F/C ratio plus the F/C ratio of the links in the neighborhood using c */
	Utot = l.second->flow / l.second->rate;
	for (vector<TLinkItrs>::iterator l0=neighborhood.begin(); l0!=neighborhood.end(); l0++)
		if ( (*l0 != l) && (CHANNEL(l0) == c) )
			Utot += l0->second->flow / l0->second->rate;
	UtotVect.push_back(pair<double,TLinkItrs>(Utot,l));

	/* the new Utot for the other links is obtained from the current Utot value */
	for (vector<TLinkItrs>::iterator l0=neighborhood.begin(); l0!=neighborhood.end(); l0++)
		if (*l0 != l)
		{
			/* by subtracting the F/C ratio of l if l escapes the collision domain of l0 */
			if ((CHANNEL(l0) == CHANNEL((&l))) && (CHANNEL(l0) != c))
				Utot = l0->second->Utot - l.second->flow / l.second->rate;
			/* by adding the F/C ratio of l if l joins the collision domain of l0 */
			else if ((CHANNEL(l0) != CHANNEL((&l))) && (CHANNEL(l0) == c))
				Utot = l0->second->Utot + l.second->flow / l.second->rate;
			else
				Utot = l0->second->Utot;
			UtotVect.push_back(pair<double,TLinkItrs>(Utot,*l0));
		}

	return UtotVect;
}


/* This function is invoked when we extract link l and want to optimize the radios of i (i.e., remove channel 1)
   */
void OptimizeRadios(Topology *T, TLinkItrs l, int i, set<TLinkItrs> Q)
{
	set<TLinkItrs> L, extracted, nonextracted;
	set<pair<TLinkItrs,unsigned int> > tomodify;

	set<unsigned int> iChannels = T->node[i].info->findChannels();
// 	findChannels((T->node.begin()+i),iChannels);

	/* if i has available radios or no radio is assigned to 1, do nothing */
	if ( (iChannels.size()<T->node[i].info->numAvailRadios()) || (find(iChannels.begin(),iChannels.end(),(unsigned int)1)==iChannels.end()) )
		return;

	/* insert all the links leaving or entering i into L */
	for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
		for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
			if ( x->info->id == i || y->info->id == i )
				L.insert(TLinkItrs(x,y));

	L.erase(l);

	/* compute L-Q, i.e., the links that have already been extracted */
	set_difference(L.begin(), L.end(), Q.begin(), Q.end(), inserter(extracted, extracted.begin()));

	/* if an extracted link has been assigned channel 1, we can do nothing */
	for (set<TLinkItrs>::iterator l=extracted.begin(); l!=extracted.end(); l++)
		if (CHANNEL(l) == 1)
			return;

	/* compute L\cap Q, i.e., the links that have not been extracted yet */
	set_difference(L.begin(), L.end(), extracted.begin(), extracted.end(), inserter(nonextracted, nonextracted.begin()));

	/* Here we can make:  nonextracted.erase(FindInverseLink(l));
	   provided that we assign the new channel selected for l to the inverse link */

	/* if a nonextracted link has no channel but 1 in common with i, we can do nothing */
	for (set<TLinkItrs>::iterator l0=nonextracted.begin(); l0!=nonextracted.end(); l0++)
	{
		set<unsigned int> lChannels, S;
		S.clear();
		// Find the set of channels of the end nodes of the link to be extracted
// 		findChannels(l0->first, iChannels);
// 		findChannels(l0->second, lChannels);
		iChannels = l0->first->info->findChannels();
		lChannels = l0->second->info->findChannels();

		set_intersection(iChannels.begin(), iChannels.end(), lChannels.begin(), lChannels.end(), inserter(S, S.begin()));
		S.erase(1);
		if (!S.empty())
			tomodify.insert(pair<TLinkItrs,unsigned int>(*l0,*S.begin()));
		else
			return;
	}

	vector<pair<double,TLinkItrs> > UtotVect;

	// if we reach this point, it means that we can remove channel 1 from the radios of i
	/* for each link to be modified */
	for (set<pair<TLinkItrs,unsigned int> >::iterator l0=tomodify.begin(); l0!=tomodify.end(); l0++)
	{
		/* update UtotVect */
		UtotVect = computeNewUtotVect(l0->first, l0->second, T->computeNeighborhood(l0->first));
		for (vector<pair<double,TLinkItrs> >::iterator i=UtotVect.begin(); i!=UtotVect.end(); i++)
			i->second.second->Utot = i->first;

		/* assign the new channel to l */
		T->AssignChannel2Link(l0->first,l0->second);
		//T->AssignChannel2Link(T->FindInverseLink(l0->first),l0->second);
	}

	/* set to channel 0 the radio of i that is set to channel 1, i.e.,make it available */
// 	pair<bool, unsigned int> foundCh = SearchChannel(T->node[i].info, 1);
	RadioIfaceIterator ri = T->node[i].info->findRadioWithChannel (1);

// 	assert( foundCh.first );
	assert ( ri != T->node[i].info->radio.end() );

// 	T->node[i].info->radio[foundCh.second].channel = 0;
	ri->channel = 0;

}





