#include "CAalgos.h"
#include "WMutils.h"

#include <math.h>
#include <iostream>
#include <algorithm>
#include <numeric>
#include <functional>
#include <cassert>

#define CHANNEL(l)              \
l.second->radio->channel


// These functions are defined in fcra.cpp

/* Compute the new Utot for every link involved (except l itself) in case c is assigned to l */
void fcra_computeNewUtot(Topology *T, TLinkItrs l, unsigned int c);

/* Binary predicate needed to find the link with the maximum Ftot */
//inline bool max_flow(const pair<int,int>& x, const pair<int,int>& y) { return x.second < y.second; }

/* Binary function needed to comput the sum of the rates for all links u->v */
// inline float sumRate(float result, TGraphNeighbor i) { return result + i.rate; }

/* This function is invoked when we extract i<->j and want to optimize the radios of i (i.e., remove channel 1)*/
/* Returns true if channel 1 is removed */
bool fcra_OptimizeRadios(Topology *T, TLinkItrs l, int i, set<TLinkItrs> Q, bool turbo);


/* constructor */
FCPRA::FCPRA (const list<pair<string,string> >& params)
 : m_optimize (true),
   m_turbo (true),
   m_ratepoweradapt (false)
{
  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("optimize", m_optimize);
    RP.read ("turbo", m_turbo);
    RP.read ("rate_power_adapt", m_ratepoweradapt);

    if (RP.notFound())
    {
      cerr << endl << "Ignoring unknown attribute (" << p->first << ") for the FCPRA algorithm" << endl << endl;
    }
  }
}


const char* FCPRA::name()
{
  stringstream ss;
  string s;

  ss << "F" << (m_ratepoweradapt ? "T" : "F");

  ss >> s;

  return s.c_str();
}


/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
double FCPRA::Run (Topology *T)
{
	/* INITIALIZATION */
	T->ResetChannels();

	//for each node, one of the radios is assigned channel 1
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		u->info->radio[0].channel = 1; 

	// each node can reach its neighbors through that radio
	// (this implies that the channel of each link is set to 1)
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			v->radio = &u->info->radio[0];

	// compute Utot for each link (initially the collision domain equals the neighborhood)
	// for every directed link u->v
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			v->Utot = T->computeUtot(TLinkItrs(u,v));

	// Set containing a pointer to all the directed links
	set<TLinkItrs> Q;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				Q.insert(TLinkItrs(u,v));

	/* END INITIALIZATION */

	set<TLinkItrs>::iterator ptr_l;
	TLinkItrs l;
	double max_Utot;
	bool ch1removed = false;
	unsigned int l_channel;
	TLinkItrs invLink;

	while (!Q.empty())
	{
		// Extract the link with the maximum Utot from Q
		max_Utot = -1.;
		for (set<TLinkItrs>::iterator p=Q.begin(); p!=Q.end(); p++)
			if (p->second->Utot > max_Utot)
			{
				max_Utot = p->second->Utot;
				ptr_l = p;
			}

		// Erase the extracted link from Q
		l = *ptr_l;
		Q.erase(ptr_l);
		l_channel = CHANNEL(l);
		if (m_optimize)
			ch1removed = fcra_OptimizeRadios(T, l, l.first->info->id, Q, m_turbo) || fcra_OptimizeRadios(T, l, l.second->info->id, Q, m_turbo);

		/* if the end nodes only share channel 1 and the optimization step removes channel 1 from one of the nodes
		   remove the inverse link (by assigning channel 0) and updates the Utot of the links involved */
		if (ch1removed)
		{
			invLink = T->FindInverseLink(l);
			RadioIface *fakeradio = new RadioIface();
			invLink.second->radio = fakeradio;
			invLink.second->radio->channel = (unsigned int)1;
			fcra_computeNewUtot(T, invLink, 0);
			invLink.second->radio->channel = 0;
		}

		set<unsigned int> uChannels, vChannels, S, Sc;
		S.clear();
		Sc.clear();
		// Find the set of channels of the end nodes of the link to be extracted
		uChannels = l.first->info->findChannels();
		vChannels = l.second->info->findChannels();

		// Consider all the cases
		// if both end nodes have available radios
		if ( uChannels.size()<l.first->info->numAvailRadios() && vChannels.size()<l.second->info->numAvailRadios() )
			// put all the channels in S
			for (unsigned int i=1; i<=T->numAvailChannels; S.insert(i++));
		// if u has all radios assigned a channel
		else if ( uChannels.size()==l.first->info->numAvailRadios() && vChannels.size()<l.second->info->numAvailRadios() )
			// put the channels of u in S
			copy(uChannels.begin(), uChannels.end(), inserter(S, S.begin()));
		// if v has all radios assigned a channel
		else if ( uChannels.size()<l.first->info->numAvailRadios() && vChannels.size()==l.second->info->numAvailRadios() )
			// put the channels of v in S
			copy(vChannels.begin(), vChannels.end(), inserter(S, S.begin()));
		// if all the radios of both end nodes have been assigned a channel
		else
			// put the common channels of u and v in S
			set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

		// S is now the set of candidate channels to be assigned to the link to be extracted
		assert( !S.empty() );

		// put the common channels of u and v in Sc
		set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(Sc, Sc.begin()));

		map<unsigned int, double> Umax, Umax2;
		map<unsigned int, set<TLinkItrs> > D1;

		// for each channel c in S
		for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
			Umax[*c] = (Umax2[*c] = 0.);

		// compute Umax[c] for all c in S
		// for every directed link x->y
		for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
			for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
				// whose potential collision domain includes the extracted link and having a channel in S
				if ((TLinkItrs(x,y) != l) && (T->PotInterfLinks(l, TLinkItrs(x,y))) && (S.find(y->radio->channel)!=S.end()))
				{
					/* remove the contribution of the extracted link from the Utot of the links having the same channel */
					if (y->radio->channel == l_channel)
						y->Utot -= l.second->flow/l.second->rate;
					if (y->Utot > Umax[y->radio->channel])
						Umax[y->radio->channel] = y->Utot;
					/* add the link to D1 */
					D1[y->radio->channel].insert(TLinkItrs(x,y));
				}

		map<unsigned int, float> rate;
		map<unsigned int, unsigned int> power;
		int m, mmin, w, wmin;
		float save_rate;
		unsigned int save_power;
		double min,save;

		/* we use a fake radio to compute Utot(u->v) for every channel */
		RadioIface *fakeradio = new RadioIface();
		l.second->radio = fakeradio;

		/* we need to save the power and rate of the original link, in order to restore them before evaluating a new channel */
		save_rate = l.second->rate;
		save_power = l.second->txpower;
		
		// for each channel c in S
		for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
		{
			// we overwrite the channel and the rate of the extracted link to compute its Utot correctly
			CHANNEL(l) = *c;
			l.second->rate = save_rate;
			l.second->txpower = save_power;

// 			for (w=0; w<NUMTXPWR && availTxPwr[w]!=l.second->power; w++);
// 			assert(w<NUMTXPWR);
			w = save_power;

			wmin = w;
			min = T->computeUtot(l) - l.second->flow/l.second->rate;

			while (m_ratepoweradapt  &&  w<numTxPowerLevels-1  &&  min>Umax[*c]  &&  Umax2[*c]<Umax[*c] )
			{
				l.second->txpower = ++w;

				// compute Umax2[c]
				// for every directed link x->y
				for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
					for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
						// not in D1 and whose potential collision domain includes the extracted link
						if ( (TLinkItrs(x,y) != l) && (T->PotInterfLinks(l, TLinkItrs(x,y))) 
						  && (D1[*c].find(TLinkItrs(x,y)) == D1[*c].end()) && (y->radio->channel == *c) )
						{cout << "INCREDIBILE!!" << endl;
							if (y->Utot > Umax2[*c])
								Umax2[*c] = y->Utot;
						}

				if ( (save = T->computeUtot(l) - l.second->flow/l.second->rate) < min  && Umax2[*c] < Umax[*c] )
				{
					min = save;
					wmin = w;
				}
			}
			// this power will be used in case a link u-> v is established on this channel
			power[*c] = (l.second->txpower = wmin);
			
if (save_power != wmin) cout << endl << "** Potenza variata sul link " << l.first->info->id << "->" << l.second->info->id << " ** da "  << TxPowerLevel(save_power) << " a " << TxPowerLevel(wmin) << endl;
			for (m=0; m<NUMRATES && availTXrates[m]!=l.second->rate; m++);
			assert(m<NUMRATES);

			mmin = m;

			while (m_ratepoweradapt && (m>0) && (min>Umax[*c]))
			{
				l.second->rate = availTXrates[--m];
				if ( (save = T->computeUtot(l) - l.second->flow/l.second->rate) < min)
				{
					min = save;
					mmin = m;
				}
			}
			// this rate will be used in case a link u-> v is established on this channel
			rate[*c] = availTXrates[mmin];
if (save_rate != availTXrates[mmin]) cout << endl << "** Rate variato sul link " << l.first->info->id << "->" << l.second->info->id << " ** da "  << save_rate << " a " << availTXrates[mmin] << endl;
			if (min > Umax[*c])
				Umax[*c] = min;
		}

		delete fakeradio;

		set<unsigned int> C;
		set_difference(S.begin(), S.end(), Sc.begin(), Sc.end(), inserter(C, C.begin()));

		/* leave in Umax only the channels in Sc and the channel in S-Sc that minimizes Umax */
		if (!C.empty())
		{
			set<unsigned int>::iterator c = C.begin();
			min = Umax[*c];
			for (set<unsigned int>::iterator c0=++C.begin(); c0!=C.end(); c0++)
				if (Umax[*c0]<min)
				{
					Umax.erase(*c);
					c = c0;
					min = Umax[*c0];
				} 
				else
					Umax.erase(*c0);
		}

		double F = l.second->flow;
		TGraphNeighbor *newlink;
		list<TGraphNeighbor>::iterator currlink;

		MeshNode* backup_info = l.second->info;
		double backup_flow = l.second->flow;
		/* save the list of end-to-end flows crossing the extracted link */
		map<pair<int,int>,double> backup_e2e =l.second->f_e2e;
		/* remove v from among u's neighbors, i.e., remove the extracted link */
		l.first->neighbor.erase(l.second);

		unsigned int c;
		float sumrates;
		double base, I;
		bool done = false;


		do
		{
			/* add a new link u-> v on the channel that minimizes Umax */
			newlink = new TGraphNeighbor(backup_info);

			l.first->neighbor.push_back(*newlink);

			currlink = --l.first->neighbor.end();

			c = min_element(Umax.begin(), Umax.end(), compareSecond<unsigned int, double>())->first;

			T->AssignChannel2Link( TLinkItrs(l.first,currlink), c );
			currlink->rate = rate[c];               /* set the previously computed rate */
			currlink->txpower = power[c];             /* set the previously computed power */
			currlink->f_e2e = backup_e2e;		/* restore the set of end-to-end flows and reset the values*/
			for (map<pair<int,int>,double>::iterator mp=currlink->f_e2e.begin(); mp!=currlink->f_e2e.end(); mp++)
			  mp->second = 0.;

			sumrates = 0.;
			for (list<TGraphNeighbor>::iterator v=l.first->neighbor.begin(); v!=l.first->neighbor.end(); v++)
				if (v->info->id == currlink->info->id)
					sumrates+=v->rate;

			base = Umax[c];
			Umax.erase(c);

			if ( (!Umax.empty()) && (F>=(min_element(Umax.begin(),Umax.end(),compareSecond<unsigned int, double>())->second - base)*sumrates) )
				I = min_element(Umax.begin(),Umax.end(),compareSecond<unsigned int, double>())->second - base;
			else
			{
				I = F / sumrates;
				done = true;
			}

			for (list<TGraphNeighbor>::iterator k=l.first->neighbor.begin(); k!=l.first->neighbor.end(); k++)
				if (k->info->id == currlink->info->id)
				{
					k->flow += I * k->rate;
					for (map<pair<int,int>,double>::iterator mp=k->f_e2e.begin(); mp!=k->f_e2e.end(); mp++)
					  mp->second += (I * k->rate) * backup_e2e[mp->first] / backup_flow;
					F -= I * k->rate;
				}

		} while (!done);

		if (ch1removed)
		// assign the same channel as the extracted link to the inverse link
		{
			fcra_computeNewUtot(T, invLink, c);
			T->AssignChannel2Link( invLink, c );
			invLink.second->Utot = T->computeUtot(invLink);
		}

		// compute the Utot for the newly established links
		for (list<TGraphNeighbor>::iterator k=l.first->neighbor.begin(); k!=l.first->neighbor.end(); k++)
			if (k->info->id == currlink->info->id)
				k->Utot = T->computeUtot(TLinkItrs(l.first,k));

		// Update the Utot for all the other links involved
		// for every directed link x->y
		for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
			for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
				// for every newly establisthed link u->v
				for (list<TGraphNeighbor>::iterator k=l.first->neighbor.begin(); k!=l.first->neighbor.end(); k++)
					if ( (k->info->id == currlink->info->id) && ( !ch1removed || TLinkItrs(x,y) != invLink ) &&
					// such that the potential collision domain of x->y includes the new link and have the same channel
					     (x!=l.first || y!=k) && (T->PotInterfLinks(TLinkItrs(l.first,k), TLinkItrs(x,y))) && (y->radio->channel==k->radio->channel) )
						/* add the contribution of the new link to the Utot of x->y */
						y->Utot += k->flow/k->rate;
	}
	return 0;
}

// These functions are defined in fcra.cpp
#if 0
/* Compute the Utot of the collision domain of a link */
double computeUtot(Topology *T, TLinkItrs l)
{
	double Utot = 0.;
	// consider all the links i->j
	for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
		for (vector<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
			// if i->j potentially interferes with l and has the same channel
			if ((T->PotInterfLinks(TLinkItrs(i,j), l)) && (j->radio->channel==CHANNEL(l)))
			// add the flow of i->j
				Utot += j->flow / j->rate;
	return Utot;
}


/* Compute the new Utot for every link involved (except l itself) in case c is assigned to l */
void fcra_computeNewUtot(Topology *T, TLinkItrs l, unsigned int c)
{
	/* the new Utot for the other links is obtained from the current Utot value */
	// for every directed link x->y
	for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
		for (vector<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
			// such that the potential collision domain of x->y includes l
			if ( (l!=TLinkItrs(x,y)) && (T->PotInterfLinks(l,TLinkItrs(x,y))) )
			{
				/* by subtracting the F/C ratio of l if l escapes the collision domain of l0 */
				if ((y->radio->channel == CHANNEL(l)) && (y->radio->channel != c))
					y->Utot -= l.second->flow / l.second->rate;
				/* by adding the F/C ratio of l if l joins the collision domain of l0 */
				else if ((y->radio->channel != CHANNEL(l)) && (y->radio->channel == c))
					y->Utot += l.second->flow / l.second->rate;
			}
}


/* This function is invoked when we extract link l and want to optimize the radios of i (i.e., remove channel 1)
   */
bool fcra_OptimizeRadios(Topology *T, TLinkItrs l, int i, set<TLinkIds> Q, bool turbo)
{
	bool invremoved = false;
	set<TLinkIds> L, extracted, nonextracted;
	set<pair<TLinkItrs,unsigned int> > tomodify;

	set<unsigned int> iChannels, lChannels, S;
	findChannels((T->node.begin()+i),iChannels);

	/* if i has available radios or no radio is assigned to 1, do nothing */
	if ( (iChannels.size()<T->node[i].info->numAvailRadios()) || (find(iChannels.begin(),iChannels.end(),(unsigned int)1)==iChannels.end()) )
		return false;

	/* insert all the links leaving or entering i into L */
	for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
		for (vector<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
			if ( x->info->id == i || y->info->id == i )
				L.insert(TLinkIds(x->info->id,y->info->id));

	L.erase(TLinkIds(l.first->info->id,l.second->info->id));

	/* compute L-Q, i.e., the links that have already been extracted */
	set_difference(L.begin(), L.end(), Q.begin(), Q.end(), inserter(extracted, extracted.begin()));

	/* if an extracted link has been assigned channel 1, we can do nothing */
	for (set<TLinkIds>::iterator l0=extracted.begin(); l0!=extracted.end(); l0++)
		if (CHANNEL(T->Link_IdsToItrs(*l0)) == 1)
			return false;

	/* compute L\cap Q, i.e., the links that have not been extracted yet */
	set_difference(L.begin(), L.end(), extracted.begin(), extracted.end(), inserter(nonextracted, nonextracted.begin()));


	if (turbo)
	{
		bool inverse = ( find(nonextracted.begin(), nonextracted.end(), TLinkIds(l.second->info->id,l.first->info->id)) != nonextracted.end() );
		if (inverse)
		{
			findChannels((T->node.begin()+l.first->info->id),iChannels);
			findChannels((T->node.begin()+l.second->info->id),lChannels);
			S.clear();
			set_intersection(iChannels.begin(), iChannels.end(), lChannels.begin(), lChannels.end(), inserter(S, S.begin()));
			S.erase(1);
			if (S.empty())
			{
				nonextracted.erase(TLinkIds(l.second->info->id,l.first->info->id));
				invremoved = true;
			}
		}
	}

	/* if a nonextracted link has no channel but 1 in common with i, we can do nothing */
	for (set<TLinkIds>::iterator l0=nonextracted.begin(); l0!=nonextracted.end(); l0++)
	{
		S.clear();
		// Find the set of channels of the end nodes of the link to be extracted
		findChannels((T->node.begin()+l0->first), iChannels);
		findChannels((T->node.begin()+l0->second), lChannels);

		set_intersection(iChannels.begin(), iChannels.end(), lChannels.begin(), lChannels.end(), inserter(S, S.begin()));
		S.erase(1);
		if (!S.empty())
			tomodify.insert(pair<TLinkItrs,unsigned int>(T->Link_IdsToItrs(*l0),*S.begin()));
		else
			return false;
	}

	// if we reach this point, it means that we can remove channel 1 from the radios of i
	/* for each link to be modified */
	for (set<pair<TLinkItrs,unsigned int> >::iterator l0=tomodify.begin(); l0!=tomodify.end(); l0++)
	{
		if (l0->first.second->radio->channel != l0->second)
		{
			/* update Utot for neighbors*/
			fcra_computeNewUtot(T, l0->first, l0->second);
			/* assign the new channel to l0 */
			T->AssignChannel2Link(l0->first,l0->second);
			/* compute Utot for the modified link*/
			l0->first.second->Utot = computeUtot(T,l0->first);
		}
	}

	/* set to channel 0 the radio of i that is set to channel 1, i.e.,make it available */
	pair<bool, unsigned int> foundCh = SearchChannel(T->node[i].info, 1);
	assert( foundCh.first );
	T->node[i].info->radio[foundCh.second].channel = 0;
	return invremoved;
}



#endif
