#include "WMtypes.h"
#include "linprogs.h"
#include <iostream>
#include <cassert>
#include <sstream>
#include <algorithm>
#include <numeric>
#include <fcntl.h>
// #include <glpk.h>
#include <CbcBranchDynamic.hpp>
#include <ClpSimplex.hpp>
#include <CoinBuild.hpp>



/*
 * Rearrange e2e flows to minimizes the total power consumption subject
 * to constraints on node power and maximum total utilization.
 * Returns the result of the LP (OPTIMAL, INFEASIBLE, etc.)
 */
int TurnOffRadio(Topology* T, map<pair<int,int>,double>* demands, const TRadio& r0, double u0, double& maxUtot, double& Ptot, vector<float> *nodePowerConstr)
{
//   lprec *lp;
  ClpSimplex model;
  int Ncol;
  float powerStandbyNodes = 0.;

  map<TRadio, set<TLinkItrs> > Lri, Lro;  // maps each radio to the set of incoming/outgoing links using it

  /* compute Lri and Lro for all the radios */
  for (unsigned int u=0; u<T->node.size(); u++)
    for (list<TGraphNeighbor>::iterator v=T->node[u].neighbor.begin(); v!=T->node[u].neighbor.end(); v++)
    {
      /* insert u->v in the set Lro of the radio on u */
      Lro[TRadio(u,distance(T->node[u].info->radio.begin(),RadioIfaceIterator(v->radio)))].insert(TLinkItrs(T->node.begin()+u,v));
      /* insert u->v in the set Lri of the radio on v */
//       pair<bool,unsigned int> ch = SearchChannel(v->info, v->radio->channel);
      RadioIfaceIterator ri = v->info->findRadioWithChannel(v->radio->channel);
      if (ri != v->info->radio.end())
	Lri[TRadio(v->info->id, distance(v->info->radio.begin(),ri))].insert(TLinkItrs(T->node.begin()+u,v));
    }

  cout << "TurnOffRadio: Outgoing links from r0: ";
  if (Lro.find(r0) != Lro.end())
    for (set<TLinkItrs>::iterator l=Lro[r0].begin(); l!=Lro[r0].end(); l++)
      cout << "(" << l->first->info->id << "," << l->second->info->id << ")[" << l->second->radio->channel << "] ";
  cout << endl << "TurnOffRadio: Incoming links to r0: ";
  if (Lri.find(r0) != Lri.end())
    for (set<TLinkItrs>::iterator l=Lri[r0].begin(); l!=Lri[r0].end(); l++)
      cout << "(" << l->first->info->id << "," << l->second->info->id << ")[" << l->second->radio->channel << "] ";
  cout << endl;
  
  
  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands->size();
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->node.size() + 1];

  cumNumNeighbs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();

  // the number of variables is numDemands*numLinks (f^ab(u->v))
  Ncol = numDemands*cumNumNeighbs[T->node.size()];

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model.resize(0, Ncol);
  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
//       int i=1;
      int i=0;
      for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++, i++) {
	stringstream s;
	s << "f(" << u->info->id << "," << v->info->id << ")[" << d->first.first << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+i, str);
      }
      linkId++;
    }

  int* colno = new int[Ncol+1];
  double *coeff = new double[Ncol+1], *coeff2 = new double[Ncol+1];
  
  vector<vector<double> > coeff_tau;
  vector<vector<int> > colno_tau;

  vector<vector<double> > coeff_power;
  vector<vector<int> > colno_power;
  vector<double> lhs_power;

  int k, pos_i;


  /* With one iteration over all the radios, we determine the coefficients used for the objective
   * function, the constraints tau_TX(r) + tau_RX(r) <= 1 (for each active radio) and node power constraints */
  for (unsigned int u=0; u<T->node.size(); u++)
    if (T->node[u].info->countRadiosWithChannel(0) == T->node[u].info->numAvailRadios())
//       numStandbyNodes++;
      powerStandbyNodes += T->node[u].info->StandbyPowerConsumption;
    else if (u == r0.first && T->node[u].info->countRadiosWithChannel(0) == T->node[u].info->numAvailRadios()-1 && T->node[u].info->radio[r0.second].channel)
//       numStandbyNodes++;
      powerStandbyNodes += T->node[u].info->StandbyPowerConsumption;
    else
    {
      k = 0;

      /* for each active radio */
      for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	if (T->node[u].info->radio[i].channel && TRadio(u,i) != r0)
	{
	  int j1=k;
	  RadioIfaceIterator ri = T->node[u].info->radio.begin() + i;

	  if (Lro.find(TRadio(u,i)) != Lro.end())
	    for (set<TLinkItrs>::iterator l=Lro[TRadio(u,i)].begin(); l!=Lro[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->rate);
	      // coeff2 is used for the power consumption constraint
	      fill_n(coeff2+k, numDemands, ((*ri->rho_tx)(DbmToW(TxPowerLevel(l->second->txpower))) - ri->rho_idle) / l->second->rate);
	      pos_i = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, counter<int>(pos_i));
	      generate_n(colno+k, numDemands, counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  int j2=k;

	  if (Lri.find(TRadio(u,i)) != Lri.end())
	    for (set<TLinkItrs>::iterator l=Lri[TRadio(u,i)].begin(); l!=Lri[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->rate);
	      fill_n(coeff2+k, numDemands, (ri->rho_rx - ri->rho_idle) / l->second->rate);
	      pos_i = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, counter<int>(pos_i));
	      generate_n(colno+k, numDemands, counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  if (k != j1)
	  {
	    /* make a copy of coeff and colno so that later we can add the tau constraint */
	    coeff_tau.push_back(vector<double>(coeff+j1,coeff+k));
	    colno_tau.push_back(vector<int>(colno+j1,colno+k));
	  }

	  /* multiply coefficients by the power consumption (needed for node power constraint) */
// 	  for (int j=j1; j<j2; j++)
// 	    coeff[j] *= (T->pm.rho_tx-T->pm.rho_idle);
// 
// 	  for (int j=j2; j<k; j++)
// 	    coeff[j] *= (T->pm.rho_rx-T->pm.rho_idle);
	}

      /* stores the coefficients for the power constraint on node u */
      if (k)
      {
	pos_i = T->node[u].info->countRadiosWithChannel(0);
	if (u == r0.first && T->node[u].info->radio[r0.second].channel)
	  pos_i++;

	coeff_power.push_back(vector<double>(coeff2,coeff2+k));
	colno_power.push_back(vector<int>(colno,colno+k));
	
	float lhs = T->node[u].info->BasePowerConsumption;

	for (RadioIfaceIterator ri = T->node[u].info->radio.begin(); ri != T->node[u].info->radio.end(); ri++)
	  lhs += (ri->channel ? ri->rho_idle : ri->rho_sleep);

// 	lhs_power.push_back(T->pm.BasePowerConsumption + pos_i*T->pm.rho_sleep + (T->node[u].info->numAvailRadios() - pos_i)*T->pm.rho_idle);
	lhs_power.push_back (lhs);
      }
    }

//   for (int i=1; i<=Ncol; i++)
  for (int i=0; i<Ncol; i++)
    coeff[i] = 0.;               // reset the elements of coeff

  for (unsigned int i=0; i<coeff_power.size(); i++)
    for (unsigned int j=0; j<coeff_power.at(i).size(); j++)
      coeff[colno_power.at(i).at(j)] += coeff_power.at(i).at(j);
    
  /* set the objective function: min sum_u\in V P(u) */
//   set_obj_fn(lp, coeff);
  
  for (int i=0; i<Ncol; i++)
    model.setObjectiveCoefficient(i, coeff[i]);

  /* set the object direction to minimize */
//   set_minim(lp);
  model.setOptimizationDirection(1);

  
//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  CoinBuild buildObject;
  
  /* constraints on the total utilizations ( |E| constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      k = 0;
      for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	  if ( TLinkItrs(x,y).IsInTheCollDomainOf(T,TLinkItrs(u,v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i;
	    k += numDemands;
	  }
//       add_constraintex(lp, k, coeff, colno, LE, max(maxUtot, u0) );
      buildObject.addRow(k, colno, coeff, max(maxUtot, u0), max(maxUtot, u0));
    }
    
  /* no flow on the links using radio r0 */
  if (Lro.find(r0) != Lro.end())
    for (set<TLinkItrs>::iterator l=Lro[r0].begin(); l!=Lro[r0].end(); l++)
//       for (unsigned int d=1; d<=numDemands; d++)
      for (unsigned int d=0; d<numDemands; d++)
// 	set_upbo(lp, (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands + d, 0.);
	model.setColumnUpper((cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands + d, 0.);

  if (Lri.find(r0) != Lri.end())
    for (set<TLinkItrs>::iterator l=Lri[r0].begin(); l!=Lri[r0].end(); l++)
//       for (unsigned int d=1; d<=numDemands; d++)
      for (unsigned int d=0; d<numDemands; d++)
// 	set_upbo(lp, (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands + d, 0.);
	model.setColUpper((cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands + d, 0.);


//   for (int i=1; i<=Ncol; i++)
  for (int i=0; i<Ncol; i++)
  {
    coeff[i] = 0.;               // reset the elements of coeff
    colno[i] = i;
  }

  /* constraint on the total power consumption */
  double lhs = 0.;
  for (unsigned int i=0; i<coeff_power.size(); i++)
  {
    for (unsigned int j=0; j<coeff_power.at(i).size(); j++)
      coeff[colno_power.at(i).at(j)] += coeff_power.at(i).at(j);
    lhs += lhs_power.at(i);
  }

//   add_constraint(lp, coeff, LE, Ptot - lhs - numStandbyNodes*T->pm.StandbyPowerConsumption);
  buildObject.addRow(Ncol, colno, coeff, -COIN_DBL_MAX, Ptot - lhs - powerStandbyNodes);

  
  /* tau constraints */
  for (unsigned int r=0; r<coeff_tau.size(); r++)
  {
    copy (coeff_tau.at(r).begin(), coeff_tau.at(r).end(), coeff);
    copy (colno_tau.at(r).begin(), colno_tau.at(r).end(), colno);

//     add_constraintex(lp, coeff_tau.at(r).size(), coeff, colno, LE, 1);
    buildObject.addRow(coeff_tau.at(r).size(), colno, coeff, -COIN_DBL_MAX, 1.);
  }

  /* node power constraints */
  for (unsigned int i=0; i<coeff_power.size(); i++)
  {
    copy (coeff_power.at(i).begin(), coeff_power.at(i).end(), coeff);
    copy (colno_power.at(i).begin(), colno_power.at(i).end(), colno);

//     add_constraintex(lp, coeff_power.at(i).size(), coeff, colno, LE, nodePowerConstr->at(i) - lhs_power.at(i));
    buildObject.addRow(coeff_power.at(i).size(), colno, coeff, -COIN_DBL_MAX, nodePowerConstr->at(i) - lhs_power.at(i));
  }

  
  vector<vector<int> > colno_v(T->node.size()*numDemands);
  vector<vector<double> > coeff_v(T->node.size()*numDemands);

  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands->begin(); di!=demands->end(); di++, d++) {
      // for (unsigned int d=0; d<numDemands; d++) {
	dist = distance(u->neighbor.begin(), v);
	/* flow leaving node u */
	coeff_v[u->info->id*numDemands+d].push_back(1);
// 	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d);
	/* flow entering node v */
	coeff_v[v->info->id*numDemands+d].push_back(-1);
// 	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d);
	/* if v is the source of the demand or u is the destination of the demand, f(u->v)[d] must be null */
	if ( (u->info->id == di->first.second) || (v->info->id == di->first.first) )
// 	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, 0.);
	  model.setColumnUpper((cumNumNeighbs[u->info->id] + dist)*numDemands + d, 0.);
//	else
//	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, di->second);
      }
    }

  /* Add the flow conservation constraints one-by-one */
  k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      /* if i is the source of the demand d */
      if (i == d->first.first)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
      /* if i is the destination of the demand d */
      else if (i == d->first.second)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
      else
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
      k++;
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because of the
   * constraints on the total utilizations
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model.addRows(buildObject);

  // write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /* Now let lpsolve calculate a solution */
//   int ret = solve(lp);
  int ret = model.initialSolve();

//   if (ret != OPTIMAL)
  if (ret)
    cerr << endl << "WARNING: TurnOffRadio: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
  else
  {
//     print_solution(lp,4);

  /* If an optimal solution has been found, then update the flows, the e2e flows and the Utot */

    cout << endl << "TurnOffRadio: the LP was able to turn off the radio!" << endl;
      
    double *ptr_var, *ptr_constr;
    
//     get_ptr_variables(lp, &ptr_var);
//     get_ptr_constraints(lp, &ptr_constr);

    ptr_var = model.primalColumnSolution();
    ptr_constr = model.primalRowSolution();

    maxUtot = 0.;

    unsigned int linkId = 0;
  
    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      {
	v->flow = 0.;
	v->f_e2e.clear();

	int i=0;
	for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++, i++)
	  if (ptr_var[linkId*numDemands + i] > 0.)
	    v->flow += (v->f_e2e[d->first] = ptr_var[linkId*numDemands + i]);

	maxUtot = max(maxUtot, (v->Utot = ptr_constr[linkId]));

	linkId++;
      }

//     Ptot = ptr_constr[linkId] + lhs + numStandbyNodes*T->pm.StandbyPowerConsumption;
    Ptot = ptr_constr[linkId] + lhs + powerStandbyNodes;
  }
  
  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;

  /* clean up such that all used memory by lpsolve is freed */
//   if (lp)
//     delete_lp(lp);

  return ret;
}


/*
 * Rearrange e2e flows to minimizes the total power consumption subject
 * to constraints on node power and maximum total utilization.
 * Returns the result of the LP (OPTIMAL, INFEASIBLE, etc.)
 */
int MinPower_eps(Topology* T, map<pair<int,int>,double>* demands, double u0, vector<float> *nodePowerConstr, const vector<ArcEpsY>& epsy)
{
//   lprec *lp;
  ClpSimplex model;

  map<TRadio, set<TLinkItrs> > Lri, Lro;  // maps each radio to the set of incoming/outgoing links using it

  /* compute Lri and Lro for all the radios */
  for (unsigned int u=0; u<T->node.size(); u++)
    for (list<TGraphNeighbor>::iterator v=T->node[u].neighbor.begin(); v!=T->node[u].neighbor.end(); v++)
    {
      /* insert u->v in the set Lro of the radio on u */
      Lro[TRadio(u,distance(T->node[u].info->radio.begin(),RadioIfaceIterator(v->radio)))].insert(TLinkItrs(T->node.begin()+u,v));
      /* insert u->v in the set Lri of the radio on v */
//       pair<bool,unsigned int> ch = SearchChannel(v->info, v->radio->channel);
      RadioIfaceIterator ri = v->info->findRadioWithChannel(v->radio->channel);
//       if (ch.first)
      if (ri != v->info->radio.end())
	Lri[TRadio(v->info->id, distance(v->info->radio.begin(),ri))].insert(TLinkItrs(T->node.begin()+u,v));
    }

  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands->size();
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->node.size() + 1];

  // compute the cumulative number of arcs of type node or radio for each node. It is such that:
  // - the number of arcs of node i is cumNumArcs[i+1] - cumNumArcs[i]
  // - the last element of cumNumArcs is the total number of arcs of type node or radio
  unsigned int cumNumArcs[T->node.size() + 1];

  cumNumNeighbs[0] = 0;
  cumNumArcs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  {
    cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();
    
    int radiosOn = u->info->numAvailRadios() - u->info->countRadiosWithChannel(0);
    cumNumArcs[u->info->id + 1] = cumNumArcs[u->info->id] + (radiosOn ? 2 + 2*radiosOn : 0);
  }

  // the number of variables is numDemands*numLinks (f^ab(u->v))
  int Ncol = numDemands*(cumNumNeighbs[T->node.size()] + cumNumArcs[T->node.size()]);

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model.resize(0, Ncol);

  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD  f_u0->u0*,d1 ... f_u0->u0*,dD f_u0*->u0,d1 ... f_u0*->u0,dD f_u0*->r^(u0,c1),d1 ... f_u0*->r^(u0,c1),dD ...
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the link variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
//       int i=1;
      int i=0;
      for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++, i++) {
	stringstream s;
	s << "f(" << u->info->id << "," << v->info->id << ")[" << d->first.first << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+i, str);
      }
      linkId++;
    }

  /* give a name to the arc variables */
  for (unsigned int u=0; u<T->node.size(); u++)
    if (T->node[u].info->countRadiosWithChannel(0) != T->node[u].info->numAvailRadios())
    {
//       int d=1;
      int d=0;
      for (map<pair<int,int>,double>::iterator dem=demands->begin(); dem!=demands->end(); dem++, d++)
      {
	stringstream s;
	s << "f(" << u << "," << u << "*)[" << dem->first.first << dem->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+d, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+d, str);
      }
      linkId++;

//       d=1;
      d=0;
      for (map<pair<int,int>,double>::iterator dem=demands->begin(); dem!=demands->end(); dem++, d++)
      {
	stringstream s;
	s << "f(" << u << "*," << u << ")[" << dem->first.first << dem->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+d, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+d, str);
      }
      linkId++;

      /* for each active radio */
      for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	if (T->node[u].info->radio[i].channel)
	{
// 	  d=1;
	  d=0;
	  for (map<pair<int,int>,double>::iterator dem=demands->begin(); dem!=demands->end(); dem++, d++)
	  {
	    stringstream s;
	    s << "f(" << u << "*,r" << u << ":" << T->node[u].info->radio[i].channel << ")[" << dem->first.first << dem->first.second << "]";
// 	    set_col_name(lp, linkId*numDemands+d, (char*)s.str().c_str());
	    string str = s.str();
	    model.setColumnName(linkId*numDemands+d, str);
	  }
	  linkId++;

// 	  d=1;
	  d=0;
	  for (map<pair<int,int>,double>::iterator dem=demands->begin(); dem!=demands->end(); dem++, d++)
	  {
	    stringstream s;
	    s << "f(r" << u << ":" << T->node[u].info->radio[i].channel << "," << u << "*" << ")[" << dem->first.first << dem->first.second << "]";
// 	    set_col_name(lp, linkId*numDemands+d, (char*)s.str().c_str());
	    string str = s.str();
	    model.setColumnName(linkId*numDemands+d, str);
	  }
	  linkId++;
	}
    }
  
    
    
  int* colno = new int[Ncol+1];
  double *coeff = new double[Ncol+1], *coeff2 = new double[Ncol+1];
  
  vector<vector<double> > coeff_tau;
  vector<vector<int> > colno_tau;

  vector<vector<double> > coeff_power;
  vector<vector<int> > colno_power;
  vector<double> lhs_power;

  int k, pos_i;

  /* With one iteration over all the radios, we determine the coefficients used for the objective
   * function, the constraints tau_TX(r) + tau_RX(r) <= 1 (for each active radio) and node power constraints */
  for (unsigned int u=0; u<T->node.size(); u++)
    if (T->node[u].info->countRadiosWithChannel(0) != T->node[u].info->numAvailRadios())
    {
      k = 0;
      int r = 0;

      float lhs = 0.f;

      /* for each active radio */
      for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	if (T->node[u].info->radio[i].channel)
	{
	  int j1=k;
	  RadioIfaceIterator ri = T->node[u].info->radio.begin() + i;

	  if (Lro.find(TRadio(u,i)) != Lro.end())
	    for (set<TLinkItrs>::iterator l=Lro[TRadio(u,i)].begin(); l!=Lro[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->rate);
	      // coeff2 is used for the power consumption constraint
	      fill_n(coeff2+k, numDemands, ((*ri->rho_tx)(DbmToW(TxPowerLevel(l->second->txpower))) - ri->rho_idle) / l->second->rate);
	      pos_i = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, counter<int>(pos_i));
	      generate_n(colno+k, numDemands, counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  int j2=k;

	  if (Lri.find(TRadio(u,i)) != Lri.end())
	    for (set<TLinkItrs>::iterator l=Lri[TRadio(u,i)].begin(); l!=Lri[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->rate);
	      fill_n(coeff2+k, numDemands, (ri->rho_rx - ri->rho_idle) / l->second->rate);
	      pos_i = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, counter<int>(pos_i));
	      generate_n(colno+k, numDemands, counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  if (k != j1)
	  {
	    /* make a copy of coeff and colno so that later we can add the tau constraint */
	    coeff_tau.push_back(vector<double>(coeff+j1,coeff+k));
	    colno_tau.push_back(vector<int>(colno+j1,colno+k));
	  }

	  /* multiply coefficients by the power consumption (needed for node power constraint) */
// 	  for (int j=j1; j<j2; j++)
// 	    coeff[j] *= (T->pm.rho_tx-T->pm.rho_idle);
// 
// 	  for (int j=j2; j<k; j++)
// 	    coeff[j] *= (T->pm.rho_rx-T->pm.rho_idle);

	  /* add the contribution of arcs of type radio */
	  if (epsy.at(u).yR.at(i))
// 	    lhs += T->pm.rho_idle;
	    lhs += ri->rho_idle;
	  else
	  {
	    pos_i = (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 2 + 2*r) * numDemands;
	    generate_n(colno+k, numDemands*2, counter<int>(pos_i-1));
// 	    fill_n(coeff+k, numDemands*2, (T->pm.rho_idle - T->pm.rho_sleep)/epsy.at(u).epsR.at(i));
	    fill_n(coeff2+k, numDemands*2, (ri->rho_idle - ri->rho_sleep)/epsy.at(u).epsR.at(i));
	    k += numDemands*2;

// 	    lhs += T->pm.rho_sleep;
	    lhs += ri->rho_sleep;
	  }

	  r++;
	}

      /* stores the coefficients for the power constraint on node u */
      if (k)
      {
// 	float node_rho_on = T->pm.BasePowerConsumption + T->pm.rho_sleep * (T->node[u].info->countRadiosWithChannel(0)),
// 	      node_rho_off = T->pm.StandbyPowerConsumption - T->pm.rho_sleep * (T->node[u].info->numAvailRadios() - T->node[u].info->countRadiosWithChannel(0));

	float node_rho_on = T->node[u].info->BasePowerConsumption;
	float node_rho_off = T->node[u].info->StandbyPowerConsumption;
 
	for (RadioIfaceIterator ri = T->node[u].info->radio.begin(); ri != T->node[u].info->radio.end(); ri++)
	{
	  node_rho_on += (ri->channel ? 0 : ri->rho_sleep);
	  node_rho_off -= (ri->channel ? ri->rho_sleep : 0);
	}
	
	/* add the contribution of arcs of type node */
	if (epsy.at(u).yN)
	  lhs += node_rho_on;
	else
	{
	  pos_i = (cumNumNeighbs[T->node.size()] + cumNumArcs[u]) * numDemands;
// 	  generate_n(colno+k, numDemands*2, counter<int>(pos_i));
	  generate_n(colno+k, numDemands*2, counter<int>(pos_i-1));
	  fill_n(coeff2+k, numDemands*2, (node_rho_on - node_rho_off)/epsy.at(u).epsN);
	  k += numDemands*2;

	  lhs += node_rho_off;
	}

	coeff_power.push_back(vector<double>(coeff2,coeff2+k));
	colno_power.push_back(vector<int>(colno,colno+k));
	lhs_power.push_back(lhs);
      }
    }

  for (int i=0; i<=Ncol; i++)
    coeff[i] = 0.;               // reset the elements of coeff

  for (unsigned int i=0; i<coeff_power.size(); i++)
    for (unsigned int j=0; j<coeff_power.at(i).size(); j++)
      coeff[colno_power.at(i).at(j)] += coeff_power.at(i).at(j);
    
  /* set the objective function: min sum_u\in V P(u) */
//   set_obj_fn(lp, coeff);

  for (int i=0; i<Ncol; i++)
    model.setObjectiveCoefficient(i, coeff[i]);
  
  /* set the object direction to minimize */
//   set_minim(lp);
  model.setOptimizationDirection(1);
  
//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  CoinBuild buildObject;
  
  /* constraints on the total utilizations ( |E| constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      k = 0;
      for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	  if ( TLinkItrs(x,y).IsInTheCollDomainOf(T,TLinkItrs(u,v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i;
	    k += numDemands;
	  }
//       add_constraintex(lp, k, coeff, colno, LE, u0);
      buildObject.addRow(k, colno, coeff, 0., u0);
    }
    
  /* tau constraints */
  for (unsigned int r=0; r<coeff_tau.size(); r++)
  {
    copy (coeff_tau.at(r).begin(), coeff_tau.at(r).end(), coeff);
    copy (colno_tau.at(r).begin(), colno_tau.at(r).end(), colno);

//     add_constraintex(lp, coeff_tau.at(r).size(), coeff, colno, LE, 1);
    buildObject.addRow(coeff_tau.at(r).size(), colno, coeff, 0., 1.);
  }

  /* node power constraints */
  for (unsigned int i=0; i<coeff_power.size(); i++)
  {
    copy (coeff_power.at(i).begin(), coeff_power.at(i).end(), coeff);
    copy (colno_power.at(i).begin(), colno_power.at(i).end(), colno);

//     add_constraintex(lp, coeff_power.at(i).size(), coeff, colno, LE, nodePowerConstr->at(i) - lhs_power.at(i));
    buildObject.addRow(coeff_power.at(i).size(), colno, coeff, 0., nodePowerConstr->at(i) - lhs_power.at(i));
  }

  /* flow conservation constraints on the vertices of type radio r^(u,c) */
  for (unsigned int u=0; u<T->node.size(); u++)
    if (T->node[u].info->countRadiosWithChannel(0) != T->node[u].info->numAvailRadios())
    {
      int r = 0;

      /* for each active radio */
      for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	if (T->node[u].info->radio[i].channel)
	{
	  // sum of the flow outgoing from r^(u,c)
// 	  for (unsigned int d=1; d<=numDemands; d++)
	  for (unsigned int d=0; d<numDemands; d++)
	  {
	    k = 0;
	    if (Lro.find(TRadio(u,i)) != Lro.end())
	      for (set<TLinkItrs>::iterator l=Lro[TRadio(u,i)].begin(); l!=Lro[TRadio(u,i)].end(); l++)
	      {
		colno[k] = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands + d;
		coeff[k++] = 1.;
	      }
	    colno[k] = (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 2 + 2*r)*numDemands + d;
	    coeff[k++] = -1.;
	    
// 	    add_constraintex(lp, k, coeff, colno, EQ, 0.);
	    buildObject.addRow(k, colno, coeff, 0., 0.);
	  }

	  // sum of the flow incoming to r^(u,c)
// 	  for (unsigned int d=1; d<=numDemands; d++)
	  for (unsigned int d=0; d<numDemands; d++)
	  {
	    k = 0;
	    if (Lri.find(TRadio(u,i)) != Lri.end())
	      for (set<TLinkItrs>::iterator l=Lri[TRadio(u,i)].begin(); l!=Lri[TRadio(u,i)].end(); l++)
	      {
		colno[k] = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands + d;
		coeff[k++] = 1.;
	      }
	    colno[k] = (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 3 + 2*r)*numDemands + d;
	    coeff[k++] = -1.;
	    
// 	    add_constraintex(lp, k, coeff, colno, EQ, 0.);
	    buildObject.addRow(k, colno, coeff, 0., 0.);
	  }

	  r++;
	}
    }

  /* flow conservation constraints on the vertices of type node u* */
  for (unsigned int u=0; u<T->node.size(); u++)
    if (T->node[u].info->countRadiosWithChannel(0) != T->node[u].info->numAvailRadios())
//       for (unsigned int d=1; d<=numDemands; d++)
      for (unsigned int d=0; d<numDemands; d++)
      {
	int r = 0;
	k = 0;

	/* for each active radio */
	for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	  if (T->node[u].info->radio[i].channel)
	  {
	    colno[k] = (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 2 + 2*r)*numDemands + d;
	    coeff[k++] = 1.;
	    r++;
	  }
	colno[k] = (cumNumNeighbs[T->node.size()] + cumNumArcs[u])*numDemands + d;
	coeff[k++] = -1.;

// 	add_constraintex(lp, k, coeff, colno, EQ, 0.);
	buildObject.addRow(k, colno, coeff, 0., 0.);

	r = 0;
	k = 0;

	/* for each active radio */
	for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	  if (T->node[u].info->radio[i].channel)
	  {
	    colno[k] = (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 3 + 2*r)*numDemands + d;
	    coeff[k++] = 1.;
	    r++;
	  }
	colno[k] = (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 1)*numDemands + d;
	coeff[k++] = -1.;

// 	add_constraintex(lp, k, coeff, colno, EQ, 0.);
	buildObject.addRow(k, colno, coeff, 0., 0.);
      }
  
  /* flow conservation constraints on nodes */
  for (unsigned int u=0; u<T->node.size(); u++)
    if (T->node[u].info->countRadiosWithChannel(0) != T->node[u].info->numAvailRadios())
    {
//       int d = 1;
      int d = 0;
      for (map<pair<int,int>,double>::iterator dem = demands->begin(); dem != demands->end(); dem++, d++)
      {
	colno[0] = (cumNumNeighbs[T->node.size()] + cumNumArcs[u])*numDemands + d;
	coeff[0] = 1.;

	colno[1] = (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 1)*numDemands + d;
	coeff[1] = -1.;

	/* if u is the source of the demand */
	if (u == dem->first.first)
	{
// 	  add_constraintex(lp, 2, coeff, colno, EQ, dem->second);
	  buildObject.addRow(2, colno, coeff, dem->second, dem->second);
//	  set_upbo(lp, (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 1)*numDemands + d, 0.);
	  model.setColumnUpper((cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 1)*numDemands + d, 0.);
	}
	/* if u is the destination of the demand */
	else if (u == dem->first.second)
	{
// 	  add_constraintex(lp, 2, coeff, colno, EQ, -dem->second);
	  buildObject.addRow(2, colno, coeff, -dem->second, -dem->second);
//	  set_upbo(lp, (cumNumNeighbs[T->node.size()] + cumNumArcs[u])*numDemands + d, 0.);
	  model.setColumnUpper((cumNumNeighbs[T->node.size()] + cumNumArcs[u])*numDemands + d, 0.);
	}
	else
// 	  add_constraintex(lp, 2, coeff, colno, EQ, 0.);
	  buildObject.addRow(2, colno, coeff, 0., 0.);
      }
    }
  

//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model.addRows(buildObject);

  // write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /* Now let lpsolve calculate a solution */
//   int ret = solve(lp);
  int ret = model.initialSolve();

//   if (ret != OPTIMAL)
  if (ret)
    cerr << endl << "WARNING: MinPower_eps: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
  else
  {
//     print_solution(lp,4);

  /* If an optimal solution has been found, then update the flows, the e2e flows and the Utot */

//     REAL *ptr_var, *ptr_constr;
    double *ptr_var, *ptr_constr;
    
//     get_ptr_variables(lp, &ptr_var);
//     get_ptr_constraints(lp, &ptr_constr);

    ptr_var = model.primalColumnSolution();
    ptr_constr = model.primalRowSolution();

    unsigned int linkId = 0;
  
    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      {
	v->Utot = ptr_constr[linkId];
	v->flow = 0.;
	v->f_e2e.clear();

	int i=0;
	for (map<pair<int,int>,double>::iterator dem = demands->begin(); dem != demands->end(); dem++, i++)
	  if (ptr_var[linkId*numDemands + i] > 0.)
	    v->flow += (v->f_e2e[dem->first] = ptr_var[linkId*numDemands + i]);

	linkId++;
      }
  }
  
  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;

  /* clean up such that all used memory by lpsolve is freed */
//   if (lp)
//     delete_lp(lp);

  return ret;
}


/** MILP methods **/

// void MILP::set_params ()
// {
//   if (timeout > 0)
//     set_timeout(lp, timeout);
// 
// //  set_epslevel(lp, EPS_BAGGY);
// //  set_epsel(lp, 1.0e-4);
// //  set_epsb(lp, 1.0e-5);
// //  set_epsd(lp, 1.0e-4);
// //  set_epsint(lp, 1.0e-3);
// //  set_break_at_first(lp, TRUE);
// 
// //  set_mip_gap(lp, FALSE, 0.05);
// //  set_break_at_value(lp, u0); 
// //  set_presolve(lp, PRESOLVE_ROWS | PRESOLVE_COLS | PRESOLVE_LINDEP, get_presolveloops(lp));
// }


// void MILP::saveMPS (char* filename)
// {
//   write_freemps(lp, filename);  
// }


int MILP::lp_solve ()
{
  CbcModel* model;
  bool done = false;
//     const char * argv2[]={"-solve"};
//     write_freemps(lp, "mps.tmp");

  while (!done)
  {
//       OsiClpSolverInterface solver1;
//       solver1.readMps("mps.tmp","");

//       model = new CbcModel(solver1);
    model = new CbcModel(*solver);

    if (timeout > 0)
      model->setMaximumSeconds (timeout);

    if (hasBreakValue)
    {
      model->setCutoff(breakValue);
      model->setMaximumSolutions(1);
    }

    callCbc("-solve", *model);
    
    //CbcMain (0,NULL,*model);

    if (hasBreakValue && model->isProvenInfeasible())
    {
      // make another iteration without break value
      hasBreakValue = false;
      delete model;
    }
    else
      done = true;
  }

  if (model->bestSolution())
  {
    obj_value = model->getObjValue();
    bound = model->getBestPossibleObjValue();

    if (apply)
      /* update the flows and the e2e flows and the Utot */
      updateFlows(model->getColSolution(), model->getRowPrice());

    delete model;

    return 0;
  }

  // no solution found
  cerr << endl << "WARNING: A (sub-)optimal solution to the MILP has NOT been found" << endl << endl;

  delete model;

  return -1;
}


int MILP::lag_solve (const set<int>& torelax, double mu_0, int max_iter, sub_grad_method* sgm)
{
//   set_params();
  relaxedConstrGroups.clear();

  set<int> relax_constr;

  for (set<int>::const_iterator s=torelax.begin(); s!=torelax.end(); s++)
  {
    // check if the group number is valid
    if (*s < 1 || *s > constrGroupSize.size()-1)
    {
      cerr << endl << "** WARNING: constraint group " << *s << " does not exist! **" << endl << endl;
      continue;
    }
    
    for (int i=constrGroupSize[(*s)-1]+1; i<=constrGroupSize[*s]; i++)
      relax_constr.insert(i);

    relaxedConstrGroups.insert(*s);
  }
  
//   Lag_Relax lag(lp, relax_constr, mu_0, max_iter);
  Lag_Relax lag(solver->getModelPtr(), relax_constr, mu_0, max_iter);

  if (sgm)
    lag.set_sgm(sgm);

//   lag.set_solver(ms);

  if (hasBreakValue)
    lag.set_breakValue(breakValue);
  
  lag.solveLAG();

  if (!lag.haveSolution())
    cerr << endl << "WARNING: A (sub-)optimal solution to the MILP has NOT been found" << endl << endl;
  else
  {
    obj_value = lag.getObjValue();

    bound = lag.getBound();

    if (apply)
      /* update the flows and the e2e flows and the Utot */
      updateFlows(lag.getSolution());
  }
  
  return (int)(lag.haveSolution());
}


MILP::~MILP ()
{
  /* clean up such that all used memory by lpsolve is freed */
//   if (lp)
//     delete_lp(lp);

  if (solver)
    delete solver;
}


/** MinPower_MILP methods **/
/* NOTE This program does not handle radios with different power consumption profiles.
   Hence, the power consumption profile of the first radio on a node is considered for
   all the radios of that node */

MinPower_MILP::MinPower_MILP(Topology* _T, double _maxUtot, vector<float>* nodePowerConstr, long _timeout, bool _apply)
  : MILP(_T, _timeout, _apply),
    maxUtot(_maxUtot)
{
  ClpSimplex* model = new ClpSimplex;
  int Ncol;
  double F = 0.;      // sum of all the demands

  T->ResetChannels();

  /* check that e2e flows are consistent and obtain the set of demands */
  if (!(T->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: MinPower_MILP: Checking E2E flows before beginning FAILED" << endl << endl;

  // compute the number of demands and the sum of all their flows
  numDemands = demands.size();
  
  for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++)
    F += d->second;

  cout << "La somma del flusso di tutte le demand e' " << F << endl;
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->node.size() + 1];

  cumNumNeighbs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();

  // the number of variables is numDemands*numLinks*numChannels (f^ab(u->v)[c]) + numNodes*numChannels (x^c(u)) + numNodes (y_u)
  Ncol = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + T->node.size();
  
  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model->resize(0, Ncol);

  /* Columns (variables) are ordered this way:
      f_e1,d1,c1 ... f_e1,d1,cC ... f_e1,dD,c1 ... f_e1,dD,cC ... f_eE,dD,c1 ... f_eE,dD,cC  x^c1(n1) ... x^cC(n1) ... x^c1(nN) ... x^cC(nN)  y(n1) ... y(nN)
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int i=0;
      for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, i++)
	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	  stringstream s;
	  s << "f(" << u->info->id << "," << v->info->id << ")[" << c << "]{" << d->first.first << d->first.second << "}";
// 	  set_col_name(lp, (linkId*numDemands+i)*T->numAvailChannels+c, (char*)s.str().c_str());
	  string str = s.str();
	  model->setColumnName((linkId*numDemands+i)*T->numAvailChannels+c-1, str);
	}
      linkId++;
    }
  
  for (unsigned int i=0; i<T->node.size(); i++)
    for (unsigned int c=1; c<=T->numAvailChannels; c++) {
      stringstream s;
      s << "x_" << i << "[" << c << "]";
      int j = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels + c;
//       set_col_name(lp, j, (char*)s.str().c_str());
// 	set_binary(lp, j, TRUE);
      string str = s.str();
      model->setColumnName(j-1, str);
      model->setInteger(j-1);
      model->setColumnBounds(j-1, 0., 1.);
    }
  
  for (unsigned int i=0; i<T->node.size(); i++)
  {
    stringstream s;
    s << "y_" << i;
    int j = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + i + 1;
//     set_col_name(lp, j, (char*)s.str().c_str());
//       set_binary(lp, j, TRUE);
    string str = s.str();
    model->setColumnName(j-1, str);
    model->setInteger(j-1);
    model->setColumnBounds(j-1, 0., 1.);
  }
  

  int* colno = new int[Ncol];
  double *coeff = new double[Ncol], *coeff2 = new double[Ncol];
  
  vector<vector<double> > coeff_tau;
  vector<vector<int> > colno_tau;

  vector<vector<double> > coeff_power;
  vector<vector<int> > colno_power;
  vector<double> lhs_power;

  map<int, set<TLinkItrs> > Linc;  // maps each node to the set of incoming links

  /* compute Linc for all the nodes */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      Linc[v->info->id].insert(TLinkItrs(u,v));

  int k, pos_i;

  /* With one iteration over all the nodes, we determine the coefficients used for the objective
   * function, the constraints tau_TX^(u,c) + tau_RX^(u,c) <= 1 (for each active radio) and node power constraints */
  for (unsigned int u=0; u<T->node.size(); u++)
  {
    k = 0;
    RadioIfaceIterator ri = T->node[u].info->radio.begin();

    /* for each channel */
//     for (unsigned int c=1; c<=T->numAvailChannels; c++)
    for (unsigned int c=0; c<T->numAvailChannels; c++)
    {
      int j1=k;

      for (list<TGraphNeighbor>::iterator v=T->node[u].neighbor.begin(); v!=T->node[u].neighbor.end(); v++)
      {
	fill_n(coeff+k, numDemands, 1. / v->rate);
	// coeff2 is used for the power consumption constraint
	fill_n(coeff2+k, numDemands, ((*ri->rho_tx)(DbmToW(TxPowerLevel(v->txpower))) - ri->rho_idle) / v->rate);
	pos_i = (cumNumNeighbs[u] + distance(T->node[u].neighbor.begin(),v))*numDemands*T->numAvailChannels + c;
	generate_n(colno+k, numDemands, counter<int>(pos_i-T->numAvailChannels, T->numAvailChannels));
	k += numDemands;
      }

      int j2=k;

      if (Linc.find(u) != Linc.end())
	for (set<TLinkItrs>::iterator l=Linc[u].begin(); l!=Linc[u].end(); l++)
	{
	  fill_n(coeff+k, numDemands, 1. / l->second->rate);
	  fill_n(coeff2+k, numDemands, (ri->rho_rx - ri->rho_idle) / l->second->rate);
	  pos_i = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands*T->numAvailChannels + c;
	  generate_n(colno+k, numDemands, counter<int>(pos_i-T->numAvailChannels, T->numAvailChannels));
	  k += numDemands;
	}

      if (k != j1)
      {
	/* make a copy of coeff and colno so that later we can add the tau constraint */
	coeff_tau.push_back(vector<double>(coeff+j1,coeff+k));
	colno_tau.push_back(vector<int>(colno+j1,colno+k));
      }

      /* multiply coefficients by the power consumption (needed for node power constraint) */
//       for (int j=j1; j<j2; j++)
// 	coeff[j] *= (T->pm.rho_tx-T->pm.rho_idle);
// 
//       for (int j=j2; j<k; j++)
// 	coeff[j] *= (T->pm.rho_rx-T->pm.rho_idle);
    }

//     float tmp = T->pm.rho_idle-T->pm.rho_sleep;
    float tmp = ri->rho_idle - ri->rho_sleep;

    /* stores the coefficients for the power constraint on node u */
    fill_n(coeff2+k, T->numAvailChannels, tmp);      // x_u^c * (rho_idle - rho_sleep)
    generate_n(colno+k, T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u*T->numAvailChannels - 1));
    k += T->numAvailChannels;
    
//     tmp = T->pm.BasePowerConsumption + T->node[u].info->numAvailRadios() * T->pm.rho_sleep - T->pm.StandbyPowerConsumption;
    tmp = T->node[u].info->BasePowerConsumption + T->node[u].info->numAvailRadios() * ri->rho_sleep - T->node[u].info->StandbyPowerConsumption;
    
    coeff2[k] = tmp;
    colno[k++] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u;   // y_u
  
    coeff_power.push_back(vector<double>(coeff2,coeff2+k));
    colno_power.push_back(vector<int>(colno,colno+k));
    lhs_power.push_back(nodePowerConstr->at(u) - T->node[u].info->StandbyPowerConsumption);
  }

  for (int i=0; i<Ncol; i++)
    coeff[i] = 0.;               // reset the elements of coeff

  for (unsigned int i=0; i<coeff_power.size(); i++)
    for (unsigned int j=0; j<coeff_power.at(i).size(); j++)
      coeff[colno_power.at(i).at(j)] += coeff_power.at(i).at(j);
    
  /* set the objective function: min sum_u\in V P(u) */
//   set_obj_fn(lp, coeff);

  for (int i=0; i<Ncol; i++)
    model->setObjectiveCoefficient(i, coeff[i]);

  /* set the object direction to minimize */
//   set_minim(lp);
  model->setOptimizationDirection(1);

//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  CoinBuild buildObject;

  /* constraints on the total utilizations ( |E|*numChannels constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) 
//       for (unsigned int c=1; c<=T->numAvailChannels; c++) {
      for (unsigned int c=0; c<T->numAvailChannels; c++) {
	k = 0;
	for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	  for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	    if ( T->PotInterfLinks(TLinkItrs(x,y),TLinkItrs(u,v)) ) {
	      fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	      for (unsigned int i=0; i<numDemands; i++)
		colno[k+i] = ((cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i)*T->numAvailChannels + c;
	      k += numDemands;
	    }
// 	add_constraintex(lp, k, coeff, colno, LE, maxUtot);
	buildObject.addRow(k, colno, coeff, 0., maxUtot);
      }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  /* tau constraints */
  for (unsigned int r=0; r<coeff_tau.size(); r++)
  {
    copy (coeff_tau.at(r).begin(), coeff_tau.at(r).end(), coeff);
    copy (colno_tau.at(r).begin(), colno_tau.at(r).end(), colno);

//     add_constraintex(lp, coeff_tau.at(r).size(), coeff, colno, LE, 1);
    buildObject.addRow(coeff_tau.at(r).size(), colno, coeff, 0., 1.);
  }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  /* node power constraints */
  for (unsigned int i=0; i<coeff_power.size(); i++)
  {
    copy (coeff_power.at(i).begin(), coeff_power.at(i).end(), coeff);
    copy (colno_power.at(i).begin(), colno_power.at(i).end(), colno);

//     add_constraintex(lp, coeff_power.at(i).size(), coeff, colno, LE, nodePowerConstr->at(i) - T->pm.StandbyPowerConsumption);
//     buildObject.addRow(coeff_power.at(i).size(), colno, coeff, 0., nodePowerConstr->at(i) - T->pm.StandbyPowerConsumption);
    buildObject.addRow(coeff_power.at(i).size(), colno, coeff, 0., lhs_power.at(i));
  }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  for (unsigned int u=0; u<T->node.size(); u++)
  {
    /* constraints y_u >= x_u^c, for each channel c */
//     for (unsigned int c=1; c<=T->numAvailChannels; c++)
    for (unsigned int c=0; c<T->numAvailChannels; c++)
    {
      coeff[0] = 1.;
      colno[0] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u*T->numAvailChannels + c;     // x_u^c

      coeff[1] = -1.;
//       colno[1] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u + 1;   // y_u
      colno[1] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u;   // y_u

//       add_constraintex(lp, 2, coeff, colno, LE, 0.);
      buildObject.addRow(2, colno, coeff, -COIN_DBL_MAX, 0.);
    }
    
    /* constraint y_u <= sum_c x_u^c */
    fill_n(coeff, T->numAvailChannels, -1.);
//     generate_n(colno, T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u*T->numAvailChannels));
    generate_n(colno, T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u*T->numAvailChannels - 1));
    
    coeff[T->numAvailChannels] = 1.;
//     colno[T->numAvailChannels] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u + 1;   // y_u
    colno[T->numAvailChannels] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u;   // y_u
    
//     add_constraintex(lp, T->numAvailChannels+1, coeff, colno, LE, 0.);
    buildObject.addRow(T->numAvailChannels+1, colno, coeff, -COIN_DBL_MAX, 0.);
  }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  vector<vector<int> > colno_v(T->node.size()*numDemands);
  vector<vector<double> > coeff_v(T->node.size()*numDemands);
  
  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands.begin(); di!=demands.end(); di++, d++) 
// 	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	for (unsigned int c=0; c<T->numAvailChannels; c++) {
	// for (unsigned int d=0; d<numDemands; d++) {
	  dist = distance(u->neighbor.begin(), v);
	  /* flow leaving node u */
	  coeff_v[u->info->id*numDemands+d].push_back(1);
	  colno_v[u->info->id*numDemands+d].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  /* flow entering node v */
	  coeff_v[v->info->id*numDemands+d].push_back(-1);
	  colno_v[v->info->id*numDemands+d].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  /* if v is the source of the demand or u is the destination of the demand, f^d(u->v)[c] must be null */
	  if ( (u->info->id == di->first.second) || (v->info->id == di->first.first) )
// 	    set_upbo(lp, ((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, 0.);
	    model->setColumnUpper(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, 0.);
//	  else
//	    set_upbo(lp, ((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, di->second);
        }
    }

  /* Add the flow conservation constraints one-by-one */
  k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      /* if i is the source of the demand d */
      if (i == d->first.first)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
      /* if i is the destination of the demand d */
      else if (i == d->first.second)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
      else
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
      buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
      k++;
    }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because the goal of minimizing M
   * will force to minimize the e2e flows
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

  /* constraints on the number of radios per node (|V| constraints in total)*/
  for (unsigned int i=0; i<T->node.size(); i++) {
    fill(coeff, coeff+T->numAvailChannels, 1);
//     generate(colno, colno+T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels));
    generate(colno, colno+T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels - 1));
    
//     add_constraintex(lp,T->numAvailChannels, coeff, colno, LE, T->node[i].info->numAvailRadios());
    buildObject.addRow(T->numAvailChannels, colno, coeff, 0., T->node[i].info->numAvailRadios());
  }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  colno_v.clear();
  colno_v.resize(T->node.size()*T->numAvailChannels);
  coeff_v.clear();
  coeff_v.resize(T->node.size()*T->numAvailChannels);

  /* constraints to make the flow null on all the links on u using channel c if x^c(u) is null  ( |V| * numChannels constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands.begin(); di!=demands.end(); di++, d++) 
	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	  dist = distance(u->neighbor.begin(), v);
	  /* flow leaving node u */
	  coeff_v[u->info->id*T->numAvailChannels+c-1].push_back(1);
// 	  colno_v[u->info->id*T->numAvailChannels+c-1].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  colno_v[u->info->id*T->numAvailChannels+c-1].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c - 1);
	  /* flow entering node v */
	  coeff_v[v->info->id*T->numAvailChannels+c-1].push_back(1);
// 	  colno_v[v->info->id*T->numAvailChannels+c-1].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  colno_v[v->info->id*T->numAvailChannels+c-1].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c - 1);
        }
    }

  /* Add the constraints one-by-one */
  k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (unsigned int c=1; c<=T->numAvailChannels; c++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

//       coeff[coeff_v[k].size()] = -F / lowbo * coeff_v[k].size();
      coeff[coeff_v[k].size()] = -F * coeff_v[k].size();
//       colno[coeff_v[k].size()] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels + c;
      colno[coeff_v[k].size()] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels + c - 1;
      
      if (coeff_v[k].size())
// 	add_constraintex(lp, coeff_v[k].size()+1, coeff, colno, LE, 0.);
	buildObject.addRow(coeff_v[k].size()+1, colno, coeff, -COIN_DBL_MAX, 0.);

      k++;
    }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  /* constraints to make flow on u->v[c] null if u and v do not use channel c (2*|E|*numChannels constraints in total) */
//   for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
//     for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) 
//       for (unsigned int c=1; c<=T->numAvailChannels; c++) {
// 	fill(coeff, coeff+numDemands, 1);
// 	coeff[numDemands] = -F;
// 	
// 	for (unsigned int i=0; i<numDemands; i++)
// 	  colno[i] = ((cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i)*T->numAvailChannels + c;
// 	colno[numDemands] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u->info->id*T->numAvailChannels + c;
// 	
// 	add_constraintex(lp, numDemands+1, coeff, colno, LE, 0.);
// 	
// 	fill(coeff, coeff+numDemands, 1);
// 	coeff[numDemands] = -F;
// 	
// 	for (unsigned int i=0; i<numDemands; i++)
// 	  colno[i] = ((cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i)*T->numAvailChannels + c;
// 	colno[numDemands] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + v->info->id*T->numAvailChannels + c;
// 	
// 	add_constraintex(lp, numDemands+1, coeff, colno, LE, 0.);
//       }


//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model->addRows(buildObject);

  // the model is used to build the solver (which will own the pointer)
  solver = new OsiClpSolverInterface(model, true);

  // write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;
}


void MinPower_MILP::updateFlows (const double* ptr_var, const double* ptr_constr)
{
  bool updateUtot = false;

  // if the Utot constraints have not been relaxed, use their values to set the Utot fields
  if (ptr_constr && (relaxedConstrGroups.find(1) == relaxedConstrGroups.end()))
  {
    updateUtot = true;
    cout << endl << "*** updateUtot true ***" << endl;
  }

  unsigned int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++) {
    int N = u->neighbor.size();
    list<TGraphNeighbor>::iterator v=u->neighbor.begin();
    for (int n=0; n<N; n++) {
      bool parallel = false;
      for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	double _f=0;
	map<pair<int,int>,double> _f_e2e;
	int i=0;
	
	for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, i++)
	  if (ptr_var[linkId*numDemands*T->numAvailChannels+i*T->numAvailChannels +c-1] > almost_null)
	    _f += (_f_e2e[pair<int,int>(d->first.first,d->first.second)] = ptr_var[linkId*numDemands*T->numAvailChannels+i*T->numAvailChannels +c-1]);

	if (_f) {
	  if (!parallel) {
	    v->flow = _f;
	    v->f_e2e = _f_e2e;
	    if (updateUtot)
	      v->Utot = ptr_constr[linkId*T->numAvailChannels+c-1];
	    
	    T->AssignChannel2Link(TLinkItrs(u,v),c);
	    
	    parallel = true;
	  }
	  else {
	    list<TGraphNeighbor>::iterator z = u->neighbor.insert(u->neighbor.end(), TGraphNeighbor(v->info,v->rate,v->txpower));
	    z->flow = _f;
	    z->f_e2e = _f_e2e;
	    if (updateUtot)
	      z->Utot = ptr_constr[linkId*T->numAvailChannels+c-1];
	    
	    T->AssignChannel2Link(TLinkItrs(u,z),c);
	  }
	}
      }
      /* if the flow is null on the link (hence no channel has been assigned to it), remove it */
      if (!parallel) {
	list<TGraphNeighbor>::iterator z = v++;
	u->neighbor.erase(z);
      }
      else
	v++;
      linkId++;
    }
  }
}
    

/** Lag_Relax methods **/

// Lag_Relax::Lag_Relax (lprec* lp, set<int> relax_constr, REAL mu_0, int max_iter, int max_noimprove)
//   : m_lp(lp),
Lag_Relax::Lag_Relax (ClpSimplex* model, set<int> relax_constr, double mu_0, int max_iter, int max_noimprove)
  : m_model(model),
    m_curr_iter(0),
    m_max_iter(max_iter),
    m_max_noimprove(max_noimprove),
    m_hasBreakValue(false),
    m_have_solution(false),
//     m_solution(new REAL[get_Ncolumns(m_lp)]),
    m_solution(new double[m_model->getNumCols()]),
    m_obj_val(0.),
    m_have_bound(false),
    m_sgm(new inverseK(this))
//     m_solver(LpSolve)
{
  int nzelem;
//   REAL* row = new REAL[1+get_Ncolumns(m_lp)];
  const double* objcoeff;
//   int* colno = new int[1+get_Ncolumns(m_lp)];
  double* row;
  int* colno;

  /* get the objective function */
//   if (!get_row(m_lp, 0, row))
//   {
//     cerr << "ERROR: Lag_Relax: Error while getting the objective function of the provided program" << endl << endl;
//     throw;
//   }
  objcoeff = model->getRowObjCoefficients();

  /* save the information in a lag_constraint object */
//   m_obj_func = vector<REAL>(row, row+1+get_Ncolumns(m_lp));
  m_obj_func = vector<double>(objcoeff, objcoeff+m_model->getNumCols());

  CoinPackedMatrix* mat = m_model->matrix();

  /* for each constraint to be relaxed */
  for (set<int>::iterator i=relax_constr.begin(); i!=relax_constr.end(); i++)
//     if (*i >= 1 && *i <= get_Nrows(m_lp))
    if (*i >= 0 && *i < m_model->getNumRows())
    {
//       /* get the corresponding row */
//       if ((nzelem = get_rowex(m_lp, *i, row, colno)) < 0 )
//       {
// 	cerr << "WARNING: Lag_Relax: Error while getting row (constraint) " << *i << " in the provided program" << endl << endl;
// 	continue;
//       }

      /* find the index of the first element of the required row. See "Detailed Description"
       * at http://www.coin-or.org/Doxygen/CoinAll/class_coin_packed_matrix.html */
      nzelem = 0;
      for (int j=0; j < *i; j++)
	nzelem += mat->getVectorSize(j);

      row = new double[mat->getVectorSize(*i)];
      colno = new int[mat->getVectorSize(*i)];

      copy (mat->getElements()+nzelem, mat->getElements()+nzelem+mat->getVectorSize(*i), row);
      copy (mat->getIndices()+nzelem, mat->getIndices()+nzelem+mat->getVectorSize(*i), colno);

      /* save the information in a lag_constraint object */
//       m_relax_constr.push_back(lag_constraint(row, colno, nzelem, get_constr_type(m_lp, *i), get_rh(m_lp, *i), get_constr_type(m_lp, *i)==GE ? -fabs(mu_0) : fabs(mu_0)));

      m_relax_constr.push_back(lag_constraint(row, colno, mat->getVectorSize(*i), m_model->rowLower()[*i], m_model->rowUpper()[*i], (m_model->rowLower()[*i] == 0. ? -fabs(mu_0) : fabs(mu_0))));

      delete row;
      delete colno;
    }
    else
      cerr << "WARNING: Lag_Relax: Cannot find row (constraint) " << *i << " in the provided program" << endl << endl;

  nzelem = 0;     // number of deleted constraints
  for (set<int>::iterator i=relax_constr.begin(); i!=relax_constr.end(); i++)
//     if (*i >= 1 && *i <= get_Nrows(m_lp))
    if (*i >= 0 && *i < m_model->getNumRows())
    {
      /* delete the row from the program */
//       del_constraint(m_lp, *i-nzelem);
      model->deleteRows(1, &(*i));
      nzelem++;
    }

//   delete row;
//   delete colno;
}


void Lag_Relax::set_sgm (sub_grad_method* sgm)
{
  delete m_sgm;
  m_sgm = sgm;
  m_sgm->set_lag(this);
}


void Lag_Relax::update_obj_func ()
{
//   REAL* row = new REAL[1+get_Ncolumns(m_lp)];
  double* row = new double[m_model->getNumCols()];

  // copy the original objective function into row
  copy(m_obj_func.begin(), m_obj_func.end(), row);

  // add the relaxed constraints
  for (vector<lag_constraint>::iterator r=m_relax_constr.begin(); r!=m_relax_constr.end(); r++)
  {
    for (unsigned int i=0; i<r->row.size(); i++)
    {
      row[r->colno[i]] += (r->row[i] * r->mu);
    }
  }

//   if (!set_obj_fn(m_lp, row))
//     cerr << "WARNING: Lag_Relax: Cannot set the objective function at iteration " << m_curr_iter << endl << endl;

  for (int i=0; i<m_model->getNumCols(); i++)
    m_model->setObjectiveCoefficient(i, row[i]);

  delete row;
}


void Lag_Relax::update_multipliers (bool failed, bool noimprove)
{
  double sum, theta;

  if (!failed)
  {
    /* compute Ax-b  */
    m_Ax_b.clear();
    for (vector<lag_constraint>::iterator r=m_relax_constr.begin(); r!=m_relax_constr.end(); r++)
    {
      sum = 0.;
      for (unsigned int i=0; i<r->row.size(); i++)
      {
	sum += r->row[i] * m_curr_solution[r->colno[i]-1];
      }
//       sum -= r->rhs;
      sum -= (r->lower == -COIN_DBL_MAX ? r->upper : r->lower);

      m_Ax_b.push_back(sum);
    }
  }

  // compute theta
  theta = m_sgm->theta(failed || noimprove);
  
  // update the multipliers
  vector<double>::iterator v = m_Ax_b.begin();
  for (vector<lag_constraint>::iterator r=m_relax_constr.begin(); r!=m_relax_constr.end(); r++, v++)
  {
    r->mu += theta * *v;

//     if (r->type == LE)
//       r->mu = max (r->mu, 0.);   // cannot be negative
//     if (r->type == GE)
//       r->mu = min (r->mu, 0.);   // cannot be positive

    if (r->lower == -COIN_DBL_MAX)
      r->mu = max (r->mu, 0.);   // cannot be negative
    if (r->upper == COIN_DBL_MAX)
      r->mu = min (r->mu, 0.);   // cannot be positive
  }
}

#if 0
// callback function used to retrieve the global bound found by glpk
void my_glpk_callback (glp_tree* tree, void* info)
{
  int i = glp_ios_best_node(tree);
  if (i)
    *(double*)info = glp_ios_node_bound(tree, i);
}
#endif

void Lag_Relax::solveSubProblem (bool &foundSolution, bool &foundOptSolution, double &objValue, double &bound)
{
  int ret;
#if 0
  if (m_solver == LpSolve || m_solver == Glpk)
  {
    ret = solve(m_lp);

    if (ret != OPTIMAL && ret != SUBOPTIMAL && ret != FEASFOUND)
    {
      foundSolution = false;
      foundOptSolution = false;
    }
    else
    {
      foundSolution = true;
      foundOptSolution = (ret == OPTIMAL) ? true : false;
      objValue = get_objective(m_lp);
      bound = (ret == OPTIMAL) ? objValue : m_lp->bb_limitOF;
      get_ptr_variables(m_lp, (double**)&m_curr_solution);
      
      // if required, try to improve the bound by using GLPK
      if (m_solver == Glpk)
      {
	write_freemps(m_lp, "mps.tmp");

	glp_prob *P;
	glp_iocp parm;
	double glpk_bound;

	P = glp_create_prob();
	glp_read_mps(P, GLP_MPS_FILE, NULL, "mps.tmp");

	glp_init_iocp(&parm);
	parm.tm_lim = 120 * 1000;
	parm.presolve = GLP_ON;
	//parm.gmi_cuts = GLP_ON;

	parm.cb_func = my_glpk_callback;
	parm.cb_info = (void*)&glpk_bound;

	// solve with branch and cut
	glp_intopt(P, &parm);

	glp_delete_prob(P);

	cout << endl << "Bound provided by branch and cut on the Lagrangian subproblem: " << glpk_bound << endl;

	if ((is_maxim(m_lp) && glpk_bound < bound) || (!is_maxim(m_lp) && glpk_bound > bound))
	  bound = glpk_bound;
      }
    }
  }
  else   // Coin-or Cbc
  {
#endif
//     write_freemps(m_lp, "mps.tmp");
// 
//     OsiClpSolverInterface solver1;
//     solver1.readMps("mps.tmp","");
// 
//     const char * argv2[]={"driver3","-solve","-quit"};
// 
//     CbcModel model(solver1);
// 
//     //model.setMaximumSeconds (600);
// 
//     CbcMain (3, argv2, model);
// 
//     // solve the model
//     OsiSolverInterface * solver = model.solver();

    OsiClpSolverInterface* solver = new OsiClpSolverInterface(m_model, true);
    
    CbcModel* cbcm = new CbcModel(*solver);
    
    callCbc("-solve", *cbcm);
    
    if (cbcm->bestSolution())
    {
      foundSolution = true;
      foundOptSolution = (!cbcm->status() && solver->isProvenOptimal()) ? true : false;
      objValue = solver->getObjValue();
      bound = cbcm->getBestPossibleObjValue();
      m_curr_solution = solver->getColSolution();
    }
    else
    {
      foundSolution = false;
      foundOptSolution = false;
    }
//   }
}

void Lag_Relax::solveLAG ()
{
  bool feasible;
  double sum;
  bool foundSolution, foundOptSolution;
  double subObjValue, subBound, fixedTerm;

  int fd = fileno(stdin);
  int fcflags = fcntl(fd,F_GETFL);
  //  set non-blocking input
  fcntl(fd,F_SETFL,fcflags | O_NONBLOCK);

  for (m_curr_iter=0; m_curr_iter<m_max_iter; m_curr_iter++)
  {
    cout << endl << "*** Iteration " << m_curr_iter+1 << " (press 'c' and 'Enter' to stop after this iteration) ***" << endl << endl;
    
    /** update the objective function with the current mu values **/
    update_obj_func();
  
    /** solve the Lagrangian subproblem **/
    solveSubProblem(foundSolution, foundOptSolution, subObjValue, subBound);

    if (!foundSolution)
    {
      if (!m_curr_iter)
      {
	cerr << endl << "ERROR: Solving the Lagrangian relaxation failed: Try with a different initial value for the multipliers" << endl << endl;
	return;
      }

      cout << endl << "* Cannot find a solution to the lagrangian subproblem. The current bound is " << m_lag_bound << " *" << endl << endl;

      // update the multipliers and proceed to the next iteration
      update_multipliers(true);
      continue;
    }

    /** compute the bound **/

    fixedTerm = 0.;
    for (vector<lag_constraint>::iterator r=m_relax_constr.begin(); r!=m_relax_constr.end(); r++)
//       fixedTerm -= r->mu * r->rhs;
      fixedTerm -= r->mu * (r->lower == -COIN_DBL_MAX ? r->upper : r->lower);

    subObjValue += fixedTerm;
    subBound += fixedTerm;
    
    cout << endl << (foundOptSolution ? "Optimal" : "Sub-Optimal") << " solution to the Lagrangian subproblem: " << subObjValue << endl;

    if (!foundOptSolution)
      cout << "Bound to the optimal solution of the Lagrangian subproblem: " << subBound << endl << endl;

    // m_curr_lag = subBound;
    m_curr_lag = subObjValue;      // use the value of the (sub-)optimal solution to the lagrangian subproblem to update the multipliers

//     if (!m_have_bound || (is_maxim(m_lp) && subBound < m_lag_bound) || (!is_maxim(m_lp) && subBound > m_lag_bound) )
    if (!m_have_bound || (m_model->optimizationDirection() == -1 && subBound < m_lag_bound) || (m_model->optimizationDirection() == 1 && subBound > m_lag_bound) )
    {
      m_lag_bound = subBound;
//       cout << "which is the new " << (is_maxim(m_lp) ? "upper" : "lower") << " bound for the original problem" << endl << endl;
      cout << "which is the new " << (m_model->optimizationDirection() == -1 ? "upper" : "lower") << " bound for the original problem" << endl << endl;
      m_noimprove = 0;
    }
    else
    {
//       cout << "The " << (is_maxim(m_lp) ? "upper" : "lower") << " bound for the original problem is still " << m_lag_bound << endl << endl;
      cout << "The " << (m_model->optimizationDirection() == -1 ? "upper" : "lower") << " bound for the original problem is still " << m_lag_bound << endl << endl;
      m_noimprove++;
    }

    if (!m_have_bound)
      m_have_bound = true;

    /** a solution has been found, check to see if it is a fesible solution for the original problem **/

    // for this purpose, check whether the relaxed constraints are verified
    feasible = true;
    for (vector<lag_constraint>::iterator r=m_relax_constr.begin(); r!=m_relax_constr.end(); r++)
    {
      sum = 0.;
      for (unsigned int i=0; i<r->row.size(); i++)
	sum += r->row[i] * m_curr_solution[r->colno[i]-1];

//       if ( (r->type==LE && sum > r->rhs) || (r->type==EQ && sum != r->rhs) || (r->type==GE && sum < r->rhs) )
      if (!(sum >= r->lower && sum <= r->upper))
      {
	feasible = false;
	cout << endl << "The solution of the Lagrangian subproblem is NOT feasible for the original problem" << endl << endl;
	break;
      }
    }
    if (feasible)
    {
      cout << endl << "The solution of the Lagrangian subproblem is feasible for the original problem" << endl;
      // compute the objective function of this solution
      sum = inner_product(m_obj_func.begin()+1, m_obj_func.end(), m_curr_solution, 0.);
      
      cout << "The value of this solution is : " << sum << endl;
      
      // if a feasible solution had not been found or this one is better than the previous one
//       if (!m_have_solution || (is_maxim(m_lp) && sum > m_obj_val) || (!is_maxim(m_lp) && sum < m_obj_val) )
      if (!m_have_solution || (m_model->optimizationDirection() == -1 && sum > m_obj_val) || (m_model->optimizationDirection() == 1 && sum < m_obj_val) )
      {
	// save this solution as the best one found so far
	m_obj_val = sum;
// 	copy(m_curr_solution, m_curr_solution+get_Ncolumns(m_lp), m_solution);
	copy(m_curr_solution, m_curr_solution + m_model->getNumCols(), m_solution);
	cout << "and it is the best solution found so far" << endl << endl;
      }
      else
	cout << "The value of the best solution found so far is still " << m_obj_val << endl << endl;

      m_have_solution = true;

      // if a breakValue has been defined, check if this solution is better
//       if (m_hasBreakValue && ((is_maxim(m_lp) && m_obj_val > m_breakValue) || (!is_maxim(m_lp) && m_obj_val < m_breakValue)))
      if (m_hasBreakValue && ((m_model->optimizationDirection() == -1 && m_obj_val > m_breakValue) || (m_model->optimizationDirection() == 1 && m_obj_val < m_breakValue)))
      {
	cout << endl << "A better solution than the required one (" << m_breakValue << ") has been found, hence we stop searching" << endl << endl;
	return;
      }
    }

    /** update the lagrangian multipliers **/
    if (m_noimprove < m_max_noimprove)
      update_multipliers();
    else
    {
      cout << endl << "* " << m_max_noimprove << " iterations without improvements in the lagrangian bound *" << endl << endl;
      update_multipliers(false, true);
      m_noimprove = 0;
    }
    
    if (getchar() == 'c')
    {
      cout << endl << "Stop searching as requested by the user" << endl << endl;
      return;
    }

  }
  //  restore blocking input
  fcntl(fd,F_SETFL,fcflags);
  
}

Lag_Relax::~Lag_Relax ()
{
  delete m_solution;
}


double newton::theta (bool repair)
{
  if (repair)
    m_lambda = m_lambda / 2;

  // compute the norm of Ax-b
  double norm2 = 0.;
  
  for (vector<double>::const_iterator v=m_Ax_b().begin(); v!=m_Ax_b().end(); v++)
    norm2 += pow (*v, 2.);

  if (m_have_solution() && m_obj_val() < m_ub)
    m_ub = m_obj_val();

  return m_lambda * (m_ub - m_curr_lag()) / norm2;
}
