#include "CAalgos.h"
#include "WMutils.h"

#include "linprogs.h"
#include <sstream>
#include <queue>
#include <map>
#include <algorithm>
#include <ext/functional>
#include <numeric>
#include <cassert>
#include <memory>
#include <ClpSimplex.hpp>
#include <CoinBuild.hpp>

#define CHANNEL(l)              \
l.second->radio->channel


/** Minimum Power-Channel Assignment and Routing Algorithm **/

namespace MP_CARA_NS
{

  /* compute the power consumption of a single node */
  float calcNodePowerConsumption(Topology* T, int n)
  {
    vector<TPower> power;
    
    T->calcAllNodesPowerConsumption(power);

    return power.at(n).power;
  }


  /** Priority function objects **/
  
  /* Abstract base class. T is the type of element in the queue */
  template <class T>
  class PrioFuncObj
  {
  public:
    /* returns the priority of an element */
    virtual double priority (const T& x) = 0;
    
    bool operator()(const T& p, const T& q) { return priority(p) < priority(q); }
  };


  /* Priority based on the sum of the utots */
  class SumUtotPrio : public PrioFuncObj<pair<int, set<TLinkItrs> > >
  {
  public:
    SumUtotPrio (Topology* _T, double _u0) : T(_T), u0(_u0) {}
    double priority (const pair<int, set<TLinkItrs> >& p);

  private:
    Topology* T;
    double u0;
  };

  /* The priority of a node is the sum of the costs of the links incident on it.
   * The cost of a link is its f/c ratio multiplied by the number of links
   * with Utot >= u0 having that link in their collision domain */
  double SumUtotPrio::priority (const pair<int, set<TLinkItrs> >& p)
  {
    double ret = 0.;
 
    for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
      for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++) 
	if (y->Utot >= u0)
	  for (set<TLinkItrs>::iterator pp=p.second.begin(); pp!=p.second.end(); pp++)
	    if (pp->IsInTheCollDomainOf(T, TLinkItrs(x,y)))
	      ret += pp->second->flow / pp->second->rate;

    return ret;
  }

  
  /* Node suffer priority */
  class NodeSufferPrio : public PrioFuncObj<pair<int, set<TLinkItrs> > >
  {
  public:
    NodeSufferPrio (vector<TPower>* _consumption, vector<float>* _constraint) : consumption(_consumption), constraint(_constraint) {}
    double priority (const pair<int, set<TLinkItrs> >& x);

  private:
    float NodeSuffer (int n) { float ret = max(consumption->at(n).power/constraint->at(n), 1.f); for (unsigned int i=0; i<consumption->at(n).tau.size(); i++) ret *= max(consumption->at(n).tau[i], 1.f); return ret; }

    vector<TPower>* consumption;
    vector<float>* constraint;
  };
  
  /* compute the "suffering" of a node, which measures the extent to which tau_tx + tau_rx exceeds 1
   * (for each radio) plus the extent to which the power consumption exceeds the power constraint */
  double NodeSufferPrio::priority (const pair<int, set<TLinkItrs> >& x)
  {
    double x_prio = 0.;
    
    for (set<TLinkItrs>::iterator l=x.second.begin(); l!=x.second.end(); l++)
      if (l->first->info->id == x.first)
	x_prio += (l->second->flow * NodeSuffer(l->second->info->id));
      
    return x_prio;
  }


  /* Sum flow priority */
  class SumFlowPrio : public PrioFuncObj<pair<TRadio, set<TLinkItrs> > >
  {
  public:
    SumFlowPrio (double F = 0.) : maxF(F) {}
    double priority (const pair<TRadio, set<TLinkItrs> >& x);

  private:
    double maxF;
  };
  
  /* computes the sum of the flows of all the links in L */
  double SumFlowPrio::priority(const pair<TRadio, set< TLinkItrs > >& x)
  {
    int numRadiosOn = 0;
    
    vector<double> F(x.second.size());

    transform(x.second.begin(), x.second.end(), F.begin(), 
	      __gnu_cxx::compose1(mem_fun(&TGraphNeighbor::getFlow), mem_fun_ref(&TLinkItrs::getNeighb_ptr) )  );

    // compute the number of active radios on the same node as the considered radio
    if (!x.second.empty())
    {
      MeshNode* info = NULL;

      if (x.second.begin()->first->info->id == x.first.first)
	info = x.second.begin()->first->info;

      if (x.second.begin()->second->info->id == x.first.first)
	info = x.second.begin()->second->info;

      if (info)
	numRadiosOn = info->numAvailRadios() - info->countRadiosWithChannel(0);
      else
	cerr << endl << "*** WARNING: SumFlowPrio: The provided link is not incident on the node owning the given radio ***" << endl << endl; 
    }

    return accumulate(F.begin(), F.end(), 0.) + maxF * (numRadiosOn-1);
  }


  /* Radio suffer priority */
  class RadioSufferPrio : public PrioFuncObj<pair<TRadio, set<TLinkItrs> > >
  {
  public:
    RadioSufferPrio (vector<TPower>* _consumption, vector<float>* _constraint) : consumption(_consumption), constraint(_constraint) {}
    double priority (const pair<TRadio, set<TLinkItrs> >& x);

  private:
    vector<TPower>* consumption;
    vector<float>* constraint;
  };

  /* compute the "suffering" of a radio, which measures the extent to which tau_tx + tau_rx exceeds 1
   * for that radio plus the extent to which the power consumption exceeds the power constraint */
  double RadioSufferPrio::priority (const pair<TRadio, set<TLinkItrs> >& x)
  {
    return 1.f / (max(consumption->at(x.first.first).power/constraint->at(x.first.first), 1.f) * max(consumption->at(x.first.first).tau.at(x.first.second), 1.f));
  }

}  /* MP_CARA_NS namespace */


using namespace MP_CARA_NS;


/* constructor */
MPCARA::MPCARA (const list<pair<string,string> >& params)
 : m_npc (NULL),
   m_seed (getSeed()),
   m_reset (false),
   m_fcpra (false),
   m_old (false),
   m_u0 (0.5),
   m_nodepc (NULL)
{
  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    if (!m_npc)
      RP.read ("npc", m_npc);
    RP.read ("seed", m_seed);
    RP.read ("reset", m_reset);
    RP.read ("fcpra", m_fcpra);
    RP.read ("old", m_old);
    RP.read ("u0", m_u0);

    if (RP.notFound())
      cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the MPCARA algorithm" << endl << endl;
  }
}


const char* MPCARA::name()
{
  stringstream ss;
  string s, tmp;

  ss << "MP";

  if (m_reset)
  {
    ss << "R";

    if (m_fcpra)
      ss << "F";
  }

  if (m_old)
    ss << "1";

  tmp = m_npc->getName();

  replace(tmp.begin(), tmp.end(), ' ', '_');

  ss << "_" << tmp << "_" << m_seed << "_" << m_u0;

  ss >> s;

  return s.c_str();
}


double MPCARA::Run (Topology* T)
{
  if (m_reset)
  {
    T->ResetChannels();
    if (!m_fcpra)
    {
      OneChannel oc;
      oc.Run(T);
    }
    else
    {
      FCPRA fcpra(true, true, false);
      fcpra.Run(T);
      //m_u0 = max(T->getUtotMax().second, m_u0);
    }
  }

//   generate the power constraints
  MotherOfAll urng(m_seed);            // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used
  
  m_nodepc = new vector<float>;
  for (unsigned int i=0; i<T->node.size(); i++)
    m_nodepc->push_back(m_npc->rv->Next());
  
  return doRun (T);
}


/* Given a network topology, attempts to minimize the whole power consumption
   while keeping the maximum Utot below u0
   It should be always true that if a link u->v exists then a link v->u exists
   (not necessarily using the same channel)
   Returns 0 if a feasible solution has been found, -1 otherwise
*/
double MPCARA::doRun(Topology *T)
{
  int r;
  double maxUtot=numeric_limits<double>::max();
  
  cout << "Starting MP_CARA" << endl << endl;

  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands;
  
  if (!(T->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: MP_CARA: Checking E2E flows before beginning FAILED" << endl << endl;
  cout << "Demands:" << endl;
  for(map<pair<int,int>,double>::iterator m=demands.begin(); m!=demands.end(); m++)
    cout << "d[" << m->first.first << "," << m->first.second << "] = " << m->second << endl;

  /* ensure that there is a power constraint value for each node */
  if (m_nodepc->size() != T->node.size())
  {
    cerr << endl << "WARNING: MP_CARA: The size of nodePowerConstr does not match the number of nodes" << endl << endl;
//     nodePowerConstr->resize(T->node.size(), T->pm.StandbyPowerConsumption);
    m_nodepc->resize(T->node.size(), T->node[0].info->StandbyPowerConsumption);
  }

  /* find the set of nodes that must be in sleep mode due to power constraints */
  /* and put them in sleep mode (remove their links and turn off their radios) */
  set<int> mustBeOff;
  list<TGraphNeighbor>::iterator nbi;
  for (unsigned int u=0; u<T->node.size(); u++)
  {
    float tmp = T->node[u].info->BasePowerConsumption;

    for (RadioIfaceIterator ri = T->node[u].info->radio.begin(); ri != T->node[u].info->radio.end(); ri++)
      tmp += ri->rho_idle;

//     if (T->pm.BasePowerConsumption + T->pm.rho_idle * T->node[u].info->numAvailRadios() > nodePowerConstr->at(u))
    if (tmp > m_nodepc->at(u))
    {
      // if the node is a source or destination of a demand, there is no feasible solution
      for(map<pair<int,int>,double>::iterator m=demands.begin(); m!=demands.end(); m++)
	if (m->first.first == u || m->first.second == u)
	{
	  cout << endl << "Node " << u << ", which is the src/dst of a demand, has a too strict power constraint."
	       << endl << "A feasible solution cannot be found" << endl << endl;
	  return -1;
	}

      mustBeOff.insert(u);
      // set all the channels to 0
      for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	T->node[u].info->radio[i].channel = 0;
      // erase all the links outgoing node u
      T->node[u].neighbor.clear();
      // erase all the links incoming node u
      for (unsigned int v=0; v<T->node.size(); v++)
	if (!T->node[v].neighbor.empty())
	  while ( (nbi = T->find_neighbor(v,u)) != T->node[v].neighbor.end() )
	    T->node[v].neighbor.erase(nbi);
    }
  }

  /* FIRST STAGE: rearrange e2e flows to minimize the maximum Utot */
  if (MinMaxUtot(T, &demands, maxUtot, EmptyVector, false, m_nodepc))
  {
    // MinMaxUtot returns a non-zero value if the optimal solution has not been found
    // Hence, we are here if the program is not feasible
    // Run MinMaxUtot again without the power constraints
    MinMaxUtot(T, &demands, maxUtot);
    // set maxUtot to infinity again
    maxUtot=numeric_limits<double>::max();
  }

  cout << "Ritornati in MP_CARA dopo MinMaxUtot" << endl;
    
  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: MP_CARA: Checking E2E flows after MinMaxUtot FAILED" << endl << endl;
  
  /* FIRST STAGE: attempt to turn on radios to minimize Utot */

  RadioIfaceIterator ri;

  /* compute a map that associates each node having radios that can be turned on with the set of its links */
  /* nodes belonging to mustBeOff are excluded */
  map<int, set<TLinkItrs> > nodeLinks;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    if (mustBeOff.find(u->info->id) == mustBeOff.end())
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      {
	if (u->info->countRadiosWithChannel(0))
	  nodeLinks[u->info->id].insert(TLinkItrs(u,v));
	if (v->info->countRadiosWithChannel(0))
	  nodeLinks[v->info->id].insert(TLinkItrs(u,v));
      }

  vector<TPower> consumption;

  while (!nodeLinks.empty() && maxUtot > m_u0) {
    
    map<int,set<TLinkItrs> >::iterator r0;
    
    if (maxUtot < numeric_limits<double>::max())
      r0 = max_element(nodeLinks.begin(), nodeLinks.end(), (m_old ? SumUtotPrio(T,m_u0) : SumUtotPrio(T,maxUtot)));
    else
    {
      T->calcAllNodesPowerConsumption(consumption);
      r0 = max_element(nodeLinks.begin(), nodeLinks.end(), NodeSufferPrio(&consumption, m_nodepc));
    }
    
    /* if the priority of the extracted node is null, it means that none of its links
     * is in the collision domain of a link with Utot > u0, hence turning on its radio is useless
     * Exit the loop because all the other nodes will have a zero priority 
     */
//     if (!calcPriority(T,u0,*r0))
//       break;

    /** BEGIN NEW FIRST STAGE **/
    
    map<TLinkItrs, pair<int,int> > lp_var;         // link entering or leaving the extracted node, pair<its position in the LP, count>
    
    if (!m_old)
      for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
	  if (u->info->id == r0->first || v->info->id == r0->first)
	  {
// 	    lp_var[TLinkItrs(u,v)] = pair<int,int>(lp_var.size(),0);
	    lp_var[TLinkItrs(u,v)] = pair<int,int>(lp_var.size()-1,0);
	  
	    for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	      for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
		if (TLinkItrs(u,v).IsInTheCollDomainOf(T,TLinkItrs(x,y)))
		  lp_var[TLinkItrs(u,v)].second++;
	  }
    
    /** END NEW FIRST STAGE **/
    
    
    map<unsigned int, double> S;            // pairs channel, maximum Utot
    map<unsigned int, double>::iterator maxS;
    list<TGraphNeighbor>::iterator v;
    map<unsigned int,map<int,list<TGraphNeighbor> > > addedNeighbors;       // set of neighbors added for each channel
    map<int,set<int> > neighbOn;        // for each channel stores the neighbor where an additional radio has been turned on
    RadioIfaceIterator ptr_radio;

    /* find the channels of the extracted node and avoid to put them in S */
    set<unsigned int> r0Channels = T->node[r0->first].info->findChannels();
//    findChannels((T->node.begin()+r0->first), r0Channels);

    /* for each available channel not in use by the extracted node */
    for (unsigned int c=1; c<=T->numAvailChannels; c++)
      if (r0Channels.find(c) == r0Channels.end())
      {
	/* find all the links that can be added on such channel */
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	  /* for each node u in the network (other than the extracted node) that can be a neighbor */
	  if ((u->info->id != r0->first) && (mustBeOff.find(u->info->id) == mustBeOff.end()) && (r = T->MaxAllowLinkRate(r0->first,u->info->id,T->maxP,T->min_rate,T->max_rate)) >= T->min_rate )
	  {
	    /* if u has a radio on channel c  or an available radio*/
	    if (u->info->countRadiosWithChannel(c) || u->info->countRadiosWithChannel(0))
	    {
	      /* a link on channel c can be established between the extracted node and u,*/
	      addedNeighbors[c][r0->first].push_front(TGraphNeighbor(u->info,availTXrates[r],T->maxP));
	      /* a link on channel c can be established between u and the extracted node */
	      addedNeighbors[c][u->info->id].push_front(TGraphNeighbor(T->node[r0->first].info,availTXrates[r],T->maxP));

	      if (!u->info->countRadiosWithChannel(c))
		neighbOn[c].insert(u->info->id);
	    }
	  }

	int numAddedLinks = 0;
	for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[c].begin(); m!=addedNeighbors[c].end(); m++)
	  numAddedLinks += m->second.size();

	if (!numAddedLinks)
	  continue;          // no new links can be established on this channel

	cout << endl << "provo ad accendere una radio del nodo " << r0->first << " sul canale " << c;

	if (m_old)
	{
	  double meanrate = 0.;
	  
	  S[c] = 0.;
	  
	  /* add the neighbors computed above for the selected channel to the topology */
	  int num = 0;
	  for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[c].begin(); m!=addedNeighbors[c].end(); m++)
	    for (list<TGraphNeighbor>::iterator n=m->second.begin(); n!=m->second.end(); n++, num++)
	    {
	      v = T->node[m->first].neighbor.insert(T->node[m->first].neighbor.end(),*n);
	      T->AssignChannel2Link(TLinkItrs(T->node.begin()+m->first,v), c);
	      
	      S[c] = max(S[c], T->computeUtot(TLinkItrs(T->node.begin()+m->first,v)));
	      meanrate += v->rate;
	    }
	  meanrate = meanrate / num;

	  /* estimate the amount of flow that can be routed on the new links */
	  if (S[c])
	    S[c] = (m_u0-S[c])*meanrate;
  // 	else
  // 	  S[c] = 100000.0;

	  cout << "  S[c] = " << S[c] << endl;

	// Optimal solution (very slow)
// 	double temp_utot=maxUtot;
//	MinMaxUtot(T, &demands, temp_utot, EmptyVector, false, nodePowerConstr);
//	S[c] = 1./temp_utot;
//	
//	cout << "  S[c] = " << temp_utot << endl;

	  /* remove the added links */
	  for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[c].begin(); m!=addedNeighbors[c].end(); m++)
	    for (list<TGraphNeighbor>::iterator n=m->second.begin(); n!=m->second.end(); n++)
	    {
	      v = T->find_neighbor(m->first, n->info->id, c);

	      if (v != T->node[m->first].neighbor.end())
		T->node[m->first].neighbor.erase(v);
	    }

	  /* remove channel c (if it was assigned to) from the channels of the extracted node */
// 	  ptr_radio = find_if(T->node[r0->first].info->radio, T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios,
// 		  __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) );
	  ptr_radio = T->node[r0->first].info->findRadioWithChannel(c);
	  
// 	  if (ptr_radio != T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios)
	  if (ptr_radio != T->node[r0->first].info->radio.end())
	    ptr_radio->channel = 0;
	  
	  for (set<int>::iterator n=neighbOn[c].begin(); n!=neighbOn[c].end(); n++) {
	    /* remove channel c (if it was assigned to) from the channels of the selected neighbors */
// 	    ptr_radio = find_if(T->node[*n].info->radio, T->node[*n].info->radio + T->node[*n].info->numAvailRadios,
// 		    __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) );
	    ptr_radio = T->node[*n].info->findRadioWithChannel(c);
	    
// 	    if (ptr_radio != T->node[*n].info->radio + T->node[*n].info->numAvailRadios)
	    if (ptr_radio != T->node[*n].info->radio.end())
	      ptr_radio->channel = 0;
	  }
	}
	else
	{
	  /** BEGIN NEW FIRST STAGE **/
	  
	  map<TLinkItrs, pair<int,int> > lp_var_c(lp_var);            // link entering or leaving the extracted node, pair<its position in the LP, count>

	  /* add the new links to lp_var_c */
	  for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[c].begin(); m!=addedNeighbors[c].end(); m++)
	    for (list<TGraphNeighbor>::iterator n=m->second.begin(); n!=m->second.end(); n++)
	    {
// 	      lp_var_c[TLinkItrs(T->node.begin()+m->first, n)] = pair<int,int>(lp_var_c.size(),0);
	      lp_var_c[TLinkItrs(T->node.begin()+m->first, n)] = pair<int,int>(lp_var_c.size()-1,0);

	      for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
		for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
		{
		  if (y->radio->channel == c && T->PotInterfLinks(TLinkItrs(T->node.begin()+m->first, n),TLinkItrs(x,y)))
		    lp_var_c[TLinkItrs(T->node.begin()+m->first, n)].second++;
		  // update the Utot of the newly added link
		  if (y->radio->channel == c && T->PotInterfLinks(TLinkItrs(x,y),TLinkItrs(T->node.begin()+m->first, n)))
		    n->Utot += y->flow / y->rate;
		}
	    }

	  /* prepare the LP*/
	  ClpSimplex model;

	  model.resize(0, lp_var_c.size());

	  /* set bounds on variables */
	  for (map<TLinkItrs,pair<int,int> >::iterator m=lp_var_c.begin(); m!=lp_var_c.end(); m++)
	    if (m->first.second->radio)
	      model.setColumnBounds(m->second.first, -m->first.second->flow, 0.);
	    else
	      model.setColumnBounds(m->second.first, 0., m->first.second->rate);

	  int *colno = new int[lp_var_c.size()], *colno1 = new int[lp_var_c.size()];
	  double *coeff = new double[lp_var_c.size()], *coeff1 = new double[lp_var_c.size()];

  // 	for (int i=1; i<=lp_var_c.size(); i++)
  // 	  coeff[i] = 0.;               // reset the elements of coeff

	  for (map<TLinkItrs,pair<int,int> >::iterator l=lp_var_c.begin(); l!=lp_var_c.end(); l++)
	    coeff[l->second.first] = l->second.second / l->first.second->rate;

	  for (map<TLinkItrs,pair<int,int> >::iterator l=lp_var_c.begin(); l!=lp_var_c.end(); l++)
	    coeff[l->second.first] += numAddedLinks / l->first.second->rate;

	  for (unsigned int i=0; i<lp_var_c.size(); i++)
	    model.setObjectiveCoefficient(i, coeff[i]);

	  /* set the object direction to minimize */
	  model.setOptimizationDirection(1);

	  CoinBuild buildObject;

	  int k=0, k1=0;

	  /* constraints to balance the flow moved from old links to new links */
	  for (map<TLinkItrs,pair<int,int> >::iterator l=lp_var_c.begin(); l!=lp_var_c.end(); l++)
	    // outgoing links from the extracted node
	    if (l->first.first->info->id == r0->first)
	    {
	      colno[k] = l->second.first;
	      coeff[k++] = 1.;
	    }
	    else
	    {
	      colno1[k1] = l->second.first;
	      coeff1[k1++] = 1.;
	    }
	  buildObject.addRow(k, colno, coeff, 0., 0.);
	  buildObject.addRow(k1, colno1, coeff1, 0., 0.);

	  model.addRows(buildObject);

	  assert (!model.initialSolve());      // this program should always be feasible

	  S[c] = model.objectiveValue();

	  /*  Before quitting, clear memory*/
	  delete colno;
	  delete colno1;
	  delete coeff;
	  delete coeff1;

	  cout << "  M = " << S[c] << "  **" << endl;
	  
	  /** END NEW FIRST STAGE **/
	}

      }

    if (addedNeighbors.empty())
    {
      nodeLinks.erase(r0);
      continue;
    }

    /* select the channel maximizing S */
    if (m_old)
      maxS = max_element(S.begin(), S.end(), compareSecond<unsigned int,double>());
    else
      maxS = min_element(S.begin(), S.end(), compareSecond<unsigned int,double>());

    cout << "il canale selezionato e' " <<  maxS->first << endl;

    /* add the neighbors computed earlier for the selected channel to the topology before calling MinMaxUtot */
    for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[maxS->first].begin(); m!=addedNeighbors[maxS->first].end(); m++)
      for (list<TGraphNeighbor>::iterator n=m->second.begin(); n!=m->second.end(); n++)
      {
	v = T->node[m->first].neighbor.insert(T->node[m->first].neighbor.end(),*n);

	T->AssignChannel2Link(TLinkItrs(T->node.begin()+m->first,v),maxS->first);

	v->Utot = T->computeUtot(TLinkItrs(T->node.begin()+m->first,v));

	// do not add the new link if its Utot is greater than maxUtot (even with no flow on it)
	// this ensures that maxUtot cannot increase after turning on this radio
	if (v->Utot > maxUtot)
	  T->node[m->first].neighbor.erase(v);
      }

    /* the extracted radio is always left on, unless no link is established with nodes already
       having the selected channel */
    bool extrRadioOn = true;
    
    if (MinMaxUtot(T, &demands, maxUtot, EmptyVector, false, m_nodepc))
    {
#if 0
      /* remove the added links involving nodes in neighbOn */
      set<pair<int,TGraphNeighbor> >::iterator tmp;

      for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); )
      {
	if (neighbOn[maxS->first].find(n->first) != neighbOn[maxS->first].end() || neighbOn[maxS->first].find(n->second.info->id) != neighbOn[maxS->first].end())
	{
	  /* find the added link and remove it */
	  v = T->find_neighbor(n->first, n->second.info->id, maxS->first);

	  if (v != T->node[n->first].neighbor.end())
	    T->node[n->first].neighbor.erase(v);
	  else
	    cerr << "** WARNING: an added link has not been found!" << endl << endl;

	  /* remove the link from addedNeighbors */
	  tmp = n++; 
	  addedNeighbors[maxS->first].erase(tmp);
	}
	else
	  n++;
      }

      /* remove channel c from the neighbors in neighbOn */
      for (set<int>::iterator n=neighbOn[maxS->first].begin(); n!=neighbOn[maxS->first].end(); n++)
      {
	ptr_radio = find_if(T->node[*n].info->radio, T->node[*n].info->radio + T->node[*n].info->numAvailRadios,
		__gnu_cxx::compose1(bind2nd(equal_to<int>(), maxS->first), mem_fun_ref(&RadioIface::getChannel) ) );
	
	if (ptr_radio != T->node[*n].info->radio + T->node[*n].info->numAvailRadios)
	  ptr_radio->channel = 0;
      }

      /* if there are no links left in addedNeighbors, remove channel c from the extracted node, too */
      /* if a feasible solution exists but leaving the radio on the extracted node on violates the power constraint on that node,
         then remove channel c from the extracted node and the remaining added links */
      if (addedNeighbors[maxS->first].empty() || (maxUtot < numeric_limits<REAL>::max() && calcNodePowerConsumption(T,r0->first) > nodePowerConstr->at(r0->first)))
      {
	extrRadioOn = false;

	for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); n++)
	{
	  /* find the added link and remove it */
	  v = T->find_neighbor(n->first, n->second.info->id, maxS->first);

	  if (v != T->node[n->first].neighbor.end())
	    T->node[n->first].neighbor.erase(v);
	  else
	    cerr << "** WARNING: an added link has not been found!" << endl << endl;
	}

	ptr_radio = find_if(T->node[r0->first].info->radio, T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios,
		__gnu_cxx::compose1(bind2nd(equal_to<int>(), maxS->first), mem_fun_ref(&RadioIface::getChannel) ) );
	
	if (ptr_radio != T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios)
	  ptr_radio->channel = 0;

	/* remove the links from addedNeighbors */
	addedNeighbors[maxS->first].clear();
      }
#endif
      /* if a feasible solution has not been found yet and the radio has been left on, then */
      if (!addedNeighbors[maxS->first].empty() && maxUtot == numeric_limits<double>::max())
      {
	// run MinMaxUtot again without the power constraints
	MinMaxUtot(T, &demands, maxUtot);
	// set maxUtot to infinity again
	maxUtot=numeric_limits<double>::max();
      }
    }
    else
      cout << "** Accendendo la radio ho abbassato la max Utot a " << maxUtot << " **" << endl << endl;

    /* add the new links to the nodeLinks map to correctly compute the priority  (the ones remaining
       in addedNeighbors, if any, in case the LP was not feasible) */
    for (map<int,list<TGraphNeighbor> >::iterator m=addedNeighbors[maxS->first].begin(); m!=addedNeighbors[maxS->first].end(); m++)
      for (list<TGraphNeighbor>::iterator n=m->second.begin(); n!=m->second.end(); n++)
	if ( (v=T->find_neighbor(m->first,n->info->id,maxS->first)) != T->node[m->first].neighbor.end() )
	{
	  if ( nodeLinks.find(m->first) != nodeLinks.end() )
	    nodeLinks[m->first].insert(TLinkItrs(T->node.begin()+m->first, v));

	  if ( nodeLinks.find(n->info->id) != nodeLinks.end() )
	    nodeLinks[n->info->id].insert(TLinkItrs(T->node.begin()+m->first, v));
	}

    /* if the radio has been turned on and there are other radios on the extracted node that can be turned on,
     * leave the node in the queue, so that the other radios can be turned on */
    if ( !extrRadioOn || !(T->node[r0->first].info->countRadiosWithChannel(0)) )
      nodeLinks.erase(r0);

    /* remove from the queue the neighbors that no more have available radios */
    if (extrRadioOn)
      for (set<int>::iterator n=neighbOn[maxS->first].begin(); n!=neighbOn[maxS->first].end(); n++)
	if (!T->node[*n].info->countRadiosWithChannel(0) && (nodeLinks.find(*n)!=nodeLinks.end()))
	  nodeLinks.erase(*n);
  }


  /* Ok, we tried our best to keep Utot <= u0 for every link 
     Now, let's try to switch some radio off, while not 
     introducing any other violation to the constraint Utot <= u0 */

  cout << endl << "** INIZIO SECOND STAGE **" << endl << endl;


  map<TRadio, set<TLinkItrs> > Lr;  // maps each radio to the set of links using it

  /* compute Lr for all the radios */
  for (unsigned int i=0; i<T->node.size(); i++)
    for (list<TGraphNeighbor>::iterator v=T->node[i].neighbor.begin(); v!=T->node[i].neighbor.end(); v++) {
//       ch = SearchChannel(v->info, v->radio->channel);
      ri = v->info->findRadioWithChannel(v->radio->channel);
      /* insert the link in the sets of the two radios at the endpoints of the link */
      Lr[TRadio(i,distance(T->node[i].info->radio.begin(),RadioIfaceIterator(v->radio)))].insert(TLinkItrs(T->node.begin()+i,v));
//       if (ch.first)
      if (ri != v->info->radio.end())
	Lr[TRadio(v->info->id, distance(v->info->radio.begin(),ri))].insert(TLinkItrs(T->node.begin()+i,v));
    }

  if (m_old)
  {
    map<TRadio, set<TLinkItrs> >::iterator minLr;
    double UtotOpt = maxUtot;
    double Ptot = numeric_limits<double>::max();

    if (maxUtot < numeric_limits<double>::max())
      Ptot = T->TotalPowerConsumption();    

    while (!Lr.empty()) {
      
      /* DEBUG */
  //     for (map<TRadio, set<TLinkItrs> >::iterator r=Lr.begin(); r!=Lr.end(); r++) {
  //       cout << "Link che usano la radio del nodo " << r->first.first << " settata sul canale " << T->node[r->first.first].info->radio[r->first.second].channel << endl;
  //       for (set<TLinkItrs>::iterator l=r->second.begin(); l!=r->second.end(); l++)
  // 	cout << l->first->info->id << "->" << l->second->info->id << "[" << l->second->radio->channel << "]" << endl;
  //       cout << endl;
  //     }

      /* extract the pair (node,radio) with the minimum flow */
      double minP = -1.;   // flow routed on the links using the extracted radio
      
      if (maxUtot < numeric_limits<double>::max())
	minLr = min_element(Lr.begin(), Lr.end(), SumFlowPrio());
      else
      {
	T->calcAllNodesPowerConsumption(consumption);
	minLr = min_element(Lr.begin(), Lr.end(), RadioSufferPrio(&consumption, m_nodepc));
      }

      SumFlowPrio sfp;
      minP = sfp.priority(*minLr);
      
      cout << "** Estraggo nodo " << minLr->first.first << " radio " << minLr->first.second << "(channel " << T->node[minLr->first.first].info->radio[minLr->first.second].channel << ") flusso " << minP << " **" << endl;

      if (UtotOpt < numeric_limits<double>::max())
	maxUtot = UtotOpt;// * 1.005;

      /* if TurnOffRadio is feasible, the radio can be turned off and all the links in Lr can be removed from the topology */
  //    if (!minP || TurnOffRadio(T, &demands, &(minLr->second), u0)) {
      if (!TurnOffRadio(T, &demands, minLr->first, m_u0, maxUtot, Ptot, m_nodepc))
      {
	cout << "** EUREKA ** sto spegnendo sul nodo " << minLr->first.first << " la radio " << minLr->first.second << "(channel " << T->node[minLr->first.first].info->radio[minLr->first.second].channel << ")  **" << endl;
	/* for all the links in minLr */
	for (set<TLinkItrs>::iterator l=minLr->second.begin(); l!=minLr->second.end(); l++)
	{
	  if (l->second->flow)
	    cerr << "WARNING: TurnOffRadio mi sta facendo eliminare un link con flusso non nullo!" << endl;
	  /* first, remove the link from the Lr set of the other radio (not the extracted one), if it is still in Lr */
// 	  ch = SearchChannel(l->second->info, l->second->radio->channel);
	  ri = l->second->info->findRadioWithChannel(l->second->radio->channel);
	  if (minLr->first.first == l->second->info->id)
	  {
	    if (Lr.find(TRadio(l->first->info->id, distance(l->first->info->radio.begin(),RadioIfaceIterator(l->second->radio)))) != Lr.end())
	      Lr[TRadio(l->first->info->id, distance(l->first->info->radio.begin(),RadioIfaceIterator(l->second->radio)))].erase(*l);
	  }
	  else if (ri != l->second->info->radio.end())
	  {
	    if (Lr.find(TRadio(l->second->info->id, distance(l->second->info->radio.begin(),ri))) != Lr.end())
	      Lr[TRadio(l->second->info->id, distance(l->second->info->radio.begin(),ri))].erase(*l);
	  }
	  /* then, remove the link from the topology*/
	  T->node[l->first->info->id].neighbor.erase(l->second);
	}
	/* switch off the radio (set channel to 0) */
	T->node[minLr->first.first].info->radio[minLr->first.second].channel = 0;
      }
      
      /* remove the whole Lr set of the extracted radio */
      Lr.erase(minLr);

      // T->printTopology();  // DEBUG
    }

    /* erase links with null flow */
    //T->eraseLinksWithNullFlow();
  }
  else
  {

    double maxF = 0.;
    for (map<pair<int,int>,double>::iterator d = demands.begin(); d != demands.end(); d++)
      maxF = max (maxF, d->second);

    vector<ArcEpsY> epsy;

    // initialize the ArcEpsY structure
    for (vector<TGraphNode>::iterator u = T->node.begin(); u != T->node.end(); u++)
      epsy.push_back(ArcEpsY(maxF/2, 1, u->info->numAvailRadios()));

    bool stop;
    set<TRadio> AdjustEpsRadioList;
    set<int> AdjustEpsNodeList;
    float alpha = 0.5;

    do
    {
      // reset y_i
      for (vector<ArcEpsY>::iterator aey = epsy.begin(); aey != epsy.end(); aey++)
      {
	aey->yN = 1;
	fill (aey->yR.begin(), aey->yR.end(), 1);
      }
      
      do
      {
	// solve MinPower_eps
	MinPower_eps(T, &demands, max(m_u0,maxUtot), m_nodepc, epsy);

	stop = true;
	AdjustEpsRadioList.clear();
	AdjustEpsNodeList.clear();
	
	// update y_i while checking if something changed
	for (unsigned int u=0; u<T->node.size(); u++)
	  if (T->node[u].info->countRadiosWithChannel(0) != T->node[u].info->numAvailRadios())
	  {
	    double nodeflow = 0.;

	    /* for each active radio */
	    for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	      if (T->node[u].info->radio[i].channel)
	      {
		double radioflow = 0.;
		if (Lr.find(TRadio(u,i)) != Lr.end())
		  for (set<TLinkItrs>::iterator l = Lr[TRadio(u,i)].begin(); l != Lr[TRadio(u,i)].end(); l++)
		    radioflow += l->second->flow;
		  
		int yi = (radioflow < epsy.at(u).epsR.at(i) ? 0 : 1);

		if (yi != epsy.at(u).yR.at(i))
		{
		  epsy.at(u).yR.at(i) = yi;
		  stop = false;
		}

		if (0. < radioflow && radioflow < epsy.at(u).epsR.at(i))
		  AdjustEpsRadioList.insert(TRadio(u,i));

		nodeflow += radioflow;
	      }

	    int yi = (nodeflow < epsy.at(u).epsN ? 0 : 1);

	    if (yi != epsy.at(u).yN)
	    {
	      epsy.at(u).yN = yi;
	      stop = false;
	    }
	    
	    if (0. < nodeflow && nodeflow < epsy.at(u).epsN)
	      AdjustEpsNodeList.insert(u);
	  }
      } while (!stop);

      for (set<TRadio>::iterator r = AdjustEpsRadioList.begin(); r != AdjustEpsRadioList.end(); r++)
	epsy.at(r->first).epsR.at(r->second) *= alpha;

      for (set<int>::iterator n = AdjustEpsNodeList.begin(); n != AdjustEpsNodeList.end(); n++)
	epsy.at(*n).epsN *= alpha;

      stop = (AdjustEpsNodeList.empty() && AdjustEpsRadioList.empty());

    } while (!stop);
    

    /* remove links with null flow and turn off radios */
    for (map<TRadio, set<TLinkItrs> >::iterator lr = Lr.begin(); lr != Lr.end(); lr++)
    {
      double sumflow = 0.;

      /* for each link incoming to or outgoing from a radio */
      for (set<TLinkItrs>::iterator l=lr->second.begin(); l!=lr->second.end(); l++)
      {
	/* if it carries no flow */
	if (!l->second->flow)
	{
	  /* first, remove the link from the Lr set of the other radio, if it is still in Lr */
// 	  ch = SearchChannel(l->second->info, l->second->radio->channel);
	  ri = l->second->info->findRadioWithChannel(l->second->radio->channel);
	  if (lr->first.first == l->second->info->id)
	  {
	    if (Lr.find(TRadio(l->first->info->id, distance(l->first->info->radio.begin(),RadioIfaceIterator(l->second->radio)))) != Lr.end())
	      Lr[TRadio(l->first->info->id, distance(l->first->info->radio.begin(),RadioIfaceIterator(l->second->radio)))].erase(*l);
	  }
	  else if (ri != l->second->info->radio.end())
	  {
	    if (Lr.find(TRadio(l->second->info->id, distance(l->second->info->radio.begin(),ri))) != Lr.end())
	      Lr[TRadio(l->second->info->id, distance(l->second->info->radio.begin(),ri))].erase(*l);
	  }
	  /* then, remove the link from the topology*/
	  T->node[l->first->info->id].neighbor.erase(l->second);
	}
	else
	  sumflow += l->second->flow;
      }

      /* if sumflow is null, the radio can be turned off */
      if (!sumflow)
	T->node[lr->first.first].info->radio[lr->first.second].channel = 0;
    }

  }

  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: MP_CARA: Checking E2E flows at the end FAILED" << endl << endl;

  if (maxUtot < numeric_limits<double>::max())
    return 0;

  return -1;
}



/* constructor */
MPCARA_OPT::MPCARA_OPT (const list<pair<string,string> >& params)
 : m_npc (NULL),
   m_seed (getSeed()),
   m_u0 (0.5),
   m_to1 (300),
   m_to2 (300),
   m_nodepc (NULL)
{
  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    if (!m_npc)
      RP.read ("npc", m_npc);
    RP.read ("seed", m_seed);
    RP.read ("u0", m_u0);
    RP.read ("to1", m_to1);
    RP.read ("to2", m_to2);

    if (RP.notFound())
      cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the MPCARA_OPT algorithm" << endl << endl;
  }
}


const char* MPCARA_OPT::name()
{
  stringstream ss;
  string s, tmp;

  ss << "MPO_";

  tmp = m_npc->getName();

  replace(tmp.begin(), tmp.end(), ' ', '_');

  ss << "_" << tmp << "_" << m_seed << "_" << m_u0 << "_" << m_to1 << "_" << m_to2;

  ss >> s;

  return s.c_str();
}

/*
 * Solves the minimum power CA and routing problem
 * Returns 0 if a feasible solution has been found, -1 otherwise
*/
double MPCARA_OPT::Run (Topology* T)
{
  MotherOfAll urng(m_seed);            // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used

  m_nodepc = new vector<float>;
  for (unsigned int i=0; i<T->node.size(); i++)
    m_nodepc->push_back(m_npc->rv->Next());

  double maxUtot;
  set<int> S;
  double mu_0 = 1.0;      // initial value of the lagrangian multipliers
  int max_iter = 1;
  // long timeout = 150;

  // parameters for the newton's method
  double lambda = 2.;
  double ub = 90;     // this should be the objective function value of a feasible solution
  
  /* clone the topology because MinMaxUtot_MILP removes links with null flow */
  //Topology* U = new Topology(*T);

  MinMaxUtot_MILP mmu(T, m_nodepc, m_to1, false);

  mmu.set_breakValue(m_u0);
  
  if (mmu.lp_solve())
    return -1;

//   S.insert(6);
// //  S.insert(4);
// 
//   if (!mmu.lag_solve (S, mu_0, max_iter, new newton(2.,2.)))
//   {
//     cerr << endl << "Error: MinMaxUtot_MILP did not find a feasible solution" << endl << endl;
//     return -1;
//   }
// 
   maxUtot = mmu.getObjValue();
//maxUtot=0.5;
  cout << endl << "Minimum maxUtot found by MinMaxUtot_MILP: " << maxUtot << endl << endl;

  MinPower_MILP mpm(T, max(maxUtot,m_u0), m_nodepc, m_to2);

  // 1-2-7 2-3 / (4-)6 for bound
//  S.clear();
  //S.insert(3);
//   S.insert(4);
//   S.insert(6);
//   S.insert(7);
  
//  if (mpm.lp_solve() == INFEASIBLE)
//  mpm.lag_solve (S, mu_0, max_iter, new newton(lambda,ub), CoinCbc);

  if (mpm.lp_solve())
    return -1;
  
  m_lbound = mpm.getBound();
  for (vector<TGraphNode>::iterator u = T->node.begin(); u != T->node.end(); u++)
    m_lbound += u->info->StandbyPowerConsumption;
  
  //printCAStats(T,"MPCARA_OPT",true);
  return 0;
}

