#include "WMtypes.h"
#include "CAalgos.h"
#include "WMutils.h"
#include "linprogs.h"
#include <sstream>
#include <queue>
#include <map>
#include <algorithm>
#include <ext/functional>
#include <numeric>
#include <cassert>
#include <memory>

#define CHANNEL(l)              \
l.second->radio->channel


namespace PACA2_NS
{

  double calcPriority(Topology* T, double u0, const pair<int, set<TLinkItrs> >& p) {
    double ret=0;
    for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
      for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++) 
	if (y->Utot >= u0)
	  for (set<TLinkItrs>::iterator pp=p.second.begin(); pp!=p.second.end(); pp++)
	    if (pp->IsInTheCollDomainOf(T, TLinkItrs(x,y))) 
	      ret += pp->second->flow / pp->second->rate;
      
    return ret;
  } 

  /* Binary predicate used to compare the priorities of nodes
  * The priority of a node is the sum of the costs of the links incident on it
  * The cost of a link is its f/c ratio multiplied by the number of links
  * with Utot >= u0 having that link in their collision domain
  */
  class SumUtotPriority {
      Topology* T;
      double u0;
    public:
      SumUtotPriority(Topology* _T, double _u0) : T(_T), u0(_u0) {}
      bool operator()(const pair<int, set<TLinkItrs> >& p, const pair<int, set<TLinkItrs> >& q) {
	return calcPriority(T,u0,p) < calcPriority(T,u0,q);
      }
  };

  /* computes the sum of the flows of all the links in L */
  double sumFlows(set<TLinkItrs> L) {
    vector<double> F(L.size());
    
    transform(L.begin(), L.end(), F.begin(), __gnu_cxx::compose1(mem_fun(&TGraphNeighbor::getFlow),
								mem_fun_ref(&TLinkItrs::getNeighb_ptr) )  );
    return accumulate(F.begin(), F.end(), 0.);
  }
}  /* PACA2_NS namespace */


using namespace PACA2_NS;


/* constructor */
PACA2::PACA2 (const list<pair<string,string> >& params)
 : m_reset (false),
   m_fcpra (false),
   m_u0 (0.5)
{
  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("reset", m_reset);
    RP.read ("fcpra", m_fcpra);
    RP.read ("u0", m_u0);

    if (RP.notFound())
      cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the PACA2 algorithm" << endl << endl;
  }
}

const char* PACA2::name()
{
  stringstream ss;
  string s;

  ss << "PACA2";

  if (m_reset)
  {
    ss << "R";

    if (m_fcpra)
      ss << "F";
  }

  ss >> s;

  return s.c_str();
}


/* Given a network topology, attempts to switch off as many radios as possible
   while keeping the maximum Utot below u0
   It should be always true that if a link u->v exists then a link v->u exists
   (not necessarily using the same channel)
*/
double PACA2::Run (Topology *T)
{
  if (m_reset)
  {
    T->ResetChannels();
    if (!m_fcpra)
    {
      OneChannel oc;
      oc.Run(T);
    }
    else
    {
      FCPRA fcpra(true, true, false);
      fcpra.Run(T);
      m_u0 = max(T->getUtotMax().second, m_u0);
    }
  }

  int r;
  double maxUtot=numeric_limits<double>::max(), prevUtot;
  
  cout << "Starting PACA2" << endl << endl;

#if 0
  set<unsigned int> Ne,U,V,Ccom,Ctemp;
  
  /* FIRST STAGE: attempt to add parallel links */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++) {
    /* find the set n of neighbors of u such that id(u) > id(v) (so that the pair is considered just once) */
    Ne.clear();
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      if (v->info->id < u->info->id)
	Ne.insert(v->info->id);
      
    /* for each neighbor */
    for (set<unsigned int>::iterator n=Ne.begin(); n!=Ne.end(); n++) {
      findChannels(u,U);
      findChannels((T->node.begin()+(*n)),V);
      Ccom.clear();
      set_intersection(U.begin(), U.end(), V.begin(), V.end(), inserter(Ccom, Ccom.begin()));
      
      /* find common channels that are not already used for a link u->n */
      Ctemp = Ccom;
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
	if (v->info->id == *n)
	  Ctemp.erase(v->radio->channel);

      if (canBeNeighbors(u,T->node.begin()+(*n),r,p))
	for (set<unsigned int>::iterator ch=Ctemp.begin(); ch!=Ctemp.end(); ch++) {
	  /* establish a parallel link u->n on ch */
	  u->neighbor.push_front(TGraphNeighbor((T->node.begin()+(*n))->info, r, p));
	  TLinkItrs lnew(u, u->neighbor.begin());
	  T->AssignChannel2Link(lnew,*ch);
	  lnew.second->Utot = T->computeUtot(lnew);
	  cout << "Aggiunto link (" << lnew.first->info->id << "," << lnew.second->info->id << ")[" << lnew.second->radio->channel << "]" << endl;
	}

      /* find common channels that are not already used for a link n->u */
      Ctemp = Ccom;
      for (list<TGraphNeighbor>::iterator v=T->node[*n].neighbor.begin(); v!=T->node[*n].neighbor.end(); v++)
	if (v->info->id == u->info->id)
	  Ctemp.erase(v->radio->channel);

      if (canBeNeighbors(T->node.begin()+(*n),u,r,p))
	for (set<unsigned int>::iterator ch=Ctemp.begin(); ch!=Ctemp.end(); ch++) {
	  /* establish a parallel link n->u on ch */
	  T->node[*n].neighbor.push_front(TGraphNeighbor(u->info, r, p));
	  TLinkItrs lnew(T->node.begin()+(*n), T->node[*n].neighbor.begin());
	  T->AssignChannel2Link(lnew,*ch);
	  lnew.second->Utot = T->computeUtot(lnew);
	  cout << "Aggiunto link (" << lnew.first->info->id << "," << lnew.second->info->id << ")[" << lnew.second->radio->channel << "]" << endl;
	}
    }
  }
#endif  
  
  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands;
  
  if (!(T->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: PACA2: Checking E2E flows before beginning FAILED" << endl << endl;
  cout << "Demands:" << endl;
  for(map<pair<int,int>,double>::iterator m=demands.begin(); m!=demands.end(); m++)
    cout << "d[" << m->first.first << "," << m->first.second << "] = " << m->second << endl;

  /* FIRST STAGE: rearrange e2e flows to minimize the maximum Utot */
  MinMaxUtot(T,&demands,maxUtot);

  cout << "Ritornati in PACA2 dopo MinMaxUtot" << endl;
    
  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: PACA2: Checking E2E flows after MinMaxUtot FAILED" << endl << endl;
  
  /* FIRST STAGE: attempt to turn on radios to minimize Utot */

  RadioIfaceIterator ri;
  
  /* compute a map that associates each node having radios that can be turned on with the set of its links */
  map<int, set<TLinkItrs> > nodeLinks;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      if (u->info->countRadiosWithChannel(0))
	nodeLinks[u->info->id].insert(TLinkItrs(u,v));
      if (v->info->countRadiosWithChannel(0))
	nodeLinks[v->info->id].insert(TLinkItrs(u,v));
    }
   

  while (!nodeLinks.empty() && maxUtot > m_u0) {
    
    map<int,set<TLinkItrs> >::iterator r0 = max_element(nodeLinks.begin(), nodeLinks.end(), SumUtotPriority(T,m_u0));
    
    /* if the priority of the extracted node is null, it means that none of its links
     * is in the collision domain of a link with Utot > u0, hence turning on its radio is useless
     * Exit the loop because all the other nodes will have a zero priority 
     */
//     if (!calcPriority(T,u0,*r0))
//       break;

    
    map<unsigned int, double> S;            // pairs channel, maximum Utot
    map<unsigned int, double>::iterator maxS;
    list<TGraphNeighbor>::iterator v;
    map<unsigned int,set<pair<int,TGraphNeighbor> > > addedNeighbors;       // set of neighbors added for each channel
    map<int,set<int> > neighbOn;        // for each channel stores the neighbor where an additional radio has been turned on
    RadioIfaceIterator ptr_radio;

    /* find the channels of the extracted node and avoid to put them in S */
    set<unsigned int> r0Channels = T->node[r0->first].info->findChannels();
//     findChannels((T->node.begin()+r0->first), r0Channels);

    /* for each available channel not in use by the extracted node */
    for (unsigned int c=1; c<=T->numAvailChannels; c++)
      if (r0Channels.find(c) == r0Channels.end()) {
	
	double meanrate = 0.;
	
	S[c] = 0.;
	
	cout << "provo ad accendere una radio del nodo " << r0->first << " sul canale " << c << endl;
	
	/* find all the links that can be added on such channel */
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	  /* for each node u in the network (other than the extracted node) that can be a neighbor */
	  //if ((u->info->id != r0->first) && T->canBeNeighbors(T->node.begin()+r0->first,u,r,p,T->maxP+1) ) {
	  if ((u->info->id != r0->first) && (r = T->MaxAllowLinkRate(r0->first,u->info->id,T->maxP,T->min_rate,T->max_rate)) >= T->min_rate ) {
	    /* if u has a radio on channel c  or an available radio*/
	    if (u->info->countRadiosWithChannel(c) || u->info->countRadiosWithChannel(0)) {
	      /* a link on channel c can be established between the extracted node and u,*/
	      addedNeighbors[c].insert(pair<int,TGraphNeighbor>(r0->first,TGraphNeighbor(u->info,availTXrates[r],T->maxP)));
	      /* a link on channel c can be established between u and the extracted node */
	      addedNeighbors[c].insert(pair<int,TGraphNeighbor>(u->info->id,TGraphNeighbor(T->node[r0->first].info,availTXrates[r],T->maxP)));

	      if (!u->info->countRadiosWithChannel(c))
		neighbOn[c].insert(u->info->id);
	    }
	  }

	/* add the neighbors computed above for the selected channel to the topology */
	for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[c].begin(); n!=addedNeighbors[c].end(); n++) {
	  v = T->node[n->first].neighbor.insert(T->node[n->first].neighbor.end(),n->second);
	  T->AssignChannel2Link(TLinkItrs(T->node.begin()+n->first,v), c);
	  
	  S[c] = max(S[c], T->computeUtot(TLinkItrs(T->node.begin()+n->first,v)));
	  meanrate += v->rate;
	}
	meanrate = meanrate / addedNeighbors[c].size();

	/* estimate the amount of flow that can be routed on the new links */
	if (S[c])
	  S[c] = (m_u0-S[c])*meanrate;

	/* remove the added links */
	for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[c].begin(); n!=addedNeighbors[c].end(); n++) {
	  v = T->find_neighbor(n->first, n->second.info->id, c);

	  if (v != T->node[n->first].neighbor.end())
	    T->node[n->first].neighbor.erase(v);
	}

	/* remove channel c (if it was assigned to) from the channels of the extracted node */
// 	ptr_radio = find_if(T->node[r0->first].info->radio, T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios,
// 		__gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) );
	
	ptr_radio = T->node[r0->first].info->findRadioWithChannel(c);
	
// 	if (ptr_radio != T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios)
	if (ptr_radio != T->node[r0->first].info->radio.end())
	  ptr_radio->channel = 0;
	
	for (set<int>::iterator n=neighbOn[c].begin(); n!=neighbOn[c].end(); n++) {
	  /* remove channel c (if it was assigned to) from the channels of the selected neighbors */
// 	  ptr_radio = find_if(T->node[*n].info->radio, T->node[*n].info->radio + T->node[*n].info->numAvailRadios,
// 		  __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) );
	  ptr_radio = T->node[*n].info->findRadioWithChannel(c);
	  
// 	  if (ptr_radio != T->node[*n].info->radio + T->node[*n].info->numAvailRadios)
	  if (ptr_radio != T->node[*n].info->radio.end())
	    ptr_radio->channel = 0;
	}
	  
      }

    if (addedNeighbors.empty()) {
      nodeLinks.erase(r0);
      continue;
    }

    /* select the channel maximizing S */
    maxS = max_element(S.begin(), S.end(), compareSecond<unsigned int,double>());

    cout << "il canale selezionato e' " <<  maxS->first << endl;

    /* add the neighbors computed earlier for the selected channel to the topology before calling MinMaxUtot */
    for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); n++) {
      v = T->node[n->first].neighbor.insert(T->node[n->first].neighbor.end(),n->second);
      
      T->AssignChannel2Link(TLinkItrs(T->node.begin()+n->first,v),maxS->first);
      
      v->Utot = T->computeUtot(TLinkItrs(T->node.begin()+n->first,v));
    }
    
    bool extrRadioOn;
    
    prevUtot = maxUtot;
    MinMaxUtot(T,&demands,maxUtot);

    if (maxUtot < prevUtot) {
      extrRadioOn = true;

      cout << "** Accendendo la radio ho abbassato la max Utot a " << maxUtot << " **" << endl << endl;
      
      /* add the new links to the nodeLinks map to correctly compute the priority */
      for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); n++)
	if ( (v=T->find_neighbor(n->first,n->second.info->id,maxS->first)) != T->node[n->first].neighbor.end() ) {
	
	  if ( nodeLinks.find(n->first) != nodeLinks.end() )
	    nodeLinks[n->first].insert(TLinkItrs(T->node.begin()+n->first, v));

	  if ( nodeLinks.find(n->second.info->id) != nodeLinks.end() )
	    nodeLinks[n->second.info->id].insert(TLinkItrs(T->node.begin()+n->first, v));
	}
    }
    else {
      extrRadioOn = false;
      /* the extracted radio can be turned off and the new links removed */
      for (set<pair<int,TGraphNeighbor> >::iterator n=addedNeighbors[maxS->first].begin(); n!=addedNeighbors[maxS->first].end(); n++) {
	/* find the added link and remove it */
	v = T->find_neighbor(n->first, n->second.info->id, maxS->first);

	if (v != T->node[n->first].neighbor.end())
	  T->node[n->first].neighbor.erase(v);
	else
	  cerr << "** WARNING: an added link has not been found!" << endl << endl;
      }
      /* remove channel c from the channels of the extracted node */
//       ptr_radio = find_if(T->node[r0->first].info->radio.begin(), T->node[r0->first].info->radio.end(),
// 	      __gnu_cxx::compose1(bind2nd(equal_to<int>(), maxS->first), mem_fun_ref(&RadioIface::getChannel) ) );
      ptr_radio = T->node[r0->first].info->findRadioWithChannel(maxS->first);
      
//       if (ptr_radio != T->node[r0->first].info->radio + T->node[r0->first].info->numAvailRadios)
      if (ptr_radio != T->node[r0->first].info->radio.end())
	ptr_radio->channel = 0;

      /* if turnOnNeighb, remove channel c from the channels of the neighbor */
	for (set<int>::iterator n=neighbOn[maxS->first].begin(); n!=neighbOn[maxS->first].end(); n++) {
// 	  ptr_radio = find_if(T->node[*n].info->radio, T->node[*n].info->radio + T->node[*n].info->numAvailRadios,
// 		  __gnu_cxx::compose1(bind2nd(equal_to<int>(), maxS->first), mem_fun_ref(&RadioIface::getChannel) ) );
	  ptr_radio = T->node[*n].info->findRadioWithChannel(maxS->first);
	  
// 	  if (ptr_radio != T->node[*n].info->radio + T->node[*n].info->numAvailRadios)
	  if (ptr_radio != T->node[*n].info->radio.end())
	    ptr_radio->channel = 0;
	}
    }

    /* if the radio has been turned on and there are other radios on the extracted node that can be turned on,
     * leave the node in the queue, so that the other radios can be turned on */
    if ( !extrRadioOn || !(T->node[r0->first].info->countRadiosWithChannel(0)) )
      nodeLinks.erase(r0);
    
    /* if turnOnNeighb, if the neighbor has no more available radios, remove it from the queue */
    if (extrRadioOn)
      for (set<int>::iterator n=neighbOn[maxS->first].begin(); n!=neighbOn[maxS->first].end(); n++)
	if (!T->node[*n].info->countRadiosWithChannel(0) && (nodeLinks.find(*n)!=nodeLinks.end()))
	  nodeLinks.erase(*n);
  }
  

  /* Ok, we tried our best to keep Utot <= u0 for every link 
     Now, let's try to switch some radio off, while not 
     introducing any other violation to the constraint Utot <= u0 */

  cout << endl << "** INIZIO SECOND STAGE **" << endl << endl;
  
  
  map<TRadio, set<TLinkItrs> > Lr;  // maps each radio to the set of links using it
  map<TRadio, set<TLinkItrs> >::iterator minLr;
  
      
  /* compute Lr for all the radios */
  for (unsigned int i=0; i<T->node.size(); i++)
    for (list<TGraphNeighbor>::iterator v=T->node[i].neighbor.begin(); v!=T->node[i].neighbor.end(); v++)
    {
//       ch = SearchChannel(v->info, v->radio->channel);
      ri = v->info->findRadioWithChannel(v->radio->channel);
      /* insert the link in the sets of the two radios at the endpoints of the link */
      Lr[TRadio(i,distance(T->node[i].info->radio.begin(),RadioIfaceIterator(v->radio)))].insert(TLinkItrs(T->node.begin()+i,v));
      if (ri != v->info->radio.end())
	Lr[TRadio(v->info->id, distance(v->info->radio.begin(),ri))].insert(TLinkItrs(T->node.begin()+i,v));
    }


  while (!Lr.empty()) {
    
    /* DEBUG */
    for (map<TRadio, set<TLinkItrs> >::iterator r=Lr.begin(); r!=Lr.end(); r++) {
      cout << "Link che usano la radio del nodo " << r->first.first << " settata sul canale " << T->node[r->first.first].info->radio[r->first.second].channel << endl;
      for (set<TLinkItrs>::iterator l=r->second.begin(); l!=r->second.end(); l++)
	cout << l->first->info->id << "->" << l->second->info->id << "[" << l->second->radio->channel << "]" << endl;
      cout << endl;
    }

    /* extract the pair (node,radio) with the minimum flow */
    minLr = Lr.begin();
    double minP = sumFlows(minLr->second), temp;
    
    for (map<TRadio, set<TLinkItrs> >::iterator i=Lr.begin(); i!=Lr.end(); i++)
      if ( (temp=sumFlows(i->second)) < minP ) {
	minP = temp;
	minLr = i;
      }
    
    cout << "** Estraggo nodo " << minLr->first.first << " radio " << minLr->first.second << "(channel " << T->node[minLr->first.first].info->radio[minLr->first.second].channel << ") flusso " << minP << " **" << endl;
    

    /* if TurnOffRadio returns true, the radio can be turned off and all the links in Lr can be removed from the topology */
    if (!minP || TurnOffRadio(T, &demands, &(minLr->second), m_u0)) {
//    if (TurnOffRadio(T, &demands, &(Lr[minQr->first]), u0)) {
      cout << "** EUREKA ** sto spegnendo sul nodo " << minLr->first.first << " la radio " << minLr->first.second << "(channel " << T->node[minLr->first.first].info->radio[minLr->first.second].channel << ")  **" << endl;
      /* for all the links in minLr */
      for (set<TLinkItrs>::iterator l=minLr->second.begin(); l!=minLr->second.end(); l++) {
	if (l->second->flow)
	  cerr << "WARNING: TurnOffRadio mi sta facendo eliminare un link con flusso non nullo!" << endl;
	/* first, remove the link from the Lr set of the other radio (not the extracted one), if it is still in Lr */
// 	ch = SearchChannel(l->second->info, l->second->radio->channel);
	ri = l->second->info->findRadioWithChannel(l->second->radio->channel);
	if (minLr->first.first == l->second->info->id) {
	  if (Lr.find(TRadio(l->first->info->id, distance(l->first->info->radio.begin(),RadioIfaceIterator(l->second->radio)))) != Lr.end())
	    Lr[TRadio(l->first->info->id, distance(l->first->info->radio.begin(),RadioIfaceIterator(l->second->radio)))].erase(*l);
	}
	else if (ri != l->second->info->radio.end()) {
	  if (Lr.find(TRadio(l->second->info->id, distance(l->second->info->radio.begin(),ri))) != Lr.end())
	    Lr[TRadio(l->second->info->id, distance(l->second->info->radio.begin(),ri))].erase(*l);
	}
	/* then, remove the link from the topology*/
	T->node[l->first->info->id].neighbor.erase(l->second);
      }
      /* switch off the radio (set channel to 0) */
      T->node[minLr->first.first].info->radio[minLr->first.second].channel = 0;
    }
    
    /* remove the whole Lr set of the extracted radio */
    Lr.erase(minLr);

    // T->printTopology();  // DEBUG
  }

  /* erase links with null flow */
  //T->eraseLinksWithNullFlow();

  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: PACA2: Checking E2E flows at the end FAILED" << endl << endl;

  return 0;
}





/* constructor */
E2CAR_OPT::E2CAR_OPT (const list<pair<string,string> >& params)
 : m_relax (false),
   m_u0 (0.5)
{
  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("relax", m_relax);
    RP.read ("u0", m_u0);

    if (RP.notFound())
      cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the E2CAR_OPT algorithm" << endl << endl;
  }
}


double E2CAR_OPT::Run (Topology* T)
{
  /* clone the topology because MinMaxUtot_MILP removes links with null flow */
  Topology* U = new Topology(*T);

  /* find the minimum maxUtot */
  double maxUtot;

  //MinMaxUtot_MILP(U, maxUtot, relax, NULL, u0);
  MinMaxUtot_MILP mmu(U, NULL, 7200);  // max 2 hours

  mmu.set_breakValue(m_u0);

  mmu.lp_solve();

  maxUtot = mmu.getObjValue();

  printCAStats(U,"MinMaxUtot_MILP",true);

  MinRadios_MILP mr(T,max(maxUtot,m_u0));

  if (mr.lp_solve())
    return -1;

  delete U;

  return 0;
}
