// -*- C++ -*-

/* myparser.cc
 *
 * Copyright (C) 2002 The libxml++ development team
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "parsers.h"
#include "WMutils.h"
#include "WMtypes.h"
#include "rndvars.h"

#include <glibmm/convert.h> //For Glib::ConvertError
#include <iostream>
#include <cassert>

using namespace std;


/** Topology parser **/

TopologyParser::TopologyParser (Topology* _T)
  : xmlpp::SaxParser(),
    defaultNradios (0),
    defaultDBgain (default_db_gain),
    defaultMinAngle (default_min_angle),
    defaultMaxAngle (default_max_angle),
    T (_T)
{
}

TopologyParser::~TopologyParser()
{
}

void TopologyParser::on_start_document()
{
  cout << "Reading topology..." << endl << endl;
}

void TopologyParser::on_end_document()
{
  cout << "Finished reading topology" << endl << endl;
}

void TopologyParser::on_start_element(const Glib::ustring& name,
                                   const AttributeList& attributes)
{
  if (name == "Topology")
  {
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("nradios", defaultNradios);
      RP.read ("DBgain", defaultDBgain);
      RP.read ("minangle", defaultMinAngle);
      RP.read ("maxangle", defaultMaxAngle);
      
      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Topology element" << endl << endl;
    }
  }
  else if (name == "Node")
  {
    // a new node is found: set attributes to default (Topology) values
    nodeDBgain = defaultDBgain;
    nodeMinAngle = defaultMinAngle;
    nodeMaxAngle = defaultMaxAngle;
    
    unsigned int nradios = defaultNradios;
    float xpos = -1.;
    float ypos = -1.;
    TMeshRouter role = router;
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("nradios", nradios);
      RP.read ("DBgain", nodeDBgain);
      RP.read ("minangle", nodeMinAngle);
      RP.read ("maxangle", nodeMaxAngle);
      RP.read ("x", xpos);
      RP.read ("y", ypos);

      if (RP.notFound())
      {
	if (iter->name == "role")
	{
	  if (iter->value == "R")
	    role = router;
	  else if (iter->value == "G")
	    role = gateway;
	  else if (iter->value == "A")
	    role = aggregation;
	  else
	    cerr << endl << "Ignoring invalid node role: " << iter->value << endl << endl;
	}
	else
	  cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Node element" << endl << endl;
      }
    }
    
    if (xpos < 0. || ypos < 0.)
    {
      cerr << endl << "Ignoring node with invalid coordinates" << endl << endl;
      return;
    }

    MeshNode* u = new MeshNode (T->node.size(), xpos, ypos, role, nradios);

    for (RadioIfaceIterator ri = u->radio.begin(); ri != u->radio.end(); ri++)
    {
      ri->DBgain = nodeDBgain;
      ri->minangle = nodeMinAngle;
      ri->maxangle = nodeMaxAngle;
    }

    // insert the new node into the topology
    T->node.push_back(TGraphNode(u));

    if (role == aggregation) 
      T->aggregation_list.push_back(T->node.back().info->id);
    else if (role == gateway)
      T->gateway_list.push_back(T->node.back().info->id);
  }
  else if (name == "Radio")
  {
    // additional radio for the current node
    float DBgain = nodeDBgain;
    unsigned int minangle = nodeMinAngle;
    unsigned int maxangle = nodeMaxAngle;

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("DBgain", DBgain);
      RP.read ("minangle", minangle);
      RP.read ("maxangle", maxangle);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Radio element" << endl << endl;
    }
    
    // at least one node must have already been added to the topology
    assert ( T->node.size() );

    // add the radio to the current node
    T->node.back().info->radio.push_back(RadioIface(DBgain, minangle, maxangle));
  }
  else
  {
    cerr << endl << "Ignoring invalid element (" << name << ")" << endl << endl;
  }
}

void TopologyParser::on_end_element(const Glib::ustring& /* name */)
{
//   std::cout << "on_end_element()" << std::endl;
}

void TopologyParser::on_characters(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_characters(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "TopologyParser::on_characters(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void TopologyParser::on_comment(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_comment(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "TopologyParser::on_comment(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void TopologyParser::on_warning(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_warning(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "TopologyParser::on_warning(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void TopologyParser::on_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "TopologyParser::on_error(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void TopologyParser::on_fatal_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_fatal_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "TopologyParser::on_characters(): Exception caught while converting value for std::cout: " << ex.what() << std::endl;
  }
}



/** WiMesh config parser **/

WiMeshParser::WiMeshParser ()
  : xmlpp::SaxParser(),
    numAvailChannels (defaultNumChannels),
    nodePCrv (NULL),
    nodePCseed (getSeed()),
    radioPCrv (NULL),
    radioPCseed (getSeed()),
    CDrv (NULL),
    CDseed (getSeed()),
    CDtype (0),
    CDid ("XX"),
    postCAopt (false),
    simulId ("xxx")
{
}

WiMeshParser::~WiMeshParser()
{
}

void WiMeshParser::on_start_document()
{
  cout << "Reading WiMesh config..." << endl << endl;
}

void WiMeshParser::on_end_document()
{
  cout << "Finished reading WiMesh config" << endl << endl;
}

void WiMeshParser::on_start_element(const Glib::ustring& name,
                                   const AttributeList& attributes)
{
  static int id;

  if (name == "NumChannels")
  {
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("n", numAvailChannels);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a NumChannels element" << endl << endl;
    }
  }
  else if (name == "Topology")
  {
    if (!xml_topo_file.empty() || !serialized_topo_file.empty())
    {
      // a topology file name has been already specified
      cerr << endl << "Ignoring duplicate Topology element" << endl << endl;
      return;
    }
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("xml_file", xml_topo_file);
      RP.read ("serialized_file", serialized_topo_file);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Topology element" << endl << endl;
    }
  }
  else if (name == "BuildGiMethod")
  {
    if (!buildGiMethod.empty())
    {
      // a buildGi method has been already specified
      cerr << endl << "Ignorind duplicate BuildGiMethod element" << endl << endl;
      return;
    }

    // "name" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring BuildGiMethod element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "name")
    {
      cerr << endl << "Ignoring BuildGiMethod element with first attribute other than name" << endl << endl;
      return;
    }

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
      buildGiParams.push_back(pair<string,string>(iter->name,iter->value));
    
    buildGiMethod = attributes.front().value;
  }
  else if (name == "PFRmethod")
  {
    if (!pfrMethod.empty())
    {
      // a pfr method has been already specified
      cerr << endl << "Ignoring duplicate PFRmethod element" << endl << endl;
      return;
    }

    // "name" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring PFRmethod element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "name")
    {
      cerr << endl << "Ignoring PFRmethod element with first attribute other than name" << endl << endl;
      return;
    }

    // save all the attributes in the params list
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
      pfrParams.push_back(pair<string,string>(iter->name,iter->value));
    
    pfrMethod = attributes.front().value;
  }
  else if (name == "NodePCprofile")
  {
    float bpc = availNodePCprofiles[0].BasePowerConsumption,
          spc = availNodePCprofiles[0].StandbyPowerConsumption;

    // "id" is expected as the first attribute, followed by the other parameters

    if (attributes.empty())
    {
      cerr << endl << "Ignoring NodePCprofile element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "id")
    {
      cerr << endl << "Ignoring NodePCprofile element with first attribute other than id" << endl << endl;
      return;
    }

    id = atoi (attributes.front().value.c_str());
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("bpc", bpc);
      RP.read ("spc", spc);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a NodePCprofile element" << endl << endl;
    }
    
    nodePCprofile.insert(pair<int,NodePowerConsumptionProfile>(id,NodePowerConsumptionProfile(bpc,spc)));
  }
  else if (name == "RadioPCprofile")
  {
    foRadioTxPower* rho_tx = availRadioPCprofiles[0].rho_tx->Clone();
    float rho_rx = availRadioPCprofiles[0].rho_rx,
          rho_idle = availRadioPCprofiles[0].rho_idle,
	  rho_sleep = availRadioPCprofiles[0].rho_sleep;

    // "id" is expected as the first attribute, followed by the other parameters

    if (attributes.empty())
    {
      cerr << endl << "Ignoring RadioPCprofile element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "id")
    {
      cerr << endl << "Ignoring RadioPCprofile element with first attribute other than id" << endl << endl;
      return;
    }

    id = atoi (attributes.front().value.c_str());

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("rho_rx", rho_rx);
      RP.read ("rho_idle", rho_idle);
      RP.read ("rho_sleep", rho_sleep);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a RadioPCprofile element" << endl << endl;
    }
    
    radioPCprofile.insert(pair<int,RadioPowerConsumptionProfile>(id,RadioPowerConsumptionProfile(rho_tx, rho_rx, rho_idle, rho_sleep)));
  }
  else if (name == "Rho_tx")
  {
    string fname;

    // "name" is expected as the first attribute, followed by the other parameters

    if (attributes.empty())
    {
      cerr << endl << "Ignoring Rho_tx element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "name")
    {
      cerr << endl << "Ignoring Rho_tx element with first attribute other than name" << endl << endl;
      return;
    }

    fname = attributes.front().value;
    
    // check if it is the name of a RadioTxPower function object
    if (fname == foRadioTxPowerLinear::name)
    {
      float a, b;
      bool hasA = false, hasB = false;

      // read foRadioTxPowerLinear parameters
      for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
      {
	if (iter->name == "a")
	{
	  a = atof(iter->value.c_str());
	  hasA = true;
	}
	else if (iter->name == "b")
	{
	  b = atof(iter->value.c_str());
	  hasB = true;
	}
	else
	  cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Rho_tx::" << fname << " element" << endl << endl;
      }

      if (!hasA || !hasB)
      {
	cerr << endl << "Ignoring Rho_tx::" << fname << " element with incomplete list of parameters (a, b)" << endl << endl;
	return;
      }

      // check whether an entry with the current id is present in radioPCprofile
      map<int, RadioPowerConsumptionProfile>::iterator mi = radioPCprofile.find(id);
      
      if (mi == radioPCprofile.end())
      {
	cerr << endl << "Ignoring Rho_tx::" << fname << " element, being unable to find a corresponding RadioPCprofile element" << endl << endl;
	return;
      }

      if (mi->second.rho_tx)
	delete mi->second.rho_tx;

      mi->second.rho_tx = new foRadioTxPowerLinear (a, b);
    }
    else
    {
      cerr << endl << "Ignoring Rho_tx element with unknown RadioTxPower function object name (" << fname << ")" << endl << endl;
      return;
    }
  }
  else if (name == "NodePCRndVar")
  {
    if (nodePCrv)
    {
      // a nodePC random variable has been already specified
      cerr << endl << "Ignoring duplicate NodePCRndVar element" << endl << endl;
      return;
    }

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      if (!nodePCrv)
	RP.read ("rv", nodePCrv);
      RP.read ("seed", nodePCseed);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid (or duplicate) attribute (" << iter->name << ") for a NodePCRndVar element" << endl << endl;
    }

    if (!nodePCrv)
    {
      cerr << endl << "Ignoring NodePCRndVar element with no valid rv attribute" << endl << endl;
      return;
    }
  }
  else if (name == "RadioPCRndVar")
  {
    if (radioPCrv)
    {
      // a radioPC random variable has been already specified
      cerr << endl << "Ignoring duplicate RadioPCRndVar element" << endl << endl;
      return;
    }
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      if (!radioPCrv)
	RP.read ("rv", radioPCrv);
      RP.read ("seed", radioPCseed);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid (or duplicate) attribute (" << iter->name << ") for a RadioPCRndVar element" << endl << endl;
    }

    if (!radioPCrv)
    {
      cerr << endl << "Ignoring RadioPCRndVar element with no valid rv attribute" << endl << endl;
      return;
    }
  }
  else if (name == "ChangeDemands")
  {
    if (CDrv)
    {
      // a CD random variable has been already specified
      cerr << endl << "Ignoring duplicate ChangeDemands element" << endl << endl;
      return;
    }
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      if (!CDrv)
	RP.read ("rv", CDrv);
      RP.read ("seed", CDseed);
      RP.read ("type", CDtype);
      RP.read ("id", CDid);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a ChangeDemands element" << endl << endl;
    }

    if (!CDrv || CDtype > 2)
    {
      cerr << endl << "Ignoring ChangeDemands element with invalid attributes" << endl << endl;
      return;
    }
  }
  else if (name == "ChannelAssign")
  {
    if (!caAlgo.empty())
    {
      // a channel assignment algorithm has been already specified
      cerr << endl << "Ignoring duplicate ChannelAssign element" << endl << endl;
      return;
    }

    // "algo" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring ChannelAssign element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "algo")
    {
      cerr << endl << "Ignoring ChannelAssign element with first attribute other than algo" << endl << endl;
      return;
    }

    // save all the attributes (but postopt) in the params list
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      if (iter->name == "postopt")
      {
	if (iter->value == "true" || iter->value == "1")
	  postCAopt = true;
	if (iter->value == "false" || iter->value == "0")
	  postCAopt = false;
      }
      else
	caParams.push_back(pair<string,string>(iter->name,iter->value));
    }

    caAlgo = attributes.front().value;
  }
  else if (name == "SimulId")
  {
    if (simulId != "xxx")
    {
      // a simulation id has been already specified
      cerr << endl << "Ignoring duplicate SimulId element" << endl << endl;
      return;
    }

    // "id" is expected as the first (and unique) attribute

    if (attributes.empty())
    {
      cerr << endl << "Ignoring SimulId element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "id")
    {
      cerr << endl << "Ignoring SimulId element with first attribute other than id" << endl << endl;
      return;
    }

    simulId = attributes.front().value;
  }
  else if (name != "WiMeshCfg")
  {
    cerr << endl << "Ignoring invalid element (" << name << ")" << endl << endl;
  }
}

void WiMeshParser::on_end_element(const Glib::ustring& name)
{
//   if (name == "WiMeshCfg")
//     // stop the parser
//     throw xmlpp::exception("Done");
}

void WiMeshParser::on_characters(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_characters(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "WiMeshParser::on_characters(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void WiMeshParser::on_comment(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_comment(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "WiMeshParser::on_comment(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void WiMeshParser::on_warning(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_warning(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshParser::on_warning(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void WiMeshParser::on_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshParser::on_error(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void WiMeshParser::on_fatal_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_fatal_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshParser::on_characters(): Exception caught while converting value for std::cout: " << ex.what() << std::endl;
  }
}

