/* This file contains the implementation of the methods of the Topology class */

#include <fstream>
#include <iostream>
#include <cmath>
#include <cstring>
#include <set>
#include <queue>
#include <cassert>
#include <numeric>
#include <algorithm>
#include <ext/functional>
#include "WMtypes.h"
#include "WMutils.h"
#include "parsers.h"
#include "subgraph.h"
#include "dag.h"
#include "DijkstraShortestPathAlg.h"
#include "YenTopKShortestPathsAlg.h"


const string foRadioTxPowerLinear::name = string("linear");


//using namespace std;


//typedef pair<vector<TGraphNeighbor>::iterator, vector<TGraphNeighbor>::iterator> TUndLink;

// inline pair<bool, unsigned int> SearchChannel(MeshNode *info, unsigned int channel)
// {
// 	bool found = false;
// 	unsigned int r=0;
// 
// 	while (r<info->numAvailRadios() && !found)
// 		if (info->radio[r].channel == channel)
// 			found = true;
// 		else
// 			r++;
// 
// 	return pair<bool,unsigned int>(found,r);
// }

  bool TLinkItrs::IsInTheCollDomainOf(Topology* T, TLinkItrs l) const {
    return T->PotInterfLinks(*this,l) && (this->second->radio->channel == l.second->radio->channel);
  }


int MeshNode::countRadiosWithChannel(int c)
{ 
  return count_if(radio.begin(), radio.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) ); 
}

RadioIfaceIterator MeshNode::findRadioWithChannel(int c)
{
  return find_if(radio.begin(), radio.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(), c), mem_fun_ref(&RadioIface::getChannel) ) ); 
}

set<unsigned int> MeshNode::findChannels ()
{
  set<unsigned int> chset;

  chset.clear();

  for (RadioIfaceIterator ri=radio.begin(); ri!=radio.end(); ri++)
    chset.insert(ri->channel);

  chset.erase(0);

  return chset;
}


set<unsigned int> Topology::haveCommonChannels (int u, int v)
{
  set<unsigned int> S, uChannels, vChannels;

  uChannels = node[u].info->findChannels();
  vChannels = node[v].info->findChannels();
  
  S.clear();
  set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

  return S;
}

#if 0
/* getNodeInfo: get node info from file 'filename' 
   on error, returns a nonzero value
*/
int Topology::getNodeInfo(FILE *f)
{
	int id=0, nradios;
	TMeshRouter _role;
	MeshNode *u;
	TGraphNode gnode;
	float xpos, ypos;
	char c;
	char msg[80]="Unexpected value or EOF while reading the ";
	
	// read the number of available channels
	//READ_PARAM(f,"%ui",&numAvailChannels,strcat(msg,"number of available channels"));

	//read nodes
	skipComment(f);
	while (fscanf(f,"%c",&c)!=EOF) 
	{
		switch (c)
		{
			case 'A':
				_role = aggregation;
				break;
			case 'G':
				_role = gateway;
				break;
			default:
				_role = router;
				break;
		}

		READ_PARAM(f,"%f",&xpos,strcat(msg,"x coordinate"));
		READ_PARAM(f,"%f",&ypos,strcat(msg,"y coordinate"));
		READ_PARAM(f,"%i",&nradios,strcat(msg,"number of radios"));
		u = new MeshNode(id++, xpos, ypos, _role, nradios);

		gnode.info = u;
		node.push_back(gnode);

		if (_role == aggregation) 
			aggregation_list.push_back(u->id); 
		else if (_role == gateway)
			gateway_list.push_back(u->id);

		skipComment(f);
	}
	return 0;
}


int Topology::getNodeInfo(const char *filename)
{
	FILE *f;

	if ((f = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "ERR: Cannot open file %s to read topology\n",filename);
		return -1;
	}
	int res = getNodeInfo(f);
	fclose(f);
	return res;
}
#endif

float Topology::dist(int x, int y)
{ 
  return sqrtf(powf(fabsf(node[x].info->xpos - node[y].info->xpos),2)+powf(fabsf(node[x].info->ypos - node[y].info->ypos),2));
}

float gain(float x)
{
	return 1 / (x*x);
}

float SINRthresh(float rate)
{
	unsigned int i=0;
	bool found = false;
	while (i<NUMRATES && !found)
		if (availTXrates[i] == rate)
			found = true;
		else
			i++;

//	assert (found);
	if (!found)
	  i = NUMRATES - 1;

	return SINRthresholds[i];
}


void Topology::addNeighbor(vector<TGraphNode>::iterator u, vector<TGraphNode>::iterator v, float rate, unsigned int power)
{
	TGraphNeighbor *gneighbor;
	list<TGraphNeighbor>::const_iterator pos;

	gneighbor = new TGraphNeighbor(v->info, rate, power);

	/* insert gneighbor in the neighbor list  */
	u->neighbor.push_back(*gneighbor);
}

/* Builds an initial graph using the given (fixed) transmission power 
 * A link exists between two nodes if there is a rate between min_rate and max_rate
 * that allows to establish the link (in such a case, the highest rate is selected) */
void Topology::buildGi_FixedPowerMaxAllowRate(unsigned int power, int minrate, int maxrate)
{
	int i;

	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (vector<TGraphNode>::iterator v=node.begin(); v!=node.end(); v++)
			if (v!=u)
			{
				i = MaxAllowLinkRate (u->info->id, v->info->id, power, minrate, maxrate);

				if (i>=minrate)
					addNeighbor(u,v,availTXrates[i],power);
			}
}


int Topology::MaxAllowLinkRate (int u, int v, unsigned int power, int minrate, int maxrate)
{
  float d = dist (u,v);

  int i = maxrate;
  while ( (i>=minrate) && (gain(d)*DbmToW(TxPowerLevel(power))/noise < SINRthresholds[i]) )
    i--;
  
  return i;
}


/* Checks whether two nodes u and v can be neighbors using some values for TX power and rate
   Starts from the highest TX rate and the lowest TX power, then increases the TX power
   If the SNR is still below the threshold, lowers the TX rate
   Thus, determines the highest possible capacity
   if c>=0 then it is possible to establish a link between u and v
*/
bool Topology::canBeNeighbors(vector<TGraphNode>::iterator u, vector<TGraphNode>::iterator v, int & c, int & p, int maxp) {
	/* compute distance between u and v */
	float d = sqrtf(pow(fabsf(u->info->xpos - v->info->xpos),2)+pow(fabsf(u->info->ypos - v->info->ypos),2));

	c = NUMRATES-1;           // start with the maximum TX rate
	do {
		p = 0;
		while ( (p<maxp) && (gain(d)*DbmToW(TxPowerLevel(p))/noise < SINRthresholds[c]) )
			p++;
		
		if (p==maxp)
			c--;

	} while (p==maxp && c>=0);
	
	return (c>=0);
}

// Searches for new links that can be added to the network graph
void Topology::findToBePossiblyAddedLinks(int maxp)
{
	int c,p;
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (vector<TGraphNode>::iterator v=node.begin(); v!=node.end(); v++)
			// if v!=u and u and v are not already neighbors
			if ( (v!=u) &&  find_if(u->neighbor.begin(), u->neighbor.end(), 
			  __gnu_cxx::compose1(bind2nd(equal_to<int>(),v->info->id), ptr_fun<TGraphNeighbor,int>(getNodeID) )) == u->neighbor.end() )
			{
				canBeNeighbors(u,v,c,p,maxp);
				if (c>=0) {
					addNeighbor(u,v,availTXrates[c],p);
					cout << "findToBePossiblyAddedLinks(): Aggiunto link " << u->info->id << "->" << v->info->id << " c " << c << " rate " << availTXrates[c] << " power  " <<
					TxPowerLevel(p) << " SNR " << gain(dist(u->info->id,v->info->id))*DbmToW(TxPowerLevel(p)) << " threshold " << SINRthresholds[c] << endl;
				}
			}
}

/* constructor */
Topology::Topology (const char* filename)
 : noise (0.00001),
   maxP (numTxPowerLevels-1)
{
  TopologyParser parser (this);
  parser.set_substitute_entities(true);
  parser.parse_file(filename);
}


/* copy constructor*/
Topology::Topology(const Topology& Tin)
{
	list<TGraphNeighbor>::const_iterator pos;

	/* simple copy of the fields (node, aggregation_list and gateway_list) */
	*this = Tin;

	/* in addition, duplicate undirectedLink structures */
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
			v->info = node[v->info->id].info;
// 			v->radio = u->info->radio + (v->radio - Tin.node[u->info->id].info->radio);
			v->radio = &u->info->radio[distance(Tin.node[u->info->id].info->radio.begin(),RadioIfaceIterator(v->radio))];
		}
}


/* set the current TX range of each interface so as to just reach the farthest neighbor reachable on the channel that interface is set to 
void Topology::AdjustPower()
{
	set<unsigned int> uChannels, vChannels, CommonChannels;

	for (vector<TGraphNode>::const_iterator u=node.begin(); u!=node.end(); u++)
	{
		float maxdist[u->info->numAvailRadios()];
		for (unsigned int i=0; i<u->info->numAvailRadios(); maxdist[i++]=0.);

// OLD CODE
// 		for (vector<TGraphNeighbor>::const_iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
// 			if (v->uLink->distance > maxdist[v->radio - u->info->radio])
// 				maxdist[v->radio - u->info->radio] = v->uLink->distance;

		uChannels.clear();
		for (unsigned int r=0; r<u->info->numAvailRadios(); uChannels.insert(u->info->radio[r++].channel));

		for (vector<TGraphNeighbor>::const_iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
		{
			vChannels.clear();
			CommonChannels.clear();
			for (unsigned int r=0; r<v->info->numAvailRadios(); vChannels.insert(v->info->radio[r++].channel));
			set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(CommonChannels, CommonChannels.begin()));
			CommonChannels.erase(0);
			assert(CommonChannels.size());
			unsigned int i;
			for (i=0; u->info->radio[i].channel != *CommonChannels.begin(); i++);
			assert(i<u->info->numAvailRadios());
			if (v->uLink->distance > maxdist[i])
				maxdist[i] = v->uLink->distance;
		}

//		for (unsigned int i=0; i<u->info->numAvailRadios(); i++)
//			u->info->radio[i].curTXrange = maxdist[i];
	}
}
*/

void Topology::ResetChannels()
{
  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
  {
    // set all channels to 0
    for (unsigned int i=0; i<u->info->numAvailRadios(); i++)
      u->info->radio[i].channel = 0;
    
    // remove parallel links
    list<TGraphNeighbor>::iterator i;
    
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
    {
      // if no link to neighbor v has been already found
      if ((i = find_neighbor(u->info->id,v->info->id)) == v)
      {
	v->radio = NULL;
	v->Utot = 0.;
      }
      else
      {
	i->flow += v->flow;
	i->rate = (i->rate > v->rate ? i->rate : v->rate);
	i->txpower = (i->txpower > v->txpower ? i->txpower : v->txpower);

	for (map<pair<int,int>,double>::iterator f = v->f_e2e.begin(); f != v->f_e2e.end(); f++)
	  if (i->f_e2e.find(f->first) != i->f_e2e.end())
	    i->f_e2e[f->first] += f->second;
	  else
	    i->f_e2e[f->first] = f->second;
	  
	// remove this neighbor
	i = v--;
	u->neighbor.erase(i);
      }
    }
  }
}


/* 
 * If type = 2, swaps all the demands, in the sense that demands with less flow are assigned the flow of the demands with more flow.
 * Each demand is multiplied (type = 0 or 2) or set (type = 1) by a random value derived from the given distribution
 * Values for type: 0 (scale), 1 (set), 2 (swap and scale)
 */
void Topology::changeDemands(RndVar* RV, double seed, unsigned int type)
{
#define SETVALUE 0x1
#define SWAP     0x2
  static MotherOfAll urng(seed);         // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used

  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands, newDemands;

  if (!(checkE2Eflows(&demands)))
    cerr << endl << "WARNING: changeDemands: Checking E2E flows before beginning FAILED" << endl << endl;

  // copy the demands in a vector and sort it based on the amount of flow
  vector<pair<pair<int,int>,double> > V(demands.size());  
  copy(demands.begin(), demands.end(), V.begin());
  if (type & SWAP)
    sort(V.begin(), V.end(), compareSecond<pair<int,int>,double>());

  throughput = 0;
  vector<pair<pair<int,int>,double> >::reverse_iterator r = V.rbegin();

  for (vector<pair<pair<int,int>,double> >::iterator v=V.begin(); v!=V.end(); v++, r++)
  {
    throughput += (newDemands[v->first] = (type & SETVALUE ? 1. : (type & SWAP ? r->second : v->second)) * RV->rv->Next());
    cout << "Demands [" << v->first.first << "," << v->first.second << "] changes from " << v->second << " to " << newDemands[v->first] << endl; 
  }

  /* update the flows */
  for (vector<TGraphNode>::iterator u=node.begin(); u!= node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      for (map<pair<int,int>,double>::iterator f=v->f_e2e.begin(); f!=v->f_e2e.end(); f++)
      {
	v->flow += f->second / demands[f->first] * (newDemands[f->first] - demands[f->first]);
	f->second += f->second / demands[f->first] * (newDemands[f->first] - demands[f->first]);
      }

  if (!(checkE2Eflows(NULL)))
    cerr << endl << "WARNING: changeDemands: Checking E2E flows at the end FAILED" << endl << endl;
}


list<TLinkItrs>* Topology::changeFlowOnRandomPath(int maxLength, bool increase, double percent) {
  MotherOfAll urng(getSeed());         // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used
  Uniform U;
  /* the path length is uniformly distributed between 1 and maxLength */
  SumRandom lenRV( (maxLength-1) * U + 1 );
  SumRandom srcRV( (node.size() - 1) * U );
  Uniform nh;
  int next, i = 0;
  double fmin = numeric_limits<double>::max();
  
  /* vector of nodes along the path (used to avoid loops) */
  vector<int> Pnode;
  
  /* the source of the path is chosen randomly */
  Pnode.push_back( (int)floor(srcRV.Next()) );
  
  list<TLinkItrs>* p = new list<TLinkItrs>;
  
  double len = lenRV.Next();
//  list<TGraphNeighbor>::iterator n = node[].
  
  while (p->size() < len) {
    next = (int)floor(nh.Next() * (node[Pnode.back()].neighbor.size() - 1));
    
    list<TGraphNeighbor>::iterator l = node[Pnode.back()].neighbor.begin();
    for (int n=0; n<next; n++)
      l++;
    
    if (find(Pnode.begin(), Pnode.end(), l->info->id) == Pnode.end()) {
      p->push_back(TLinkItrs(node.begin() + Pnode.back(), l));
      
      Pnode.push_back(l->info->id);
      
      fmin = min(fmin, p->back().second->flow);
    }
    
    if (i++ > 2 * len)
      break;
  }
  
//  fmin = fmin * (U.Next() * percent) * (increase ? 1 : -1);
  fmin = fmin * (U.Next() * percent) * 3;
  
  for (list<TLinkItrs>::iterator l=p->begin(); l!=p->end(); l++) {
    l->second->flow += fmin;
    
    /* update the Utot of all the affected links */
    for (vector<TGraphNode>::iterator x=node.begin(); x!=node.end(); x++)
      for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	// if u->v belongs to the collision domain of x->y
	if ( (PotInterfLinks(*l, TLinkItrs(x,y))) && (l->second->radio->channel == y->radio->channel) )
	  y->Utot += fmin/l->second->rate;

    cerr << l->first->info->id << "->" << l->second->info->id << " ";
  }
  cerr << endl;
  
  return p;
}

/*  Erases links with null flow provided that 
 *  either there is a parallel link on another channel
 *  or there are only inverse links with null flow
 */
void Topology::eraseLinksWithNullFlow() {
  list<TGraphNeighbor>::iterator next;
  /* for all the links in the network */
  for (vector<TGraphNode>::iterator x=node.begin(); x!=node.end(); x++)
    for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); ) {
      /* with null flow */
      if (!y->flow) {
	/* if there is a parallel link on another channel */
	if (find_if(x->neighbor.begin(), x->neighbor.end(), 
		__gnu_cxx::compose2(logical_and<bool>(),
		__gnu_cxx::compose1(bind2nd(equal_to<int>(),y->info->id), ptr_fun<TGraphNeighbor,int>(getNodeID) ),
		__gnu_cxx::compose1(bind2nd(not_equal_to<unsigned int>(),y->radio->channel), ptr_fun<TGraphNeighbor,unsigned int>(getRadioCh) )) )
		!= x->neighbor.end() ) {
	  /* erase x->y */
	  next = y;
	  next++;
	  cout << "Erasing " << x->info->id << "->" << y->info->id << "[" << y->radio->channel <<"] with flow = " << y->flow << endl;
	  x->neighbor.erase(y);
	  y = next;
	}
	else 
	  /* there are no parallel links */
	  /* if there is no inverse link with non-null flow */
	  if (find_if(node[y->info->id].neighbor.begin(), node[y->info->id].neighbor.end(), 
		  __gnu_cxx::compose2(logical_and<bool>(),
		  __gnu_cxx::compose1(bind2nd(equal_to<int>(),x->info->id), ptr_fun<TGraphNeighbor,int>(getNodeID) ),
		  __gnu_cxx::compose1(bind2nd(not_equal_to<double>(),0.), mem_fun_ref(&TGraphNeighbor::getFlow) )) )
		  == node[y->info->id].neighbor.end() ) {
	    /* erase x->y */
	    next = y;
	    next++;
	    cout << "Erasing " << x->info->id << "->" << y->info->id << "[" << y->radio->channel <<"] with flow = " << y->flow << endl;
	    x->neighbor.erase(y);
	    y = next;
	  }
	  else
	    y++;
	}
      else
	y++;
    }
}

/* Given a TLinkItrs u->v returns the TLinkItrs v->u */
TLinkItrs Topology::FindInverseLink(TLinkItrs l, unsigned int channel)
{
	vector<TGraphNode>::iterator node1 = find_if(node.begin(), node.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(),l.second->info->id), ptr_fun<TGraphNode,int>(getNodeID) ));

//cout << "FindInverseLink " << l.first->info->id << "->" << l.second->info->id << endl;

	assert( node1 != node.end() );
	
	list<TGraphNeighbor>::iterator node2;
	
	if (!channel)
		node2 = find_if(node1->neighbor.begin(), node1->neighbor.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(),l.first->info->id), ptr_fun<TGraphNeighbor,int>(getNodeID) ));
	else
		node2 = find_if(node1->neighbor.begin(), node1->neighbor.end(), 
				__gnu_cxx::compose2(logical_and<bool>(),
				__gnu_cxx::compose1(bind2nd(equal_to<int>(),l.first->info->id), ptr_fun<TGraphNeighbor,int>(getNodeID) ),
				__gnu_cxx::compose1(bind2nd(equal_to<unsigned int>(),channel), ptr_fun<TGraphNeighbor,unsigned int>(getRadioCh) )) );

// 	if (!channel)
// 		assert( node2 != node1->neighbor.end() );

/*	TLinkItrs* res = NULL;
	if ( node2 != node1->neighbor.end() )
		res = new TLinkItrs(node1,node2);
	return res;*/
	
	return TLinkItrs(node1,node2);
}



/* Assign channel c to the (directed) link l
   PRE: for both end nodes, either a radio assigned to c already exists or an available radio exists */
void Topology::AssignChannel2Link(TLinkItrs l, unsigned int c)
{
// 	pair<bool, unsigned int> foundCh;
	RadioIfaceIterator ri;

// 	foundCh = SearchChannel(l.first->info, c);
	ri = l.first->info->findRadioWithChannel(c);
	
// 	if (!foundCh.first)
	if (ri == l.first->info->radio.end())
	{
// 		foundCh = SearchChannel(l.first->info, 0);
		ri = l.first->info->findRadioWithChannel(0);
// 		assert( foundCh.first );
		assert( ri != l.first->info->radio.end() );
// 		l.first->info->radio[foundCh.second].channel = c;
		ri->channel = c;
	}
// 	l.second->radio = l.first->info->radio + foundCh.second;
	l.second->radio = &(*ri);

// 	foundCh = SearchChannel(l.second->info, c);
	ri = l.second->info->findRadioWithChannel(c);
// 	if (!foundCh.first)
	if (ri == l.second->info->radio.end())
	{
// 		foundCh = SearchChannel(l.second->info, 0);
		ri = l.second->info->findRadioWithChannel(0);
// 		assert( foundCh.first );
		assert ( ri != l.second->info->radio.end() );
// 		l.second->info->radio[foundCh.second].channel = c;
		ri->channel = c;
	}

/*	TLinkItrs InvLink = FindInverseLink(l);
	InvLink.second->radio = l.second->info->radio + foundCh.second;*/
}


/* returns true if x->y potentially interferes with u->v */
bool Topology::PotInterfLinks(TLinkItrs xy, TLinkItrs uv)
{
/*	if (xy == uv || FindInverseLink(xy) == uv)
		return true;*/
#if 0
	// x == u and y == v (the same link or a parallell link)
	if (xy.first->info->id == uv.first->info->id && xy.second->info->id == uv.second->info->id)
		return true;

	// x == v and y == u (an inverse link)
	if (xy.first->info->id == uv.second->info->id && xy.second->info->id == uv.first->info->id)
		return true;
#endif
	// x == u or y == v (links have the same transmitter or receiver)
	if (xy.first->info->id == uv.first->info->id || xy.second->info->id == uv.second->info->id)
		return true;

	// x == v or y == u (the transmitter of a link is the receiver of the other)
	if (xy.first->info->id == uv.second->info->id || xy.second->info->id == uv.first->info->id)
		return true;

	float power1 = DbmToW(TxPowerLevel(uv.second->txpower)),
	      power2 = DbmToW(TxPowerLevel(xy.second->txpower));

	// x is close to u, so when x transmits u senses the channel busy and does not transmit
	// the received power at 90m when transmitting at 15dbm is taken as reference
	if (gain(dist(xy.first->info->id, uv.first->info->id)) * power2 > gain(90) * 0.03162) 
		return true;

	// the SINR is affected by the distance between x and v (relative to the distance between u and v)
	return gain(dist(uv.first->info->id,uv.second->info->id)) * power1 / ( gain(dist(xy.first->info->id,uv.second->info->id)) * power2 + noise ) < SINRthresh(uv.second->rate);

}


list<TGraphNeighbor>::iterator Topology::find_neighbor(int u, int v)
{
  return find_if(node[u].neighbor.begin(), node[u].neighbor.end(),
		  __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), ptr_fun<TGraphNeighbor,int>(getNodeID) ));
  
}

list<TGraphNeighbor>::iterator Topology::find_neighbor(int u, int v, list<TGraphNeighbor>::iterator n)
{
  return find_if(n, node[u].neighbor.end(),
		  __gnu_cxx::compose1(bind2nd(equal_to<int>(),v), ptr_fun<TGraphNeighbor,int>(getNodeID) ));
  
}

list<TGraphNeighbor>::iterator Topology::find_neighbor(int u, int v, unsigned int c)
{
  return find_if(node[u].neighbor.begin(), node[u].neighbor.end(), 
		__gnu_cxx::compose2(logical_and<bool>(),
		__gnu_cxx::compose1(bind2nd(equal_to<int>(),v), ptr_fun<TGraphNeighbor,int>(getNodeID) ),
		__gnu_cxx::compose1(bind2nd(equal_to<unsigned int>(),c), ptr_fun<TGraphNeighbor,unsigned int>(getRadioCh) )) );
}

/* Find all the links that potentially interfere with l */
vector<TLinkItrs> Topology::computeNeighborhood(TLinkItrs l)
{
	vector<TLinkItrs> neighborhood;

	// consider all the links i<->j
	for (vector<TGraphNode>::iterator i=node.begin(); i!=node.end(); i++)
		for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
			// if i<->j is in the neighborhood of the extracted link
			if (PotInterfLinks(TLinkItrs(i,j),l))
			// add i<->j in neighborhood
				neighborhood.push_back(TLinkItrs(i,j));

	return neighborhood;
}



/* computes the set of channels used in the topology */
void Topology::getChannelSet(set<unsigned int> &channelSet)
{
	channelSet.clear();
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (v->radio)
				channelSet.insert(v->radio->channel);
}

#define CHANNEL(l)              \
l.second->radio->channel

/* computes the Utot of the collision domain of a link */
double Topology::computeUtot(TLinkItrs l)
{
	double Utot = 0.;
	// consider all the links i->j
	for (vector<TGraphNode>::iterator i=node.begin(); i!=node.end(); i++)
		for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
			// if i->j potentially interferes with l and has the same channel
			if ((PotInterfLinks(TLinkItrs(i,j), l)) && (j->radio) && (j->radio->channel==CHANNEL(l)))
			// add the flow of i->j
				Utot += j->flow / j->rate;
	return Utot;
}


bool Topology::checkUtot (int verbose, vector<pair<TLinkItrs,double> >** vec)
{
  bool ok = true;
  double Utot;
  
  if (vec)
    *vec = new vector<pair<TLinkItrs,double> >;
  
  for (vector<TGraphNode>::iterator u = node.begin(); u != node.end(); u++)
    for (list<TGraphNeighbor>::iterator v = u->neighbor.begin(); v != u->neighbor.end(); v++)
      if (v->radio)
      {
	Utot = computeUtot(TLinkItrs(u,v));
	
	if (vec)
	  (*vec)->push_back (pair<TLinkItrs,double>(TLinkItrs(u,v),Utot));
	
	if (verbose==2 || (verbose==1 && fabs(v->Utot - Utot)>0.0001))
	  cout << "Link (" << u->info->id << "," << v->info->id << ")" << "[" << v->radio->channel << "]" << " Utot = " << Utot;
	
	if (verbose >= 1 && fabs(v->Utot - Utot)>0.0001)
	  cout << "  instead the Utot field contains " << v->Utot;

	if (verbose==2 || (verbose==1 && fabs(v->Utot - Utot)>0.0001))
	  cout << endl;

	if (fabs(v->Utot - Utot)>0.0001)
	  ok = false;
      }

  return ok;
}


/* returns the maximum among the utilization of the collision domains of all the links */
pair<TLinkItrs,double> Topology::getUtotMax()
{
	vector< pair<TLinkItrs,double> >* UtotVect;
	
	checkUtot(0, &UtotVect);
	
	pair<TLinkItrs,double> p = *max_element(UtotVect->begin(), UtotVect->end(), compareSecond<TLinkItrs,double>());

	delete UtotVect;

	return p;
}


/* scales all the precomputed flow rates and the total utilizations */
void Topology::scaleFlowsByUtotMax()
{
	double UtotMax = getUtotMax().second;
	
	if (UtotMax > 1.)
	{
		// consider all the links i<->j
		for (vector<TGraphNode>::iterator i=node.begin(); i!=node.end(); i++)
			for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
			{
				j->flow /= UtotMax;
				j->Utot /= UtotMax;
			}
	}
}


/* Update the Utot for every link involved (except l itself) in case l is assigned channel c and rate r.
 * Must be invoked before making the change 
 */
void Topology::updateUtot (TLinkItrs l, unsigned int c, float r)
{
	/* the new Utot for the other links is obtained from the current Utot value */
	// for every directed link x->y
	for (vector<TGraphNode>::iterator x = node.begin(); x != node.end(); x++)
		for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
			// such that the potential collision domain of x->y includes l
			if ( (l!=TLinkItrs(x,y)) && (y->radio) && (PotInterfLinks(l,TLinkItrs(x,y))) )
			{
				/* by subtracting the F/C ratio of l if l escapes the collision domain of l0 */
				if (l.second->radio && (y->radio->channel == CHANNEL(l)) && (y->radio->channel != c))
					y->Utot -= l.second->flow / l.second->rate;
				/* by adding the F/C ratio of l if l joins the collision domain of l0 */
				else if ((!l.second->radio || y->radio->channel != CHANNEL(l)) && (y->radio->channel == c))
					y->Utot += l.second->flow / r;
			}
}



/* computes the degree of the network (i.e. average number of neighbors per node) */
double Topology::getDegree()
{
	double degree = 0.;
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		degree += u->neighbor.size();
	degree = degree / node.size();

	return degree;
}

class TNodeVisit {
  public:
	int predecessor; // -1 = NIL
	int color;	// 0 = white, 1 = gray, 2 = black
	int component;	// a number identifying each connected component (-1 = no component yet)

	/* constructor */
	TNodeVisit() : predecessor(-1), color(0), component(-1) {}
};

void DFS_Visit(Topology* T, vector<TNodeVisit> & nodes, unsigned int u, int compId) {
	nodes[u].color = 1;
	nodes[u].component = compId;
	for (list<TGraphNeighbor>::iterator v=T->node[u].neighbor.begin(); v!=T->node[u].neighbor.end(); v++)
		if (!nodes[v->info->id].color) {
			nodes[v->info->id].predecessor = u;
			DFS_Visit(T, nodes, v->info->id, compId);
		}
	nodes[u].color = 2;
}

/* Performs a depth-first-visit and returns the number of connected components */
int DFS(Topology* T, vector<TNodeVisit> & nodes) {
	int compId = 0;
	for (unsigned int i=0; i<nodes.size(); i++)
		if (!nodes[i].color) {
			DFS_Visit(T, nodes, i, compId);
			compId++;
		}
	return compId;
}

/* Given two connected components in a graph, attempts to merge them by adding a link between
   a pair of nodes belonging to the two distinct components
   If the merge succeeds, nodes in the component comp2 are assigned to comp1 and 
   the function returns true 
*/
bool connectComponents(Topology* T, vector<TNodeVisit> & nodes, int comp1, int comp2) {
	int c, p, imin, imax;
	vector<TGraphNode>::iterator u = T->node.begin();
	
	vector<int> nodes_comp1, nodes_comp2;
	
	for (unsigned int i=0; i<nodes.size(); i++)
		if (nodes[i].component == comp1)
			nodes_comp1.push_back(i);
	// cout << "Componente " << comp1 << ": "; for (vector<int>::iterator i=nodes_comp1.begin(); i!=nodes_comp1.end(); i++) cout << *i << " "; cout << endl;
	for (unsigned int i=0; i<nodes.size(); i++)
		if (nodes[i].component == comp2)
			nodes_comp2.push_back(i);
	// cout << "Componente " << comp2 << ": "; for (vector<int>::iterator i=nodes_comp2.begin(); i!=nodes_comp2.end(); i++) cout << *i << " "; cout << endl;
	
	for (vector<int>::iterator i1=nodes_comp1.begin(); i1!=nodes_comp1.end(); i1++)
		for (vector<int>::iterator i2=nodes_comp2.begin(); i2!=nodes_comp2.end(); i2++) {
			T->canBeNeighbors(u + *i1, u + *i2, c, p);
			if (c>=0) {
				imin = min (*i1, *i2);
				imax = max (*i1, *i2);
				T->addNeighbor(u + imin,u + imax,availTXrates[c],p);
				T->addNeighbor(u + imax,u + imin,availTXrates[c],p);

				// cout << "connectComponents: Aggiungo link " << *i1 << "->" << *i2 << " per il merge di componenti " << comp1 << " e " << comp2 << endl;
				
				for (vector<int>::iterator i=nodes_comp2.begin(); i!=nodes_comp2.end(); i++)
					nodes[*i].component = comp1;
				return true;
			}
		}
	return false;
}

/* If connect is false, just checks whether the graph is connected
   Otherwise, if there are multiple connected components, it attempts to add links
   in order to get a connected graph
*/

bool Topology::isConnected(bool connect) {
	vector<TNodeVisit> nodes(node.size());
	
	int numComp = DFS(this, nodes);
	cout << "Number of connected components in the graph: " << numComp << endl;
	
	if (!connect || numComp == 1)
		return (numComp == 1);
	
	set<int> comp_set, erased;
	for (int i=1; i<numComp; i++)
		comp_set.insert(i);
	
	do {
		erased.clear();
		for (set<int>::iterator comp2=comp_set.begin(); comp2!=comp_set.end(); comp2++)
			if (connectComponents(this, nodes, 0, *comp2)) 
				erased.insert(*comp2);

		for (set<int>::iterator i=erased.begin(); i!=erased.end(); i++)
			comp_set.erase(*i);
	} while (!comp_set.empty() && !erased.empty());
	
	return (comp_set.empty());
}


class lessDistant {
  vector<TNodeProp>* v;
  public:
    lessDistant(vector<TNodeProp>* _v) : v(_v) {}
    bool operator()(const int x, const int y) { return (v->at(x)).distance < (v->at(y)).distance; }
};

/*
   Applies Dijkstra's algorithm to the topology
   Returns a vector containing the distance and the predecessor of each node
   Links for which prune is true are not considered
   The link cost function is computed by the cost function object
*/
vector<TNodeProp> Topology::Dijkstra(int source, linkPruneFunc& prune, linkCostFunc& cost)
{
	int current;
	TLinkItrs l;

	/* initialize the vector of node properties */
	/* the i-th element of state is related to node i */
	vector<TNodeProp> state(node.size());
	state[source].distance = 0.;

	/* create the queue containing node IDs */
	set<int> Q;
	
	for (unsigned int i=0; i<node.size(); i++)
		Q.insert(i);

	while (!Q.empty())
	{
		current = *(min_element(Q.begin(),Q.end(),lessDistant(&state)));

		for (list<TGraphNeighbor>::iterator j=node[current].neighbor.begin(); j!=node[current].neighbor.end(); j++)
			if ((!(prune(l = TLinkItrs(node.begin()+current, j))))  && cost(l)>0. &&  state[j->info->id].distance > state[current].distance + 
							cost(l))
			{
				state[j->info->id].distance = state[current].distance + cost(l);
				state[j->info->id].prev_link= l;
			}
/*			else if (prune(l))
			  cout << "pruning " << current << "->" << j->info->id << endl;*/
		Q.erase(current);
	}
	return state;
}


vector<TLinkItrs>* Topology::ShortestPath(int source, int dest, linkPruneFunc& prune, linkCostFunc& cost)
/*
   Applies Dijkstra algorithm to topology T
   Returns the shortest feasible path between source and dest
*/
{
	vector<TLinkItrs>* path = new vector<TLinkItrs>;
	vector<TNodeProp> state = Dijkstra(source, prune, cost);

	// path calculation
	int k=dest;

	if (state[k].distance < numeric_limits<float>::max())
		while (k!=source)
		{
			path->insert(path->begin(),state[k].prev_link);
			k = state[k].prev_link.first->info->id;
		}

	return path;
}

/*
 * Find the first (at most k) shortest paths with length at most alpha times the length
 * of the shortest path
 * Returns the set of links on those paths
 */

set<pair<int,int> > Topology::find_path(pair<int,int> d, float alpha, int k)
{
  set<pair<int,int> > fksp;
  int count = 1;

  SubGraph my_graph(this);
  
  YenTopKShortestPathsAlg yenAlg(my_graph, my_graph.get_vertex(d.first), my_graph.get_vertex(d.second));

  BasePath* shortest = yenAlg.get_shortest_path(my_graph.get_vertex(d.first), my_graph.get_vertex(d.second));
  // shortest->PrintOut(cout);
  
  while(yenAlg.has_next()) {
    BasePath* path = yenAlg.next();
    
    if (path->length() > alpha*shortest->length() || count > k)
      break;
    
    // path->PrintOut(cout);

    for (int ver=0; ver<path->length()-1; ver++)
      fksp.insert(pair<int,int>(path->GetVertex(ver)->getID(),path->GetVertex(ver+1)->getID()));
    
    count++;
  }
  
  return fksp;
}


/* Helper function to add a BasePath to a DAG */
static void addPathToDAG (DAG& dag, BasePath* path) {
  vector<int> tmp;

  for (int ver=0; ver < path->length(); ver++)
    tmp.push_back(path->GetVertex(ver)->getID());

  dag.addLinks(tmp);
}


set<pair<int,int> > Topology::FindLooplessResilientDAG(int s, int d, float alpha)
{

  SubGraph my_graph(this);
  
  YenTopKShortestPathsAlg yenAlg(my_graph, my_graph.get_vertex(s), my_graph.get_vertex(d));

  BasePath* path;
  bool found;
  int Lsu, Lud;
  
  BasePath* shortest = yenAlg.get_shortest_path(my_graph.get_vertex(s), my_graph.get_vertex(d));

  cout << "shortest path: ";
  shortest->PrintOut(cout);
  
  int minlen = shortest->length() - 1;  // length returns the number of vertices, not the number of hops
  
  // create a DAG including the shortest path
  DAG dag(node.size());

  addPathToDAG(dag, shortest);
  
  // run a DFS to sort the nodes, find predecessors
  dag.DFS(s);
  
  // dag.Print(cout);
  
  // a node is "done" when it is protected against single link/node failures
  vector<bool> done(node.size(), false);
  
  /*
   * nodes are protected starting from the penultimate node in the topological ordering
   */ 
  
  list<int>::const_reverse_iterator u = ++dag.sorted_rbegin();
  set<int>::const_iterator v;
  
  while (u != dag.sorted_rend())
  {

#ifdef DEBUG_RDAS
    cout << endl << "Analyzing node " << *u << endl;
    cout << "num adj: " << dag.getNumAdj(*u) <<  endl;
#endif
    
    // if u is done, continue with its predecessor
    if (done[*u])
    {
      u++;
      continue;
    }
    
    assert(dag.getNumAdj(*u)>0);
    
    // we check whether u is able to reach d despite the failure of each of its neighbors.
    // As soon as this is not true, an alternative path is sought and added to the dag; then,
    // the while loop starts again from the penultimate node
    // for (set<int>::const_iterator v=dag.adj_begin(*u); v!=dag.adj_end(*u); v++) {
    if (dag.getNumAdj(*u) == 1)
    {
      // copy the current dag and prune the neighbor
      // DAG pruned_dag(dag);

      v = dag.adj_begin(*u);
      
/*      if (*v == d)
	pruned_dag.removeLink(*u,d);
      else
	pruned_dag.removeNode(*v);*/
      
      // check whether a path still exists in pruned_dag between u and d
      // this can be done by running DFS and checking whether d has a valid predecessor
      // pruned_dag.DFS(*u);
	
      // if a path is not found
      // if (pruned_dag.getPred(d) == pruned_dag.getNumNodes()) {
	// we try to find an alternative path in the graph
	SubGraph pruned_graph(my_graph);
	
	if (*v == d)
	{
	  pruned_graph.removeEdge(*u,d);
#ifdef DEBUG_RDAS
	  cout << "prune link " << *u << "->" << d << endl;
#endif
	}
	else
	{
	  pruned_graph.removeVertex(*v);
#ifdef DEBUG_RDAS
	  cout << "prune node " << *v << endl;
#endif
	}

	YenTopKShortestPathsAlg yen(pruned_graph, pruned_graph.get_vertex(*u), pruned_graph.get_vertex(d));

//	cout << (yen.has_next() ? "true" : "false") << endl;
	
	found = false;
	Lsu = dag.getMaxDistFromSource(*u);
	Lud = 0;

//	cout << "Lsu = " << Lsu << endl;

	while(yen.has_next() && !found && (Lsu + Lud <= alpha * minlen) )
	{
		path = yen.next();
		
#ifdef DEBUG_RDAS
		cout << "Trying path... ";
		path->PrintOut(cout);
#endif
		
		// try to add path to the dag and check for loops and maximum length
		DAG augm_dag(dag);

		addPathToDAG(augm_dag, path);
		
		augm_dag.DFS(s);

		if (augm_dag.isAcyclic() && augm_dag.getMaxDistToDest(s) <= alpha * minlen) {
		  found = true;
#ifdef DEBUG_RDAS
		  cout << "OK" << endl;
#endif
		}
#ifdef DEBUG_RDAS
		else if (!augm_dag.isAcyclic())
		  cout << "Loop in the DAG" << endl;
		else
		  cout << "Too long!" << endl;
#endif

		Lud = path->length() - 1;
#ifdef DEBUG_RDAS
		cout << "Lud = " << Lud << endl;
#endif
	}

	if (found)
	{
	  done[*u] = true;

	  // add the path to the DAG
	  addPathToDAG(dag, path);
	  
	  // sort the nodes in the new dag
	  dag.DFS(s);
	  
	  // restart from the penultimate node
	  u = ++dag.sorted_rbegin();
	  
//	  dag.Print(cout);
	  
	  continue;
	}
      //}
    }

    done[*u] = true;
    u++;
    
    // dag.Print(cout);
  }
  
  // dagLinks contains all the links of the dag
  set<pair<int,int> > dagLinks;
  
  for (list<int>::const_iterator p=dag.sorted_begin(); p!=dag.sorted_end(); p++)
    for (set<int>::const_iterator q=dag.adj_begin(*p); q!=dag.adj_end(*p); q++)
      dagLinks.insert(pair<int,int>(*p,*q));
  
  return dagLinks;
}


/* copies the flow value in the availBW field */
void Topology::InitRouting()
{
	// for every link
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			v->availBW = v->flow;
}


/*
   Route flows along the shortest path
   AvailBW is used as link capacity
   Returns the cross-section goodput
 */
double Topology::RouteFlows(int NumFlows, RndVar *Source, RndVar *Dest, RndVar *Bwdr, double seed)
{
	MotherOfAll urng(seed);              // declare uniform random number generator
	Random::Set(urng);                   // set urng as generator to be used

	vector<TLinkItrs>* path;
	int src, dst, Accepted = 0;
	double bw, CSgoodput = 0.;

	for (int flowcounter=0; flowcounter<NumFlows; flowcounter++)
	{
		src = aggregation_list[(int)floor(Source->rv->Next())];
		dst = gateway_list[(int)floor(Dest->rv->Next())];
		bw = (double)Bwdr->rv->Next();
		cout << "Flow " << flowcounter << ": src=" << src << " dst=" << dst << " bw=" << bw << "\t Path: ";

		linkPrune_Bwd prune(bw);
		path = ShortestPath(src, dst, prune);

		if (!path->empty())
		{
			Accepted ++;
			CSgoodput += bw;

			/* decrease availBW along the path and print the path */
			// vector<int>::iterator cur_node = path.begin();
			// vector<TGraphNeighbor>::iterator nexthop;

			cout << path->begin()->first->info->id << " ";

			for (vector<TLinkItrs>::iterator l=path->begin(); l!=path->end(); l++) {
				l->second->availBW -= 1.0;
				cout << l->second->info->id << " ";
			}

/*			while (cur_node+1 != path.end() )
			{
				nexthop = find_if(node[*cur_node].neighbor.begin(), node[*cur_node].neighbor.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(),*(cur_node+1)), ptr_fun<TGraphNeighbor,int>(getNodeID) ));
		
				assert( nexthop != node[*cur_node].neighbor.end() );
		
				nexthop->availBW -= bw;
		
				cout << *++cur_node << " ";
			}*/
		}
		else
			cout << "Not found";
		cout << endl;
	}
	cout << endl << "Accepted: " << Accepted << " of " << NumFlows << endl << "Cross-section Goodput: " << CSgoodput << endl << endl;
	return CSgoodput;
}

int Topology::countRadiosOff() {
  int i = 0;
  for (vector<TGraphNode>::iterator n=node.begin(); n!=node.end(); n++)
    i += n->info->countRadiosWithChannel(0);
  return i;
}

int Topology::countNodesOff() {
  int i=0;
  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
    if (u->info->countRadiosWithChannel(0) == u->info->numAvailRadios())
      i++;
  return i;
}

int Topology::countNumRadios() {
  int i = 0;
  for (vector<TGraphNode>::iterator n=node.begin(); n!=node.end(); n++)
    i += n->info->numAvailRadios();
  return i;
}


int Topology::countRadioChanged (Topology* Tnew, vector<multiset<unsigned int> >& oldChSet, vector<multiset<unsigned int> >& newChSet)
{
  multiset<unsigned int> commonCh;
  int changes = 0;
  
  // the old and new topologies must have the same number of nodes
  assert (node.size() == Tnew->node.size());
  
  oldChSet.resize(node.size());
  newChSet.resize(node.size());

  for (int u = 0; u < (int)node.size(); u++)
  {
    // this node must have the same number of radios in both topologies
    assert (Tnew->node[u].info->numAvailRadios() == node[u].info->numAvailRadios());
    
    oldChSet[u].clear();
    newChSet[u].clear();
    commonCh.clear();

    for (unsigned int j=0; j<node[u].info->numAvailRadios(); j++)
    {
      oldChSet[u].insert (node[u].info->radio[j].channel);
      newChSet[u].insert (Tnew->node[u].info->radio[j].channel);
    }
    
    set_intersection (oldChSet[u].begin(), oldChSet[u].end(), newChSet[u].begin(), newChSet[u].end(), inserter(commonCh, commonCh.begin()));
    
    for (multiset<unsigned int>::iterator c = commonCh.begin(); c != commonCh.end(); c++)
    {
      oldChSet[u].erase (*c);
      newChSet[u].erase (*c);
    }
    
    changes += newChSet[u].size();
  }

  return changes;
}


/* Computes the power consumption of all the nodes and the time usage (tau) for each of their radios.
  * Results are stored in the vector consumption (which can also be empty when passed) */  
void Topology::calcAllNodesPowerConsumption(vector<TPower>& consumption)
{
  consumption.clear();
  
  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
//     consumption.push_back(TPower(u->info->numAvailRadios(), pm.BasePowerConsumption));
    consumption.push_back(TPower(u->info->numAvailRadios(), u->info->BasePowerConsumption));
  
  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
    {
//       consumption[u->info->id].power += v->flow / v->rate * (pm.rho_tx - pm.rho_idle);
      consumption[u->info->id].power += v->flow / v->rate * ((*v->radio->rho_tx)(DbmToW(TxPowerLevel(v->txpower))) - v->radio->rho_idle);

//       consumption[u->info->id].tau.at(v->radio - u->info->radio) += v->flow / v->rate;
      consumption[u->info->id].tau.at(v->radio - &u->info->radio[0]) += v->flow / v->rate;

//       pair<bool,unsigned int> ch = SearchChannel(v->info, v->radio->channel);
      RadioIfaceIterator ri = v->info->findRadioWithChannel(v->radio->channel);
//       if (ch.first)
      if (ri != v->info->radio.end())
      {
// 	consumption[v->info->id].power += v->flow / v->rate * (pm.rho_rx - pm.rho_idle);
	consumption[v->info->id].power += v->flow / v->rate * (ri->rho_rx - ri->rho_idle);

// 	consumption[v->info->id].tau.at(ch.second) += v->flow / v->rate;
	consumption[v->info->id].tau.at(distance(v->info->radio.begin(),ri)) += v->flow / v->rate;
      }
    }

  for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
  {
    if (u->info->countRadiosWithChannel(0) == u->info->numAvailRadios())
//       consumption.at(u->info->id).power = pm.StandbyPowerConsumption;
      consumption.at(u->info->id).power = u->info->StandbyPowerConsumption;
    else
//       consumption.at(u->info->id).power += u->info->countRadiosWithChannel(0)*pm.rho_sleep + (u->info->numAvailRadios()-u->info->countRadiosWithChannel(0))*pm.rho_idle;
      for (RadioIfaceIterator ri = u->info->radio.begin(); ri != u->info->radio.end(); ri++)
// 	consumption.at(u->info->id).power += u->info->countRadiosWithChannel(0)*pm.rho_sleep + (u->info->numAvailRadios()-u->info->countRadiosWithChannel(0))*pm.rho_idle;
	consumption.at(u->info->id).power += (ri->channel ? ri->rho_idle : ri->rho_sleep);
  }
}


/* checks whether the power constraints on all the nodes and the time constraints (tau) on all the radios are satisfied.
 * If the vector of power consumptions is available (i.e., calcAllNodesPowerConsumption has been called), its address
 * can be passed so that calcAllNodesPowerConsumption is not called again
 */
bool Topology::checkPowerConstraints (vector<float>* nodePowerConstr, vector<TPower>* consumption)
{
  bool computed = false;

  if (!consumption)
  {
    consumption = new vector<TPower>;
    computed = true;
    calcAllNodesPowerConsumption(*consumption);
  }

  for (unsigned int u=0; u<node.size(); u++)
  {
    if (consumption->at(u).power > nodePowerConstr->at(u))
    {
      if (computed)
	delete consumption;
      return false;
    }

    for (unsigned int r=0; r<node[u].info->numAvailRadios(); r++)
      if (consumption->at(u).tau.at(r) > 1.f)
      {
	if (computed)
	  delete consumption;
	return false;
      }
  }

  if (computed)
    delete consumption;

  return true;
}


/* computes the total power consumption of the whole network.
 * If the vector of power consumptions is available (i.e., calcAllNodesPowerConsumption has been called), its address
 * can be passed so that calcAllNodesPowerConsumption is not called again
 */
float Topology::TotalPowerConsumption (vector<TPower>* consumption)
{
  bool computed = false;

  if (!consumption)
  {
    consumption = new vector<TPower>;
    computed = true;
    calcAllNodesPowerConsumption(*consumption);
  }

  float Ptot = 0.f;
  for (vector<TPower>::iterator i=consumption->begin(); i!=consumption->end(); i++)
    Ptot += i->power;
  
  if (computed)
    delete consumption;

  return Ptot;
}


void Topology::printTopology(FILE *f)
/* file f must be already open for writing*/
{
	fprintf(f,"Number of available channels: %i\n",numAvailChannels);
	fprintf(f,"Node id\t type\t x_pos\t y_pos\t #radios #radio_off\t channels\t \n");
	for (vector<TGraphNode>::const_iterator i=node.begin(); i!=node.end(); i++)
	{
		fprintf(f,"%i\t %c\t %.2f\t %.2f\t %i\t %i\t \t",i->info->id, MeshTypes[i->info->role], i->info->xpos, i->info->ypos, i->info->numAvailRadios(), i->info->countRadiosWithChannel(0));
		for (unsigned int j=0; j<i->info->numAvailRadios(); j++)
			fprintf(f,"%d ",i->info->radio[j].channel);
		fprintf(f,"\t\t ");
//		for (unsigned int j=0; j<i->info->numAvailRadios(); j++)
//			fprintf(f,"%.2f ",i->info->radio[j].maxTXrange);
		fprintf(f,"\n");
	}

	fprintf(f,"\nNumber of radios off: %d\n",countRadiosOff());

	fprintf(f,"\nTotal number of radios: %d\n",countNumRadios());

	fprintf(f,"\nAggregation devices\n");
	for (vector<int>::const_iterator i=aggregation_list.begin(); i!=aggregation_list.end(); i++)
		fprintf(f,"%i ",*i);

	fprintf(f,"\nGateways\n");
	for (vector<int>::const_iterator i=gateway_list.begin(); i!=gateway_list.end(); i++)
		fprintf(f,"%i ",*i);

	fprintf(f,"\n\nNeighbors\n");
	for (vector<TGraphNode>::const_iterator i=node.begin(); i!=node.end(); i++)
	{
		fprintf(f,"%i:",i->info->id);
		for (list<TGraphNeighbor>::const_iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
		{
// 		  unsigned int w;
		  fprintf(f,"\t%i (d=%.2f",j->info->id, dist(i->info->id,j->info->id));
//			fprintf(f,", B=%.2lf",j->availBW);
// 			for (w=0; w<NUMTXPWR && availTxPwr[w]!=j->power; w++);
			fprintf(f,", f=%.2lf, C=%.2lf, P=%f dbm", j->flow, j->rate, TxPowerLevel(j->txpower));
			if (j->radio != NULL)
				fprintf(f,", c=%i",j->radio->channel);
//				fprintf(f,", r=%i, TXr=%.2f",j->radio - i->info->radio, j->radio->curTXrange);
			fprintf(f,") {");
		for (map<pair<int,int>,double>::const_iterator q=j->f_e2e.begin(); q!=j->f_e2e.end(); q++)
		  fprintf(f,"f[%i,%i] = %lf ", q->first.first, q->first.second, q->second);
		fprintf(f,"}\n");
		}
	}
}


void Topology::printTopology(const char *filename)
{
	FILE *f;

	if ((f = fopen(filename, "w")) == NULL)
	{
		fprintf(stderr, "ERR: Cannot open file %s to write topology\n",filename);
		return;
	}
	printTopology(f);
	fclose(f);
}


/* returns the distance between two nodes */
// float Topology::dist(int x, int y)
// {
// 	return sqrtf(pow(fabsf(node[x].info->xpos - node[y].info->xpos),2)+pow(fabsf(node[x].info->ypos - node[y].info->ypos),2));
// }


/* creates an octave file named draw_<toponame>.m to draw the topology */
void Topology::drawTopology()
{
	const float ratio = 0.05;
	float xmin = node[0].info->xpos,
	      xmax = 0.,
	      ymin = node[0].info->xpos,
	      ymax = 0.,
	      radius, dist, Dx, Dy;

	/* compute the minimum and maximum coordinates */
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
	{
		if (u->info->xpos < xmin)
			xmin = u->info->xpos;
		else if (u->info->xpos > xmax)
			xmax = u->info->xpos;

		if (u->info->ypos < ymin)
			ymin = u->info->ypos;
		else if (u->info->ypos > ymax)
			ymax = u->info->ypos;
	}

	radius = ratio * (xmax - xmin);

	char fname[100];
	sprintf(fname,"draw_%s.m",toponame);

	ofstream f;
	f.open(fname);

	f << "printf(" << '"' << "Legenda:\\n\\n 54Mbps -> green   (thick) \\n 48Mbps -> cyan    (thick) \\n 36Mbps -> magenta (thick) \\n 24Mbps -> blue    (thick)\\n" << '"' << ")" << endl;
	f << "printf(" << '"' << " 18Mbps -> magenta \\n 12Mbps -> blue \\n  9Mbps -> cyan \\n  6Mbps -> green\\n\\n" << '"' << ")" << endl;

	/* clear the current figure and hold on */
	//f << "#! /usr/bin/octave -qf" << endl << endl;
	f << "clf" << endl << "hold on" << endl;

	/* draw a labelled circle for each node */
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		f << "x = " << u->info->xpos - radius << ":0.1:" << u->info->xpos + radius << ";" << endl
		  << "plot(x, " << u->info->ypos << " + sqrt( " << radius*radius << " - (x - " << u->info->xpos << ").^2), \"" << 3-u->info->role << "\" )" << endl
		  << "plot(x, " << u->info->ypos << " - sqrt( " << radius*radius << " - (x - " << u->info->xpos << ").^2), \"" << 3-u->info->role << "\" )" << endl
		  << "text(" << u->info->xpos << ", " << u->info->ypos << ", \"" << u->info->id << "\")" << endl;

	/* draw links */
	for (vector<TGraphNode>::iterator u=node.begin(); u!=node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->xpos <= v->info->xpos)          /* draw one undirected link between two connected nodes */
			{
				/* distance between the centers of the nodes */
				dist = sqrt(pow(u->info->xpos - v->info->xpos,2) + pow(u->info->ypos - v->info->ypos,2));
				if (dist <= 2 * radius)            /* nodes intersect with each other */
					continue;
				Dx = radius * fabs(u->info->xpos - v->info->xpos) / dist;
				Dy = radius * fabs(u->info->ypos - v->info->ypos) / dist;

				if (v->info->xpos - u->info->xpos > 0.1 + 2 * Dx)  /* there is at least one element in the vector x (non-vertical lines) */
					f << "x = " << u->info->xpos + Dx << ":0.1:" << v->info->xpos - Dx << ";" << endl
					<< "plot(x, " << u->info->ypos << " + " << (v->info->ypos - u->info->ypos) / (v->info->xpos - u->info->xpos)
					<< " * (x - " << u->info->xpos << ")";
				else
				{
					if (u->info->ypos < v->info->ypos)
						f << "x = " << u->info->ypos + Dy << ":0.1:" << v->info->ypos - Dy << ";" << endl;
					else
						f << "x = " << v->info->ypos + Dy << ":0.1:" << u->info->ypos - Dy << ";" << endl;
					f << "plot(" << u->info->xpos << " + " << (v->info->xpos - u->info->xpos) / (v->info->ypos - u->info->ypos)
					<< " * (x - " << u->info->ypos << "), x";
				}
				/* set line color and width */
				f << "," << '"' << (v->rate < 20 ? ((int)v->rate) % 5 + 1 : ((int)v->rate) % 11 % 9 + 1 ) << '"' << ", " << '"' <<
				     "linewidth" << '"' << ", " << (v->rate < 20 ? 1 : 3) << " )" << endl;
			}

	f << "hold off" << endl;

	f.close();
}

#if 0
int createTopology(Topology **T)
{
	char f_topology[60],    /* name of the file containing the topology */
		msg[80]="Unexpected value or EOF while reading ",
		msg_power[30],
		msg_rate[30],
		buildGiFuncStr[30],
		filenameLinks[256];

	unsigned int NumChannels, power;
	int min_rate, max_rate;
	
	/* read the number of available channels */
	READ_PARAM(stdin,"%ud",&NumChannels,strcat(msg,"the number of channels"));
	
	/* read topology file name */
	READ_PARAM(stdin,"%59s",f_topology,strcat(msg,"topology file name"));
	
// 	*T = new Topology(NumChannels);
// 	
// 	(*T)->getNodeInfo(f_topology);

	*T = new Topology (f_topology);

	(*T)->numAvailChannels = NumChannels;

	string s(f_topology);
	strcpy((*T)->topopath, s.c_str());
	string::size_type pos2 = s.rfind(".");
	string::size_type pos1 = s.rfind("/");
	strcpy((*T)->toponame, s.substr(pos1+1,pos2-pos1-1).c_str());

	/* read buildGi function */
	READ_PARAM(stdin,"%29s",buildGiFuncStr,strcat(msg,"buildGiFunc name"));
	
	sprintf(msg_power,"the power level (0-%ld expected)",NUMTXPWR-1);
	sprintf(msg_rate,"the rate index (0-%ld expected)",NUMRATES-1);

	if (!strcmp(buildGiFuncStr,"FixedPowerMaxAllowRate"))
	{
		/* read the power value (dbm) */
		READ_PARAM(stdin,"%ud",&power,strcat(msg,msg_power));
		if ( (power < 0) || (power > NUMTXPWR-1) )
			PRINT_ERROR(strcat(msg,msg_power));
		/* read the min rate index */
		READ_PARAM(stdin,"%d",&min_rate,strcat(msg,msg_rate));
		if ( (min_rate < 0) || (min_rate > (int)(NUMRATES-1)) )
			PRINT_ERROR(strcat(msg,msg_rate));
		/* read the max rate index */
		READ_PARAM(stdin,"%d",&max_rate,strcat(msg,msg_rate));
		if ( (max_rate < 0) || (max_rate > (int)(NUMRATES-1)) )
			PRINT_ERROR(strcat(msg,msg_rate));
		(*T)->buildGi_FixedPowerMaxAllowRate(availTxPwr[power], min_rate, max_rate);
		(*T)->maxP = power;
		(*T)->min_rate = min_rate;
		(*T)->max_rate = max_rate;
	} else if (!strcmp(buildGiFuncStr,"FromFile"))
	{
		READ_PARAM(stdin,"%255s",filenameLinks,strcat(msg,"Filename links missing"));
		cout<<"building topo from file\n";
		(*T)->buildGi_FromFile(filenameLinks);
	}  
/*	else if (!strcmp(buildGiFuncStr,"FixedPowerMaxRate"))
	{
		// read the power value (dbm)
		READ_PARAM(stdin,"%ud",&power,strcat(msg,msg_power));
		if ( (power < 0) || (power > NUMTXPWR-1) )
			PRINT_ERROR(strcat(msg,msg_power));
		(*T)->buildGi_FixedPowerMaxRate(availTxPwr[power]);
		(*T)->maxP = power;
	}*/
/*	else if (!strcmp(buildGiFuncStr,"StepTXrate"))
		buildGiFunc = StepTXrate;
	else // default
		buildGiFunc = OnOffTXrate;
	*/
	return 1;
}
#endif

/* The caller must declare a map<pair<int,int>,double> and
 *  pass its address. The map will be filled in with the amount
 *  of flow for each demand
 * Otherwise, pass ret=NULL */
bool Topology::checkE2Eflows(map<pair<int,int>,double>* ret) {
  vector<map<pair<int,int>,double > > in(node.size()), out(node.size()), diff(node.size());
  
  if (ret)
    ret->clear();
  
  for (vector<TGraphNode>::const_iterator u=node.begin(); u!=node.end(); u++)
    for (list<TGraphNeighbor>::const_iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      for (map<pair<int,int>,double>::const_iterator f=v->f_e2e.begin(); f!=v->f_e2e.end(); f++) {
	out[u->info->id][f->first] += f->second;
	in[v->info->id][f->first] += f->second;
	diff[u->info->id][f->first] += f->second;
	diff[v->info->id][f->first] -= f->second;
      }
    }

#if 0
  /*uncomment for detailed DEBUG */
  for (int i=0; i<node.size(); i++) {
    cout << "Nodo " << i << ": in: ";
    for (map<pair<int,int>,double>::const_iterator f=in[i].begin(); f!=in[i].end(); f++)
      cout << "[" << f->first.first << "," << f->first.second << "] = " << f->second << "  ";
    cout << endl << "Nodo " << i << ": out: ";
    for (map<pair<int,int>,double>::const_iterator f=out[i].begin(); f!=out[i].end(); f++)
      cout << "[" << f->first.first << "," << f->first.second << "] = " << f->second << "  ";
    cout << endl << "Nodo " << i << ": diff: ";
    for (map<pair<int,int>,double>::const_iterator f=diff[i].begin(); f!=diff[i].end(); f++)
      cout << "[" << f->first.first << "," << f->first.second << "] = " << f->second << "  ";
    cout << endl << endl;
  }
  cout << endl << endl;
#endif

  bool ok = true;
  
  cout << endl;
 
  for (int i=0; i<(int)node.size(); i++) {
    /* if node i is an edge node (aggregated or gateway) */
    if ( (find(aggregation_list.begin(), aggregation_list.end(), i) != aggregation_list.end() )  
      || (find(gateway_list.begin(), gateway_list.end(), i) != gateway_list.end() )  ) {

      for (map<pair<int,int>,double>::const_iterator f=in[i].begin(); f!=in[i].end(); f++) {
      /* if the in entry has node i as source node, it must be null  */
	if ( f->first.first == i  && fabs(f->second) > almost_null ) {
	  cerr << "WARNING: Node " << i << ": entering non-null amount (" << f->second << ") of e2e flow [" << f->first.first << "," << f->first.second << "]" << endl;
	  ok = false;
	}
      /* if the in entry has node i as destination node, print the amount of flow entering node i */
	else if ( f->first.second == i )
	  cout << "Node " << i << ": entering " << f->second << " of e2e flow [" << f->first.first << "," << f->first.second << "]" << endl;
      }
      for (map<pair<int,int>,double>::const_iterator f=out[i].begin(); f!=out[i].end(); f++) {
      /* if the out entry has node i as destination node, it must be null  */
	if ( f->first.second== i  && fabs(f->second) > almost_null ) {
	  cerr << "WARNING: Node " << i << ": leaving non-null amount (" << f->second << ") of e2e flow [" << f->first.first << "," << f->first.second << "]" << endl;
	  ok = false;
	}
      /* if the out entry has node i as source node, print the amount of flow leaving node i */
	else if ( f->first.first == i ) {
	  cout << "Node " << i << ": leaving " << f->second << " of e2e flow [" << f->first.first << "," << f->first.second << "]" << endl;
	  if (ret)
	    ret->insert(*f);
	}
      }
      for (map<pair<int,int>,double>::const_iterator f=diff[i].begin(); f!=diff[i].end(); f++) {
      /* if the diff entry hasn't node i as source or destination node, the diff must be null */
	if ( f->first.first != i  && f->first.second != i  && fabs(f->second) > almost_null ) {
	  cerr << "WARNING: Node " << i << ": the amount of e2e flow [" << f->first.first << "," << f->first.second << "] leaving the node does not equal the amount of flow entering" << endl;
	  ok = false;
	}
      }
    }
    else {
      /* node i is an intermediate node */
      for (map<pair<int,int>,double>::const_iterator f=diff[i].begin(); f!=diff[i].end(); f++)
      /* if the diff entry hasn't node i as source or destination node, the diff must be null */
	if ( f->first.first != i  && f->first.second != i  && fabs(f->second) > almost_null ) {
	  cerr << "WARNING: Node " << i << ": the amount of e2e flow [" << f->first.first << "," << f->first.second << "] leaving the node does not equal the amount of flow entering" << endl;
	  ok = false;
	}
      }
  }

return ok;
}

#if 0
void Topology::saveConfig(const char* fname, PFR_method* pfr, bool commented)
{
  fstream f(fname, fstream::out);
  char str[3];
  
  if (commented)
    strcpy (str,"# ");
  else
    strcpy (str, " ");
  
  f << str << numAvailChannels << endl << str << topopath << endl;
  
  f << str << "FixedPowerMaxAllowRate " << maxP << " " << min_rate << " " << max_rate << endl;
  
  f << str << pfr->getStr() << endl;

//   f << str << "Power Consumption model: " << pm.name << endl;

  f.close();
}
#endif

void Topology::SerializeTo (const char *filename)
{
  ofstream ofs (filename, ofstream::binary);
  int tmp, n;
  
  // write member variables
  ofs.write ((const char*)&numAvailChannels, sizeof(unsigned int));
  ofs.write ((const char*)&maxP, sizeof(int));
  ofs.write ((const char*)&min_rate, sizeof(int));
  ofs.write ((const char*)&max_rate, sizeof(int));
//   ofs.write ((const char*)&pm, sizeof(PowerConsumptionModel));
  ofs.write ((const char*)&throughput, sizeof(float));

  tmp = strlen (toponame);
  ofs.write ((const char*)&tmp, sizeof(int));
  ofs.write (toponame, tmp);
  
  tmp = strlen (topopath);
  ofs.write ((const char*)&tmp, sizeof(int));
  ofs.write (topopath, tmp);

  // write aggregation list
  tmp = aggregation_list.size();
  ofs.write ((const char*)&tmp, sizeof(int));
  for (int i=0; i<tmp; i++)
    ofs.write ((const char*)&aggregation_list[i], sizeof(int));
  
  // write gateway list
  tmp = gateway_list.size();
  ofs.write ((const char*)&tmp, sizeof(int));
  for (int i=0; i<tmp; i++)
    ofs.write ((const char*)&gateway_list[i], sizeof(int));
  
  // write node info (MeshNode objects)
  tmp = node.size();
  ofs.write ((const char*)&tmp, sizeof(int));
  
  for (int i=0; i<tmp; i++)
    node[i].info->SerializeTo (ofs);
    
  for (int i=0; i<tmp; i++)
  {
    n = node[i].neighbor.size();
    ofs.write ((const char*)&n, sizeof(int));
    for (list<TGraphNeighbor>::iterator j=node[i].neighbor.begin(); j!=node[i].neighbor.end(); j++)
      j->SerializeTo (ofs);
  }
    
  ofs.close ();
}

void Topology::DeserializeFrom (const char *filename)
{
  int tmp, elem, n;
  ifstream ifs (filename, ifstream::binary);

  if (!ifs)
  {
    cerr << endl << "*** ERROR: Cannot find the file storing the serialized topology (" << filename << ") ***" << endl << endl;
    return;
  }

  // read member variables
  ifs.read ((char*)&numAvailChannels, sizeof(unsigned int));
  ifs.read ((char*)&maxP, sizeof(int));
  ifs.read ((char*)&min_rate, sizeof(int));
  ifs.read ((char*)&max_rate, sizeof(int));
//   ifs.read ((char*)&pm, sizeof(PowerConsumptionModel));
  ifs.read ((char*)&throughput, sizeof(float));

  ifs.read ((char*)&tmp, sizeof(int));
  ifs.read (toponame, tmp);
  toponame[tmp] = '\0';

  ifs.read ((char*)&tmp, sizeof(int));
  ifs.read (topopath, tmp);
  topopath[tmp] = '\0';

  // read aggregation list
  ifs.read ((char*)&tmp, sizeof(int));
  for (int i=0; i<tmp; i++)
  {
    ifs.read ((char*)&elem, sizeof(int));
    aggregation_list.push_back (elem);
  }

  // read gateway list
  ifs.read ((char*)&tmp, sizeof(int));
  for (int i=0; i<tmp; i++)
  {
    ifs.read ((char*)&elem, sizeof(int));
    gateway_list.push_back (elem);
  }

  // resize the node vector
  ifs.read ((char*)&tmp, sizeof(int));
  node.clear();

  TGraphNode gnode;
  for (int i=0; i<tmp; i++)
  {
    gnode.info = MeshNode::DeserializeFrom (ifs);
    node.push_back (gnode);
  }

  for (int i=0; i<tmp; i++)
  {
    ifs.read ((char*)&n, sizeof(int));
    for (int j=0; j<n; j++)
      node[i].neighbor.push_back (*TGraphNeighbor::DeserializeFrom (ifs, this, i));
  }

  ifs.close ();
}


void MeshNode::SerializeTo (ofstream& ofs)
{
  ofs.write ((const char*)&id, sizeof(int));
  ofs.write ((const char*)&xpos, sizeof(float));
  ofs.write ((const char*)&ypos, sizeof(float));
  ofs.write ((const char*)&role, sizeof(int));
  ofs.write ((const char*)&BasePowerConsumption, sizeof(float));
  ofs.write ((const char*)&StandbyPowerConsumption, sizeof(float));

  unsigned int numAvailRadios = radio.size();
  ofs.write ((const char*)&numAvailRadios, sizeof(unsigned int));
  
  for (unsigned int i=0; i<numAvailRadios; i++)
    radio[i].SerializeTo (ofs);
}

MeshNode* MeshNode::DeserializeFrom (ifstream& ifs)
{
  int _id;
  TMeshRouter _role;
  float _xpos, _ypos, bpc, spc;
  unsigned int _numAvailRadios;
  
  ifs.read ((char*)&_id, sizeof(int));
  ifs.read ((char*)&_xpos, sizeof(float));
  ifs.read ((char*)&_ypos, sizeof(float));
  ifs.read ((char*)&_role, sizeof(int));
  ifs.read ((char*)&bpc, sizeof(float));
  ifs.read ((char*)&spc, sizeof(float));
  ifs.read ((char*)&_numAvailRadios, sizeof(unsigned int));
  
  MeshNode* m = new MeshNode (_id, _xpos, _ypos, _role, _numAvailRadios, bpc, spc);
  
  for (unsigned int i=0; i<_numAvailRadios; i++)
    m->radio[i].DeserializeFrom (ifs);
  
  return m;
}

void foRadioTxPowerLinear::SerializeTo (ofstream& ofs)
{
  int slen = name.length();
  ofs.write ((const char*)&slen, sizeof(int));
  ofs.write (name.c_str(), slen);
  
  ofs.write ((const char*)&a, sizeof(float));
  ofs.write ((const char*)&b, sizeof(float));
}

foRadioTxPower* foRadioTxPower::DeserializeFrom (ifstream& ifs)
{
  int slen;

  ifs.read ((char*)&slen, sizeof(int));
  
  char tmp[slen+1];
  
  ifs.read (tmp, slen);
  tmp[slen] = '\0';
  
  if (!strcmp(tmp, foRadioTxPowerLinear::name.c_str()))
    return foRadioTxPowerLinear::DeserializeFrom (ifs);

  cerr << endl << "** ERROR: Unknown RadioTxPower function object name! **" << endl << endl;
  return NULL;
}

foRadioTxPowerLinear* foRadioTxPowerLinear::DeserializeFrom (ifstream& ifs)
{
  float a, b;
  
  ifs.read ((char*)&a, sizeof(float));
  ifs.read ((char*)&b, sizeof(float));
  
  return new foRadioTxPowerLinear (a, b);
}

void RadioIface::SerializeTo (ofstream& ofs)
{
  ofs.write ((const char*)&channel, sizeof(unsigned int));
  ofs.write ((const char*)&DBgain, sizeof(float));
  ofs.write ((const char*)&minangle, sizeof(unsigned int));
  ofs.write ((const char*)&maxangle, sizeof(unsigned int));
  rho_tx->SerializeTo (ofs);
  ofs.write ((const char*)&rho_rx, sizeof(float));
  ofs.write ((const char*)&rho_idle, sizeof(float));
  ofs.write ((const char*)&rho_sleep, sizeof(float));
}

void RadioIface::DeserializeFrom (ifstream& ifs)
{
  ifs.read ((char*)&channel, sizeof(unsigned int));
  ifs.read ((char*)&DBgain, sizeof(float));
  ifs.read ((char*)&minangle, sizeof(unsigned int));
  ifs.read ((char*)&maxangle, sizeof(unsigned int));
  rho_tx = foRadioTxPower::DeserializeFrom (ifs);
  ifs.read ((char*)&rho_rx, sizeof(float));
  ifs.read ((char*)&rho_idle, sizeof(float));
  ifs.read ((char*)&rho_sleep, sizeof(float));
}

void TGraphNeighbor::SerializeTo (ofstream& ofs)
{
  int tmp;
  
  // write the id of the neighbor
  ofs.write ((const char*)&info->id, sizeof(int));

  if (!radio)
    cerr << "ERROR: Trying to serialize a TGraphNeighbor node with NULL radio pointer" << endl << endl;

  assert (radio != 0);
  
  ofs.write ((const char*)&radio->channel, sizeof(unsigned int));
  
  ofs.write ((const char*)&availBW, sizeof(double));
  ofs.write ((const char*)&flow, sizeof(double));
  ofs.write ((const char*)&Utot, sizeof(double));
  ofs.write ((const char*)&rate, sizeof(float));
  ofs.write ((const char*)&txpower, sizeof(float));

  tmp = f_e2e.size();
  ofs.write ((const char*)&tmp, sizeof(int));
  
  for (map<pair<int,int>,double>::iterator m=f_e2e.begin(); m!=f_e2e.end(); m++)
  {
    ofs.write ((const char*)&m->first.first, sizeof(int));
    ofs.write ((const char*)&m->first.second, sizeof(int));
    ofs.write ((const char*)&m->second, sizeof(double));
  }
}

TGraphNeighbor* TGraphNeighbor::DeserializeFrom (ifstream& ifs, Topology* T, int me)
{
  int tmp, s, d;
  unsigned int ch;
  double dem;
  TGraphNeighbor* n;
  
  // read the id of the neighbor;
  ifs.read ((char*)&tmp, sizeof(int));
  
  n = new TGraphNeighbor (T->node[tmp].info);
  
  ifs.read ((char*)&ch, sizeof(unsigned int));
  
//   for (tmp=0; tmp<(int)T->node[me].info->numAvailRadios() && T->node[me].info->radio[tmp].channel != ch; tmp++);
  RadioIfaceIterator ri = T->node[me].info->findRadioWithChannel (ch);
  
//   assert (tmp < (int)T->node[me].info->numAvailRadios());
  assert ( ri != T->node[me].info->radio.end() );
  
//   n->radio = T->node[me].info->radio + tmp;
  n->radio = &(*ri);
  
  ifs.read ((char*)&n->availBW, sizeof(double));
  ifs.read ((char*)&n->flow, sizeof(double));
  ifs.read ((char*)&n->Utot, sizeof(double));
  ifs.read ((char*)&n->rate, sizeof(float));
  ifs.read ((char*)&n->txpower, sizeof(unsigned int));

  ifs.read ((char*)&tmp, sizeof(int));
  
  for (int i=0; i<tmp; i++)
  {
    ifs.read ((char*)&s, sizeof(int));
    ifs.read ((char*)&d, sizeof(int));
    ifs.read ((char*)&dem, sizeof(double));
    
    n->f_e2e[pair<int,int>(s,d)] = dem;
  }
  
  return n;
}



int Topology::buildGi_FromFile(const char * filename)
{
	vector<TGraphNode>::iterator u, v;
	int rate;
	int idu,idv;
	char c;
	string s;

	// read the number of available channels
	//READ_PARAM(f,"%ui",&numAvailChannels,strcat(msg,"number of available channels"));

	ifstream ifs;

	ifs.open (filename);
	
	if (!ifs)
	{
	  cerr << endl << "ERROR: Cannot open file " << filename << " to read topology links" << endl << endl;
	  return -1;
	}

// 	skipComment(f);
	while (ifs >> c)
	{
		switch (c)
		{
			case 'L':
			  readValue (ifs, idu, "the id of the first endpoint");
			  readValue (ifs, idv, "the id of the second endpoint");
			  readValue (ifs, rate, "the rate");
		
			  //cout << "Name "<<nameu<<endl;
			  u=node.begin()+idu;
			  v=node.begin()+idv;
			  cout << "Node "<< idu << " and " << idv <<" are set to be neighbors and the rate is "<< rate<<".\n";
			  addNeighbor(u,v,rate,defaultTxPowerLevel);
			  //addNeighbor(v,u,rate,availTxPwr[15]);
			  break;

			case '#':
			  getline (ifs, s);
			  break;

			default:
			  break;
		}
// 		skipComment(f);
	}

	return 0;
}

