// -*- C++ -*-

/* myparser.cc
 *
 * Copyright (C) 2002 The libxml++ development team
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mpls-parser.h"
#include "WMutils.h"

#include <glibmm/convert.h> //For Glib::ConvertError
#include <iostream>
#include <cassert>

using namespace std;




/** WiMesh config parser **/

SplitRatiosParser::SplitRatiosParser ()
  : xmlpp::SaxParser(),
    rdas (true),
    alpha (1.5f)
{
}

SplitRatiosParser::~SplitRatiosParser()
{
}

void SplitRatiosParser::on_start_document()
{
  cout << "Reading calcMplsRatios config..." << endl << endl;
}

void SplitRatiosParser::on_end_document()
{
  cout << "Finished reading calcMplsRatios config" << endl << endl;
}

void SplitRatiosParser::on_start_element(const Glib::ustring& name,
                                   const AttributeList& attributes)
{

  if (name == "Topology")
  {
    if (!topofile.empty())
    {
      // a topology file name has been already specified
      cerr << endl << "Ignoring duplicate Topology element" << endl << endl;
      return;
    }

    // "serialized_file" is expected as the first (and unique) attribute

    if (attributes.empty())
    {
      cerr << endl << "Ignoring Topology element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "serialized_file")
    {
      cerr << endl << "Ignoring Topology element with first attribute other than serialized_file" << endl << endl;
      return;
    }

    topofile = attributes.front().value;

  }  /* end Topology */

  else if (name == "SubgraphMethod")
  {
    if (attributes.empty())
    {
      cerr << endl << "Ignoring SubgraphMethod element with no attributes" << endl << endl;
      return;
    }

    string sgm;

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("name", sgm);
      RP.read ("alpha", alpha);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a SubgraphMethod element" << endl << endl;
    }

    if (sgm == "rdas")
      rdas = true;
    else if (sgm == "ksp")
      rdas = false;
    else if (!sgm.empty())
      cerr << endl << "Ignoring invalid value (" << sgm << ") for the name attribute of a SubgraphMethod element" << endl << endl;

  }  /* end SubgraphMethod */

  else if (name == "numericOption")
  {
    string optname;
    double optvalue;
    bool hasValue = false;
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("name", optname);
      RP.read ("value", optvalue);

      if (iter->name == "value")
	hasValue = true;

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a numericOption element" << endl << endl;
    }

    if (optname.empty() || !hasValue)
    {
      cerr << endl << "Ignoring numericOption element with missing attributes" << endl << endl;
      return;
    }

    numericValues.push_back(pair<string,double>(optname,optvalue));

  }  /* end numericOption */

  else if (name == "integerOption")
  {
    string optname;
    int optvalue;
    bool hasValue = false;
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("name", optname);
      RP.read ("value", optvalue);

      if (iter->name == "value")
	hasValue = true;

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for an integerOption element" << endl << endl;
    }

    if (optname.empty() || !hasValue)
    {
      cerr << endl << "Ignoring integerOption element with missing attributes" << endl << endl;
      return;
    }

    integerValues.push_back(pair<string,int>(optname,optvalue));

  }  /* end integerOption */

  else if (name == "stringOption")
  {
    string optname;
    string optvalue;
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("name", optname);
      RP.read ("value", optvalue);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a stringOption element" << endl << endl;
    }

    if (optname.empty() || optvalue.empty())
    {
      cerr << endl << "Ignoring stringOption element with missing attributes" << endl << endl;
      return;
    }

    stringValues.push_back(pair<string,string>(optname,optvalue));

  }  /* end stringOption */

  else if (name != "MplsRatiosCfg")
  {
    cerr << endl << "Ignoring invalid element (" << name << ")" << endl << endl;
  }
}

void SplitRatiosParser::on_end_element(const Glib::ustring& name)
{
//   if (name == "WiMeshCfg")
//     // stop the parser
//     throw xmlpp::exception("Done");
}

void SplitRatiosParser::on_characters(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_characters(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "SplitRatiosParser::on_characters(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void SplitRatiosParser::on_comment(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_comment(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "SplitRatiosParser::on_comment(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void SplitRatiosParser::on_warning(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_warning(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "SplitRatiosParser::on_warning(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void SplitRatiosParser::on_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "SplitRatiosParser::on_error(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void SplitRatiosParser::on_fatal_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_fatal_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "SplitRatiosParser::on_characters(): Exception caught while converting value for std::cout: " << ex.what() << std::endl;
  }
}

