/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 Network Security Lab, University of Washington, Seattle.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Sidharth Nabar <snabar@uw.edu>, He Wu <mdzz@u.washington.edu>
 */

#include "wimesh-energy-source.h"
#include "ns3/log.h"
#include "ns3/assert.h"
#include "ns3/double.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/simulator.h"


namespace ns3 {

WiMeshEnergySourceHelper::WiMeshEnergySourceHelper ()
{
  m_wimeshEnergySource.SetTypeId ("ns3::WiMeshEnergySource");
}

WiMeshEnergySourceHelper::~WiMeshEnergySourceHelper ()
{
}

void
WiMeshEnergySourceHelper::Set (std::string name, const AttributeValue &v)
{
  m_wimeshEnergySource.Set (name, v);
}

Ptr<EnergySource>
WiMeshEnergySourceHelper::DoInstall (Ptr<Node> node) const
{
  NS_ASSERT (node != NULL);
  // check if energy source already exists
  Ptr<EnergySource> source = node->GetObject<EnergySource> ();
  if (source != NULL)
    {
      NS_FATAL_ERROR ("Energy source already installed!");
    }
  source = m_wimeshEnergySource.Create<EnergySource> ();
  NS_ASSERT (source != NULL);
  source->SetNode (node);
  return source;
}

} // namespace ns3




/* ------------------------------------------------------------ */





NS_LOG_COMPONENT_DEFINE ("WiMeshEnergySource");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (WiMeshEnergySource);

TypeId
WiMeshEnergySource::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::WiMeshEnergySource")
    .SetParent<EnergySource> ()
    .AddConstructor<WiMeshEnergySource> ()
    .AddAttribute ("WiMeshEnergySourceInitialEnergyJ",
                   "Initial energy stored in basic energy source.",
                   DoubleValue (10),  // in Joules
                   MakeDoubleAccessor (&WiMeshEnergySource::SetInitialEnergy,
                                       &WiMeshEnergySource::GetInitialEnergy),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("WiMeshEnergySupplyVoltageV",
                   "Initial supply voltage for basic energy source.",
                   DoubleValue (3.0), // in Volts
                   MakeDoubleAccessor (&WiMeshEnergySource::SetSupplyVoltage,
                                       &WiMeshEnergySource::GetSupplyVoltage),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("PeriodicEnergyUpdateInterval",
                   "Time between two consecutive periodic energy updates.",
                   TimeValue (Seconds (1.0)),
                   MakeTimeAccessor (&WiMeshEnergySource::SetEnergyUpdateInterval,
                                     &WiMeshEnergySource::GetEnergyUpdateInterval),
                   MakeTimeChecker ())
    .AddAttribute ("ActiveNodePowerConsumptionW",
                   "The power consumption (Watts) when the node is active.",
                   DoubleValue (2.29), // in Watts
                   MakeDoubleAccessor (&WiMeshEnergySource::SetActiveNodePowerConsumption,
                                       &WiMeshEnergySource::GetActiveNodePowerConsumption),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("SleepNodePowerConsumptionW",
                   "The power consumption (Watts) when the node is sleeping.",
                   DoubleValue (0.50), // in Watts
                   MakeDoubleAccessor (&WiMeshEnergySource::SetSleepNodePowerConsumption,
                                       &WiMeshEnergySource::GetSleepNodePowerConsumption),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("EnergyRechargeRateW",
                   "The amount of energy recharged per second.",
                   DoubleValue (3.00), // in Watts
                   MakeDoubleAccessor (&WiMeshEnergySource::SetEnergyRechargeRate,
                                       &WiMeshEnergySource::GetEnergyRechargeRate),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("BatteryCapacityJ",
                   "The maximum amount of energy that the battery can store.",
                   DoubleValue (5.00), // in Joules
                   MakeDoubleAccessor (&WiMeshEnergySource::SetMaximumEnergyJ,
                                       &WiMeshEnergySource::GetMaximumEnergyJ),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("BatteryThresholdToSuspend",
                   "Percentage of battery capacity under which the node goes into sleep mode.",
                   DoubleValue (0.05), // between 0 and 1
                   MakeDoubleAccessor (&WiMeshEnergySource::SetLowThreshold,
                                       &WiMeshEnergySource::GetLowThreshold),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("BatteryThresholdToResume",
                   "Percentage of battery capacity required for a sleeping node to resume.",
                   DoubleValue (0.15), // between 0 and 1
                   MakeDoubleAccessor (&WiMeshEnergySource::SetHighThreshold,
                                       &WiMeshEnergySource::GetHighThreshold),
                   MakeDoubleChecker<double> ())
    .AddTraceSource ("RemainingEnergy",
                     "Remaining energy at WiMeshEnergySource.",
                     MakeTraceSourceAccessor (&WiMeshEnergySource::m_remainingEnergyJ))
  ;
  return tid;
}

WiMeshEnergySource::WiMeshEnergySource ()
{
  NS_LOG_FUNCTION (this);
  m_lastUpdateTime = Seconds (0.0);
  m_sleep = false;
}

WiMeshEnergySource::~WiMeshEnergySource ()
{
  NS_LOG_FUNCTION (this);
}

void
WiMeshEnergySource::SetInitialEnergy (double initialEnergyJ)
{
  NS_LOG_FUNCTION (this << initialEnergyJ);
  NS_ASSERT (initialEnergyJ >= 0);
  m_initialEnergyJ = initialEnergyJ;
  m_remainingEnergyJ = m_initialEnergyJ;
}

void
WiMeshEnergySource::SetSupplyVoltage (double supplyVoltageV)
{
  NS_LOG_FUNCTION (this << supplyVoltageV);
  m_supplyVoltageV = supplyVoltageV;
}

void
WiMeshEnergySource::SetEnergyUpdateInterval (Time interval)
{
  NS_LOG_FUNCTION (this << interval);
  m_energyUpdateInterval = interval;
}

void
WiMeshEnergySource::SetActiveNodePowerConsumption (double activeNodePowerConsumption)
{
  NS_LOG_FUNCTION (this << activeNodePowerConsumption);
  m_activeNodePowerConsumption = activeNodePowerConsumption;
}

void
WiMeshEnergySource::SetSleepNodePowerConsumption (double sleepNodePowerConsumption)
{
  NS_LOG_FUNCTION (this << sleepNodePowerConsumption);
  m_sleepNodePowerConsumption = sleepNodePowerConsumption;
}

void
WiMeshEnergySource::SetEnergyRechargeRate (double energyRechargeRate)
{
  NS_LOG_FUNCTION (this << energyRechargeRate);
  m_energyRechargeRate = energyRechargeRate;
}

void
WiMeshEnergySource::SetMaximumEnergyJ (double maximumEnergyJ)
{
  NS_LOG_FUNCTION (this << maximumEnergyJ);
  m_maximumEnergyJ = maximumEnergyJ;
}

void
WiMeshEnergySource::SetLowThreshold (double lowThreshold)
{
  lowThreshold = std::min (lowThreshold, 1.0);
  lowThreshold = std::max (lowThreshold, 0.0);
  NS_LOG_FUNCTION (this << lowThreshold);
  m_lowThreshold = lowThreshold;
}

void
WiMeshEnergySource::SetHighThreshold (double highThreshold)
{
  highThreshold = std::min (highThreshold, 1.0);
  highThreshold = std::max (highThreshold, 0.0);
  NS_LOG_FUNCTION (this << highThreshold);
  m_highThreshold = highThreshold;
}


Time
WiMeshEnergySource::GetEnergyUpdateInterval (void) const
{
  NS_LOG_FUNCTION (this);
  return m_energyUpdateInterval;
}

double
WiMeshEnergySource::GetSupplyVoltage (void) const
{
  NS_LOG_FUNCTION (this);
  return m_supplyVoltageV;
}

double
WiMeshEnergySource::GetInitialEnergy (void) const
{
  NS_LOG_FUNCTION (this);
  return m_initialEnergyJ;
}

double
WiMeshEnergySource::GetRemainingEnergy (void)
{
  NS_LOG_FUNCTION (this);
  // update energy source to get the latest remaining energy.
  UpdateEnergySource ();
  return m_remainingEnergyJ;
}

double
WiMeshEnergySource::GetEnergyFraction (void)
{
  NS_LOG_FUNCTION (this);
  // update energy source to get the latest remaining energy.
  UpdateEnergySource ();
  return m_remainingEnergyJ / m_initialEnergyJ;
}

double
WiMeshEnergySource::GetActiveNodePowerConsumption (void) const
{
  NS_LOG_FUNCTION (this);
  return m_activeNodePowerConsumption;
}

double
WiMeshEnergySource::GetSleepNodePowerConsumption (void) const
{
  NS_LOG_FUNCTION (this);
  return m_sleepNodePowerConsumption;
}

double
WiMeshEnergySource::GetEnergyRechargeRate (void) const
{
  NS_LOG_FUNCTION (this);
  return m_energyRechargeRate;
}

double
WiMeshEnergySource::GetMaximumEnergyJ (void) const
{
  NS_LOG_FUNCTION (this);
  return m_maximumEnergyJ;
}

double
WiMeshEnergySource::GetLowThreshold (void) const
{
  NS_LOG_FUNCTION (this);
  return m_lowThreshold;
}

double
WiMeshEnergySource::GetHighThreshold (void) const
{
  NS_LOG_FUNCTION (this);
  return m_highThreshold;
}


void
WiMeshEnergySource::UpdateEnergySource (void)
{
  NS_LOG_FUNCTION (this);
  NS_LOG_DEBUG ("WiMeshEnergySource:Updating remaining energy.");

  // do not update if simulation has finished
  if (Simulator::IsFinished ())
  {
    return;
  }

  m_energyUpdateEvent.Cancel ();

  CalculateRemainingEnergy ();

  if (!m_sleep && m_remainingEnergyJ / m_maximumEnergyJ <= m_lowThreshold)
  {
    m_sleep = true;
    HandleEnergyDrainedEvent ();
  }
  
  if (m_sleep && m_remainingEnergyJ / m_maximumEnergyJ >= m_highThreshold)
  {
    m_sleep = false;
    // TODO HandleNodeResume ??
  }

  m_lastUpdateTime = Simulator::Now ();

  m_energyUpdateEvent = Simulator::Schedule (m_energyUpdateInterval,
                                             &WiMeshEnergySource::UpdateEnergySource,
                                             this);
}

/*
 * Private functions start here.
 */

void
WiMeshEnergySource::DoInitialize (void)
{
  NS_LOG_FUNCTION (this);
  UpdateEnergySource ();  // start periodic update
}

void
WiMeshEnergySource::DoDispose (void)
{
  NS_LOG_FUNCTION (this);
  BreakDeviceEnergyModelRefCycle ();  // break reference cycle
}

void
WiMeshEnergySource::HandleEnergyDrainedEvent (void)
{
  NS_LOG_FUNCTION (this);
  NS_LOG_DEBUG ("WiMeshEnergySource:Energy depleted!");
  NotifyEnergyDrained (); // notify DeviceEnergyModel objects
  if (m_remainingEnergyJ < 0)
    m_remainingEnergyJ = 0; // energy never goes below 0
}

void
WiMeshEnergySource::CalculateRemainingEnergy (void)
{
  NS_LOG_FUNCTION (this);
  double totalCurrentA = CalculateTotalCurrent ();
  Time duration = Simulator::Now () - m_lastUpdateTime;
  NS_ASSERT (duration.GetSeconds () >= 0);
  // energy = current * voltage * time
  double energyToDecreaseJ;

  if (m_sleep)
    energyToDecreaseJ = m_sleepNodePowerConsumption * duration.GetSeconds ();
  else
    energyToDecreaseJ = (totalCurrentA * m_supplyVoltageV + m_activeNodePowerConsumption) * duration.GetSeconds ();

  double energyToIncreaseJ = m_energyRechargeRate * duration.GetSeconds ();
  
  m_remainingEnergyJ = TracedValue<double> (std::min (m_remainingEnergyJ.Get() - energyToDecreaseJ + energyToIncreaseJ, m_maximumEnergyJ));
  NS_LOG_DEBUG ("WiMeshEnergySource:Remaining energy = " << m_remainingEnergyJ);
}

} // namespace ns3
