// -*- C++ -*-

/* myparser.cc
 *
 * Copyright (C) 2002 The libxml++ development team
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wimesh-ns3-parser.h"
#include "WMutils.h"
#include "WMtypes.h"
#include "rndvars.h"

#ifdef HAVE_MPLS
#include "wimesh-mpls-ns3.h"
#endif

#include "wimesh-rtproto-ns3.h"

#include <glibmm/convert.h> //For Glib::ConvertError
#include <iostream>
#include <cassert>

using namespace std;




/** WiMesh config parser **/

WiMeshNs3Parser::WiMeshNs3Parser ()
  : xmlpp::SaxParser(),
    ns3 (NULL)
{
}

WiMeshNs3Parser::~WiMeshNs3Parser()
{
}

void WiMeshNs3Parser::on_start_document()
{
  cout << "Reading WiMeshNs3 config..." << endl << endl;
}

void WiMeshNs3Parser::on_end_document()
{
  cout << "Finished reading WiMeshNs3 config" << endl << endl;
}

void WiMeshNs3Parser::on_start_element(const Glib::ustring& name,
                                   const AttributeList& attributes)
{
  static string split_ratios_file;

  if (name == "Routing")
  {
    if (ns3)
    {
      // a routing protocol has been already specified
      cerr << endl << "Ignoring duplicate Routing element" << endl << endl;
      return;
    }

    // "proto" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring Routing element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "proto")
    {
      cerr << endl << "Ignoring Routing element with first attribute other than proto" << endl << endl;
      return;
    }

    if (attributes.front().value == "dot11s")
      ns3 = new WiMeshRtProtoNs3 ("ns3::Dot11sStack", false);
#ifdef HAVE_LAYER2_5
    else if (attributes.front().value == "l2.5")
      ns3 = new WiMeshRtProtoNs3 ("ns3::LayerStack", false);
#endif
#ifdef HAVE_MPLS
    else if (attributes.front().value == "mpls")
      ns3 = new WiMeshMplsNs3 (false);
#endif
    else if (attributes.front().value == "flame")
      ns3 = new WiMeshRtProtoNs3 ("ns3::FlameStack", false);
    else
    {
      cerr << endl << "Ignoring Routing element with unknown or unavailable proto value: " << attributes.front().value << endl << endl;
      return;
    }

    // additional attributes
    double out_time = 0.0;

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      if (attributes.front().value == "mpls")
      {
	RP.read ("out_time", out_time);
	RP.read ("split_ratios_file", split_ratios_file);
      }

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Routing element" << endl << endl;
    }
#ifdef HAVE_MPLS
    if (attributes.front().value == "mpls")
    {
      WiMeshMplsNs3* mpls = dynamic_cast<WiMeshMplsNs3*>(ns3);

      assert ( mpls );

      // split ratios are set once the topology is known

      if (out_time > 0.0)
      {
	mpls->SetDisableNhlfe (true);
	mpls->SetOutTime (out_time);
      }
    }
#endif
    return;   // no further actions are required in case of a Routing element
  }  /* end Routing */

  // for sure, this is not a Routing element

  if (!ns3 && name != "WiMeshNs3Cfg")
  {
    cerr << endl << "A valid Routing element must be inserted before a " << name << " element" << endl << endl;
    return;
  }

  if (name == "Topology")
  {
    if (ns3->GetTopology())
    {
      // a topology file name has been already specified
      cerr << endl << "Ignoring duplicate Topology element" << endl << endl;
      return;
    }

    // "serialized_file" is expected as the first (and unique) attribute

    if (attributes.empty())
    {
      cerr << endl << "Ignoring Topology element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "serialized_file")
    {
      cerr << endl << "Ignoring Topology element with first attribute other than serialized_file" << endl << endl;
      return;
    }

    Topology* T = new Topology;

    T->DeserializeFrom (attributes.front().value.c_str());

    if (T->node.empty())
    {
      cerr << endl << "Ignoring Topology element (Error while reading the topology from " << attributes.front().value << ")" << endl << endl;
      delete T;
      return;
    }

    ns3->SetTopology (T);

    // (part of) the topology file name is used as the ns-3 simulation id

    string simulId;

    if (attributes.front().value.substr(0, string("topodata_").length()) == "topodata_")
      // simulId is what follows topodata_
      simulId = attributes.front().value.substr(string("topodata_").length());
    else
    {
      cout << endl << "Warning: the serialized file name does not start with topodata_" << endl << endl;
      simulId = attributes.front().value;
    }

    ns3->SetSimulationId(simulId);
#ifdef HAVE_MPLS
    // mpls requires to set the split ratios once the topology is known

    WiMeshMplsNs3* mpls = dynamic_cast<WiMeshMplsNs3*>(ns3);

    if (mpls)
    {
      if (!split_ratios_file.empty())
      {
	ifstream ifs (split_ratios_file.c_str());
	
	if (ifs)
	{
	  mpls->ReadSplitRatiosFromFile (ifs);
	  ifs.close();
	  return;
	}
	cerr << endl << "*** WARNING: Unable to read split ratios from file " << split_ratios_file << ". As a fallback, split ratios are calculated from the WiMesh topology ***" << endl << endl;
      }

      mpls->CalcSplitRatiosFromTopology();
    }
#endif
  }  /* end Topology */

  else if (name == "TrafficLoad")
  {
    // "type" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring TrafficLoad element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "type")
    {
      cerr << endl << "Ignoring TrafficLoad element with first attribute other than type" << endl << endl;
      return;
    }

    if (attributes.front().value == "onoff")
      ns3->SetTrafficType (string("OnOff"));
    else if (attributes.front().value == "trace_based")
      ns3->SetTrafficType (string("TraceBased"));
    else
    {
      cerr << endl << "Ignoring TrafficLoad element with unknown type value: " << attributes.front().value << endl << endl;
      return;
    }

    // additional attributes
    string proto, files;

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      if (attributes.front().value == "onoff")
      {
	RP.read ("t_on_min", ns3->t_on_min);
	RP.read ("t_on_max", ns3->t_on_max);
	RP.read ("t_off_min", ns3->t_off_min);
	RP.read ("t_off_max", ns3->t_off_max);
	RP.read ("pktsize", ns3->packetsize);
	RP.read ("proto", proto);
      }

      if (attributes.front().value == "trace_based")
	RP.read ("files", files);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a TrafficLoad element" << endl << endl;
    }

    if (attributes.front().value == "onoff")
    {
      if (proto == "udp")
	ns3->SetTransportProto (string("ns3::UdpSocketFactory"));
      else if (proto == "tcp")
	ns3->SetTransportProto (string("ns3::TcpSocketFactory"));
      else if (!proto.empty())
	cerr << endl << "Ignoring invalid value (" << proto << ") for the proto attribute of a TrafficLoad element" << endl << endl;
    }

    if (attributes.front().value == "trace_based")
    {
      stringstream ss (files);
      string s;

      while (ss >> s)
	ns3->AddTraceFile (s);
      
      ns3->SetTransportProto(string("ns3::UdpSocketFactory"));
    }
  }  /* end TrafficLoad */

  else if (name == "Options")
  {
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("channelsep", ns3->channelsep);
      RP.read ("duration", ns3->duration);
      RP.read ("enablePcap", ns3->pcapEnabled);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for an Options element" << endl << endl;
    }
  }  /* end Options */

  else if (name == "RemoteStaManager")
  {
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("name", ns3->RSMname);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a RemoteStaManager element" << endl << endl;
    }
  }  /* end RemoteStaManager */

  else if (name == "RSMoption")
  {
    WiMeshNs3::RemoteStaManagerOption option;
    enum { ns3_unspec, ns3_uint, ns3_double, ns3_bool, ns3_string, ns3_time } opt_type = ns3_unspec;
    bool hasValue = false;
    unsigned int uint_val;
    double double_val;
    bool bool_val;
    string string_val;

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      if (iter->name == "name")
      {
	option.m_name = iter->value;
      }
      else if (iter->name == "type")
      {
	// detect the parameter type
	if (iter->value == "uint")
	  opt_type = ns3_uint;
	else if (iter->value == "double")
	  opt_type = ns3_double;
	else if (iter->value == "bool")
	  opt_type = ns3_bool;
	else if (iter->value == "string")
	  opt_type = ns3_string;
	else if (iter->value == "time")
	  opt_type = ns3_time;
	else
	  opt_type = ns3_unspec;
      }
      else if (iter->name == "value")
      {
	switch (opt_type)
	{
	  case ns3_uint:
	    RP.read ("value", uint_val);
	    option.m_ptr_value = Create<UintegerValue> (uint_val);
	    break;
	  case ns3_double:
	    RP.read ("value", double_val);
	    option.m_ptr_value = Create<DoubleValue> (double_val);
	    break;
	  case ns3_bool:
	    RP.read ("value", bool_val);
	    option.m_ptr_value = Create<BooleanValue> (bool_val);
	    break;
	  case ns3_string:
	    RP.read ("value", string_val);
	    option.m_ptr_value = Create<StringValue> (string_val);
	    break;
	  case ns3_time:
	    RP.read ("value", double_val);
	    option.m_ptr_value = Create<TimeValue> (Seconds (double_val));
	    break;
	  case ns3_unspec:
	    cerr << endl << "Ignoring invalid RSMoption element (value declared before type)" << endl << endl;
	    return;
	}

	hasValue = true;
      }
      else
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for an RSMoption element" << endl << endl;
    }

    if (!option.m_name.empty() && hasValue)
      ns3->RSMoptions.push_back (option);
    else
      cerr << endl << "Ignoring invalid RSMoption element" << endl << endl;

  }  /* end RSMoption */

  else if (name == "Failure")
  {
    // "type" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring Failure element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "type")
    {
      cerr << endl << "Ignoring Failure element with first attribute other than type" << endl << endl;
      return;
    }

    int type = atoi(attributes.front().value.c_str());

    int node = -1, channel = -1, radio_id = -1;
    double time = -1.0, noise = 100.0;

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("time", time);
      RP.read ("noise", noise);
      RP.read ("node", node);
      RP.read ("channel", channel);
      RP.read ("radio_id", radio_id);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Failure element" << endl << endl;
    }

    switch (type)
    {
      case 1:
	if (time < 0. || node  < 0 || channel < 0)
	{
	  cerr << endl << "Ignoring Failure element with invalid or missing parameter values" << endl << endl;
	  return;
	}
	ns3->failures.push_back (WiMeshNs3::FailureEvent (time, noise, type, node, channel));
	break;

      case 2:
	if (time < 0. || radio_id  < 0)
	{
	  cerr << endl << "Ignoring Failure element with invalid or missing parameter values" << endl << endl;
	  return;
	}
	ns3->failures.push_back (WiMeshNs3::FailureEvent (time, noise, type, radio_id));
	break;

      case 3:
	if (time < 0. || node < 0)
	{
	  cerr << endl << "Ignoring Failure element with invalid or missing parameter values" << endl << endl;
	  return;
	}
	ns3->failures.push_back (WiMeshNs3::FailureEvent (time, noise, type, node));
	break;

      default:
	cerr << endl << "Ignoring Failure element with invalid type" << endl << endl;
	return;
    }
  }  /* end Failure */
  
  else if (name == "FlowMonitor")
  {
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("interval", ns3->measure_interval);
      RP.read ("perflow", ns3->measure_perflow);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a FlowMonitor element" << endl << endl;
    }

    ns3->enableFlowMonitor = true;
  }  /* end FlowMonitor */

  else if (name == "EnergyMeter")
  {
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      RP.read ("source", ns3->initEnergy);
      RP.read ("seed", ns3->initEnergySeed);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a EnergyMeter element" << endl << endl;
    }

    ns3->enableEnergyMeter = true;
  }  /* end MeasureEnergy */

  else if (name != "WiMeshNs3Cfg")
  {
    cerr << endl << "Ignoring invalid element (" << name << ")" << endl << endl;
  }
}

void WiMeshNs3Parser::on_end_element(const Glib::ustring& name)
{
//   if (name == "WiMeshCfg")
//     // stop the parser
//     throw xmlpp::exception("Done");
}

void WiMeshNs3Parser::on_characters(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_characters(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "WiMeshNs3Parser::on_characters(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void WiMeshNs3Parser::on_comment(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_comment(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "WiMeshNs3Parser::on_comment(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void WiMeshNs3Parser::on_warning(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_warning(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshNs3Parser::on_warning(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void WiMeshNs3Parser::on_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshNs3Parser::on_error(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void WiMeshNs3Parser::on_fatal_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_fatal_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshNs3Parser::on_characters(): Exception caught while converting value for std::cout: " << ex.what() << std::endl;
  }
}

