#include "wimesh-ns3.h"
#include "WMutils.h"
#include "udp-trace-client.h"
#include "wimesh-radio-energy-model.h"
#include "ns3/point-to-point-module.h"
#include <sstream>
#include <cassert>
#include <cstdarg>

WiMeshNs3::WiMeshNs3 (bool _useWiMeshModels)
  : topology (0),
    useWiMeshModels (_useWiMeshModels),
    wifiPhy (YansWifiPhyHelper::Default ()),
    wifiChannel (YansWifiChannelHelper::Default ()),
    simul_id (string("xxx")),
    channelsep (4),
    trafficType (string("OnOff")),
    transportProto (string("ns3::UdpSocketFactory")),
    t_on_min (0.5),
    t_on_max (1.5),
    t_off_min (0.25),
    t_off_max (0.75),
    traceFiles (),
    packetsize (1400),
    duration (60.0),
    pcapEnabled (false),
    failures (),
    ch_switches(),
    enableFlowMonitor (false),
    measure_interval (1.0),
    measure_perflow (false),
    enableEnergyMeter (false),
    initEnergySeed (getSeed()),
    RSMname (string("ns3::WiMeshWifiManager"))
{
  Config::SetDefault ("ns3::TcpSocket::SegmentSize", UintegerValue (packetsize));
  DoInitChannel ();

  stringstream ss(string("C 100000.0"));
  initEnergy = new RndVar (ss);
}

WiMeshNs3::WiMeshNs3 (const WiMeshNs3& w)
{
  *this = w;
  DoInitChannel ();
}

void WiMeshNs3::DoInitChannel()
{
  Config::SetDefault ("ns3::YansWifiPhy::TxGain", DoubleValue (default_db_gain));
  Config::SetDefault ("ns3::YansWifiPhy::RxGain", DoubleValue (default_db_gain));

  //Configure the WiMesh available power levels
  Config::SetDefault ("ns3::YansWifiPhy::TxPowerStart", DoubleValue (minTxPowerLevel));
  Config::SetDefault ("ns3::YansWifiPhy::TxPowerEnd", DoubleValue (maxTxPowerLevel));
  Config::SetDefault ("ns3::YansWifiPhy::TxPowerLevels", UintegerValue(numTxPowerLevels));

  if (useWiMeshModels)
    Config::SetDefault ("ns3::YansWifiPhy::RxNoiseFigure", DoubleValue(0.));

  //the default propagation loss model for YansWifi is LogDistance: set the parameter so that the loss it gives is zero
  Config::SetDefault("ns3::LogDistancePropagationLossModel::Exponent", DoubleValue(0.));
  Config::SetDefault("ns3::LogDistancePropagationLossModel::ReferenceLoss", DoubleValue(0.));
  
  // increase the maximum number of peers in dot11sStack
  Config::SetDefault("ns3::dot11s::PeerManagementProtocol::MaxNumberOfPeerLinks", UintegerValue (70));
  
  // Initializes Channel and Phy
  if (useWiMeshModels)
    wifiChannel.AddPropagationLoss("WiMeshPropagationLossModel");
  
  wifiPhy.SetChannel (wifiChannel.Create ());
  if (useWiMeshModels)
    wifiPhy.SetErrorRateModel("WiMeshErrorRateModel");
  wifiPhy.SetPcapDataLinkType (YansWifiPhyHelper::DLT_IEEE802_11_RADIO); 
}


void WiMeshNs3::SetTopology(Topology* _topology)
{
  topology = _topology;
}

Topology* WiMeshNs3::GetTopology()
{
  return topology;
}

void WiMeshNs3::SetSimulationId (string sid)
{
  simul_id = sid;
}

string WiMeshNs3::GetSimulationId()
{
  return simul_id;
}

void WiMeshNs3::SetChannelSep (unsigned int cs)
{
  channelsep = cs;
}

unsigned int WiMeshNs3::GetChannelSep()
{
  return channelsep;
}

void WiMeshNs3::SetTrafficType (std::string tt)
{
  trafficType = tt;
}

void WiMeshNs3::AddTraceFile (std::string f)
{
  traceFiles.push_back (f);
}

void WiMeshNs3::SetTransportProto (std::string tp)
{
  transportProto = tp;
}


void WiMeshNs3::SetTOn(double min, double max)
{
  t_on_min = min;
  t_on_max = max;
}


void WiMeshNs3::SetTOff(double min, double max)
{
  t_off_min = min;
  t_off_max = max;
}

void WiMeshNs3::GetTOn(double& min, double& max)
{
  min = t_on_min;
  max = t_on_max;
}

void WiMeshNs3::GetTOff(double& min, double& max)
{
  min = t_off_min;
  max = t_off_max;
}

void WiMeshNs3::SetPacketSize(unsigned int ps)
{
  packetsize = ps;
}

unsigned int WiMeshNs3::GetPacketSize()
{
  return packetsize;
}

void WiMeshNs3::SetDuration(double d)
{
  duration = d;
}

double WiMeshNs3::GetDuration()
{
  return duration;
}

void WiMeshNs3::SetPcapEnabled(bool enable)
{
  pcapEnabled = enable;
}

void WiMeshNs3::SetFlowMonitorEnabled(bool flowmon)
{
  enableFlowMonitor = flowmon;
}

void WiMeshNs3::SetEnergyMeterEnabled(bool emeter)
{
  enableEnergyMeter = emeter;
}

void WiMeshNs3::SetInitialEnergy(RndVar* source)
{
  initEnergy = source;
}

void WiMeshNs3::SetInitialEnergySeed(double seed)
{
  initEnergySeed = seed;
}

WiMeshNs3::FailureEvent::FailureEvent (double _time, double _noise, int _type, ...)
  : time (_time),
    noise (_noise),
    type (_type)
{
  va_list params;
  va_start (params, _type);
  
  switch (type)
  {
    case 1:
      node = va_arg (params, int);
      channel = va_arg (params, int);
      break;
      
    case 2:
      radio = va_arg (params, int);
      break;
      
    case 3:
      node = va_arg (params, int);
      break;
      
    default:
      throw;
  }
}

int WiMeshNs3::GetWiMeshRadioTXpower (unsigned int node, unsigned int channel)
{
  list<TGraphNeighbor>::iterator v = topology->node[node].neighbor.begin();
  for ( ; v != topology->node[node].neighbor.end() && v->radio->channel != channel; v++);

  if (v != topology->node[node].neighbor.end())
    return v->txpower;

  cerr << endl << "** WARNING: No link outgoing node " << node << "  using channel " << channel 
       << " has been found in the WiMesh topology **" << endl << endl;

  return topology->maxP;
}

WiMeshNs3::RemoteStaManagerOption::RemoteStaManagerOption (string n, Ptr<AttributeValue> p)
  : m_name (n),
    m_ptr_value (p)
{
}

void WiMeshNs3::SetRSMoptions ()
{
  Ptr<WifiNetDevice> wnd;
  Ptr<WifiRemoteStationManager> rsm;

  for (NodeContainer::Iterator n = meshnodes.Begin(); n != meshnodes.End(); n++)
    for (uint32_t i = 0; i < (*n)->GetNDevices (); ++i)
    {
      wnd = (*n)->GetDevice (i)->GetObject<WifiNetDevice> ();
      // if it is a WifiNetDevice
      if (wnd != 0)
      {
	unsigned int channel = wnd->GetPhy ()->GetChannelNumber () / channelsep;

	rsm = wnd->GetRemoteStationManager ();

	if (!rsm)         // no wifi remote station manager ??
	{
	  cerr << endl << "** WARNING: No Wifi remote station manager found on the radio using channel " << channel
	       << " on node " << (*n)->GetId() << " **" << endl << endl;
	  continue;
	}

	// set the options specified in the XML config file
	for (vector<RemoteStaManagerOption>::iterator i = RSMoptions.begin(); i != RSMoptions.end(); i++)
	  rsm->SetAttribute (i->m_name, *PeekPointer(i->m_ptr_value));

	// set the TX power level for this WifiNetDevice
	rsm->SetDefaultTxPowerLevel (GetWiMeshRadioTXpower((*n)->GetId(), channel));
      }
    }
}


void WiMeshNs3::DoAttachEnergySources ()
{
  BasicEnergySourceHelper basicSourceHelper;
  WiMeshRadioEnergyModelHelper radioEnergyHelper;
  Ptr<WifiNetDevice> wnd;
  MotherOfAll urng(initEnergySeed);    // declare uniform random number generator
  Random::Set(urng);                   // set urng as generator to be used

  // install an energy source on each node
  for (NodeContainer::Iterator n = meshnodes.Begin(); n != meshnodes.End(); n++)
  {
    basicSourceHelper.Set ("BasicEnergySourceInitialEnergyJ", DoubleValue (initEnergy->rv->Next()));

    ESnodes.Add (basicSourceHelper.Install (*n));

    for (uint32_t i = 0; i < (*n)->GetNDevices (); ++i)
    {
      wnd = (*n)->GetDevice (i)->GetObject<WifiNetDevice> ();
      // if it is a WifiNetDevice
      if (wnd != 0)
      {
	int channel = wnd->GetPhy ()->GetChannelNumber () / channelsep;

	// find the radio with such a channel
	RadioIfaceIterator ri = topology->node[(*n)->GetId()].info->findRadioWithChannel(channel);

	assert ( ri != topology->node[(*n)->GetId()].info->radio.end() );

	double voltage = ESnodes.Get(ESnodes.GetN()-1)->GetSupplyVoltage();

	radioEnergyHelper.Set("IdleCurrentA", DoubleValue (ri->rho_idle / voltage) );
	radioEnergyHelper.Set("CcaBusyCurrentA", DoubleValue (ri->rho_idle / voltage) );
	radioEnergyHelper.Set("SwitchingCurrentA", DoubleValue (ri->rho_idle / voltage) );
	radioEnergyHelper.Set("RxCurrentA", DoubleValue (ri->rho_rx / voltage) );
	radioEnergyHelper.Set("TxCurrentA", DoubleValue ((*ri->rho_tx)(DbmToW(TxPowerLevel(GetWiMeshRadioTXpower((*n)->GetId(), channel)))) / voltage) );
	radioEnergyHelper.Set("WiMeshNs3Ptr", WiMeshNs3PtrValue (WiMeshNs3Ptr (this)));

	// this device draws power from the last created energy source
	radioEnergyHelper.Install (wnd, ESnodes.Get (ESnodes.GetN()-1));
      }
    }
  }
}


void WiMeshNs3::GetNodesInitialEnergy (vector<double>* energy)
{
  Simulator::Schedule (Seconds (0.), &WiMeshNs3::DoGetNodesInitialEnergy, this, energy);
}


void WiMeshNs3::DoGetNodesInitialEnergy (vector<double>* energy)
{
  energy->resize(ESnodes.GetN());

  for (uint32_t i = 0; i < ESnodes.GetN(); i++)
    energy->at(i) = ESnodes.Get(i)->GetInitialEnergy();
}


void WiMeshNs3::GetNodesRemainingEnergyAtTime (double time, vector<double>* energy)
{
  Simulator::Schedule (Seconds (time), &WiMeshNs3::DoGetNodesRemainingEnergy, this, energy);
}


void WiMeshNs3::DoGetNodesRemainingEnergy (vector<double>* energy)
{
  energy->resize(ESnodes.GetN());

  for (uint32_t i = 0; i < ESnodes.GetN(); i++)
    energy->at(i) = ESnodes.Get(i)->GetRemainingEnergy();
}


void WiMeshNs3::ScheduleFailureEvent (FailureEvent& f, vector<int> &cumRadios, stringstream &fname)
{
  int _node;
  unsigned int _channel;
  vector<int>::iterator res;
  
  switch (f.type)
  {
    case 1:
      Simulator::Schedule (Seconds (f.time), &WiMeshNs3::DoGenerateNoise, this, f.node, f.channel, f.noise);
      fname << "_" << f.node << "_" << f.channel;
      break;
      
    case 2:
      res = find_if(cumRadios.begin(), cumRadios.end(), bind2nd(greater<int>(), f.radio));
      if (res == cumRadios.end())
      {
	cerr << "No radio found with the given index (" << f.radio << "). Noise is not added" << endl << endl;
	return;
      }
      
      _node = res - cumRadios.begin();
      _channel = topology->node[_node].info->radio[f.radio-(*res-topology->node[_node].info->numAvailRadios())].channel;
      
      if (!_channel)
      {
	cout << "Warning: the radio is already turned off. Noise is not added" << endl << endl;
	return;
      }

      Simulator::Schedule (Seconds (f.time), &WiMeshNs3::DoGenerateNoise, this, _node, _channel, f.noise);
      fname << "_" << _node << "_" << _channel;
      break;
      
    case 3:
      for (unsigned int i = 0; i < topology->node[f.node].info->numAvailRadios(); i++)
	if ((_channel = topology->node[f.node].info->radio[i].channel))
	{
	  Simulator::Schedule (Seconds (f.time), &WiMeshNs3::DoGenerateNoise, this, f.node, _channel, f.noise);
	}
      fname << "_n" << f.node;
      break;
  }
}

WiMeshNs3::ChannelSwitch::ChannelSwitch (double time, int node, unsigned int fromCh, unsigned int toCh)
  : m_time (time),
    m_node (node),
    m_fromCh (fromCh),
    m_toCh (toCh)
{
}

void WiMeshNs3::ScheduleChannelSwitch (double time, int node, unsigned int fromCh, unsigned int toCh)
{
  ch_switches.push_back (ChannelSwitch (time, node, fromCh, toCh));
}


RadioIface*
WiMeshNs3::GetRadioPtr (Ptr<WifiNetDevice> wnd)
{
  int n = wnd->GetNode()->GetId();
  unsigned int channel = wnd->GetPhy()->GetChannelNumber() / channelsep;

  RadioIfaceIterator ri = topology->node[n].info->findRadioWithChannel(channel);

  if (ri == topology->node[n].info->radio.end())
    return NULL;

  return &(*ri);
}


void WiMeshNs3::DoCreateNodes()
{
  // create the nodes
  meshnodes.Create (topology->node.size());
}


void WiMeshNs3::CreateNs3MeshNodes ()
{  
  DoCreateNodes ();

  // set the position of nodes
  MobilityHelper mobility;
  Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
  
  for (vector<TGraphNode>::iterator n=topology->node.begin(); n!=topology->node.end(); n++)
    positionAlloc->Add (Vector (n->info->xpos, n->info->ypos, 0.0));
  
  mobility.SetPositionAllocator (positionAlloc);
  mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
  mobility.Install (meshnodes);
}

void WiMeshNs3::CreateNs3Hosts ()
{
  PointToPointHelper pointToPoint;
  InternetStackHelper internet;
  NetDeviceContainer hostdevices;
  Ipv4AddressHelper address;
  Ipv4StaticRoutingHelper ipv4RoutingHelper;
  ApplicationContainer apps;
  Ptr<Ipv4StaticRouting> staticRouting;
  Ptr<Node> nodep;
  uint16_t port = 9000;
  string prefix("192.168.");
  
  pointToPoint.SetDeviceAttribute ("DataRate", StringValue ("100Mbps"));
  pointToPoint.SetChannelAttribute ("Delay", StringValue ("1ms"));  
  
  // check that e2e flows are consistent and obtain the set of demands
  topology->checkE2Eflows(&demands);

  set<int> edges;
  
  // find all the edge nodes
  for (TDemands::iterator d = demands.begin (); d != demands.end (); d++)
  {
    edges.insert (d->first.first);
    edges.insert (d->first.second);
  }
  
  // associates each WiMesh node id with the index in the node container
  vector<uint32_t> index (topology->node.size ());

  // create a host for each edge node and a point-to-point channel between them
  for (set<int>::iterator i = edges.begin (); i != edges.end (); i++)
  {
    index[*i] = hostnodes.GetN ();
    nodep = CreateObject<Node> ();
    hostnodes.Add (nodep);
    internet.Install (nodep);
    
    hostdevices = pointToPoint.Install (nodep, meshnodes.Get (*i));

    // assign IP addresses
    stringstream netaddr;
    netaddr << prefix << *i;
    address.SetBase ((netaddr.str() + ".0").c_str(), "255.255.255.0");
    address.Assign (hostdevices);
    
    // set default route on the host node
    staticRouting = ipv4RoutingHelper.GetStaticRouting (nodep->GetObject<Ipv4> ());
    staticRouting->SetDefaultRoute (Ipv4Address ((netaddr.str() + ".2").c_str ()), 1);
  }
    
  double mean_t_on = (t_on_min + t_on_max) / 2,
         mean_t_off = (t_off_min + t_off_max) / 2;
  double coeff = (mean_t_on + mean_t_off) / mean_t_on;
  int dem_id = 0;
    
  // install an OnOff application on the hosts
  for (TDemands::iterator d = demands.begin (); d != demands.end (); d++, dem_id++)
  {
    stringstream remote;
    remote << prefix << d->first.second << ".1";
    
    if (trafficType == "TraceBased")
    {
      UdpTraceBasedClientHelper udpTr (Ipv4Address (remote.str().c_str()), port, traceFiles[dem_id % traceFiles.size()]);
      
      apps = udpTr.Install (hostnodes.Get (index[d->first.first]));
    }
    else if (d->second) // the default is OnOff
    {
      OnOffHelper onOff(transportProto, InetSocketAddress (Ipv4Address (remote.str().c_str()), port));

      onOff.SetAttribute ("DataRate", DataRateValue (DataRate ((uint64_t)(d->second * 1000000 * coeff))));
      onOff.SetAttribute ("PacketSize", UintegerValue (packetsize));
//    For ns < 3.18 
//    onOff.SetAttribute ("OnTime", RandomVariableValue (UniformVariable (t_on_min, t_on_max)));
//    onOff.SetAttribute ("OffTime", RandomVariableValue (UniformVariable (t_off_min, t_off_max)));
     
      stringstream ontime, offtime;
      
      ontime << "ns3::UniformRandomVariable[Min=" << t_on_min << "|Max=" << t_on_max << "]";
      offtime << "ns3::UniformRandomVariable[Min=" << t_off_min << "|Max=" << t_off_max << "]";
      
      onOff.SetAttribute ("OnTime", StringValue (ontime.str()));
      onOff.SetAttribute ("OffTime", StringValue (offtime.str()));
      
      apps = onOff.Install (hostnodes.Get (index[d->first.first]));
    }
    
    apps.Start (Seconds (0.01));
    apps.Stop (Seconds (duration));
    
    // Create a packet sink to receive these packets
    PacketSinkHelper sink (transportProto, 
			  InetSocketAddress (Ipv4Address::GetAny (), port++));
    apps = sink.Install (hostnodes.Get (index[d->first.second]));
    apps.Start (Seconds (0.01));
  }
  
  if (pcapEnabled)
    pointToPoint.EnablePcapAll("host");
}

int WiMeshNs3::HostIPaddressToNodeId (const Ipv4Address& addr)
{
  uint8_t buf[4];

  addr.Serialize (buf);

  // the node id is stored in the 3rd byte

  return (int)(buf[2]);
}

bool WiMeshNs3::CheckDemand (int src, int dst, double f, bool strict)
{
  if (demands.find (pair<int,int> (src,dst)) == demands.end ())
    return false;
  
  if (strict && trunc(demands[pair<int,int> (src,dst)]) != trunc(f))
    return false;
  
  return true;
}

void WiMeshNs3::RunSimulation ()
{
  CreateNs3MeshNodes ();
  CreateNs3MeshDevices ();
  SetRSMoptions ();
  CreateNs3Hosts ();
  ConfigureForwarding ();

  if (pcapEnabled)
    wifiPhy.EnablePcap ("mesh", meshdevices);

  Mac48Address ma("00:00:00:00:00:01");
  
  stringstream fname;
  
  fname << (transportProto == "ns3::UdpSocketFactory" ? "udp_" : "tcp_") << simul_id << "_";
  if (stack == "MplsSplitting")
    fname << "mpls";
  else if (stack == "ns3::LayerStack")
    fname << "layer";
  else if (stack ==  "ns3::FlameStack")
    fname << "flame";
  else if (stack == "ns3::Dot11sStack")
    fname << "dot11s";
  
  // find the vector of cumulative number of radios
  vector<int> cumRadios;
  int cum = 0;
  for (vector<TGraphNode>::iterator u = topology->node.begin(); u != topology->node.end(); u++)
    cumRadios.push_back (cum += u->info->numAvailRadios());
  
  
  // schedules failures, if any
  for (vector<FailureEvent>::iterator e = failures.begin (); e != failures.end (); ++e)
  {
    ScheduleFailureEvent (*e, cumRadios, fname);
  }

  // schedules channel switches, if any
  for (vector<ChannelSwitch>::iterator cs = ch_switches.begin(); cs != ch_switches.end(); cs++)
    Simulator::Schedule (Seconds (cs->m_time), &WiMeshNs3::DoChannelSwitch, this, cs->m_node, cs->m_fromCh, cs->m_toCh);

  if (enableFlowMonitor)
    WMflowmon.Enable (measure_interval, duration, measure_perflow, string("flow_") + fname.str ());

  vector<double> energy_i, energy_f;

  if (enableEnergyMeter)
  {
    DoAttachEnergySources();
    GetNodesInitialEnergy (&energy_i);
    GetNodesRemainingEnergyAtTime (duration, &energy_f);
  }

  // force the simulator to terminate
  Simulator::Stop (Seconds(duration));
  Simulator::Run ();
  
  PrintConfiguration ();
  
  if (enableEnergyMeter)
  {
    // compute the energy consumed by each node
    transform(energy_i.begin(), energy_i.end(), energy_f.begin(), energy_f.begin(), minus<double>());

    // add the energy consumed by sleeping radios or nodes
    for (vector<TGraphNode>::iterator u = topology->node.begin(); u != topology->node.end(); u++)
    {
      unsigned int n = u->info->countRadiosWithChannel(0);
      if (n == u->info->numAvailRadios())
	energy_f.at(u->info->id) += u->info->StandbyPowerConsumption * duration;
      else
      {
	energy_f.at(u->info->id) += u->info->BasePowerConsumption * duration;
	for (vector<RadioIface>::iterator ri = u->info->radio.begin(); ri != u->info->radio.end(); ri++)
	  if (!ri->channel)
	    energy_f.at(u->info->id) += ri->rho_sleep * duration;
      }
    }

    // write results to file
    ofstream ofs ((string("energy_")+fname.str()).c_str());
    copy (energy_f.begin(), energy_f.end(), ostream_iterator<double>(ofs, "\n"));
    ofs.close();
  }

  Simulator::Destroy ();
}

void
WiMeshNs3::PrintObjConf (ostream& os)
{
  if (topology)
    os << "Topology name: " << topology->toponame << endl;
  os << "Channel separation: " << channelsep << endl << endl;

  if (trafficType == "TraceBased")
  {
    os << "Traffic type: " << trafficType << endl;
    for (vector<string>::iterator f = traceFiles.begin(); f != traceFiles.end(); f++)
      os << *f << endl;
  }
  else
  {
    os << "Traffic type: OnOff" << endl
       << "Ton ~ U(" << t_on_min << "," << t_on_max << ")     Toff ~ U(" << t_off_min << "," << t_off_max << ")" 
       << endl << endl << "Packet size: " << packetsize;
  }
  
  os << endl << "Transport protocol: " << transportProto << endl
     << "Duration: " << duration << endl << endl;
     
  for (vector<ChannelSwitch>::iterator cs = ch_switches.begin(); cs != ch_switches.end(); cs++)
    os << "Channel switch at time " << cs->m_time << " on node " << cs->m_node << " from channel " << cs->m_fromCh 
       << " to channel " << cs->m_toCh << endl;
}


void
WiMeshNs3::GetTxRatePower (int myId, Mac48Address neighbor, uint32_t &rate, uint32_t &power)
{
  int channel;
  int neighbId = DoFindNeighborByMacAddress (myId, neighbor, channel);
  
  // if the link is not present in topology, return the basic rate
  if (myId < 0 || myId >= (int)topology->node.size () || neighbId < 0 || neighbId >= (int)topology->node.size () )
  {
    rate = 0;
    power = topology->maxP;
    return;
  }
  
  list<TGraphNeighbor>::iterator n = topology->find_neighbor (myId, neighbId, channel);
  
  if (n == topology->node[myId].neighbor.end ()) 
  {
    rate = 0;
    power = topology->maxP;
    return;
  }

  rate = n->rate;
  power = n->txpower;
}

void
WiMeshNs3::ReceiveFinalDataFailed (const WifiRemoteStationManager* me, Mac48Address neighbor)
{
  int myId = DoFindNodeByStationManager (me);
  DoReceiveFinalDataFailed (myId, neighbor);
}


int
WiMeshNs3::DoFindNodeByStationManager (const WifiRemoteStationManager* me)
{
  Ptr<Node> n;
  bool found = false;
  
  for (NetDeviceContainer::Iterator i = meshdevices.Begin (); i != meshdevices.End () && !found; ++i)
    if (GetPointer<WifiRemoteStationManager>((*i)->GetObject<WifiNetDevice> ()->GetRemoteStationManager ()) == me)
    {
      found = true;
      n = (*i)->GetNode ();
    }
    
  assert (found);
  
  for (uint32_t i = 0; i < meshnodes.GetN (); ++i)
    if (meshnodes.Get(i)->GetId () == n->GetId ())
      return i;
  
  // should never get here
  return topology->node.size ();
}

int
WiMeshNs3::DoFindNeighborByMacAddress (int me, Mac48Address neighbor, int& channel, Ptr<WifiNetDevice>* device)
{
  Ptr<WifiNetDevice> wnd;

  // for each neighbor of mine
  for (list<TGraphNeighbor>::iterator n = topology->node[me].neighbor.begin (); n != topology->node[me].neighbor.end (); n++)
  {
    // check the address of all of its Wifi interfaces
    for (uint32_t i = 0; i < meshnodes.Get (n->info->id)->GetNDevices (); ++i)
    {
      wnd = meshnodes.Get (n->info->id)->GetDevice (i)->GetObject<WifiNetDevice> ();
      // if it is a WifiNetDevice and it has neighbor as address
      if (wnd != 0 && Mac48Address::ConvertFrom (wnd->GetAddress ()) == neighbor)
      {
	channel = wnd->GetPhy ()->GetChannelNumber () / channelsep;
	if (device)
	  *device = wnd;
	return n->info->id;
      }
    }
  }
  // we might get here when not using Mpls or Layer
  return topology->node.size ();
}

void
WiMeshNs3::DoGenerateNoise (int node, unsigned int channel, double noise)
{
  Ptr<WifiNetDevice> wnd;
  
  for (uint32_t i = 0; i < meshnodes.Get (node)->GetNDevices (); ++i)
  {
    wnd = meshnodes.Get (node)->GetDevice (i)->GetObject<WifiNetDevice> ();
    // if it is a WifiNetDevice and it is set to the given channel
    if (wnd != 0 && wnd->GetPhy ()->GetChannelNumber () / channelsep == channel)
    {
      cerr << "Adding a " << noise << "db noise around the radio on node " << node << " using channel " << channel << endl;
      wnd->GetPhy ()->GetObject<YansWifiPhy> ()->SetRxNoiseFigure (1e10);
    }
  }
}

void
WiMeshNs3::DoChannelSwitch (int node, unsigned int fromCh, unsigned int toCh)
{
  Ptr<WifiNetDevice> wnd;
  
  for (uint32_t i = 0; i < meshnodes.Get (node)->GetNDevices (); ++i)
  {
    wnd = meshnodes.Get (node)->GetDevice (i)->GetObject<WifiNetDevice> ();
    // if it is a WifiNetDevice and it is set to the given channel
    if (wnd != 0 && wnd->GetPhy ()->GetChannelNumber () / channelsep == fromCh)
    {
      cerr << "Switching a radio on node " << node << " from channel " << fromCh << " to channel ";
      wnd->GetPhy ()->SetChannelNumber (toCh * channelsep);
      cerr << wnd->GetPhy ()->GetChannelNumber () / channelsep  << endl;
      break;
    }
  }
}


WiMeshNs3Ptr::WiMeshNs3Ptr ()
  : m_ptr (0)
{
}

WiMeshNs3Ptr::WiMeshNs3Ptr (WiMeshNs3* ptr)
  : m_ptr (ptr)
{
}

WiMeshNs3*
WiMeshNs3Ptr::GetPtr ()
{
  return m_ptr;
}


ostream&
operator<< (ostream& os, const WiMeshNs3Ptr& p)
{
  os << "this does not make sense";
  return os;
}

istream&
operator>>(istream& is, WiMeshNs3Ptr& p)
{
  return is;
}


ATTRIBUTE_HELPER_CPP (WiMeshNs3Ptr);

