#include "ns3/log.h"
#include "ns3/double.h"
#include "ns3/simulator.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/wifi-radio-energy-model-helper.h"
#include "ns3/basic-energy-source-helper.h"
#include "ns3/wifi-net-device.h"
// #include "energy-source.h"
#include "wimesh-radio-energy-model.h"
#include "wimesh-wifi-manager.h"

NS_LOG_COMPONENT_DEFINE ("WiMeshRadioEnergyModel");

namespace ns3 {

WiMeshRadioEnergyModelHelper::WiMeshRadioEnergyModelHelper ()
{
  m_radioEnergy.SetTypeId ("ns3::WiMeshRadioEnergyModel");
  m_depletionCallback.Nullify ();
}

WiMeshRadioEnergyModelHelper::~WiMeshRadioEnergyModelHelper ()
{
}

void
WiMeshRadioEnergyModelHelper::Set (std::string name, const AttributeValue &v)
{
  m_radioEnergy.Set (name, v);
}

void
WiMeshRadioEnergyModelHelper::SetDepletionCallback (
  WifiRadioEnergyModel::WifiRadioEnergyDepletionCallback callback)
{
  m_depletionCallback = callback;
}

/*
 * Private function starts here.
 */

Ptr<DeviceEnergyModel>
WiMeshRadioEnergyModelHelper::DoInstall (Ptr<NetDevice> device,
                                       Ptr<EnergySource> source) const
{
  NS_ASSERT (device != NULL);
  NS_ASSERT (source != NULL);
  // check if device is WifiNetDevice
  std::string deviceName = device->GetInstanceTypeId ().GetName ();
  if (deviceName.compare ("ns3::WifiNetDevice") != 0)
    {
      NS_FATAL_ERROR ("NetDevice type is not WifiNetDevice!");
    }
  Ptr<Node> node = device->GetNode ();
  Ptr<WiMeshRadioEnergyModel> model = m_radioEnergy.Create ()->GetObject<WiMeshRadioEnergyModel> ();
  NS_ASSERT (model != NULL);
  // set energy source pointer
  model->SetEnergySource (source);
  // set energy depletion callback
  model->SetEnergyDepletionCallback (m_depletionCallback);
  // add model to device model list in energy source
  source->AppendDeviceEnergyModel (model);
  // create and register energy model phy listener
  Ptr<WifiNetDevice> wifiDevice = DynamicCast<WifiNetDevice> (device);
  Ptr<WifiPhy> wifiPhy = wifiDevice->GetPhy ();
  wifiPhy->RegisterListener (model->GetPhyListener ());

  WiMeshNs3Ptr wmn = model->GetWiMeshNs3Ptr();
  model->SetWifiNetDevice (wifiDevice);
  model->SetWiMeshRadioPtr (wmn.GetPtr()->GetRadioPtr(wifiDevice));
  model->SetVoltage (source->GetSupplyVoltage());
  return model;
}



//---------------------------------------------------------------------------//


  
NS_OBJECT_ENSURE_REGISTERED (WiMeshRadioEnergyModel);

TypeId
WiMeshRadioEnergyModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::WiMeshRadioEnergyModel")
    .SetParent<DeviceEnergyModel> ()
    .AddConstructor<WiMeshRadioEnergyModel> ()
    .AddAttribute ("IdleCurrentA",
                   "The default radio Idle current in Ampere.",
                   DoubleValue (0.000426),  // idle mode = 426uA
                   MakeDoubleAccessor (&WifiRadioEnergyModel::SetIdleCurrentA,
                                       &WifiRadioEnergyModel::GetIdleCurrentA),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("CcaBusyCurrentA",
                   "The default radio CCA Busy State current in Ampere.",
                   DoubleValue (0.000426),  // default to be the same as idle mode
                   MakeDoubleAccessor (&WifiRadioEnergyModel::SetCcaBusyCurrentA,
                                       &WifiRadioEnergyModel::GetCcaBusyCurrentA),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("TxCurrentA",
                   "The radio Tx current in Ampere.",
                   DoubleValue (0.0174),    // transmit at 0dBm = 17.4mA
                   MakeDoubleAccessor (&WifiRadioEnergyModel::SetTxCurrentA,
                                       &WifiRadioEnergyModel::GetTxCurrentA),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("RxCurrentA",
                   "The radio Rx current in Ampere.",
                   DoubleValue (0.0197),    // receive mode = 19.7mA
                   MakeDoubleAccessor (&WifiRadioEnergyModel::SetRxCurrentA,
                                       &WifiRadioEnergyModel::GetRxCurrentA),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("SwitchingCurrentA",
                   "The default radio Channel Switch current in Ampere.",
                   DoubleValue (0.000426),  // default to be the same as idle mode
                   MakeDoubleAccessor (&WifiRadioEnergyModel::SetSwitchingCurrentA,
                                       &WifiRadioEnergyModel::GetSwitchingCurrentA),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("WiMeshNs3Ptr", "The pointer to the WiMeshNs3 object holding the topology",
                   WiMeshNs3PtrValue (WiMeshNs3Ptr ()),
                   MakeWiMeshNs3PtrAccessor (&WiMeshRadioEnergyModel::m_WiMeshNs3),
                   MakeWiMeshNs3PtrChecker ())
//     .AddTraceSource ("TotalEnergyConsumption",
//                      "Total energy consumption of the radio device.",
//                      MakeTraceSourceAccessor (&WifiRadioEnergyModel::m_totalEnergyConsumption))
  ; 
  return tid;
}


WiMeshRadioEnergyModel::WiMeshRadioEnergyModel()
{
  NS_LOG_FUNCTION (this);
}

WiMeshRadioEnergyModel::~WiMeshRadioEnergyModel()
{
  NS_LOG_FUNCTION (this);
}

void 
WiMeshRadioEnergyModel::SetWifiNetDevice(Ptr< WifiNetDevice > wnd)
{
  m_WifiDevice = wnd;
}

void
WiMeshRadioEnergyModel::SetWiMeshRadioPtr(RadioIface* ri)
{
  m_WiMeshRadioPtr = ri;
}

void
WiMeshRadioEnergyModel::SetVoltage(double v)
{
  m_voltage = v;
}


WiMeshNs3Ptr
WiMeshRadioEnergyModel::GetWiMeshNs3Ptr ()
{
  return m_WiMeshNs3;
}


void
WiMeshRadioEnergyModel::ChangeState (int newState)
{
  if (newState == WifiPhy::TX)
  {
    // set txCurrent
    
    NS_ASSERT (m_WifiDevice != NULL);

    // find the remote station manager
    Ptr<WifiRemoteStationManager> staman = m_WifiDevice->GetRemoteStationManager();
    
    // if the remote station manager is of type WiMeshWifiManager
    std::string managerName = staman->GetInstanceTypeId ().GetName ();
    if (managerName.compare ("ns3::WiMeshWifiManager") == 0)
    {
      Ptr<WiMeshWifiManager> sta = DynamicCast<WiMeshWifiManager> (staman);

      // get the current TX power level
      uint32_t txpower = sta->GetCurrTxPower();

      if (m_WiMeshRadioPtr)
	SetTxCurrentA ((*m_WiMeshRadioPtr->rho_tx)(DbmToW(TxPowerLevel(txpower))) / m_voltage);
    }
  }

  WifiRadioEnergyModel::ChangeState (newState);
}


/*
 * Private functions start here.
 */

void
WiMeshRadioEnergyModel::DoDispose (void)
{
//   NS_LOG_FUNCTION (this);
//   m_source = NULL;
//   m_energyDepletionCallback.Nullify ();
}


} // namespace ns3
