#ifndef WIMESH_RADIO_ENERGY_MODEL
#define WIMESH_RADIO_ENERGY_MODEL

#include "ns3/device-energy-model.h"
#include "ns3/nstime.h"
#include "ns3/event-id.h"
#include "ns3/traced-value.h"
#include "ns3/wifi-phy.h"
#include "ns3/energy-model-helper.h"
#include "ns3/wifi-radio-energy-model.h"
#include "wimesh-ns3.h"

namespace ns3 {

  
/**
 * \ingroup energy
 * \brief Assign WiMeshRadioEnergyModel to wifi devices.
 *
 * This installer installs WiMeshRadioEnergyModel for only WifiNetDevice objects.
 *
 */
class WiMeshRadioEnergyModelHelper : public DeviceEnergyModelHelper
{
public:
  /**
   * Construct a helper which is used to add a radio energy model to a node
   */
  WiMeshRadioEnergyModelHelper ();

  /**
   * Destroy a RadioEnergy Helper
   */
  ~WiMeshRadioEnergyModelHelper ();

  /**
   * \param name the name of the attribute to set
   * \param v the value of the attribute
   *
   * Sets an attribute of the underlying PHY object.
   */
  void Set (std::string name, const AttributeValue &v);

  /**
   * \param callback Callback function for energy depletion handling.
   *
   * Sets the callback to be invoked when energy is depleted.
   */
  void SetDepletionCallback (
    WifiRadioEnergyModel::WifiRadioEnergyDepletionCallback callback);


private:
  /**
   * \param device Pointer to the NetDevice to install DeviceEnergyModel.
   * \param source Pointer to EnergySource to install.
   *
   * Implements DeviceEnergyModel::Install.
   */
  virtual Ptr<DeviceEnergyModel> DoInstall (Ptr<NetDevice> device,
                                            Ptr<EnergySource> source) const;

private:
  ObjectFactory m_radioEnergy;
  WifiRadioEnergyModel::WifiRadioEnergyDepletionCallback m_depletionCallback;

};





/**
 * \ingroup energy
 * \brief A WiFi radio energy model.
 * 
 * This class inherits from WifiRadioEnergyModel. If used in conjunction with
 * the WiMesh wifi radio station manager (WiMeshWifiManager), it allows to
 * compute the energy consumed during a packet transmission according to the
 * rho_tx function of the corresponding WiMesh radio
 */
class WiMeshRadioEnergyModel : public WifiRadioEnergyModel
{
public:
  /**
   * Callback type for energy depletion handling.
   */
//  typedef Callback<void> WifiRadioEnergyDepletionCallback;

public:
  static TypeId GetTypeId (void);
  WiMeshRadioEnergyModel ();
  virtual ~WiMeshRadioEnergyModel ();

  /**
   * \brief Changes state of the WifiRadioEnergyMode.
   *
   * \param newState New state the wifi radio is in.
   *
   * Implements DeviceEnergyModel::ChangeState.
   */
  virtual void ChangeState (int newState);

  void SetWifiNetDevice (Ptr<WifiNetDevice> wnd);
  void SetWiMeshRadioPtr (RadioIface* ri);
  void SetVoltage (double v);

  WiMeshNs3Ptr GetWiMeshNs3Ptr ();

private:
  void DoDispose (void);

  Ptr<WifiNetDevice> m_WifiDevice;
  RadioIface* m_WiMeshRadioPtr;
  double m_voltage;
  WiMeshNs3Ptr m_WiMeshNs3;

};

} // namespace ns3

#endif /* WIMESH_RADIO_ENERGY_MODEL */
