#include "wimesh-wifi-manager.h"
#include <cassert>

namespace ns3 {

// records the tx rate and power used to send packets to a remote station
struct WiMeshWifiRemoteStation : public WifiRemoteStation
{
  uint32_t m_rate;
  uint32_t m_power;
};

NS_OBJECT_ENSURE_REGISTERED (WiMeshWifiManager);

TypeId
WiMeshWifiManager::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::WiMeshWifiManager")
    .SetParent<WifiRemoteStationManager> ()
    .AddConstructor<WiMeshWifiManager> ()
    .AddAttribute ("WiMeshNs3Ptr", "The pointer to the WiMeshNs3 object holding the topology",
                   WiMeshNs3PtrValue (WiMeshNs3Ptr ()),
                   MakeWiMeshNs3PtrAccessor (&WiMeshWifiManager::m_WiMeshNs3),
                   MakeWiMeshNs3PtrChecker ())
    ;
  return tid;
}

WiMeshWifiManager::WiMeshWifiManager ()
 : m_WiMeshNodeId (-1),        // uninitialized
   m_currTxPower (GetDefaultTxPowerLevel ())
{
}

WiMeshWifiManager::~WiMeshWifiManager ()
{
}

uint32_t
WiMeshWifiManager::GetCurrTxPower ()
{
  return m_currTxPower;
}


WifiRemoteStation*
WiMeshWifiManager::DoCreateStation (void) const
{
  WiMeshWifiRemoteStation *station = new WiMeshWifiRemoteStation ();
  
  station->m_rate = NUMRATES;           // uninitialized
  station->m_power = numTxPowerLevels;  // uninitialized
  return station;
}

void
WiMeshWifiManager::DoReportRxOk (WifiRemoteStation* station,
				 double rxSnr, WifiMode txMode)
{
}

void
WiMeshWifiManager::DoReportRtsFailed (WifiRemoteStation* station)
{
}

void
WiMeshWifiManager::DoReportDataFailed (WifiRemoteStation* station)
{
}

void
WiMeshWifiManager::DoReportRtsOk (WifiRemoteStation* station,
				  double ctsSnr, WifiMode ctsMode, double rtsSnr)
{
}

void
WiMeshWifiManager::DoReportDataOk (WifiRemoteStation* station,
				   double ackSnr, WifiMode ackMode, double dataSnr)
{
}

void
WiMeshWifiManager::DoReportFinalRtsFailed (WifiRemoteStation* station)
{
}

void
WiMeshWifiManager::DoReportFinalDataFailed (WifiRemoteStation* station)
{
  m_WiMeshNs3.GetPtr ()->ReceiveFinalDataFailed (this, station->m_state->m_address);
}

// WifiMode
// WiMeshWifiManager::DoGetDataMode (WifiRemoteStation* sta, uint32_t size)
WifiTxVector
WiMeshWifiManager::DoGetDataTxVector (WifiRemoteStation* sta, uint32_t size)
{
  WiMeshWifiRemoteStation *station = (WiMeshWifiRemoteStation*) sta;

  // if we have already found rate and power for this neighbor/WifiRemoteStation, return them
  if (station->m_rate < NUMRATES)
  {
    m_currTxPower = station->m_power;
//     return GetSupported (station, station->m_rate);
    return WifiTxVector (GetSupported (station, station->m_rate), station->m_power, GetLongRetryCount (station), GetShortGuardInterval (station), min (GetNumberOfReceiveAntennas (station),GetNumberOfTransmitAntennas()), GetNumberOfTransmitAntennas (station), GetStbc (station));
  }

  // we cannot go on if a WiMesh topology has not been specified
  assert (m_WiMeshNs3.GetPtr () != 0);

  // if we have not yet found the id of the node where this WifiManager is installed on, do it
  if (m_WiMeshNodeId < 0)
    m_WiMeshNodeId = m_WiMeshNs3.GetPtr()->DoFindNodeByStationManager(this);

  // we must initialize m_rate and m_power

  uint32_t wimesh_rate, wimesh_power;

  m_WiMeshNs3.GetPtr ()->GetTxRatePower (m_WiMeshNodeId, sta->m_state->m_address, wimesh_rate, wimesh_power);

  for (uint32_t i = 0; i < GetNSupported (sta); ++i)
    if (GetSupported (sta, i).GetDataRate () == wimesh_rate * 1e6)
    {
      station->m_rate = i;
      station->m_power = wimesh_power;
      m_currTxPower = wimesh_power;
//       return GetSupported (station, station->m_rate);
      return WifiTxVector (GetSupported (station, station->m_rate), station->m_power, GetLongRetryCount (station), GetShortGuardInterval (station), min (GetNumberOfReceiveAntennas (station),GetNumberOfTransmitAntennas()), GetNumberOfTransmitAntennas (station), GetStbc (station));
    }

  // we should not get here...
  station->m_rate = 0;
  station->m_power = wimesh_power;
  m_currTxPower = wimesh_power;
//   return GetSupported (station, station->m_rate);
  return WifiTxVector (GetSupported (station, station->m_rate), station->m_power, GetLongRetryCount (station), GetShortGuardInterval (station), min (GetNumberOfReceiveAntennas (station),GetNumberOfTransmitAntennas()), GetNumberOfTransmitAntennas (station), GetStbc (station));
}

// WifiMode
// WiMeshWifiManager::DoGetRtsMode (WifiRemoteStation* sta)
WifiTxVector
WiMeshWifiManager::DoGetRtsTxVector (WifiRemoteStation* sta)
{
  m_currTxPower = GetDefaultTxPowerLevel();
//   return GetSupported (sta, 0);
  return WifiTxVector (GetSupported (sta, 0), GetDefaultTxPowerLevel (), GetShortRetryCount (sta), GetShortGuardInterval (sta), min (GetNumberOfReceiveAntennas (sta),GetNumberOfTransmitAntennas()), GetNumberOfTransmitAntennas (sta), GetStbc (sta));
}

uint8_t
WiMeshWifiManager::DoGetCtsTxPowerLevel (Mac48Address address, WifiMode ctsMode)
{
  m_currTxPower = GetDefaultTxPowerLevel();
  return m_currTxPower;
}

uint8_t
WiMeshWifiManager::DoGetAckTxPowerLevel (Mac48Address address, WifiMode ackMode)
{
//   m_currTxPower = GetDefaultTxPowerLevel();
  return m_currTxPower;
}

uint8_t
WiMeshWifiManager::DoGetBlockAckTxPowerLevel (Mac48Address address, WifiMode blockAckMode)
{
//   m_currTxPower = GetDefaultTxPowerLevel();
  return m_currTxPower;
}

bool
WiMeshWifiManager::IsLowLatency (void) const
{
  return true;
}


} // namespace ns3
