#ifndef WIMESH_WIFI_MANAGER
#define WIMESH_WIFI_MANAGER


#include "ns3/wifi-remote-station-manager.h"
#include "wimesh-ns3.h"

namespace ns3 {

/*
 * ns3::WifiRemoteStationManager is an abstract base class. An object whose type is a class 
 * derived from WifiRemoteStationManager must be associated with a WifiNetDevice and is in 
 * charge to determine transmission rate and power for every frame the device has to transmit.
 * 
 * WiMeshWifiManager is a class derived from WifiRemoteStationManager which selects,
 * for each data frame to transmit, the transmission rate and power associated with the
 * corresponding link in the WiMesh topology (TGraphNeighbor::rate & txpower).
 * Control frames (ack, rts, cts) are sent with the ns-3 default values for rate and power.
 * 
 * WiMeshWifiManager also stores (in m_currTxPower) the TX power level used to transmit
 * the current frame. Such a value is needed by WiMeshRadioEnergyModel to correctly
 * compute the energy consumption.
 */


class WiMeshWifiManager : public WifiRemoteStationManager
{
public:
  static TypeId GetTypeId (void);
  WiMeshWifiManager ();
  virtual ~WiMeshWifiManager ();

  uint32_t GetCurrTxPower ();

private:
  // overriden from base class
  virtual WifiRemoteStation *DoCreateStation (void) const;
  virtual void DoReportRxOk (WifiRemoteStation *station, 
                             double rxSnr, WifiMode txMode);
  virtual void DoReportRtsFailed (WifiRemoteStation *station);
  virtual void DoReportDataFailed (WifiRemoteStation *station);
  virtual void DoReportRtsOk (WifiRemoteStation *station,
                              double ctsSnr, WifiMode ctsMode, double rtsSnr);
  virtual void DoReportDataOk (WifiRemoteStation *station,
                               double ackSnr, WifiMode ackMode, double dataSnr);
  virtual void DoReportFinalRtsFailed (WifiRemoteStation *station);
  virtual void DoReportFinalDataFailed (WifiRemoteStation *station);

//   virtual WifiMode DoGetDataMode (WifiRemoteStation *sta, uint32_t size);   // REPLACED since ns-3.18
  virtual WifiTxVector DoGetDataTxVector (WifiRemoteStation *sta, uint32_t size);

//   virtual WifiMode DoGetRtsMode (WifiRemoteStation *sta);   // REPLACED since ns-3.18
  virtual WifiTxVector DoGetRtsTxVector (WifiRemoteStation *sta);

  virtual uint8_t DoGetCtsTxPowerLevel (Mac48Address address, WifiMode ctsMode);

  virtual uint8_t DoGetAckTxPowerLevel (Mac48Address address, WifiMode ackMode);

  virtual uint8_t DoGetBlockAckTxPowerLevel (Mac48Address address, WifiMode blockAckMode);

  virtual bool IsLowLatency (void) const;

  WiMeshNs3Ptr m_WiMeshNs3;

  // the WiMesh id of the node where this WifiManager is installed on
  int m_WiMeshNodeId;

  uint32_t m_currTxPower;
};



} // namespace ns3


#endif
