#include "parsers.h"
#include "pfr_methods.h"
#include "linprogs.h"
#include "CAalgos.h"
#include <fstream>
#include <ext/functional>
#include <algorithm>
#include <ctime>

/*
 * Usage example:
 * 
 * cat configWM.xml | ./WMrun
 * 
 * This program runs a WiMesh simulation, as specified in the XML configuration file.
 * The topology can be either deserialized from a binary file or read from an XML
 * file. In the latter case, the config file should also specify number of available
 * channels, method to build the initial graph, method to compute the initial flow
 * rates and a string identifying the simulation.
 * In both cases, the power consumption profile of nodes and radios can be set (or
 * altered) as specified in the config file.
 * Traffic demands can be altered before running the specified channel assignment
 * algorithm. A post-optimization stage (minimization of the maximum Utot given the
 * current channel assignment) can be performed after running the channel assignment
 * algorithm.
 * 
 * If the topology is read from a text file, the simulation id is provided in the
 * config file (or it is the default value "xxx" if not specified). If the topology
 * is deserialized from a binary file, the simulation id is taken from the name of
 * such file (which should begin with "topodata_<id>"). In the latter case, the id of
 * the demands change and/or the name of the channel assignment algorithm are appended.
 * 
 * Two files are generated: topodata_<simulId>, which contains the serialization of
 * the current topology status (after running the channel assignment algorithm), and
 * utot_<simulId>, which contains some statistics (e.g., Utot values) related to the
 * current topology status. The former can be passed as input to the RunNs3 utility,
 * which performs an ns-3 simulation of the stored topology.
 */
 
 

int main()
{

  WiMeshParser WMparser;
  string simulId;
  bool serializedTopo = false;
  double ret_value;
  Topology* T;
  PFR_method* pfr = NULL;
  CAalgo* caAlgo = NULL;
  double elapsed_time;

  
  try
  {
    WMparser.set_substitute_entities(true);
    WMparser.parse_stream(std::cin);
  }
  catch(const xmlpp::exception& ex)
  {
    std::cerr << "libxml++ exception: " << ex.what() << std::endl;
  }
  
  
  if (!WMparser.getSerializedTopoFile().empty())
  {
    /*** Read the topology serialized in a binary file ***/
    
    // check whether the specified file exists
    ifstream topofile;

    // starting from a serialized topology
    cout << endl << "** Starting from the topology stored in " << WMparser.getSerializedTopoFile() << " **" << endl;

    T = new Topology;

    T->DeserializeFrom(WMparser.getSerializedTopoFile().c_str());

    if (T->node.empty())
    {
      cerr << endl << "Error while reading the topology. Exiting." << endl << endl;
      return -1;
    }

    serializedTopo = true;

    if (WMparser.getSerializedTopoFile().substr(0, string("topodata_").length()) == "topodata_")
      // simulId is what follows topodata_
      simulId = WMparser.getSerializedTopoFile().substr(string("topodata_").length());
    else
    {
      cout << endl << "Warning: the serialized file name does not start with topodata_" << endl << endl;
      simulId = WMparser.getSerializedTopoFile();
    }

    simulId += "~";
  }
  else if (!WMparser.getXmlTopoFile().empty())
  {
    /*** Read the topology stored in an XML file ***/

    T = new Topology (WMparser.getXmlTopoFile().c_str());
    
    T->numAvailChannels = WMparser.getNumAvailChannels();

    strcpy(T->topopath, WMparser.getXmlTopoFile().c_str());
    string::size_type pos2 = WMparser.getXmlTopoFile().rfind(".");
    string::size_type pos1 = WMparser.getXmlTopoFile().rfind("/");
    strcpy(T->toponame, WMparser.getXmlTopoFile().substr(pos1+1,pos2-pos1-1).c_str());

    simulId = WMparser.getSimulId();


    /*** Build the initial graph ***/

    if (WMparser.getBuildGiMethod().empty())
    {
      cerr << "** ERROR: No BuildGi method has been specified **" << endl;
      return 0;
    }

    if (WMparser.getBuildGiMethod() == "FixedPowerMaxAllowRate")
    {
      int power_index = defaultTxPowerLevel, min_rate = 3, max_rate = 7;
      float power_dbm = TxPowerLevel(power_index);

      for (list<pair<string,string> >::const_iterator p = WMparser.getBuildGiParams().begin(); p != WMparser.getBuildGiParams().end(); p++)
      {
	if (p->first == "power_dbm")
	  power_dbm = atof (p->second.c_str());
	else if (p->first == "min_rate")
	  min_rate = atoi (p->second.c_str());
	else if (p->first == "max_rate")
	  max_rate = atoi (p->second.c_str());
      }

      // if the user specified power_dbm, find the highest TX power level not greater than power_dbm
      if (power_dbm != TxPowerLevel(power_index))
      {
	power_index = numTxPowerLevels - 1;
	while (power_index > 0 && TxPowerLevel(power_index) > power_dbm)
	  power_index--;
      }

      if (min_rate < 0 || min_rate > (int)(NUMRATES-1) ||
	  max_rate < 0 || max_rate > (int)(NUMRATES-1) )
      {
	cerr << endl << "ERROR: Inadmissible rate values for FixedPowerMaxAllowRate" << endl << endl;
	return 0;
      }

      T->buildGi_FixedPowerMaxAllowRate(power_index, min_rate, max_rate);
      T->maxP = power_index;
      T->min_rate = min_rate;
      T->max_rate = max_rate;
    }
    else if (WMparser.getBuildGiMethod() == "FromFile")
    {
      string filename;
      list<pair<string,string> > params = WMparser.getBuildGiParams();

      for (list<pair<string,string> >::iterator p = params.begin(); p != params.end(); p++)
	if (p->first == "filename")
	  filename = p->second;

      if (filename.empty())
      {
	cerr << endl << "ERROR: Missing file name for FromFile buildGi method" << endl << endl;
	return 0;
      }

      T->buildGi_FromFile(filename.c_str());
    }
    else
    {
      cerr << "** ERROR: " << WMparser.getBuildGiMethod() << " is not a valid BuildGi method **" << endl;
      return 0;
    }

    if (!T->isConnected(true))
    {
      cout << "The initial graph is not connected!" << endl;
      return 0;
    }

    cout << "The average node degree of the initial graph is " << T->getDegree() << endl;


    /*** Compute the set of flow rates ***/

    if (WMparser.getPfrMethod().empty())
    {
      cerr << "** ERROR: No PFR method has been specified **" << endl;
      return 0;
    }
    
    pfr = PFR_method::Create (WMparser.getPfrMethod(), WMparser.getPfrParams());
    
    if (!pfr)
    {
      cerr << endl << "** ERROR: No valid PFR method has been specified **" << endl << endl;
      return 0;
    }
    
    pfr->pfr_calc(T);

  }
  else
  {
    cerr << "** ERROR: No topology file has been specified **" << endl;
    return 0;
  }
  

  /*** Set the specified power consumption profile (if any) for each node ***/
  
  if (WMparser.getNodePCrv())
  {
    int id;
    MotherOfAll* urng = new MotherOfAll(WMparser.getNodePCseed());         // declare uniform random number generator
    Random::Set(*urng);                   // set urng as generator to be used

    for (vector<TGraphNode>::iterator u = T->node.begin(); u != T->node.end(); u++)
    {
      id = (int)WMparser.getNodePCrv()->rv->Next();

      // find a profile with such an id
      map<int, NodePowerConsumptionProfile>::const_iterator mi;

      mi = WMparser.getNodePCprofile().find (id);

      if (mi == WMparser.getNodePCprofile().end())
      {
	cerr << endl << "Warning: node profile id=" << id << " not found. Leaving node " << u->info->id << " unchanged" << endl << endl;
      }
      else
      {
	u->info->BasePowerConsumption = mi->second.BasePowerConsumption;
	u->info->StandbyPowerConsumption = mi->second.StandbyPowerConsumption;
      }
    }
    
    delete urng;
  }


  /*** Set the specified power consumption profile (if any) for each radio ***/
  
  if (WMparser.getRadioPCrv())
  {
    int id;
    MotherOfAll* urng = new MotherOfAll(WMparser.getRadioPCseed());         // declare uniform random number generator
    Random::Set(*urng);                   // set urng as generator to be used

    for (vector<TGraphNode>::iterator u = T->node.begin(); u != T->node.end(); u++)
      for (RadioIfaceIterator r = u->info->radio.begin(); r != u->info->radio.end(); r++)
      {
	id = (int)WMparser.getRadioPCrv()->rv->Next();

	// find a profile with such an id
	map<int, RadioPowerConsumptionProfile>::const_iterator mi;

	mi = WMparser.getRadioPCprofile().find (id);

	if (mi == WMparser.getRadioPCprofile().end())
	{
	  cerr << endl << "Warning: radio profile id=" << id << " not found. Leaving radio on node " << u->info->id << " unchanged" << endl << endl;
	}
	else
	{
	  r->rho_tx = mi->second.rho_tx->Clone();
	  r->rho_rx = mi->second.rho_rx;
	  r->rho_idle = mi->second.rho_idle;
	  r->rho_sleep = mi->second.rho_sleep;
	}
      }
    
    delete urng;
  }


  /*** Change demands, if required ***/

  if (WMparser.getCDrv())
  {
    cout << endl << "** Changing the traffic demands (seed = " << WMparser.getCDseed() << ", rv = ";
    WMparser.getCDrv()->print(cout);
    cout << ", type = " << WMparser.getCDtype() << ") **" << endl;

    T->changeDemands (WMparser.getCDrv(), WMparser.getCDseed(), WMparser.getCDtype());

    if (serializedTopo)
      simulId += "cd" + WMparser.getCDid();
  }
  
  
  T->printTopology();


  /*** Run the specified channel assignment algorithm ***/

  if (WMparser.getCAalgo().empty())
  {
    cerr << "** ERROR: No CA algorithm has been specified **" << endl;
    return 0;
  }

  caAlgo = CAalgo::Create (WMparser.getCAalgo(), WMparser.getCAparams());

  if (!caAlgo)
  {
    cerr << endl << "** ERROR: No valid channel assignment algorithm has been specified **" << endl << endl;
    return 0;
  }

  vector<set<unsigned int> > prevChannels(T->node.size());

  if (serializedTopo)
  {
    // update simulId
    if (WMparser.getCDrv())
      simulId += "+";

    simulId += caAlgo->name();

    // record the current channels assigned to all the radios (used to detect channel switches)
    for (vector<TGraphNode>::iterator u = T->node.begin(); u != T->node.end(); u++)
      for (RadioIfaceIterator ri = u->info->radio.begin(); ri != u->info->radio.end(); ri++)
	prevChannels[u->info->id].insert(ri->channel);
  }

  cout << endl << "** Running the CA algorithm " << caAlgo->name() << " **" << endl;

  clock_t startt = clock();

  ret_value = caAlgo->Run(T);

  if (WMparser.getpostCAopt())
  {
    cout << endl << "Running MinMaxUtot to optimize the distribution of flows" << endl << endl;
    
    /* check that e2e flows are consistent and obtain the set of demands */
    map<pair<int,int>,double> demands;
    
    if (!(T->checkE2Eflows(&demands)))
      cerr << endl << "WARNING: Checking E2E flows before post-optimization FAILED" << endl << endl;

    // re-route flows in order to minimize the maximum total utilization
    double maxUtot = numeric_limits<double>::max();
    MinMaxUtot (T, &demands, maxUtot);

    // add the suffix "O" to the algo name
    if (serializedTopo)
      simulId += "O";
  }

  elapsed_time = (double)(clock() - startt) / CLOCKS_PER_SEC;

  // quit in case of problems
  if (ret_value)
  {
    string fname = string("utot_") + simulId;
    ofstream ofs;
    ofs.open(fname.c_str(), ofstream::out);
    ofs.close();
    return -1;
  }

  int rchanges = 0;

  if (serializedTopo)
  {
    // detect channel switches
    for (vector<TGraphNode>::iterator u = T->node.begin(); u != T->node.end(); u++)
      for (RadioIfaceIterator ri = u->info->radio.begin(); ri != u->info->radio.end(); ri++)
	if (prevChannels[u->info->id].find(ri->channel) == prevChannels[u->info->id].end())
	  rchanges++;
  }


  /************************/

  printCAStats(T, caAlgo->name(), true);

  T->drawTopology();

  cout << "Throughput = " << T->throughput << endl;


  /*** Save the results to files ***/  

  string fname = string("utot_") + simulId;
  ofstream ofs;

  ofs.open(fname.c_str(), ofstream::out);

  if (!serializedTopo)
  {
    ofs << "# Number of channels: " << WMparser.getNumAvailChannels() << endl;
    ofs << "# Topology file: " << WMparser.getXmlTopoFile() << endl;
    ofs << "# " << WMparser.getBuildGiMethod() << " ";
    for (list<pair<string,string> >::const_iterator p = WMparser.getBuildGiParams().begin(); p != WMparser.getBuildGiParams().end(); p++)
      ofs << p->first << " = " << p-> second << "  ";
    ofs << endl;
    ofs << "# " << pfr->getStr() << endl;
    ofs << "# CA algo: " << caAlgo->name();
    if (WMparser.getpostCAopt())
      ofs << "O";
    ofs << endl;
  }
  
  if (WMparser.getCDrv())
  {
    ofs << "# Change demands: seed = " << WMparser.getCDseed() << "  rv = ";
    WMparser.getCDrv()->print(ofs);
    ofs << "  type = " << WMparser.getCDtype() << endl << endl;
  }

  // save all the Utots and the sum of the demands in a file
  vector< pair<TLinkItrs,double> >* UtotVect;

  T->checkUtot(0, &UtotVect);

  ofs << "Utot = [" << endl;
  transform (UtotVect->begin(), UtotVect->end(), ostream_iterator<double>(ofs, "\n"), __gnu_cxx::select2nd<pair<TLinkItrs,double> >());
  ofs << "];" << endl << endl << "demands = " << T->throughput << ";" << endl;

  if (serializedTopo)
    ofs << endl << "NumRadioChanged = " << rchanges << ";" << endl;
  
  ofs << endl << "NumRadiosOff = " << T->countRadiosOff() << ";" << endl;
  ofs << endl << "NumNodesOff = " << T->countNodesOff() << ";" << endl;

  /* compute the power consumption of each node */
  vector<TPower> pconsume;
  T->calcAllNodesPowerConsumption(pconsume);

  ofs << endl << "PowerConsumption = " << T->TotalPowerConsumption(&pconsume) << ";" << endl;

  ofs << endl << "NodeConsumption = [" << endl;
  for (unsigned int i=0; i<pconsume.size(); i++)
    ofs << i << "  " << pconsume[i].power << endl;
  ofs << "];" << endl << endl;

  /* check the power constraints, if any */
  vector<float>* nodepc = caAlgo->getPowerConstraints();
  if (nodepc)
  {
    bool exceeded = false;
    for (unsigned int i=0; i<T->node.size(); i++)
      if (pconsume.at(i).power > nodepc->at(i))
      {
	ofs << "# " << i << " (" << nodepc->at(i) << ")" << endl;
	exceeded = true;
      }
    if (exceeded)
      ofs << endl << "# The power constraint(s) on the above node(s) have NOT been met" << endl;
    else
      ofs << "# The power constraints on all the nodes have been met" << endl;
  }

  ofs << endl << "ElapsedTime = " << elapsed_time << ";" << endl;

  caAlgo->printFurtherInfo(ofs);

  ofs.close();

  string f1name = string("topodata_") + simulId;

  T->SerializeTo(f1name.c_str());
}

