#include <coin/IpIpoptApplication.hpp>
#include "split_ratios.h"
#include "mpls-parser.h"
#include "WMtypes.h"
#include "linprogs.h"
#include <cassert>
#include <sstream>
#include <fstream>
#include <string>


using namespace Ipopt;


/*
 * Usage example:
 * 
 * cat configMpls.xml | ./calcMplsRatios
 * 
 * This program computes the MPLS split ratios for a topology serialized in a given file.
 * The XML config file also allows to specify the method to compute the directed subgraph
 * (RDAS or KSP) and to set the value for some IpOpt options.
 */
 

int main()
{
  SplitRatiosParser parser;
  stringstream simulId;

  try
  {
    parser.set_substitute_entities(true);
    parser.parse_stream(std::cin);
  }
  catch(const xmlpp::exception& ex)
  {
    std::cerr << "libxml++ exception: " << ex.what() << std::endl;
  }

  if (parser.topofile.empty())
  {
    cerr << endl << "A file containing a serialized topology must be specified in the configuration. Exiting." << endl << endl;
    return -1;
  }

  // deserialize the topology from the specified file

  Topology* T = new Topology;

  T->DeserializeFrom (parser.topofile.c_str());

  if (T->node.empty())
  {
    cerr << endl << "Error while reading the topology from " << parser.topofile << ".. Exiting." << endl << endl;
    delete T;
    return -1;
  }

  if (parser.rdas)
    simulId << "RDAS_";
  else
    simulId << "KSP_";
  
  // (part of) the topology file name is used as identifier

  if (parser.topofile.substr(0, string("topodata_").length()) == "topodata_")
    // simulId is what follows topodata_
    simulId << parser.topofile.substr(string("topodata_").length());
  else
  {
    cout << endl << "Warning: the serialized file name does not start with topodata_" << endl << endl;
    simulId << parser.topofile;
  }

//   T->printTopology();

  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> mapD;
  
  if (!(T->checkE2Eflows(&mapD)))
    cerr << endl << "WARNING: main: Checking E2E flows before beginning FAILED" << endl << endl;
  cout << "Demands:" << endl;
  for(map<pair<int,int>,double>::iterator m=mapD.begin(); m!=mapD.end(); m++)
    cout << "d[" << m->first.first << "," << m->first.second << "] = " << m->second << endl;


  vector<TDemand> demands;
  for (map<pair<int,int>,double>::iterator d=mapD.begin(); d!=mapD.end(); d++)
      demands.push_back(TDemand(d->first.first, d->first.second, d->second));

  /* find the set of links carrying the traffic of each demand */
  vector< set< pair< int, int > > > Sd;
  
  for (vector<TDemand>::iterator d=demands.begin(); d!=demands.end(); d++)
    if (parser.rdas)
      Sd.push_back(T->FindLooplessResilientDAG(d->src(), d->dst(), parser.alpha));      
    else 
      Sd.push_back(T->find_path(pair<int,int>(d->src(),d->dst()), parser.alpha, 10000));

  for (int i=0; i<demands.size(); i++)
  {
    cout << "Sd per demand " << demands[i].src() << "-" << demands[i].dst() << endl;
    for (set<pair<int,int> >::iterator l=Sd[i].begin(); l!=Sd[i].end(); l++)
      cout << l->first << "->" << l->second << " ";
    cout << endl;
  }


  /* Call MinMaxUtot to reroute flows to minimize the maximum Utot now that channels
   * have been assigned. The capacity constraints are included to check that the 
   * current set of demands is feasible. If the LP can be optimally solved, the 
   * solution provides the initial solution for the convex optimization problem
   */
  double maxUtot = numeric_limits<double>::max();
  if (MinMaxUtot(T, &mapD, maxUtot, Sd, true))
  {
    cerr << "The set of demands is not feasible given the current channel assignment, program terminates." << endl;
    return 0;
  }

  
  // Create a new instance of your nlp
  //  (use a SmartPtr, not raw)
  SmartPtr<TNLP> mynlp = new SplitRatios(T,demands,Sd);

  SplitRatios* mynlp_ptr = (SplitRatios*)GetRawPtr(mynlp);
  
  // verbosely print the problem
  // cout << *mynlp_ptr;
  
  // Create a new instance of IpoptApplication
  //  (use a SmartPtr, not raw)
  // We are using the factory, since this allows us to compile this
  // example with an Ipopt Windows DLL
  SmartPtr<IpoptApplication> app = IpoptApplicationFactory();

  app->Options()->SetStringValue("output_file", simulId.str()+".ipopt_log");

  // Set the options specified in the configuration
  for (vector<pair<string,double> >::iterator i = parser.numericValues.begin(); i != parser.numericValues.end(); i++)
    app->Options()->SetNumericValue(i->first, i->second);

  for (vector<pair<string,int> >::iterator i = parser.integerValues.begin(); i != parser.integerValues.end(); i++)
    app->Options()->SetIntegerValue(i->first, i->second);

  for (vector<pair<string,string> >::iterator i = parser.stringValues.begin(); i != parser.stringValues.end(); i++)
    app->Options()->SetStringValue(i->first, i->second);


  // Intialize the IpoptApplication and process the options
  ApplicationReturnStatus status;
  status = app->Initialize();
  if (status != Solve_Succeeded) {
    cout << endl << endl << "*** Error during initialization!" << endl;
    return (int) status;
  }

  // Ask Ipopt to solve the problem
  status = app->OptimizeTNLP(mynlp);

  if (status == Solve_Succeeded)
    cout << endl << endl << "*** The problem solved!" << endl;
  else 
    cout << endl << endl << "*** The problem FAILED!" << endl;

  fstream f ((simulId.str()+".cfg").c_str(), ofstream::out);
  
  f << "# " << parser.topofile << endl;

  f << "# " << (parser.rdas ? "RDAS  " : "KSP  ") << parser.alpha << endl;

  // write split ratios to file
  mynlp_ptr->saveRatios((simulId.str()+".cfg").c_str());

  // append exit status, throughput and objective value
  f << "Exit status: " << (int) status << endl << endl;
  
  f << "Objective value: " << mynlp_ptr->getObjValue() << endl << endl;

  f << "Throughput: " << T->throughput << endl << endl;
  
  f.close();

  vector<map<TLinkItrs,double> > ratios = mynlp_ptr->getRatios();
  
  for (int i=0; i<ratios.size(); i++) {
    cout << "Demand: (" << demands[i].src() << "," << demands[i].dst() << ") Fmax = " << demands[i].Fmax() << endl;
    
    for (map<TLinkItrs,double>::iterator m=ratios[i].begin(); m!=ratios[i].end(); m++)
      cout << m->first.first->info->id << "->" << m->first.second->info->id << ": " << m->second << endl;
    cout << endl;
  }

  
  // As the SmartPtrs go out of scope, the reference count
  // will be decremented and the objects will automatically
  // be deleted.

  return (int) status;
}

