#include "WMtypes.h"
#include "linprogs.h"
#include <iostream>
#include <cassert>
#include <sstream>
#include <algorithm>
#include <ext/functional>
#include <ClpSimplex.hpp>
#include <CoinBuild.hpp>


/*
 * Rearrange e2e flows of given demands to minimize the maximum Utot
 * based on a Linear Program
 * Sd is an (optional) vector including the set of links, for each demand (in the
 * same order as in the demands map),
 * that are allowed to carry the flow of the corresponding demand
 * If capacity is true, the capacity constraints are included (which may
 * make the LP infeasible)
 * Includes the constraint M<=maxUtot
 * Changes are applied only if an optimal solution is found, hence
 * the resulting max Utot < maxUtot
 * In that case, maxUtot is updated to the new max Utot
 * Returns the result of the LP (OPTIMAL, INFEASIBLE, etc.)
 */
int MinMaxUtot(Topology* T, map<pair<int,int>,double>* demands, double& maxUtot, vector< set< pair< int, int > > > Sd, bool capacity, vector<float> *nodePowerConstr) {
//   lprec *lp;
  ClpSimplex model;
  int Ncol;

  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands->size();
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->node.size() + 1];

  cumNumNeighbs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();

  // the number of variables is numDemands*numLinks (f^ab(u->v)) + 1 (M)
  Ncol = numDemands*cumNumNeighbs[T->node.size()]+1;

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model.resize(0, Ncol);

  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD  M
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
//      int i=1;
      int i=0;
      for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++, i++) {
	stringstream s;
	s << "f(" << u->info->id << "," << v->info->id << ")[" << d->first.first << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+i, str);
      }
      linkId++;
    }
  
  int* colno = new int[Ncol];
  double *coeff = new double[Ncol], *coeff2 = new double[Ncol];
  
  /* set the objective function: min M*/
//   coeff[0] = 1;
//   colno[0] = Ncol;
//   set_obj_fnex(lp, 1, coeff, colno);
  model.setObjectiveCoefficient(Ncol-1, 1);

  /* set the object direction to minimize */
//   set_minim(lp);
  model.setOptimizationDirection(1);
  
//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  /* M < maxUtot (through M <= maxUtot - 0.001) */
//    if (maxUtot < numeric_limits<REAL>::max())
//   {
//     coeff[0] = 1;
//     colno[0] = Ncol;
//     add_constraintex(lp, 1, coeff, colno, LE, maxUtot+0.0001);
//   }

  vector<vector<int> > colno_v(T->node.size()*numDemands);
  vector<vector<double> > coeff_v(T->node.size()*numDemands);

  CoinBuild buildObject;

  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands->begin(); di!=demands->end(); di++, d++) {
      // for (unsigned int d=0; d<numDemands; d++) {
	dist = distance(u->neighbor.begin(), v);
	/* flow leaving node u */
	coeff_v[u->info->id*numDemands+d].push_back(1);
// 	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d);
	/* flow entering node v */
	coeff_v[v->info->id*numDemands+d].push_back(-1);
// 	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d);
	/* if v is the source of the demand or u is the destination of the demand, f(u->v)[d] must be null */
	if ( (u->info->id == di->first.second) || (v->info->id == di->first.first) )
// 	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, 0.);
	  model.setColumnUpper((cumNumNeighbs[u->info->id] + dist)*numDemands + d, 0.);
//	else
//	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, di->second);
      }
    }

  /* Add the flow conservation constraints one-by-one */
  int k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      /* if i is the source of the demand d */
      if (i == d->first.first)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
      /* if i is the destination of the demand d */
      else if (i == d->first.second)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
      else
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
      k++;
    }

  /* constraints on the total utilizations ( |E| constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      k = 0;
      for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	  if ( TLinkItrs(x,y).IsInTheCollDomainOf(T,TLinkItrs(u,v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i;
	    k += numDemands;
	  }
      coeff[k] = -1;
//       colno[k++] = Ncol;         // -M
      colno[k++] = Ncol-1;         // -M
//       add_constraintex(lp, k, coeff, colno, LE, 0 );
      buildObject.addRow(k, colno, coeff, -COIN_DBL_MAX, 0.);
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because the goal of minimizing M
   * will force to minimize the e2e flows
   */
  if (capacity)
    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
	fill(coeff, coeff+numDemands, 1);
	for (int i=0; i<numDemands; i++)
// 	  colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i +1;
	  colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i;
	
// 	add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
	buildObject.addRow(numDemands, colno, coeff, 0., v->rate);
      }

  /* constraints to only have non-null flow on the links belonging to Sd */
  if (!Sd.empty())
    for (int d=0; d<numDemands; d++) {
      for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
	  if (Sd[d].find(pair<int,int>(u->info->id,v->info->id)) == Sd[d].end()) {
	    dist = distance(u->neighbor.begin(), v);
// 	    set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, 0.);
	    model.setColumnUpper((cumNumNeighbs[u->info->id] + dist)*numDemands + d, 0.);
	  }
    }

  /* constraints on the node power consumption */
  if (nodePowerConstr)
  {
    map<TRadio, set<TLinkItrs> > Lri, Lro;  // maps each radio to the set of incoming/outgoing links using it

    /* compute Lri and Lro for all the radios */
    for (unsigned int u=0; u<T->node.size(); u++)
      for (list<TGraphNeighbor>::iterator v=T->node[u].neighbor.begin(); v!=T->node[u].neighbor.end(); v++)
      {
	/* insert u->v in the set Lro of the radio on u */
	Lro[TRadio(u,distance(T->node[u].info->radio.begin(),RadioIfaceIterator(v->radio)))].insert(TLinkItrs(T->node.begin()+u,v));
	/* insert u->v in the set Lri of the radio on v */
// 	pair<bool,unsigned int> ch = SearchChannel(v->info, v->radio->channel);
	RadioIfaceIterator ri = v->info->findRadioWithChannel(v->radio->channel);
// 	if (ch.first)
	if (ri != v->info->radio.end())
	  Lri[TRadio(v->info->id, distance(v->info->radio.begin(),ri))].insert(TLinkItrs(T->node.begin()+u,v));
      }

    int pos_i;
    /* constraints tau_TX(r) + tau_RX(r) <= 1 (for each active radio) and node power constraints */
    for (unsigned int u=0; u<T->node.size(); u++)
    {
      k=0;

      /* add the constraint tau_TX(r) + tau_RX(r) <= 1 for each active radio */
      for (unsigned int i=0; i<T->node[u].info->numAvailRadios(); i++)
	if (T->node[u].info->radio[i].channel)
	{
	  int j1=k;
	  RadioIfaceIterator ri = T->node[u].info->radio.begin() + i;

	  if (Lro.find(TRadio(u,i)) != Lro.end())
	    for (set<TLinkItrs>::iterator l=Lro[TRadio(u,i)].begin(); l!=Lro[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->rate);
	      // coeff2 is used for the power consumption constraint
	      fill_n(coeff2+k, numDemands, ((*ri->rho_tx)(DbmToW(TxPowerLevel(l->second->txpower))) - ri->rho_idle) / l->second->rate);
	      pos_i = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, counter<int>(pos_i));
	      generate_n(colno+k, numDemands, counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  int j2=k;

	  if (Lri.find(TRadio(u,i)) != Lri.end())
	    for (set<TLinkItrs>::iterator l=Lri[TRadio(u,i)].begin(); l!=Lri[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->rate);
	      fill_n(coeff2+k, numDemands, (ri->rho_rx - ri->rho_idle) / l->second->rate);
	      pos_i = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, counter<int>(pos_i));
	      generate_n(colno+k, numDemands, counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  if (k != j1)
	  {
	    /* make a copy of coeff and colno because add_constraintex sorts the element in the passed vectors */
	    double coeff_c[k-j1];
	    int colno_c[k-j1];
	    
	    copy (coeff+j1, coeff+k, coeff_c);
	    copy (colno+j1, colno+k, colno_c);
	    
// 	    add_constraintex(lp, k-j1, coeff_c, colno_c, LE, 1);
	    buildObject.addRow(k-j1, colno_c, coeff_c, 0., 1.);
	  }

	  /* multiply coefficients by the power consumption (needed for node power constraint) */
// 	  for (int j=j1; j<j2; j++)
// 	    coeff[j] *= (T->pm.rho_tx-T->pm.rho_idle);
// 
// 	  for (int j=j2; j<k; j++)
// 	    coeff[j] *= (T->pm.rho_rx-T->pm.rho_idle);
	}

      /* if node u has active radios, add the node power constraint */
      if ((pos_i=T->node[u].info->countRadiosWithChannel(0)) < T->node[u].info->numAvailRadios())
// 	add_constraintex(lp, k, coeff, colno, LE, nodePowerConstr->at(u) - T->pm.BasePowerConsumption - pos_i*T->pm.rho_sleep - (T->node[u].info->numAvailRadios() - pos_i)*T->pm.rho_idle);
// 	buildObject.addRow(k, colno, coeff, 0., nodePowerConstr->at(u) - T->pm.BasePowerConsumption - pos_i*T->pm.rho_sleep - (T->node[u].info->numAvailRadios() - pos_i)*T->pm.rho_idle);
      {
	float constr = nodePowerConstr->at(u) - T->node[u].info->BasePowerConsumption;

	for (RadioIfaceIterator ri = T->node[u].info->radio.begin(); ri != T->node[u].info->radio.end(); ri++)
	    constr -= (ri->channel ? ri->rho_idle : ri->rho_sleep);

	buildObject.addRow(k, colno, coeff2, 0., constr);
      }
    }
  }
    

//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model.addRows(buildObject);

  // write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /* Now let lpsolve calculate a solution */
//   int ret = solve(lp);
  int ret = model.initialSolve();

//   if (ret != OPTIMAL)
  if (ret)
    cerr << endl << "WARNING: MinMaxUtot: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
  else
  {
//    print_solution(lp,4);

    /* If an optimal solution has been found, then update the flows, the e2e flows and the Utot */

//     maxUtot = get_objective(lp);
    maxUtot = model.objectiveValue();
    
    double *ptr_var, *ptr_constr;
    
//     get_ptr_variables(lp, &ptr_var);
//     get_ptr_constraints(lp, &ptr_constr);
    
    ptr_var = model.primalColumnSolution();
    ptr_constr = model.primalRowSolution();
    
    unsigned int linkId = 0;
    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      {
	v->flow = 0.;
	v->f_e2e.clear();

	int i=0;
	for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++, i++)
	  if (ptr_var[linkId*numDemands + i] > 0.)
	    v->flow += (v->f_e2e[pair<int,int>(d->first.first,d->first.second)] = ptr_var[linkId*numDemands + i]);
	
	v->Utot = ptr_constr[T->node.size()*numDemands+linkId] + ptr_var[Ncol-1];

	linkId++;
      }
  }
  
  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;

  /* clean up such that all used memory by lpsolve is freed */
//   if (lp)
//     delete_lp(lp);

  return ret;
}



/*
 * Rearrange e2e flows to minimize the sum of the flows on the links L of a radio
 * such that no Utot above u0 increases nor exceeds u0
 * If the minimum is zero, returns true (the radio can be turned off)
 */
bool TurnOffRadio(Topology* T, map<pair<int,int>,double>* demands, set<TLinkItrs>* L, double u0) {
//   lprec *lp;
  ClpSimplex model;
  int Ncol;
  bool success = false;

  cout << "TurnOffRadio: set L: ";
  for (set<TLinkItrs>::iterator l=L->begin(); l!=L->end(); l++)
    cout << "(" << l->first->info->id << "," << l->second->info->id << ")[" << l->second->radio->channel << "] ";
  cout << endl;
  
  
  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands->size();
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->node.size() + 1];

  cumNumNeighbs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();

  // the number of variables is numDemands*numLinks (f^ab(u->v))
  Ncol = numDemands*cumNumNeighbs[T->node.size()];

  /* find the subset of demands routed on the links in L
   * (moving the other demands is just overhead)
   * subdemands is a map containing ALL the demands, each of which
   * is associated with TRUE if it is routed on the links in L
   */
  map<pair<int,int>,bool> subdemands;
  
  for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++)
    subdemands[d->first] = false;
  for (set<TLinkItrs>::iterator l=L->begin(); l!=L->end(); l++)
    for (map<pair<int,int>,double>::iterator f=l->second->f_e2e.begin(); f!=l->second->f_e2e.end(); f++) {
      subdemands[f->first] = true;
      cout << "demand [" << f->first.first << f->first.second << "] true" << endl;
    }

  unsigned int numSubDemands = count_if(subdemands.begin(), subdemands.end(), __gnu_cxx::compose1(bind2nd(equal_to<bool>(),true), __gnu_cxx::select2nd<map<pair<int,int>, bool>::value_type>() ) );

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model.resize(0, Ncol);

  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD  M
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
//       int i=1;
      int i=0;
      for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++, i++) {
	stringstream s;
	s << "f(" << u->info->id << "," << v->info->id << ")[" << d->first.first << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+i, str);
      }
      linkId++;
    }

  int* colno = new int[Ncol];
  double* coeff = new double[Ncol];
  
  /* set the objective function: min sum_l\in L sum_ab\in subDemands f^ab(l) */
  int k = 0;
  for (set<TLinkItrs>::iterator l=L->begin(); l!=L->end(); l++) {
//     int i=1;
    int i=0;
    for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++, i++)
      if (subdemands[d->first]) {
	coeff[k] = 1;
	colno[k++] = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(), l->second))*numDemands + i;
      }
  }

//   set_obj_fnex(lp, k, coeff, colno);

  for (int i=0; i<k; i++)
    model.setObjectiveCoefficient(colno[i], coeff[i]);

  /* set the object direction to minimize */
//   set_minim(lp);

  model.setOptimizationDirection(1);

//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  CoinBuild buildObject;

  vector<vector<int> > colno_v(T->node.size()*numSubDemands);
  vector<vector<double> > coeff_v(T->node.size()*numSubDemands);
  
  /* flow conservation constraints  ( |V| * numSubDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
//       int d=1, sd=0;
      int d=0, sd=0;
      for (map<pair<int,int>,double>::iterator di=demands->begin(); di!=demands->end(); di++, d++) 
	/* if the demand is routed on the current link u->v */
	if (subdemands[di->first]) {
	  dist = distance(u->neighbor.begin(), v);
	  /* flow leaving node u */
	  coeff_v[u->info->id*numSubDemands+sd].push_back(1);
	  colno_v[u->info->id*numSubDemands+sd].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d);
	  /* flow entering node v */
	  coeff_v[v->info->id*numSubDemands+sd].push_back(-1);
	  colno_v[v->info->id*numSubDemands+sd].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d);
	  /* if v is the source of the demand or u is the destination of the demand, f(u->v)[d] must be null */
	  if ( (u->info->id == di->first.second) || (v->info->id == di->first.first) )
// 	    set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d, 0.);
	    model.setColumnUpper((cumNumNeighbs[u->info->id] + dist)*numDemands + d, 0.);
//	  else
//	    set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d, di->second);
	  sd++;
      }
      else {
	/* the corresponding variable must be set to the current value */
	dist = distance(u->neighbor.begin(), v);
// 	set_bounds(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d, v->f_e2e[di->first], v->f_e2e[di->first]);
	model.setColumnBounds((cumNumNeighbs[u->info->id] + dist)*numDemands + d, v->f_e2e[di->first], v->f_e2e[di->first]);
	// DEBUG cout << "set_bounds " << cumNumNeighbs[u->info->id] << " " << dist << " " << numDemands << " " << d << endl;
      }
    }

  /* Add the flow conservation constraints one-by-one */
  k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++)
      if (subdemands[d->first]) {
	copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
	copy(colno_v[k].begin(), colno_v[k].end(), colno);

	/* if i is the source of the demand d */
	if (i == d->first.first)
// 	  add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	  buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
	/* if i is the destination of the demand d */
	else if (i == d->first.second)
// 	  add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	  buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
	else
// 	  add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
	  buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
	k++;
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because of the
   * constraints on the total utilizations
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

  /* constraints on the total utilizations ( |E| constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      k = 0;
      for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	  if ( TLinkItrs(x,y).IsInTheCollDomainOf(T,TLinkItrs(u,v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i;
	    k += numDemands;
	  }
//       add_constraintex(lp, k, coeff, colno, LE, max(v->Utot, u0) );
      buildObject.addRow(k, colno, coeff, 0., max(v->Utot, u0));
    }
    
//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model.addRows(buildObject);

  // write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /* Now let lpsolve calculate a solution */
//   int ret = solve(lp);
  int ret = model.initialSolve();

//   if (ret != OPTIMAL)
  if (ret)
    cerr << endl << "WARNING: TurnOffRadio: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
//   else
//     print_solution(lp,4);

  /* If an optimal solution has been found that zeroes the flow
   * on all the links of the extracted radio,
   * update the flows, the e2e flows and the Utot */
//   if ((ret == OPTIMAL) && (get_objective(lp) < almost_null)) {
  if ((!ret) && (model.objectiveValue() < almost_null)) {

    cout << endl << "TurnOffRadio: the LP was able to turn off the radio!" << endl;
    success = true;
      
    double *ptr_var, *ptr_constr;
    
//     get_ptr_variables(lp, &ptr_var);
//     get_ptr_constraints(lp, &ptr_constr);

    ptr_var = model.primalColumnSolution();
    ptr_constr = model.primalRowSolution();

    unsigned int linkId = 0;
  
    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      {
	v->flow = 0.;
	v->f_e2e.clear();

	int i=0;
	for (map<pair<int,int>,double>::iterator d=demands->begin(); d!=demands->end(); d++, i++)
	  if (ptr_var[linkId*numDemands + i])
	    v->flow += (v->f_e2e[pair<int,int>(d->first.first,d->first.second)] = ptr_var[linkId*numDemands + i]);

	v->Utot = ptr_constr[T->node.size()*numSubDemands+linkId];
	  
	linkId++;
      }
  }
  
  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;

  /* clean up such that all used memory by lpsolve is freed */
//   if (lp)
//     delete_lp(lp);

  return success;
}



/*
 * Rearrange e2e flows of given demands to minimize the maximum Utot
 * while maximizing the number of radios turned off
 * based on a Mixed Integer Linear Program
 * Changes are applied only if the model is valid (M0 != 0)
 */
MRMU_MILP::MRMU_MILP(Topology* _T, long int _timeout, bool _apply): MILP(_T, _timeout, _apply)
{
//   lprec *lp;
  ClpSimplex* model = new ClpSimplex;
  double F = 0.;      // sum of all the demands
  int Nradios = T->countNumRadios();
//   pair<bool, unsigned int> ch;
  RadioIfaceIterator ri;
  double alpha = 20.;    // upper bound on M0

  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands;
  
  if (!(T->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: mrmu: Checking E2E flows before beginning FAILED" << endl << endl;
  
  // compute the number of demands and the sum of all their flows
  unsigned int numDemands = demands.size();
  
  for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++)
    F += d->second;

  cout << "La somma del flusso di tutte le demand e' " << F << endl;
      
  /* compute Lr for all the radios */
  for (unsigned int i=0; i<T->node.size(); i++)
    for (list<TGraphNeighbor>::iterator v=T->node[i].neighbor.begin(); v!=T->node[i].neighbor.end(); v++)
    {
//       ch = SearchChannel(v->info, v->radio->channel);
      ri = v->info->findRadioWithChannel(v->radio->channel);
      /* insert the link in the sets of the two radios at the endpoints of the link */
      Lr[TRadio(i,distance(T->node[i].info->radio.begin(),RadioIfaceIterator(v->radio)))].insert(TLinkItrs(T->node.begin()+i,v));
//       if (ch.first)
      if (ri != v->info->radio.end())
	Lr[TRadio(v->info->id, distance(v->info->radio.begin(),ri))].insert(TLinkItrs(T->node.begin()+i,v));
    }

  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  cumNumNeighbs.resize(T->node.size() + 1);

  cumNumNeighbs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();
  
  // Likewise, for the number of radios
  cumNumRadios.resize(T->node.size() + 1);

  cumNumRadios[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumRadios[u->info->id + 1] = cumNumRadios[u->info->id] + u->info->numAvailRadios();

  
  // the number of variables is numDemands*numLinks (g^ab(u->v)) + numRadios*2 (yi e zi) + 1 (M0)
  Ncol = numDemands*cumNumNeighbs[T->node.size()] + 2*Nradios + 1;

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model->resize(0, Ncol);
  
  /* Columns (variables) are ordered this way:
      g_e1,d1 ... g_e1,dD ... g_eE,d1 ... g_eE,dD  y1 ... yN  z1 ... zN  M0
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int i=1;
      for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, i++) {
	stringstream s;
	s << "g(" << u->info->id << "," << v->info->id << ")[" << d->first.first << "-" << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model->setColumnName(linkId*numDemands+i-1, str);
      }
      linkId++;
    }
  for (unsigned int i=0; i<T->node.size(); i++)
    for (unsigned int r=0; r<T->node[i].info->numAvailRadios(); r++) {
      stringstream s1, s2;
      s1 << "y[" << i << "." << r << "]";
//       set_col_name(lp, numDemands*cumNumNeighbs[T->node.size()]+cumNumRadios[i]+r+1, (char*)s1.str().c_str());
      string str1 = s1.str();
      model->setColumnName(numDemands*cumNumNeighbs[T->node.size()]+cumNumRadios[i]+r, str1);
      s2 << "z[" << i << "." << r << "]";
//       set_col_name(lp, numDemands*cumNumNeighbs[T->node.size()]+Nradios+cumNumRadios[i]+r+1, (char*)s2.str().c_str());
      string str2 = s2.str();
      model->setColumnName(numDemands*cumNumNeighbs[T->node.size()]+Nradios+cumNumRadios[i]+r, str2);
    }
//   set_col_name(lp, Ncol, "M0");
  {
    string str("MO");
    model->setColumnName(Ncol-1, str);
  }

  
  for (int i=numDemands*cumNumNeighbs[T->node.size()] + Nradios + 1;  i<Ncol; i++)
//     set_binary(lp, i, TRUE);  // sets variable zi to binary
  {
    model->setInteger(i);  // sets variable zi to binary
    model->setColumnBounds(i, 0., 1.);
  }

  int* colno = new int[Ncol];
  double* coeff = new double[Ncol];
  
  /* set the objective function: max M0 - sum_i 1/N yi */
  fill(coeff, coeff+Nradios, -1./Nradios);
//   generate(colno, colno+Nradios, counter<int>(numDemands*cumNumNeighbs[T->node.size()]));
  generate(colno, colno+Nradios, counter<int>(numDemands*cumNumNeighbs[T->node.size()]-1));
  
  coeff[Nradios] = 1;
//   colno[Nradios] = Ncol;
  colno[Nradios] = Ncol-1;

//   set_obj_fnex(lp, Nradios+1, coeff, colno);
  for (int i=0; i<=Nradios; i++)
    model->setObjectiveCoefficient(colno[i], coeff[i]);
  
  /* set the object direction to maximize */
//   set_maxim(lp);
  model->setOptimizationDirection(-1);
  
//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  vector<vector<int> > colno_v(T->node.size()*numDemands);
  vector<vector<double> > coeff_v(T->node.size()*numDemands);

  CoinBuild buildObject;

  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands.begin(); di!=demands.end(); di++, d++) {
      // for (unsigned int d=0; d<numDemands; d++) {
	dist = distance(u->neighbor.begin(), v);
	/* flow leaving node u */
	coeff_v[u->info->id*numDemands+d].push_back(1);
// 	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d);
	/* flow entering node v */
	coeff_v[v->info->id*numDemands+d].push_back(-1);
// 	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d);
	/* if v is the source of the demand or u is the destination of the demand, f(u->v)[d] must be null */
	if ( (u->info->id == di->first.second) || (v->info->id == di->first.first) )
// 	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, 0.);
	  model->setColumnUpper((cumNumNeighbs[u->info->id] + dist)*numDemands + d, 0.);
      }
    }

  /* Add the flow conservation constraints one-by-one */
  int k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, k++) {

      /* if i is the source of the demand d */
      if (i == d->first.first) {
	coeff_v[k].push_back(-d->second);
// 	colno_v[k].push_back(Ncol);
	colno_v[k].push_back(Ncol-1);
      }
      /* if i is the destination of the demand d */
      else if (i == d->first.second) {
	coeff_v[k].push_back(d->second);
// 	colno_v[k].push_back(Ncol);
	colno_v[k].push_back(Ncol-1);
      }
      
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

//       add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
      buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because of the
   * constraints on the total utilizations
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

  /* constraints on the total utilizations ( |E| constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      k = 0;
      for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	  if ( TLinkItrs(x,y).IsInTheCollDomainOf(T,TLinkItrs(u,v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i;
	    k += numDemands;
	  }
//       add_constraintex(lp, k, coeff, colno, LE, 1. );
      buildObject.addRow(k, colno, coeff, -COIN_DBL_MAX, 1.);
    }
    
  /* constraints to force a null flow on the radios turned off (Nradios constraints in total) */
  for (unsigned int i=0; i<T->node.size(); i++)
    for (unsigned int r=0; r<T->node[i].info->numAvailRadios(); r++)
      /* if this radio is used by some links */
      if (Lr.find(TRadio(i,r)) != Lr.end()) {
	k = 0;     // counts the variables
	/* for each link, add the variables for all the demands */
	for (set<TLinkItrs>::iterator l=Lr[TRadio(i,r)].begin(); l!=Lr[TRadio(i,r)].end(); l++, k+=numDemands) {
	  fill(coeff+k, coeff+k+numDemands, 1);
// 	  generate(colno+k, colno+k+numDemands, 
// 		   counter<int>((cumNumNeighbs[l->first->info->id]+distance(l->first->neighbor.begin(),l->second))*numDemands) );
	  generate(colno+k, colno+k+numDemands, 
		   counter<int>((cumNumNeighbs[l->first->info->id]+distance(l->first->neighbor.begin(),l->second))*numDemands-1) );
	}
	coeff[k] = -F * Lr[TRadio(i,r)].size();
// 	colno[k++] = numDemands*cumNumNeighbs[T->node.size()]+cumNumRadios[i]+r+1;    // yi
	colno[k++] = numDemands*cumNumNeighbs[T->node.size()]+cumNumRadios[i]+r;      // yi
// 	add_constraintex(lp, k, coeff, colno, LE, 0.);
	buildObject.addRow(k, colno, coeff, -COIN_DBL_MAX, 0.);
      }
      else {
	/* if no link uses this radio, it can be considered off (yi = 0, zi = 0) */
// 	set_bounds(lp, numDemands*cumNumNeighbs[T->node.size()]+cumNumRadios[i]+r+1, 0., 0. );            // yi
	model->setColumnBounds(numDemands*cumNumNeighbs[T->node.size()]+cumNumRadios[i]+r, 0., 0.);         // yi
// 	set_bounds(lp, numDemands*cumNumNeighbs[T->node.size()]+Nradios+cumNumRadios[i]+r+1, 0., 0. );    // zi
	model->setColumnBounds(numDemands*cumNumNeighbs[T->node.size()]+Nradios+cumNumRadios[i]+r, 0., 0.);       // zi
      }

  /* constraints to make yi variables binary (3*Nradios constraints in total) */
//   for (int r=1; r<=Nradios; r++) {
  for (int r=0; r<Nradios; r++) {
    /* yi - alpha*zi <= 0 */
    coeff[0] = 1;
    colno[0] = numDemands*cumNumNeighbs[T->node.size()] + r;  // yi
    
    coeff[1] = -alpha;
    colno[1] = numDemands*cumNumNeighbs[T->node.size()] + Nradios + r;  // zi
    
//     add_constraintex(lp, 2, coeff, colno, LE, 0.);
    buildObject.addRow(2, colno, coeff, -COIN_DBL_MAX, 0.);
    
    /* yi - M0 - alpha*zi >= -alpha */
    coeff[0] = 1;
    colno[0] = numDemands*cumNumNeighbs[T->node.size()] + r;  // yi
    
    coeff[1] = -1;
    colno[1] = Ncol;   // M0
    
    coeff[2] = -alpha;
    colno[2] = numDemands*cumNumNeighbs[T->node.size()] + Nradios + r;  // zi
    
//     add_constraintex(lp, 3, coeff, colno, GE, -alpha);
    buildObject.addRow(3, colno, coeff, -alpha, COIN_DBL_MAX);
    
    /* yi - M0 + alpha*zi <= alpha */
    coeff[0] = 1;
    colno[0] = numDemands*cumNumNeighbs[T->node.size()] + r;  // yi
    
    coeff[1] = -1;
    colno[1] = Ncol;   // M0
    
    coeff[2] = alpha;
    colno[2] = numDemands*cumNumNeighbs[T->node.size()] + Nradios + r;  // zi
    
//     add_constraintex(lp, 3, coeff, colno, LE, alpha);
    buildObject.addRow(3, colno, coeff, -COIN_DBL_MAX, alpha);
  }
  

//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model->addRows(buildObject);

  // the model is used to build the solver (which will own the pointer)
  solver = new OsiClpSolverInterface(model, true);

//   write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;
}

void MRMU_MILP::updateFlows(const double* ptr_var, const double* ptr_constr)
{
//   get_ptr_variables(lp, &ptr_var);
//   get_ptr_constraints(lp, &ptr_constr);
  
  /* If M0 == 0, the solution is invalid */
//   if (ptr_var[Ncol-1]) {
    RadioIfaceIterator ri;
    unsigned int linkId = 0;
    int Nradios = T->countNumRadios();

    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++, linkId++)
      {
	v->flow = 0.;
	v->f_e2e.clear();

	int i=0;
	for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, i++)
	  if (ptr_var[linkId*numDemands + i])
	    v->flow += (v->f_e2e[pair<int,int>(d->first.first,d->first.second)] = ptr_var[linkId*numDemands + i] / ptr_var[Ncol-1]);

	v->Utot = ptr_constr[T->node.size()*numDemands+linkId] / ptr_var[Ncol-1];
      }
    
    /* turn off the radios with zi=0 */
    for (unsigned int i=0; i<T->node.size(); i++)
      for (unsigned int r=0; r<T->node[i].info->numAvailRadios(); r++)
	if (!ptr_var[numDemands*cumNumNeighbs[T->node.size()]+Nradios+cumNumRadios[i]+r]) {
	  
	  if (Lr.find(TRadio(i,r)) != Lr.end())
	    /* for all the links */
	    for (set<TLinkItrs>::iterator l=Lr[TRadio(i,r)].begin(); l!=Lr[TRadio(i,r)].end(); l++) {
	      if (l->second->flow)
		cerr << "WARNING: mrmu mi sta facendo eliminare un link con flusso non nullo!" << endl;
	      /* first, remove the link from the Lr set of the other radio (not the extracted one) */
// 	      ch = SearchChannel(l->second->info, l->second->radio->channel);
	      ri = l->second->info->findRadioWithChannel(l->second->radio->channel);
	      if (i == l->second->info->id)
		Lr[TRadio(l->first->info->id, distance(l->first->info->radio.begin(),RadioIfaceIterator(l->second->radio)))].erase(*l);
	      else if (ri != l->second->info->radio.end())
		Lr[TRadio(l->second->info->id, distance(l->second->info->radio.begin(),ri))].erase(*l);
	      /* then, remove the link from the topology*/
	      T->node[l->first->info->id].neighbor.erase(l->second);
	    }
	  /* switch off the radio (set channel to 0) */
	  T->node[i].info->radio[r].channel = 0;
	}
//   }
//   else
//     cerr << endl << "WARNING: the solution of the MILP is invalid (M0 = 0)" << endl << endl;

  /* clean up such that all used memory by lpsolve is freed */
//   if (lp)
//     delete_lp(lp);
  
  if (!(T->checkE2Eflows(NULL)))
    cerr << endl << "WARNING: mrmu: Checking E2E flows FAILED" << endl << endl;
}



/** MinMaxUtot_MILP methods **/
/* NOTE This program does not handle radios with different power consumption profiles.
   Hence, the power consumption profile of the first radio on a node is considered for
   all the radios of that node */

MinMaxUtot_MILP::MinMaxUtot_MILP(Topology* _T, vector<float> *nodePowerConstr, long _timeout, bool _apply)
  : MILP(_T, _timeout, _apply)
{
  ClpSimplex* model = new ClpSimplex;
  int Ncol;
  double F = 0.;      // sum of all the demands

  T->ResetChannels();

  /* check that e2e flows are consistent and obtain the set of demands */
  if (!(T->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: MinMaxUtot_MILP: Checking E2E flows before beginning FAILED" << endl << endl;

  // compute the number of demands and the sum of all their flows
  numDemands = demands.size();
  
  for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++)
    F += d->second;

  cout << "La somma del flusso di tutte le demand e' " << F << endl;
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->node.size() + 1];

  cumNumNeighbs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();

  // the number of variables is numDemands*numLinks*numChannels (f^ab(u->v)[c]) + numNodes*numChannels (x^c(u)) + 1 (M)
  Ncol = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + 1;
  
  // if power constraints are supplied, we consider additional numNodes variables (y_u)
  if (nodePowerConstr)
    Ncol += T->node.size();

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model->resize(0, Ncol);
  
  /* Columns (variables) are ordered this way:
      f_e1,d1,c1 ... f_e1,d1,cC ... f_e1,dD,c1 ... f_e1,dD,cC ... f_eE,dD,c1 ... f_eE,dD,cC  x^c1(n1) ... x^cC(n1) ... x^c1(nN) ... x^cC(nN) [ y(n1) ... y(nN) ]  M
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int i=0;
      for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, i++)
	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	  stringstream s;
	  s << "f(" << u->info->id << "," << v->info->id << ")[" << c << "]{" << d->first.first << d->first.second << "}";
// 	  set_col_name(lp, (linkId*numDemands+i)*T->numAvailChannels+c, (char*)s.str().c_str());
	  string str = s.str();
	  model->setColumnName((linkId*numDemands+i)*T->numAvailChannels+c-1, str);
	}
      linkId++;
    }
  
  for (unsigned int i=0; i<T->node.size(); i++)
    for (unsigned int c=1; c<=T->numAvailChannels; c++) {
      stringstream s;
      s << "x_" << i << "[" << c << "]";
      int j = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels + c;
//       set_col_name(lp, j, (char*)s.str().c_str());
// 	set_binary(lp, j, TRUE);
      string str = s.str();
      model->setColumnName(j-1, str);
      model->setInteger(j-1);
      model->setColumnBounds(j-1, 0., 1.);
    }

  if (nodePowerConstr)
    for (unsigned int i=0; i<T->node.size(); i++)
    {
      stringstream s;
      s << "y_" << i;
//       int j = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + i + 1;
      int j = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + i;
//       set_col_name(lp, j, (char*)s.str().c_str());
// 	set_binary(lp, j, TRUE);
      string str = s.str();
      model->setColumnName(j, str);
      model->setInteger(j);
      model->setColumnBounds(j, 0., 1.);
    }

//   set_col_name(lp, Ncol, "M");
  {
    string str("M");
    model->setColumnName(Ncol-1, str);
  }

  int* colno = new int[Ncol];
  double *coeff = new double[Ncol], *coeff2 = new double[Ncol];
  
  /* set the objective function: min M*/
//   coeff[0] = 1;
//   colno[0] = Ncol;
//   set_obj_fnex(lp, 1, coeff, colno);

  model->setObjectiveCoefficient(Ncol-1, 1);

  /* set the object direction to minimize */
//   set_minim(lp);
  model->setOptimizationDirection(1);

//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  vector<vector<int> > colno_v(T->node.size()*numDemands);
  vector<vector<double> > coeff_v(T->node.size()*numDemands);

  CoinBuild buildObject;

  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands.begin(); di!=demands.end(); di++, d++) 
// 	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	for (unsigned int c=0; c<T->numAvailChannels; c++) {
	// for (unsigned int d=0; d<numDemands; d++) {
	  dist = distance(u->neighbor.begin(), v);
	  /* flow leaving node u */
	  coeff_v[u->info->id*numDemands+d].push_back(1);
	  colno_v[u->info->id*numDemands+d].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  /* flow entering node v */
	  coeff_v[v->info->id*numDemands+d].push_back(-1);
	  colno_v[v->info->id*numDemands+d].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  /* if v is the source of the demand or u is the destination of the demand, f^d(u->v)[c] must be null */
	  if ( (u->info->id == di->first.second) || (v->info->id == di->first.first) )
// 	    set_upbo(lp, ((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, 0.);
	    model->setColumnUpper(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, 0.);
//	  else
//	    set_upbo(lp, ((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, di->second);
        }
    }

  /* Add the flow conservation constraints one-by-one */
  int k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      /* if i is the source of the demand d */
      if (i == d->first.first)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
      /* if i is the destination of the demand d */
      else if (i == d->first.second)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
      else
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
      k++;
    }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because the goal of minimizing M
   * will force to minimize the e2e flows
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

  /* constraints on the total utilizations ( |E|*numChannels constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) 
//       for (unsigned int c=1; c<=T->numAvailChannels; c++) {
      for (unsigned int c=0; c<T->numAvailChannels; c++) {
	k = 0;
	for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	  for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	    if ( T->PotInterfLinks(TLinkItrs(x,y),TLinkItrs(u,v)) ) {
	      fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	      for (unsigned int i=0; i<numDemands; i++)
		colno[k+i] = ((cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i)*T->numAvailChannels + c;
	      k += numDemands;
	    }
	coeff[k] = -1;
// 	colno[k++] = Ncol;         // -M
	colno[k++] = Ncol-1;       // -M
// 	add_constraintex(lp, k, coeff, colno, LE, 0 );
	buildObject.addRow(k, colno, coeff, -COIN_DBL_MAX, 0.);
      }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  /* constraints on the number of radios per node (|V| constraints in total)*/
  for (unsigned int i=0; i<T->node.size(); i++) {
    fill(coeff, coeff+T->numAvailChannels, 1);
//     generate(colno, colno+T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels));
    generate(colno, colno+T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels - 1));
    
//     add_constraintex(lp,T->numAvailChannels, coeff, colno, LE, T->node[i].info->numAvailRadios());
    buildObject.addRow(T->numAvailChannels, colno, coeff, 0., T->node[i].info->numAvailRadios());
  }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  /* constraints to make flow on u->v[c] null if u and v do not use channel c (2*|E|*numChannels constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) 
//       for (unsigned int c=1; c<=T->numAvailChannels; c++) {
      for (unsigned int c=0; c<T->numAvailChannels; c++) {
	fill(coeff, coeff+numDemands, 1);
	coeff[numDemands] = -F;
	
	for (unsigned int i=0; i<numDemands; i++)
	  colno[i] = ((cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i)*T->numAvailChannels + c;
	colno[numDemands] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u->info->id*T->numAvailChannels + c;

// 	add_constraintex(lp, numDemands+1, coeff, colno, LE, 0.);
	buildObject.addRow(numDemands+1, colno, coeff, -COIN_DBL_MAX, 0.);

	fill(coeff, coeff+numDemands, 1);
	coeff[numDemands] = -F;
	
	for (unsigned int i=0; i<numDemands; i++)
	  colno[i] = ((cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i)*T->numAvailChannels + c;
	colno[numDemands] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + v->info->id*T->numAvailChannels + c;

// 	add_constraintex(lp, numDemands+1, coeff, colno, LE, 0.);
	buildObject.addRow(numDemands+1, colno, coeff, -COIN_DBL_MAX, 0.);
      }

//   constrGroupSize.push_back(get_Nrows(lp));
  constrGroupSize.push_back(buildObject.numberRows());

  if (nodePowerConstr)
  {
    for (unsigned int u=0; u<T->node.size(); u++)
    {
      /* constraints y_u >= x_u^c, for each channel c */
//       for (unsigned int c=1; c<=T->numAvailChannels; c++)
      for (unsigned int c=0; c<T->numAvailChannels; c++)
      {
	coeff[0] = 1.;
	colno[0] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u*T->numAvailChannels + c;     // x_u^c

	coeff[1] = -1.;
// 	colno[1] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u + 1;   // y_u
	colno[1] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u;   // y_u

// 	add_constraintex(lp, 2, coeff, colno, LE, 0.);
	buildObject.addRow(2, colno, coeff, -COIN_DBL_MAX, 0.);
      }
      
      /* constraint y_u <= sum_c x_u^c */
      fill_n(coeff, T->numAvailChannels, -1.);
//       generate_n(colno, T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u*T->numAvailChannels));
      generate_n(colno, T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u*T->numAvailChannels - 1));
      
      coeff[T->numAvailChannels] = 1.;
//       colno[T->numAvailChannels] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u + 1;   // y_u
      colno[T->numAvailChannels] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u;   // y_u

//       add_constraintex(lp, T->numAvailChannels+1, coeff, colno, LE, 0.);
      buildObject.addRow(T->numAvailChannels+1, colno, coeff, -COIN_DBL_MAX, 0.);
    }

//     constrGroupSize.push_back(get_Nrows(lp));
    constrGroupSize.push_back(buildObject.numberRows());

    map<int, set<TLinkItrs> > Linc;  // maps each node to the set of incoming links

    /* compute Linc for all the nodes */
    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
	Linc[v->info->id].insert(TLinkItrs(u,v));

    int pos_i;
    /* constraints tau_TX^(u,c) + tau_RX^(u,c) <= 1 (for each channel) and node power constraints */
    for (unsigned int u=0; u<T->node.size(); u++)
    {
      k=0;
      RadioIfaceIterator ri = T->node[u].info->radio.begin();

      /* add the constraint tau_TX^(u,c) + tau_RX^(u,c) <= 1 for each active radio */
//       for (unsigned int c=1; c<=T->numAvailChannels; c++)
      for (unsigned int c=0; c<T->numAvailChannels; c++)
      {
	int j1=k;

	for (list<TGraphNeighbor>::iterator v=T->node[u].neighbor.begin(); v!=T->node[u].neighbor.end(); v++)
	{
	  fill_n(coeff+k, numDemands, 1. / v->rate);
	  // coeff2 is used for the power consumption constraint
	  fill_n(coeff2+k, numDemands, ((*ri->rho_tx)(DbmToW(TxPowerLevel(v->txpower))) - ri->rho_idle) / v->rate);
	  pos_i = (cumNumNeighbs[u] + distance(T->node[u].neighbor.begin(),v))*numDemands*T->numAvailChannels + c;
	  generate_n(colno+k, numDemands, counter<int>(pos_i-T->numAvailChannels, T->numAvailChannels));
	  k += numDemands;
	}

	int j2=k;

	if (Linc.find(u) != Linc.end())
	  for (set<TLinkItrs>::iterator l=Linc[u].begin(); l!=Linc[u].end(); l++)
	  {
	    fill_n(coeff+k, numDemands, 1. / l->second->rate);
	    fill_n(coeff2+k, numDemands, (ri->rho_rx - ri->rho_idle) / l->second->rate);
	    pos_i = (cumNumNeighbs[l->first->info->id] + distance(l->first->neighbor.begin(),l->second))*numDemands*T->numAvailChannels + c;
	    generate_n(colno+k, numDemands, counter<int>(pos_i-T->numAvailChannels, T->numAvailChannels));
	    k += numDemands;
	  }

	if (k != j1)
	{
	  /* make a copy of coeff and colno because add_constraintex sorts the element in the passed vectors */
	  double coeff_c[k-j1];
	  int colno_c[k-j1];

	  copy (coeff+j1, coeff+k, coeff_c);
	  copy (colno+j1, colno+k, colno_c);

// 	  add_constraintex(lp, k-j1, coeff_c, colno_c, LE, 1);
	  buildObject.addRow(k-j1, colno_c, coeff_c, 0., 1.);
	}

	/* multiply coefficients by the power consumption (needed for node power constraint) */
// 	for (int j=j1; j<j2; j++)
// 	  coeff[j] *= (T->pm.rho_tx-T->pm.rho_idle);
// 
// 	for (int j=j2; j<k; j++)
// 	  coeff[j] *= (T->pm.rho_rx-T->pm.rho_idle);
      }

//       float tmp = T->pm.rho_idle-T->pm.rho_sleep;
      float tmp = ri->rho_idle - ri->rho_sleep;

      fill_n(coeff2+k, T->numAvailChannels, tmp);      // x_u^c * (rho_idle - rho_sleep)
      generate_n(colno+k, T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u*T->numAvailChannels - 1));
      k += T->numAvailChannels;

//       tmp = T->pm.BasePowerConsumption + T->node[u].info->numAvailRadios() * T->pm.rho_sleep - T->pm.StandbyPowerConsumption;
      tmp = T->node[u].info->BasePowerConsumption + T->node[u].info->numAvailRadios() * ri->rho_sleep - T->node[u].info->StandbyPowerConsumption;

      coeff2[k] = tmp;
      colno[k++] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels + u;   // y_u

//       add_constraintex(lp, k, coeff, colno, LE, nodePowerConstr->at(u) - T->pm.StandbyPowerConsumption);
//       buildObject.addRow(k, colno, coeff, 0., nodePowerConstr->at(u) - T->pm.StandbyPowerConsumption);
      buildObject.addRow(k, colno, coeff2, 0., nodePowerConstr->at(u) - T->node[u].info->StandbyPowerConsumption);
    }

//     constrGroupSize.push_back(get_Nrows(lp));
    constrGroupSize.push_back(buildObject.numberRows());
  }

//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model->addRows(buildObject);

  // the model is used to build the solver (which will own the pointer)
  solver = new OsiClpSolverInterface(model, true);

  // write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;
}


void MinMaxUtot_MILP::updateFlows (const double* ptr_var, const double* ptr_constr)
{
  bool updateUtot = false;

  // TODO Utots could be updated it the utot constraints were the first in the program
  // if the Utot constraints have not been relaxed, use their values to set the Utot fields
//   if (ptr_constr && (relaxedConstrGroups.find(1) == relaxedConstrGroups.end()))
//   {
//     updateUtot = true;
//     cout << endl << "*** updateUtot true ***" << endl;
//   }
  
  unsigned int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++) {
    int N = u->neighbor.size();
    list<TGraphNeighbor>::iterator v=u->neighbor.begin();
    for (int n=0; n<N; n++) {
      bool parallel = false;
      for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	double _f=0;
	map<pair<int,int>,double> _f_e2e;
	int i=0;
	
	for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, i++)
	  if (ptr_var[linkId*numDemands*T->numAvailChannels+i*T->numAvailChannels +c-1] > almost_null)
	    _f += (_f_e2e[pair<int,int>(d->first.first,d->first.second)] = ptr_var[linkId*numDemands*T->numAvailChannels+i*T->numAvailChannels +c-1]);

	if (_f) {
	  if (!parallel) {
	    v->flow = _f;
	    v->f_e2e = _f_e2e;
// 	    if (updateUtot)
// 	      v->Utot = ptr_constr[T->node.size()*numDemands+linkId*T->numAvailChannels+c-1] + ptr_var[get_Ncolumns(lp)-1];
	    
	    T->AssignChannel2Link(TLinkItrs(u,v),c);
	    
	    parallel = true;
	  }
	  else {
	    list<TGraphNeighbor>::iterator z = u->neighbor.insert(u->neighbor.end(), TGraphNeighbor(v->info,v->rate,v->txpower));
	    z->flow = _f;
	    z->f_e2e = _f_e2e;
// 	    if (updateUtot)
// 	      z->Utot = ptr_constr[T->node.size()*numDemands+linkId*T->numAvailChannels+c-1] + ptr_var[get_Ncolumns(lp)-1];
	    
	    T->AssignChannel2Link(TLinkItrs(u,z),c);
	  }
	}
      }
      /* if the flow is null on the link (hence no channel has been assigned to it), remove it */
      if (!parallel) {
	list<TGraphNeighbor>::iterator z = v++;
	u->neighbor.erase(z);
      }
      else
	v++;
      linkId++;
    }
  }
cout << endl << "*** uscendo da updateFlows ***" << endl;
}


/*
 * Assign channels and rearrange e2e flows of given demands to minimize the number
 * of radios turned on given a constraint on the maximum total utilization
 * Based on a Mixed Integer Linear Program
 * ResetChannels is called to avoid duplicate neighbors in TGraphNeighbor lists and
 * reset the channel of every radio to zero (they are set after solving the MILP)
 * Returns true if the model is infeasible, false otherwise
 * The current configuration of T is used as initial basis
 */
MinRadios_MILP::MinRadios_MILP (Topology* _T, double u0, long _timeout, bool _apply)
 : MILP (_T, _timeout, _apply)
{
//   lprec *lp;
  ClpSimplex* model = new ClpSimplex;
  int Ncol;
  double F = 0.;      // sum of all the demands
  bool infeasible = true;

  // clone the topology before resetting channels to save a feasible solution used as initial basis
  Topology* U = new Topology(*T);
  
  T->ResetChannels();
  
  /* check that e2e flows are consistent and obtain the set of demands */
  map<pair<int,int>,double> demands;
  
  if (!(T->checkE2Eflows(&demands)))
    cerr << endl << "WARNING: E2CAR_MILP: Checking E2E flows before beginning FAILED" << endl << endl;

  // compute the number of demands and the sum of all their flows
  unsigned int numDemands = demands.size();
  
  for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++)
    F += d->second;

  cout << "La somma del flusso di tutte le demand e' " << F << endl;
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->node.size() + 1];

  cumNumNeighbs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();

  // the number of variables is numDemands*numLinks*numChannels (f^ab(u->v)[c]) + numNodes*numChannels (x^c(u))
  Ncol = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + T->node.size()*T->numAvailChannels;

  // initial feasible solution
//   REAL guessvector[1+Ncol];
//   for (int i=0; i<=Ncol; i++)
//     guessvector[i] = 0.0;
  
  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;
  model->resize(0, Ncol);

  /* Columns (variables) are ordered this way:
      f_e1,d1,c1 ... f_e1,d1,cC ... f_e1,dD,c1 ... f_e1,dD,cC ... f_eE,dD,c1 ... f_eE,dD,cC  x^c1(n1) ... x^cC(n1) ... x^c1(nN) ... x^cC(nN)  M
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int di=0;
      for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, di++)
	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	  stringstream s;
	  s << "f(" << u->info->id << "," << v->info->id << ")[" << c << "]{" << d->first.first << d->first.second << "}";
// 	  set_col_name(lp, (linkId*numDemands+di)*T->numAvailChannels+c, (char*)s.str().c_str());
	  string str = s.str();
	  model->setColumnName((linkId*numDemands+di)*T->numAvailChannels+c-1, str);
	  
	  // store the value of this variable in the feasible solution
	  // check if a link u->v on channel c exists
// 	  list<TGraphNeighbor>::iterator nb = U->find_neighbor(u->info->id, v->info->id, c);
// 	  // and has some flow of the current demand
// 	  if (nb != U->node[u->info->id].neighbor.end() && nb->f_e2e.find(d->first) != nb->f_e2e.end() )
// 	    guessvector[(linkId*numDemands+di)*T->numAvailChannels+c] = nb->f_e2e[d->first];
	}
      linkId++;
    }
  
  // find the maximum number of radios per node
  int rpn = 0;
  for (vector<TGraphNode>::iterator n=T->node.begin(); n!=T->node.end(); n++)
    if (n->info->numAvailRadios() > rpn)
      rpn = n->info->numAvailRadios();
  
  double lowbo = rpn/(rpn+1.0)+0.01;

  for (int i=0; i<T->node.size(); i++)
    for (unsigned int c=1; c<=T->numAvailChannels; c++) {
      stringstream s;
      s << "x_" << i << "[" << c << "]";
      int j = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels + c;
//       set_col_name(lp, j, (char*)s.str().c_str());
      string str =s.str();
      model->setColumnName(j-1, str);
      model->setInteger(j-1);
      model->setColumnBounds(j-1, 0., 1.);
//       if (!relax)
// 	set_binary(lp, j, TRUE);
//       else
//       {
// 	set_semicont(lp, j, TRUE);
// 	set_lowbo(lp, j, lowbo);
// 	set_upbo(lp, j, 1.);      // sets xi <= 1
//       }

      // store the value of this variable in the feasible solution
//       if (U->node[i].info->countRadiosWithChannel(c))
// 	guessvector[j] = lowbo;
    }

  
  int* colno = new int[Ncol];
  double* coeff = new double[Ncol];
  
  /* set the objective function: min sum x^c(u) */
  fill(coeff, coeff+T->node.size()*T->numAvailChannels, 1);
//   generate(colno, colno+T->node.size()*T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels));
  generate(colno, colno+T->node.size()*T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels-1));
  
//   set_obj_fnex(lp, T->node.size()*T->numAvailChannels, coeff, colno);
  for (int i=0; i<T->node.size()*T->numAvailChannels; i++)
    model->setObjectiveCoefficient(colno[i], coeff[i]);

  /* set the object direction to minimize */
//   set_minim(lp);
  model->setOptimizationDirection(1);

//   set_epslevel(lp, EPS_BAGGY);
//   set_epsel(lp, 1.0e-2);
//   set_epsb(lp, 1.0e-5);
//   set_epsd(lp, 1.0e-4);
//   set_epsint(lp, 1.0e-3);
//   set_break_at_first(lp, TRUE);
//   set_timeout(lp, 1800);
  
//  set_scaling(lp, SCALE_MEAN | SCALE_EQUILIBRATE | SCALE_INTEGERS );

//    set_improve(lp, IMPROVE_DUALFEAS | IMPROVE_BBSIMPLEX);
    //set_basiscrash(lp, CRASH_LEASTDEGENERATE);
//    set_anti_degen(lp, ANTIDEGEN_FIXEDVARS | ANTIDEGEN_RHSPERTURB);
//  set_BFP(lp,"bfp_LUSOL");
  
//   if (relax)
//     set_mip_gap(lp, FALSE, 0.1);
  
   //set_presolve(lp, PRESOLVE_ROWS | PRESOLVE_COLS | PRESOLVE_LINDEP, get_presolveloops(lp));
   //default_basis(lp);
  
//   set_pivoting(lp, PRICER_FIRSTINDEX | PRICE_ADAPTIVE);
//    set_bb_rule(lp, NODE_PSEUDORATIOSELECT | NODE_AUTOORDER | NODE_BREADTHFIRSTMODE);
  
//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  vector<vector<int> > colno_v(T->node.size()*numDemands);
  vector<vector<double> > coeff_v(T->node.size()*numDemands);

  CoinBuild buildObject;

  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;
  set<int> cancel;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands.begin(); di!=demands.end(); di++, d++) 
// 	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	for (unsigned int c=0; c<T->numAvailChannels; c++) {
	// for (unsigned int d=0; d<numDemands; d++) {
	  dist = distance(u->neighbor.begin(), v);
	  /* flow leaving node u */
	  coeff_v[u->info->id*numDemands+d].push_back(1);
	  colno_v[u->info->id*numDemands+d].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  /* flow entering node v */
	  coeff_v[v->info->id*numDemands+d].push_back(-1);
	  colno_v[v->info->id*numDemands+d].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  /* if v is the source of the demand or u is the destination of the demand, f^d(u->v)[c] must be null */
	  if ( (u->info->id == di->first.second) || (v->info->id == di->first.first) )
	    //cancel.insert(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
// 	    set_upbo(lp, ((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, 0.);
	    model->setColumnUpper(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, 0.);
//	  else
//	    set_upbo(lp, ((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c, di->second);
        }
    }

  /* Add the flow conservation constraints one-by-one */
  int k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      /* if i is the source of the demand d */
      if (i == d->first.first)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
      /* if i is the destination of the demand d */
      else if (i == d->first.second)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
      else
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
      k++;
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because the goal of minimizing M
   * will force to minimize the e2e flows
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

  /* constraints on the total utilizations ( |E|*numChannels constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) 
//       for (unsigned int c=1; c<=T->numAvailChannels; c++) {
      for (unsigned int c=0; c<T->numAvailChannels; c++) {
	k = 0;
	for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	  for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	    if ( T->PotInterfLinks(TLinkItrs(x,y),TLinkItrs(u,v)) ) {
	      fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	      for (unsigned int i=0; i<numDemands; i++)
		colno[k+i] = ((cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i)*T->numAvailChannels + c;
	      k += numDemands;
	    }
// 	add_constraintex(lp, k, coeff, colno, LE, u0 );
	buildObject.addRow(k, colno, coeff, -COIN_DBL_MAX, u0);
      }

  /* constraints on the number of radios per node (|V| constraints in total)*/
  for (int i=0; i<T->node.size(); i++) {
    fill(coeff, coeff+T->numAvailChannels, 1);
//     generate(colno, colno+T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels));
    generate(colno, colno+T->numAvailChannels, counter<int>(numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels - 1));

//     add_constraintex(lp,T->numAvailChannels, coeff, colno, LE, T->node[i].info->numAvailRadios());
    buildObject.addRow(T->numAvailChannels, colno, coeff, -COIN_DBL_MAX, T->node[i].info->numAvailRadios());
  }


  colno_v.clear();
  colno_v.resize(T->node.size()*T->numAvailChannels);
  coeff_v.clear();
  coeff_v.resize(T->node.size()*T->numAvailChannels);
  
  /* constraints to make the flow null on all the links on u using channel c if x^c(u) is null  ( |V| * numChannels constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands.begin(); di!=demands.end(); di++, d++) 
	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	  dist = distance(u->neighbor.begin(), v);
	  /* flow leaving node u */
	  coeff_v[u->info->id*T->numAvailChannels+c-1].push_back(1);
// 	  colno_v[u->info->id*T->numAvailChannels+c-1].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  colno_v[u->info->id*T->numAvailChannels+c-1].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c - 1);
	  /* flow entering node v */
	  coeff_v[v->info->id*T->numAvailChannels+c-1].push_back(1);
// 	  colno_v[v->info->id*T->numAvailChannels+c-1].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c);
	  colno_v[v->info->id*T->numAvailChannels+c-1].push_back(((cumNumNeighbs[u->info->id] + dist)*numDemands + d)*T->numAvailChannels + c - 1);
        }
    }

  /* Add the constraints one-by-one */
  k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
//     for (unsigned int c=1; c<=T->numAvailChannels; c++) {
    for (unsigned int c=0; c<T->numAvailChannels; c++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      coeff[coeff_v[k].size()] = -F / lowbo * coeff_v[k].size();
      colno[coeff_v[k].size()] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + i*T->numAvailChannels + c;
      
      if (coeff_v[k].size())
// 	add_constraintex(lp, coeff_v[k].size()+1, coeff, colno, LE, 0.);
	buildObject.addRow(coeff_v[k].size()+1, colno, coeff, -COIN_DBL_MAX, 0.);
      
      k++;
    }
  
  
  
  /* constraints to make flow on u->v[c] null if u and v do not use channel c (2*|E|*numChannels constraints in total) */
//   for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
//     for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) 
//       for (unsigned int c=1; c<=T->numAvailChannels; c++) {
// 	fill(coeff, coeff+numDemands, 1);
// 	coeff[numDemands] = (relax ? -F/lowbo : -F);
// 	
// 	for (unsigned int i=0; i<numDemands; i++)
// 	  colno[i] = ((cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i)*T->numAvailChannels + c;
// 	colno[numDemands] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + u->info->id*T->numAvailChannels + c;
// 	
// 	add_constraintex(lp, numDemands+1, coeff, colno, LE, 0.);
// 	
// 	fill(coeff, coeff+numDemands, 1);
// 	coeff[numDemands] = (relax ? -F/lowbo : -F);
// 	
// 	for (unsigned int i=0; i<numDemands; i++)
// 	  colno[i] = ((cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i)*T->numAvailChannels + c;
// 	colno[numDemands] = numDemands*cumNumNeighbs[T->node.size()]*T->numAvailChannels + v->info->id*T->numAvailChannels + c;
// 	
// 	add_constraintex(lp, numDemands+1, coeff, colno, LE, 0.);
//       }


//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model->addRows(buildObject);

  // the model is used to build the solver (which will own the pointer)
  solver = new OsiClpSolverInterface(model, true);

  // remove the variables that are known to be null
//   for (set<int>::iterator i=cancel.begin(); i!=cancel.end(); i++)
//     if (del_column(lp, *i) == FALSE)
//       cerr << endl << "WARNING: Cannot delete a column" << endl << endl;
  
//   int basis[1+Ncol+get_Nrows(lp)];
//   
//   if (guess_basis(lp, guessvector, basis) == TRUE)
//     cout << endl <<  "A valid base has been determined" << endl << endl;
//   else
//     cout << endl <<  "WARNING: A valid base could have not been determined" << endl << endl;
//   
//   if (set_basis(lp, basis, TRUE) == TRUE)
//     cout << endl <<  "The provided basis has been set" << endl << endl;
//   else
//     cout << endl <<  "WARNING: The provided basis could have not been set" << endl << endl;
  
  //write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;
}

void MinRadios_MILP::updateFlows(const double* ptr_var, const double* ptr_constr)
{
  /* Now let lpsolve calculate a solution */
//   int ret = solve(lp);
// 
//   if (ret != OPTIMAL)
//     cerr << endl << "WARNING: MinRadiosOpt: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
// 
//   print_solution(lp,4);

  /* If the solution is feasible (though suboptimal) apply the changes */
//   if (ret!=INFEASIBLE)  {
//     infeasible = false;
// 
//     REAL *ptr_var, *ptr_constr;
//     
//     get_ptr_variables(lp, &ptr_var);
//     get_ptr_constraints(lp, &ptr_constr);
    
    unsigned int linkId = 0;
    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++) {
      int N = u->neighbor.size();
      list<TGraphNeighbor>::iterator v=u->neighbor.begin();
      for (int n=0; n<N; n++) {
	bool parallel = false;
	for (unsigned int c=1; c<=T->numAvailChannels; c++) {
	  double _f=0;
	  map<pair<int,int>,double> _f_e2e;
	  int i=0;
	  
	  for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++, i++)
	  {
	    stringstream s;
	    s << "f(" << u->info->id << "," << v->info->id << ")[" << c << "]{" << d->first.first << d->first.second << "}";
// 	    int col = get_nameindex(lp,(char*)s.str().c_str(),FALSE);
	    const vector<string>* names = solver->getModelPtr()->columnNames();
	    vector<string>::const_iterator ni = find (names->begin(), names->end(), s.str());
	    // if (ptr_var[linkId*numDemands*T->numAvailChannels+i*T->numAvailChannels +c-1] > almost_null)
// 	    if (col != -1 && ptr_var[col-1] > almost_null)
	    if (ni != names->end()) 
	    {
	      int col = distance(names->begin(), ni);
	      if (ptr_var[col] > almost_null)
	      _f += (_f_e2e[pair<int,int>(d->first.first,d->first.second)] = ptr_var[col]);
	    }
	  }

	  if (_f) {
	    if (!parallel) {
	      v->flow = _f;
	      v->f_e2e = _f_e2e;
	      // TODO v->Utot = ptr_constr[T->node.size()*numDemands+linkId*T->numAvailChannels+c-1] + ptr_var[Ncol-1];
	      
	      T->AssignChannel2Link(TLinkItrs(u,v),c);
	      
	      parallel = true;
	    }
	    else {
	      list<TGraphNeighbor>::iterator z = u->neighbor.insert(u->neighbor.end(), TGraphNeighbor(v->info,v->rate,v->txpower));
	      z->flow = _f;
	      z->f_e2e = _f_e2e;
	      // TODO z->Utot = ptr_constr[T->node.size()*numDemands+linkId*T->numAvailChannels+c-1] + ptr_var[Ncol-1];
	      
	      T->AssignChannel2Link(TLinkItrs(u,z),c);
	    }
	  }
	}
	/* if the flow is null on the link (hence no channel has been assigned to it), remove it */
	if (!parallel) {
	  list<TGraphNeighbor>::iterator z = v++;
	  u->neighbor.erase(z);
	}
	else
	  v++;
	linkId++;
      }
    }
//   }
//   else
//     cerr << endl << "WARNING: the solution of the MILP is infeasible" << endl << endl;  
  
  /*  Before quitting, clear memory*/
//   delete U;
// 
//   /* clean up such that all used memory by lpsolve is freed */
//   if (lp)
//     delete_lp(lp);
// 
//   return infeasible;
}



/* Given the initial graph, builds an LP to solve the 
   maximum (throughput) multi-commodity flow problem
*/
double solveMaxMultiCommFlow(Topology *T)
{
# if defined ERROR
#  undef ERROR
# endif
# define ERROR() { cerr << "Error" << endl; return(1); }
// 	lprec *lp;
	ClpSimplex model;

	// compute the cumulative number of neighbors for each node. It is such that:
	// - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
	// - the last element of cumuNumNeighbs is the total number of links
	unsigned int cumNumNeighbs[T->node.size() + 1];

	cumNumNeighbs[0] = 0;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	{
		cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id];
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			cumNumNeighbs[u->info->id + 1]++;
	}

	// compute the number of commodities (each aggregation router/gateway acts both as a source and a sink)
	unsigned int numComms = T->aggregation_list.size() * T->gateway_list.size() * 2;


	char var_name[100];
	unsigned int num_vars = cumNumNeighbs[T->node.size()] * numComms;
	int colno[numComms][num_vars];
	double coeff[num_vars];
	double temp[num_vars];

	// create the LP problem
	// Number of variables: |E|*|V_A|*|V_G|*2
// 	if ((lp = make_lp(0, num_vars)) == NULL)
// 		ERROR();
	model.resize(0, num_vars);

	/* Columns (variables) are ordered this way:
	   f_e1,c1 ... f_e1,cM ... f_eE,c1 ... f_eE,cM
	   where links e1 ... eE are given by
	   neighbors(node 0) neighbors(node 1) ... neighbors(node N)
	   and commodities c1 ... cM are given by
	   a_1,g_1 ... a_1,g_G ... a_A,g_1 ... a_A,g_G ... g_1,a_1 ... g_1,a_A ... g_G,a_1 ... g_G,a_A
	*/

	// give a name to the variables
	int linkId = 0;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
		{
			for (unsigned int i=1; i <= numComms; i++)
			{
// 				sprintf(var_name,"f%d(%d->%d)",i,u->info->id,v->info->id);
// 				if (!set_col_name(lp, linkId*numComms + i, var_name))
// 					ERROR();
				stringstream s;
				s << "f" << i << "(" << u->info->id << "->" << v->info->id << ")";
				string str = s.str();
				model.setColumnName(linkId*numComms + i - 1, str);
			}
			linkId++;
		}

// 	set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

	CoinBuild buildObject;

	// add |E|/2 capacity constraints: \sum_{i=1}^{k} f_i(u->v) / c(u->v) + \sum_{i=1}^{k} f_i(v->u) / c(v->u) <= 1 for each u<->v
	TLinkItrs invLink;
	unsigned int dist;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->id < v->info->id)          /* each undirected link just once */
			{
				/* link u->v */
				dist = distance(u->neighbor.begin(), v);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					coeff[i-1] = 1. / v->rate;
					coeff[i] = 1. / v->rate;
// 					colno[0][i-1] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
					colno[0][i] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
				}
				/* link v->u */
				invLink = T->FindInverseLink(TLinkItrs(u,v));
				dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					coeff[numComms+i-1] = 1. / invLink.second->rate;
					coeff[numComms+i] = 1. / invLink.second->rate;
// 					colno[0][numComms+i-1] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					colno[0][numComms+i] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
				}
				/* add constraint */
// 				if (!add_constraintex(lp, numComms*2, coeff, colno[0], LE, 1))
// 					ERROR();
				buildObject.addRow(numComms*2, colno[0], coeff, -COIN_DBL_MAX, 1.);
			}

	// add (|V|-|V_A|-|V_G|)*numComms flow conservation constraints
	unsigned int j, k, pos;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		if ( ( find(T->aggregation_list.begin(), T->aggregation_list.end(), u->info->id) == T->aggregation_list.end() ) &&
		     ( find(T->gateway_list.begin(), T->gateway_list.end(), u->info->id) == T->gateway_list.end() ) )
			{
				j = 0;
				/* outgoing links */
// 				pos = cumNumNeighbs[u->info->id]*numComms + 1;
				pos = cumNumNeighbs[u->info->id]*numComms;
				for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				{
					for (unsigned int i=0; i < numComms; i++)
						colno[i][j] = pos++;
					coeff[j++] = 1;
				}
				/* incoming links */
				for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				{
					invLink = T->FindInverseLink(TLinkItrs(u,v));
					dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 					for (unsigned int i=1; i <= numComms; i++)
					for (unsigned int i=0; i < numComms; i++)
// 						colno[i-1][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
						colno[i][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					coeff[j++] = -1;
				}
				/* add constraints */
				for (unsigned int i=0; i < numComms; i++)
				{
					// Need to copy because add_constraintex modifies the vector of coefficients
					copy(coeff, coeff+j, temp);
// 					if (!add_constraintex(lp, j, temp, colno[i], EQ, 0))
// 						ERROR();
					buildObject.addRow(j, colno[i], temp, 0., 0.);
				}
			}

	// add numComms constraints to ensure that the flow cannot return back to the source: \sum_{v\in Neighbs(s_i)} f_i(v->s_i) = 0 for i=1,..numComms
	vector<TGraphNode>::iterator n;
	/* aggregation nodes */
	for (vector<int>::iterator a=T->aggregation_list.begin(); a!=T->aggregation_list.end(); a++)
	{
		j = 0;
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			n = T->node.begin();
			advance(n,*a);
			invLink = T->FindInverseLink( TLinkItrs(n,v) );
			dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->aggregation_list.begin(), a) + 1;
			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->aggregation_list.begin(), a);
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], EQ, 0))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0., 0.);
	}
	/* gateways */
	for (vector<int>::iterator a=T->gateway_list.begin(); a!=T->gateway_list.end(); a++)
	{
		j = 0;
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			n = T->node.begin();
			advance(n,*a);
			invLink = T->FindInverseLink( TLinkItrs(n,v) );
			dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size() + 1;
			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size();
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], EQ, 0))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0., 0.);
	}

	// add numComms constraints to ensure that each commodity has a strictly positive flow: \sum_{v\in Neighbs(s_i)} f_i(s_i->v) > 0 for i=1,..numComms
	/* aggregation nodes */
	for (vector<int>::iterator a=T->aggregation_list.begin(); a!=T->aggregation_list.end(); a++)
	{
		j = 0;
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			dist = distance(T->node[*a].neighbor.begin(), v);
// 			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(), a) + 1;
			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(), a);
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], GE, 0.00001))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0.00001, COIN_DBL_MAX);
	}
	/* gateways */
	for (vector<int>::iterator a=T->gateway_list.begin(); a!=T->gateway_list.end(); a++)
	{
		j = 0;
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			dist = distance(T->node[*a].neighbor.begin(), v);
// 			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size() + 1;
			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size();
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], GE, 0.00001))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0.00001, COIN_DBL_MAX);
	}

// 	set_add_rowmode(lp, FALSE); /* rowmode should be turned off again when done building the model */

	model.addRows(buildObject);

	// set objective function: max \sum_{i=1}^{k} \sum_{v\in Neighbs(s_i)} f_i(s_i->v)
	j = 0;
	unsigned int al_size = T->aggregation_list.size();
	unsigned int gl_size = T->gateway_list.size();
	for (vector<int>::iterator a=T->aggregation_list.begin(); a!=T->aggregation_list.end(); a++)
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			dist = distance(T->node[*a].neighbor.begin(), v);
			/* there is a commodity between aggregation node (*a)->id and every gateway */
// 			for (unsigned int i=1; i<=gl_size; i++)
			for (unsigned int i=0; i<gl_size; i++)
			{
// 				colno[0][j*gl_size+i-1] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(),a)*gl_size + i;
				colno[0][j*gl_size+i] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(),a)*gl_size + i;
// 				coeff[j*gl_size+i-1] = 1;
				coeff[j*gl_size+i] = 1;
			}
			j++;
		}
	k = 0;
	for (vector<int>::iterator a=T->gateway_list.begin(); a!=T->gateway_list.end(); a++)
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			dist = distance(T->node[*a].neighbor.begin(), v);
			/* there is a commodity between gateway (*a)->id and every aggregation node */
// 			for (unsigned int i=1; i<=al_size; i++)
			for (unsigned int i=0; i<al_size; i++)
			{
// 				colno[0][j*gl_size+k*al_size+i-1] = (cumNumNeighbs[*a] + dist)*numComms + al_size*gl_size + distance(T->gateway_list.begin(),a)*al_size + i;
				colno[0][j*gl_size+k*al_size+i] = (cumNumNeighbs[*a] + dist)*numComms + al_size*gl_size + distance(T->gateway_list.begin(),a)*al_size + i;
// 				coeff[j*gl_size+k*al_size+i-1] = 1;
				coeff[j*gl_size+k*al_size+i] = 1;
			}
			k++;
		}

// 	set_obj_fnex(lp,j*gl_size+k*al_size,coeff,colno[0]);
	
	for (int i=0; i<j*gl_size+k*al_size; i++)
	  model.setObjectiveCoefficient(colno[0][i], coeff[i]);

// 	set_maxim(lp);
	model.setOptimizationDirection(-1);

//	print_lp(lp);

	cout << "Solving the LP..." << endl;

// 	if ( solve(lp) == 0 )
// 	{
// //		print_objective(lp);
// 		print_solution(lp, 4);
// //		print_constraints(lp, 4);
// 	}

// 	REAL result = get_objective(lp);

	int ret = model.initialSolve();

	if (ret)
	{
	  cerr << endl << "WARNING: solveMaxMultiCommFlow: An optimal solution to the LP has NOT been found" << endl << endl;

	  return 0;
	}

	/* clean up such that all used memory by lpsolve is freed */
// 	if(lp != NULL)
// 		delete_lp(lp);

	return model.getObjValue();
}



/* Given the initial graph and maxF, builds an LP to solve the 
   maximum minimum congestion factor problem
   Returns the value of all the variables
*/
double *solveMaxMinCongFactor(Topology *T, double maxF)
{
# if defined ERROR
#  undef ERROR
# endif
# define ERROR() { cerr << "Error" << endl; if(lp != NULL) delete_lp(lp); return(NULL); }
// 	lprec *lp;

	ClpSimplex model;

	// compute the cumulative number of neighbors for each node. It is such that:
	// - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
	// - the last element of cumuNumNeighbs is the total number of links
	unsigned int cumNumNeighbs[T->node.size() + 1];

	cumNumNeighbs[0] = 0;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	{
		cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id];
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			cumNumNeighbs[u->info->id + 1]++;
	}

	// compute the number of commodities (each aggregation router/gateway acts both as a source and a sink)
	unsigned int numComms = T->aggregation_list.size() * T->gateway_list.size() * 2;


	char var_name[100];
	unsigned int num_vars = cumNumNeighbs[T->node.size()] * numComms * 1.5 + 1;
	int colno[numComms][num_vars];
	double coeff[num_vars];
	double temp[num_vars];
	double * vars = new double[num_vars];

	// create the LP problem
	// Number of variables: |E|*|V_A|*|V_G|*2 + 1
// 	if ((lp = make_lp(0, num_vars)) == NULL)
// 		ERROR();
	
	model.resize(0, num_vars);

	/* Columns (variables) are ordered this way:
	   f_e1,c1 ... f_e1,cM ... f_eE,c1 ... f_eE,cM B_1,c1 ... B_1,cM ... B_(E/2),c1 ... B_(E/2),cM alpha
	   where links e1 ... eE are given by
	   neighbors(node 0) neighbors(node 1) ... neighbors(node N)
	   and commodities c1 ... cM are given by
	   a_1,g_1 ... a_1,g_G ... a_A,g_1 ... a_A,g_G ... g_1,a_1 ... g_1,a_A ... g_G,a_1 ... g_G,a_A
	*/

	// give a name to the variables
	int linkId = 0;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
		{
			for (unsigned int i=1; i <= numComms; i++)
			{
// 				sprintf(var_name,"f_(%d->%d),%d",u->info->id,v->info->id,i);
// 				if (!set_col_name(lp, linkId*numComms + i, var_name))
// 					ERROR();
				stringstream s;
				s << "f_(" << u->info->id << "->" << v->info->id << ")," << i;
				string str = s.str();
				model.setColumnName(linkId*numComms + i - 1, str);
			}
			linkId++;
		}
	linkId = 0;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->id < v->info->id)       /* each undirected link just once */
			{
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					sprintf(var_name,"B_(%d->%d),%d",u->info->id,v->info->id,i);
// 					if (!set_col_name(lp, (cumNumNeighbs[T->node.size()]+linkId)*numComms + i, var_name))
// 						ERROR();
					stringstream s;
					s << "B_(" << u->info->id << "->" << v->info->id << ")," << i+1;
					string str = s.str();
					model.setColumnName((cumNumNeighbs[T->node.size()]+linkId)*numComms + i, str);
#ifdef MILP
// 					if (!set_binary(lp, (cumNumNeighbs[T->node.size()]+linkId)*numComms + i, TRUE))
					model.setInteger((cumNumNeighbs[T->node.size()]+linkId)*numComms + i);
					model.setColumnBounds((cumNumNeighbs[T->node.size()]+linkId)*numComms + i, 0., 1.);
#else
// 					if (!set_bounds(lp, (cumNumNeighbs[T->node.size()]+linkId)*numComms + i, 0, 1))
					model.setColumnBounds((cumNumNeighbs[T->node.size()]+linkId)*numComms + i, 0., 1.);
#endif
// 						ERROR();
				}
				linkId++;
			}
// 	if (!set_col_name(lp, num_vars, "alpha"))
// 		ERROR();
	{
		string str("alpha");
		model.setColumnName(num_vars-1, str);
	}


// 	set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

	CoinBuild buildObject;

	// add |E|/2 capacity constraints: \sum_{i=1}^{k} f_i(u->v) / c(u->v) + \sum_{i=1}^{k} f_i(v->u) / c(v->u) <= 1 for each u<->v
	TLinkItrs invLink;
	unsigned int dist;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->id < v->info->id)          /* each undirected link just once */
			{
				/* link u->v */
				dist = distance(u->neighbor.begin(), v);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					coeff[i-1] = 1. / v->rate;
					coeff[i] = 1. / v->rate;
// 					colno[0][i-1] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
					colno[0][i] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
				}
				/* link v->u */
				invLink = T->FindInverseLink(TLinkItrs(u,v));
				dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					coeff[numComms+i-1] = 1. / invLink.second->rate;
					coeff[numComms+i] = 1. / invLink.second->rate;
// 					colno[0][numComms+i-1] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					colno[0][numComms+i] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
				}
				/* add constraint */
// 				if (!add_constraintex(lp, numComms*2, coeff, colno[0], LE, 1))
// 					ERROR();
				buildObject.addRow(numComms*2, colno[0], coeff, -COIN_DBL_MAX, 1.);
			}

	// add |E|*numComms constraints on the minimum congestion factor: \sum_{i=1}^{k} f_i(u->v) >= alpha * c(u->v) for each u->v
	linkId = 0;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
		{
// 			for (unsigned int i=1; i <= numComms; i++)
			for (unsigned int i=0; i < numComms; i++)
			{
// 				coeff[i-1] = 1;
				coeff[i] = 1;
// 				colno[0][i-1] = linkId*numComms + i;
				colno[0][i] = linkId*numComms + i;
			}
			coeff[numComms] = - v->rate;
// 			colno[0][numComms] = num_vars;
			colno[0][numComms] = num_vars-1;
// 			if (!add_constraintex(lp, numComms+1, coeff, colno[0], GE, 0))
// 				ERROR();
			buildObject.addRow(numComms+1, colno[0], coeff, 0., COIN_DBL_MAX);
			linkId++;
		}

	// add (|V|-|V_A|-|V_G|)*numComms flow conservation constraints
	unsigned int j, k, pos;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		if ( ( find(T->aggregation_list.begin(), T->aggregation_list.end(), u->info->id) == T->aggregation_list.end() ) &&
		     ( find(T->gateway_list.begin(), T->gateway_list.end(), u->info->id) == T->gateway_list.end() ) )
			{
				j = 0;
				/* outgoing links */
// 				pos = cumNumNeighbs[u->info->id]*numComms + 1;
				pos = cumNumNeighbs[u->info->id]*numComms;
				for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				{
					for (unsigned int i=0; i < numComms; i++)
						colno[i][j] = pos++;
					coeff[j++] = 1;
				}
				/* incoming links */
				for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				{
					invLink = T->FindInverseLink(TLinkItrs(u,v));
					dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 					for (unsigned int i=1; i <= numComms; i++)
					for (unsigned int i=0; i < numComms; i++)
// 						colno[i-1][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
						colno[i][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					coeff[j++] = -1;
				}
				/* add constraints */
				for (unsigned int i=0; i < numComms; i++)
				{
					// Need to copy because add_constraintex modifies the vector of coefficients
					copy(coeff, coeff+j, temp);
// 					if (!add_constraintex(lp, j, temp, colno[i], EQ, 0))
// 						ERROR();
					buildObject.addRow(j, colno[i], temp, 0., 0.);
				}
			}

#ifndef MILP
	// add numComms constraints to ensure that the flow cannot return back to the source: \sum_{v\in Neighbs(s_i)} f_i(v->s_i) = 0 for i=1,..numComms
	vector<TGraphNode>::iterator n;
	/* aggregation nodes */
	for (vector<int>::iterator a=T->aggregation_list.begin(); a!=T->aggregation_list.end(); a++)
	{
		j = 0;
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			n = T->node.begin();
			advance(n,*a);
			invLink = T->FindInverseLink( TLinkItrs(n,v) );
			dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->aggregation_list.begin(), a) + 1;
			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->aggregation_list.begin(), a);
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], EQ, 0))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0., 0.);
	}
	/* gateways */
	for (vector<int>::iterator a=T->gateway_list.begin(); a!=T->gateway_list.end(); a++)
	{
		j = 0;
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			n = T->node.begin();
			advance(n,*a);
			invLink = T->FindInverseLink( TLinkItrs(n,v) );
			dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size() + 1;
			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size();
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], EQ, 0))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0., 0.);
	}

	// add numComms constraints to ensure that each commodity has a strictly positive flow: \sum_{v\in Neighbs(s_i)} f_i(s_i->v) > 0 for i=1,..numComms
	/* aggregation nodes */
	for (vector<int>::iterator a=T->aggregation_list.begin(); a!=T->aggregation_list.end(); a++)
	{
		j = 0;
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			dist = distance(T->node[*a].neighbor.begin(), v);
// 			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(), a) + 1;
			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(), a);
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], GE, 0.00001))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0.00001, COIN_DBL_MAX);
	}
	/* gateways */
	for (vector<int>::iterator a=T->gateway_list.begin(); a!=T->gateway_list.end(); a++)
	{
		j = 0;
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			dist = distance(T->node[*a].neighbor.begin(), v);
// 			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size() + 1;
			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size();
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], GE, 0.00001))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0.00001, COIN_DBL_MAX);
	}
#endif

#if 0
	// add |E|*numComms special ordered sets (SOS)
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (vector<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->id < v->info->id)          /* each undirected link just once */
			{
				/* link u->v */
				dist = distance(u->neighbor.begin(), v);
				for (unsigned int i=1; i <= numComms; i++)
					colno[i-1][0] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
				/* link v->u */
				invLink = T->FindInverseLink(TLinkItrs(u,v));
				dist = distance(invLink.first->neighbor.begin(), invLink.second);
				for (unsigned int i=1; i <= numComms; i++)
					colno[i-1][1] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
				/* add constraint */
				for (unsigned int i=1; i <= numComms; i++)
					add_SOS(lp, "SOS", 1, 1, 2, colno[i-1], NULL);
			}
#endif

	// add |E|/2*numComms pairs of constraints: f_i(u->v) <= B * c(u->v) and f_i(v->u) <= (1-B) * c(v->u)
	linkId = 0;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->id < v->info->id)          /* each undirected link just once */
			{
				/* link u->v */
				dist = distance(u->neighbor.begin(), v);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
					coeff[0] = 1;
// 					colno[i-1][0] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
					colno[i][0] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
					coeff[1] = - v->rate;
// 					colno[i-1][1] = (cumNumNeighbs[T->node.size()]+linkId)*numComms + i;
					colno[i][1] = (cumNumNeighbs[T->node.size()]+linkId)*numComms + i;
// 					if (!add_constraintex(lp, 2, coeff, colno[i-1], LE, 0))
// 						ERROR();
					buildObject.addRow(2, colno[i], coeff, -COIN_DBL_MAX, 0.);
				}
				/* link v->u */
				invLink = T->FindInverseLink(TLinkItrs(u,v));
				dist = distance(invLink.first->neighbor.begin(), invLink.second);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
					coeff[0] = 1;
// 					colno[i-1][0] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					colno[i][0] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					coeff[1] = invLink.second->rate;
// 					colno[i-1][1] = (cumNumNeighbs[T->node.size()]+linkId)*numComms + i;
					colno[i][1] = (cumNumNeighbs[T->node.size()]+linkId)*numComms + i;
// 					if (!add_constraintex(lp, 2, coeff, colno[i-1], LE, invLink.second->rate))
// 						ERROR();
					buildObject.addRow(2, colno[i], coeff, -COIN_DBL_MAX, invLink.second->rate );
				}
				linkId++;
			}

	// add the constraint on the maximum throughput: \sum_{i=1}^{k} \sum_{v\in Neighbs(s_i)} f_i(s_i->v) = maxF
	j = 0;
	unsigned int al_size = T->aggregation_list.size();
	unsigned int gl_size = T->gateway_list.size();
	for (vector<int>::iterator a=T->aggregation_list.begin(); a!=T->aggregation_list.end(); a++)
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			dist = distance(T->node[*a].neighbor.begin(), v);
			/* there is a commodity between aggregation node (*a)->id and every gateway */
// 			for (unsigned int i=1; i<=gl_size; i++)
			for (unsigned int i=0; i<gl_size; i++)
			{
// 				colno[0][j*gl_size+i-1] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(),a)*gl_size + i;
// 				coeff[j*gl_size+i-1] = 1;
				colno[0][j*gl_size+i] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(),a)*gl_size + i;
				coeff[j*gl_size+i] = 1;
			}
			j++;
		}
	k = 0;
	for (vector<int>::iterator a=T->gateway_list.begin(); a!=T->gateway_list.end(); a++)
		for (list<TGraphNeighbor>::iterator v=T->node[*a].neighbor.begin(); v!=T->node[*a].neighbor.end(); v++)
		{
			dist = distance(T->node[*a].neighbor.begin(), v);
			/* there is a commodity between gateway (*a)->id and every aggregation node */
// 			for (unsigned int i=1; i<=al_size; i++)
			for (unsigned int i=0; i<al_size; i++)
			{
// 				colno[0][j*gl_size+k*al_size+i-1] = (cumNumNeighbs[*a] + dist)*numComms + al_size*gl_size + distance(T->gateway_list.begin(),a)*al_size + i;
// 				coeff[j*gl_size+k*al_size+i-1] = 1;
				colno[0][j*gl_size+k*al_size+i] = (cumNumNeighbs[*a] + dist)*numComms + al_size*gl_size + distance(T->gateway_list.begin(),a)*al_size + i;
				coeff[j*gl_size+k*al_size+i] = 1;
			}
			k++;
		}

// 	if (!add_constraintex(lp, j*gl_size+k*al_size, coeff, colno[0], EQ, maxF))
// 		ERROR();
	buildObject.addRow(j*gl_size+k*al_size, colno[0], coeff, maxF, maxF);

// 	set_add_rowmode(lp, FALSE); /* rowmode should be turned off again when done building the model */

	model.addRows(buildObject);

	// set the objective function: maximize alpha
	coeff[0] = 1;
// 	colno[0][0] = num_vars;
	colno[0][0] = num_vars-1;
// 	set_obj_fnex(lp,1,coeff,colno[0]);
	model.setObjectiveCoefficient(colno[0][0], coeff[0]);

// 	set_maxim(lp);
	model.setOptimizationDirection(-1);

//	print_lp(lp);

	cout << "Solving the LP..." << endl;

// 	if ( solve(lp) == 0 )
// 	{
// //		print_objective(lp);
// 		print_solution(lp, 4);
// //		print_constraints(lp, 4);
// 	}

	int ret = model.initialSolve();

	if (ret)
	{
		cerr << endl << "WARNING: solveMaxMinCongFactor: An optimal solution to the LP has NOT been found" << endl << endl;
		return NULL;
	}
	else
	{
// 		if (!get_variables(lp, vars))
// 			ERROR();
		vars = model.primalColumnSolution();

// 		/* clean up such that all used memory by lpsolve is freed */
// 		if(lp != NULL)
// 			delete_lp(lp);

		return vars;
	}
}


